/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/

package javax.bpelj;

import javax.xml.namespace.QName;

/**
 * BPELJException.java
 * Code called from a Java snippet may generate a BPEL fault by throwing a
 * javax.bpelj.BPELJException exception. The constructor of this exception
 * requires the QName of the BPEL fault that should be thrown. The exception
 * may also, optionally, contain an object to represent the fault data.
 *
 * @author Jean-Noel Rouvignac
 * @version $Id: BPELJException.java 5 2006-10-20 14:37:44Z lejeuneg $
 */
public class BPELJException extends Exception {

    /**
     * the mandatory QName of the BPELJ fault that should be thrown
     */
    private QName  faultName;
    /**
     * an optional object representing the fault data
     */
    private Object faultData = null;

    /**
     * Constructs a new BPELJException. It requires the QName of the BPEL
     * fault that should be thrown. The exception may also, optionally,
     * contain an object to represent the fault data.
     * @param faultName the mandatory QName of the BPELJ fault that should be thrown
     * @param faultData an optional object representing the fault data
     * @throws IllegalArgumentException if faultName is null
     */
    public BPELJException(QName faultName, Object faultData) throws Exception {
        super(faultName.toString());
        init(faultName, faultData);
    }

    /**
     * Constructs a new BPELJException. It requires the QName of the BPEL
     * fault that should be thrown. The exception may also, optionally,
     * contain an object to represent the fault data.
     * @param faultName the mandatory QName of the BPELJ fault that should be thrown
     * @param faultData an optional object representing the fault data
     */
    public BPELJException(QName faultName, Throwable faultData) {
        super(faultName.toString(), faultData);
        init(faultName, faultData);
    }

    /**
     * Intializes the faultName and the faultData
     * @param faultName the mandatory QName of the BPELJ fault that should be thrown
     * @param faultData an optional object representing the fault data
     */
    private void init(QName faultName, Object faultData) {
        if (faultName == null) {
            throw new IllegalArgumentException("The faultName of a javax.bpelj.BPELJException must not be null.");
        }
        this.faultName = faultName;
        this.faultData = faultData;
    }

    /**
     * Returns the QName of the BPEL fault that should be thrown
     * @return the QName of the BPEL fault that should be thrown
     */
    public QName getFaultName() {
        return this.faultName;
    }

    /**
     * Returns the object representing the fault data of this BPEL fault
     * @return the object representing the fault data
     */
    public Object getFaultData() {
        return this.faultData;
    }

}
