/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.objectweb.orchestra.bindingframework;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.Map;
import java.util.Vector;

import javax.xml.namespace.QName;
import javax.xml.rpc.Call;

import org.objectweb.orchestra.engine.BPELMessage;
import org.objectweb.orchestra.exception.BFUtilException;
import org.objectweb.orchestra.exception.BPELBindingFrameworkException;
import org.objectweb.orchestra.util.BPELTrace;


/**
 * BCSoapWork is the BindingFramework SOAP binding component class.  This supports
 * invokes using Soap.
 *
 * @author Eric Hardesty
 * @version $Id: BCSoapWork.java 39 2006-11-21 10:09:54Z lejeuneg $
 *
 */
public class BCSoapWork extends BCBaseWork {


    /**
     * Constructor.
     *
     * @param classL Classloader to find classes
     * @param partLinkName String partner link name
     * @param qnamePortType String QName port type
     * @param invokeOp String invoke operation
     * @param jmsgin BPELMessage input data
     * @param jmsgout BPELMessage output data
     * @param endpoint String endpoint to use
     * @param address String address to use
     * @param outVar boolean to deteremine if output variables are needed
     * @param processName String process name
     * @param processId String process Id
     * @param sl ServiceListener to get events
     */
    public BCSoapWork(ClassLoader classL, String partLinkName, String qnamePortType, String port, String serviceLocatorClassName,
            String invokeOp, BPELMessage jmsgin, BPELMessage jmsgout,
            String endpoint, String address, boolean outVar, String processName, int processId, ServiceListener sl) {
        super(classL, partLinkName, qnamePortType, port, serviceLocatorClassName, invokeOp, jmsgin, jmsgout,
                endpoint, address, outVar, processName, processId, sl);
    }

    /**
     * Release call from work manager, nothing to disable at this time.
     */
    public void release() {
        // TODO
    }

    /**
     * Invocation from the WorkManager.
     */
    public void run() {
        ServiceData result = null;
        ServiceException invokeEx = null;
        try {
            result = (ServiceData) invoke(this.classL, this.partLinkName, this.qnamePortType,
                    this.port, this.serviceLocatorClassName, this.invokeOp, this.jmsgin,
                    this.jmsgout, this.endpoint, this.address, this.outVar, this.processName, this.processId);

        } catch (Exception ex) {
            ex.printStackTrace();
            invokeEx = new ServiceException(ex.getMessage(), ex.getCause());
            if (BPELTrace.isBFDebugEnabled()) {
                BPELTrace.traceBF("BF BCSoapWork invokeEx: " + invokeEx);
            }
        }

        // if the invoke failed, setup a ServiceData object to hold exception information
        if (result == null) {
            result = new ServiceData();
            result.processId = this.processId;
        }

        ServiceEvent se = new ServiceEvent(this, result, invokeEx);
        if (invokeEx != null) {
            this.sl.faultOccurred(se);
        } else {
            this.sl.responseReceived(se);
        }
    }

    /**
     * Invoke the webservice.
     *
     * @param classL
     * @param partLinkName
     * @param qnamePortType
     * @param port
     * @param serviceLocatorClassName
     * @param invokeOp
     * @param jmsgin
     * @param jmsgout
     * @param endpoint
     * @param address
     * @param outVar
     * @param processName
     * @param processId
     * @return
     * @throws BPELBindingFrameworkException
     * @throws InvocationTargetException
     */
    private Object invoke(final ClassLoader classL, final String partLinkName, final String qnamePortType, final String port,
            final String serviceLocatorClassName, final String invokeOp, final BPELMessage jmsgin, final BPELMessage jmsgout,
            final String endpoint, final String address, final boolean outVar, final String processName, final int processId)
    throws BPELBindingFrameworkException, InvocationTargetException {
        if (BPELTrace.isBFDebugEnabled()) {
            BPELTrace.traceBF("BF BCSoapWork invoke, classL = " + classL + "   partnerLink = " + partLinkName
                    + "   qnamePortType = " + qnamePortType + "   invokdeOp = " + invokeOp + "   jmsgin = " + jmsgin
                    + "  jmsgout = " + jmsgout + "   serviceLocatorClassName = " + serviceLocatorClassName + ", port = "+port
                    + "   address = " + address);
        }

        if (BPELTrace.isBFDebugEnabled()) {
            BPELTrace.traceBF("BF BCSoapWork invoke serviceLocatorClassName = "+serviceLocatorClassName);
            BPELTrace.traceBF("BF BCSoapWork invoke port = "+port);
            BPELTrace.traceBF("BF BCSoapWork invoke endpoint = "+endpoint);
        }

        //get the service class
        Object result = null;
        Method met = null;
        Class cl = null;
        Object stub = null;
        Map outputMap = null;
        Object [] tabInput = null;
        ClassLoader origCL = null;
        try {
            origCL = Thread.currentThread().getContextClassLoader();
            Thread.currentThread().setContextClassLoader(classL);
            cl = Class.forName(serviceLocatorClassName, true, classL);
            Constructor constr = cl.getConstructor((Class[]) null);
            //ser is an instance of the service
            Object ser = constr.newInstance((Object[]) null);
            met = cl.getMethod("getServiceName", (Class[]) null);
            met.invoke(ser, (Object[]) null);
            met = cl.getMethod("setEndpointAddress", new Class[]{java.lang.String.class, java.lang.String.class});
//          met.invoke(ser, new Object[]{port,address});
            met.invoke(ser, new Object[]{port, endpoint});
            //String wsdllocation = address + "?wsdl";

            met = ser.getClass().getMethod("get"+port, (Class[]) null);
            stub = met.invoke(ser, (Object[]) null);

            cl = stub.getClass();
            met = getMethodByName(cl, invokeOp);

            // construct the input parameters tab
            Vector values = BFUtil.getInputValues(jmsgin, jmsgout, met, processName);
            tabInput = new Object[values.size()];
            tabInput = values.toArray();

            try {
                result = met.invoke(stub, tabInput);
            } catch (IllegalArgumentException ill) {
                String err = "IllegalArgumentException BF BCSoapWork invoke when performing the WS call :\n";
                err += "classL = "+classL+", partLinkName = "+partLinkName+", qnamePortType = "+qnamePortType;
                err += ", invokeOp = "+invokeOp+", jmsgin = "+jmsgin+", jmsgout = "+jmsgout;
                err += ", serviceLocatorClassName = "+serviceLocatorClassName+", port = "+port+", endpoint = "+endpoint
                +", address = "+address;
                err += ", outVar = "+outVar+", processName = "+processName+", processId = "+processId;
                err += "cl = "+cl+", met = "+met+", stub = "+stub+", values = "+values;
                err += "values.size = "+values.size()+", tabInput = "+tabInput+", invokeOp = "+invokeOp;
                err += " : \n";
                ill.printStackTrace();
                throw new BPELBindingFrameworkException(err, ill);
            }

            if (outVar) {
                Call call = null;
                met = cl.getMethod("getMyCalls", (Class[]) null);
                Vector myCalls = (Vector) met.invoke(stub, (Object[]) null);
                //get the call corresponding to the one we make just before
                for (int e = 0; e < myCalls.size(); e++) {
                    QName tmpCall = (QName) myCalls.get(e);
                    e++;
                    if (tmpCall.getLocalPart().equals(invokeOp)) {
                        call = (Call) myCalls.get(e);
                    }
                }
                //get outputparams
                outputMap = call.getOutputParams();
            }
        } catch (IllegalArgumentException ex) {
            ex.printStackTrace();
            throw new BPELBindingFrameworkException("IllegalArgumentException BF BCSoapWork invoke :\n" , ex);
        } catch (IllegalAccessException ex) {
            throw new BPELBindingFrameworkException("IllegalAccessException BF BCSoapWork invoke :\n" , ex);
        } catch (InvocationTargetException ex) {
            if (BPELTrace.isBFDebugEnabled()) {
                BPELTrace.traceBF("InvocationTargetException BF BCSoapWork cause: " + ex.getCause());
            }
            ex.printStackTrace();
            throw ex;
        } catch (NoSuchMethodException ex) {
            throw new BPELBindingFrameworkException("NoSuchMethodException BF BCSoapWork invoke :\n" , ex);
        } catch (InstantiationException ex) {
            throw new BPELBindingFrameworkException("InstantiationException BF BCSoapWork invoke :\n" , ex);
        } catch (ClassNotFoundException ex) {
            throw new BPELBindingFrameworkException("ClassNotFoundException BF BCSoapWork invoke :\n" , ex);
        } catch (BFUtilException ex) {
            throw new BPELBindingFrameworkException("BFUtilException BF BCSoapWork invoke :\n" , ex);
        } finally {
            if (origCL != null) {
                Thread.currentThread().setContextClassLoader(origCL);
            }
        }

        // Build output data item
        ServiceData sd = new ServiceData();
        sd.result = result;
        sd.outputMap = outputMap;
        sd.processId = processId;
        return sd;
    }

}
