/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.objectweb.orchestra.bindingframework;

import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.Map;
import java.util.Vector;

import javax.resource.spi.work.ExecutionContext;
import javax.resource.spi.work.Work;
import javax.resource.spi.work.WorkCompletedException;
import javax.resource.spi.work.WorkEvent;
import javax.resource.spi.work.WorkException;
import javax.resource.spi.work.WorkListener;
import javax.resource.spi.work.WorkManager;
import javax.resource.spi.work.WorkRejectedException;

import org.objectweb.orchestra.util.BPELTrace;
import org.objectweb.transaction.jta.TransactionManager;


/**
 * BFWorkManager implements the WorkManager interface of the Connector Architecture
 * 1.5 specification.
 *
 * @author Eric Hardesty
 * @version $Id: BFWorkManager.java 5 2006-10-20 14:37:44Z lejeuneg $
 *
 */
public class BFWorkManager implements WorkManager {

    int          minThreads;
    //private int          maxThreads;
    long         waitTimeout;
    long         execTimeout;
    Vector       vThreads;

    LinkedList   workObjs;
    int freeThreads;  // free threads ready to work

    //private final static int WO_REJECTED = 1;
    //private final static int WO_ACCEPTED = 2;
    //private final static int WO_STARTED = 3;
    //private final static int WO_COMPLETED = 4;

    /**
     * The transaction manager in the server
     */
    //private TransactionManager tm = null ;

    // Static hashtable: Xid ---> work Object
    static Map wrkXids = Collections.synchronizedMap(new HashMap());

    public BFWorkManager(TransactionManager tmgr, int minThreads, int maxThreads,
            int waitTimeout, int execTimeout) {
        this.minThreads  = minThreads;
        //this.maxThreads  = maxThreads;
        this.waitTimeout = waitTimeout * 1000;
        this.execTimeout = execTimeout * 1000;
        this.vThreads    = new Vector();
        this.workObjs    = new LinkedList();
        //tm          = tmgr;

        // Start the specified number of threads
        for (int i = 0; i < minThreads; i++) {
            WorkerPool wp = new WorkerPool();
            this.vThreads.add(wp);
            wp.start();
        }

        // Need the Monitor thread to check on timeouts
        (new BFManagerMonitor()).start();
    }


    /**
     * Run the specified work object and wait until it completes to return
     * @param work the Work object to execute
     * @throws WorkException if an error occurs during
     * the acceptance, starting or execution
     */
    public void doWork(Work work) throws WorkException {
        doWork(work, INDEFINITE, null, null);
    }

    /**
     * Run the specified work object and wait until it completes to return
     * @param work the Work object to execute
     * @param timeout the max number of milliseconds before Work object must start
     * @param eCtx the ExecutionContext to use for the Work object
     * @param wListen the WorkListener to use
     * @throws WorkException if an error occurs during
     * the acceptance, starting or execution
     */
    public void doWork(Work work, long timeout, ExecutionContext eCtx,
            WorkListener wListen) throws WorkException {

        if (BPELTrace.isBFDebugEnabled())
            BPELTrace.traceBF("work=" + work + " timeout=" + timeout
                    + " execCtx=" + eCtx + " wrkListen=" + wListen);
        int rStat = WorkEvent.WORK_COMPLETED;
        WorkObject wo = new WorkObject(work, timeout, eCtx, wListen, rStat);
        synchronized (wo.lock) {
            add(wo, rStat);
            try {
                wo.lock.wait();
            } catch (InterruptedException ie) {
                // Nothing
            }
        }
        if (wo.wException != null) {
            if (BPELTrace.isBFDebugEnabled())
                BPELTrace.traceBF("" + wo.wException);
            throw wo.wException;
        }
    }

    /**
     * Run the specified work object and return when it has been started
     * @param work the Work object to execute
     * @return long number of milliseconds elapsed between work acceptance
     *   and work start
     * @throws WorkException if an error occurs during
     * the acceptance, starting or execution
     */
    public long startWork(Work work) throws WorkException {
        return startWork(work, INDEFINITE, null, null);
    }

    /**
     * Run the specified work object and return when it has started
     * @param work the Work object to execute
     * @param timeout the max number of milliseconds before Work object must start
     * @param eCtx the ExecutionContext to use for the Work object
     * @param wListen the WorkListener to use
     * @return long number of milliseconds elapsed between work acceptance
     *   and work start
     * @throws WorkException if an error occurs during
     * the acceptance, starting or execution
     */
    public long startWork(Work work, long timeout, ExecutionContext eCtx,
            WorkListener wListen) throws WorkException {

        if (BPELTrace.isBFDebugEnabled())
            BPELTrace.traceBF("work=" + work + " timeout=" + timeout
                    + " execCtx=" + eCtx + " wrkListen=" + wListen);
        int rStat = WorkEvent.WORK_STARTED;
        WorkObject wo = new WorkObject(work, timeout, eCtx, wListen, rStat);
        synchronized (wo.lock) {
            add(wo, rStat);
            try {
                wo.lock.wait();
            } catch (InterruptedException ie) {
                // Nothing
            }
        }

        if (wo.wException != null) {
            throw wo.wException;
        }
        return (wo.startTime - wo.acceptTime);
    }

    /**
     * Run the specified work object and return when it has been accepted
     * @param work the Work object to execute
     * @throws WorkException if an error occurs during
     * the acceptance, starting or execution
     */
    public void scheduleWork(Work work) throws WorkException {
        scheduleWork(work, INDEFINITE, null, null);
    }

    /**
     * Run the specified work object and return when it has been accepted
     * @param work the Work object to execute
     * @param timeout the max number of milliseconds before Work object must start
     * @param eCtx the ExecutionContext to use for the Work object
     * @param wListen the WorkListener to use
     * @throws WorkException if an error occurs during
     * the acceptance, starting or execution
     */
    public void scheduleWork(Work work, long timeout,
            ExecutionContext eCtx,
            WorkListener wListen) throws WorkException {

        if (BPELTrace.isBFDebugEnabled())
            BPELTrace.traceBF("work=" + work + " timeout=" + timeout
                    + " execCtx=" + eCtx + " wrkListen=" + wListen);
        int rStat = WorkEvent.WORK_ACCEPTED;
        WorkObject wo = new WorkObject(work, timeout, eCtx, wListen, rStat);
        add(wo, rStat);
    }

    /**
     * Used for testing to cause the threads to go away
     */
    public void releasePool() {
        for (int i = 0; i < this.vThreads.size(); i++) {
            WorkerPool wp = (WorkerPool) this.vThreads.get(i);
            wp.shutdown = true;
        }
        synchronized (this.workObjs) {
            this.workObjs.notifyAll();
        }
    }

    /**
     * Accept the Work object and post a WorkEvent on the WorkListener if it
     * was configured.
     */
    private void add(WorkObject wo, int rStat) throws WorkException {

        if (BPELTrace.isBFDebugEnabled())
            BPELTrace.traceBF("workobj=" + wo + " rStat=" + rStat);

        // Add last to list, set accepted time and send the event if necessary
        synchronized (this.workObjs) {
            // If an EnvironmentContext and an Xid exists, then check Xid/Work hash
            if (wo.eCtx != null && wo.eCtx.getXid() != null) {
                if (wrkXids.containsKey(wo.eCtx.getXid())) {
                    WorkRejectedException wex = new WorkRejectedException(
                            "Xid is already being processed by another work object",
                            WorkException.TX_CONCURRENT_WORK_DISALLOWED);
                    WorkEvent we = new WorkEvent(wo.work, WorkEvent.WORK_REJECTED, wo.work, wex);
                    if (wo.wListen != null) {
                        wo.wListen.workRejected(we);
                    }
                    if (BPELTrace.isBFDebugEnabled())
                        BPELTrace.traceBF("" + wex);
                    throw wex;
                }
                wrkXids.put(wo.eCtx.getXid(), wo.work);
            }
            this.workObjs.addLast(wo);
            wo.acceptTime = System.currentTimeMillis();
            if (wo.wListen != null) {
                WorkEvent we = new WorkEvent(wo.work, WorkEvent.WORK_ACCEPTED, wo.work, null);
                wo.wListen.workAccepted(we);
            }
            if (this.freeThreads <= 0) {
                WorkerPool wp = new WorkerPool();
                this.vThreads.add(wp);
                wp.start();
            }
            this.workObjs.notify();
        }
    }


    // Inner classes for ResourceWorkManager
    /* Wrapper class for the Work object, keeps state and time information
     */
    private class WorkObject {
        // Work object information
        Work work;
        ExecutionContext eCtx;
        WorkListener wListen;
        WorkException wException;

        // Time values
        long acceptTime;
        long executionTimeout;
        long startTime;
        long timeout;

        // State information
        Object lock;
        int rState;
        int cState;
        boolean rejected;

        public WorkObject(Work work, long timeout,
                ExecutionContext eCtx,
                WorkListener wListen,
                int rState) {

            this.eCtx = eCtx;
            this.work = work;
            this.wException = null;
            this.wListen = wListen;

            this.executionTimeout = 0;
            this.acceptTime = 0;
            this.startTime = 0;
            long curTime = System.currentTimeMillis();
            if (timeout == INDEFINITE) {
                this.timeout = timeout;
            } else {
                this.timeout = curTime + timeout;
            }
            if (BFWorkManager.this.execTimeout > 0) {
                this.executionTimeout = curTime + BFWorkManager.this.execTimeout;
            }

            this.lock = new Object();
            this.rState = rState;
            this.cState = -1;
            this.rejected = false;
        }
    }


    // Thread pool runs the work objects
    private class WorkerPool extends Thread {
        WorkObject wo = null;
        boolean shutdown = false;

        public WorkerPool() {
            super();
            setDaemon(true);
            if (BPELTrace.isBFDebugEnabled())
                BPELTrace.traceBF(getName() + " constructor");
        }

        public void run() {

            while (true) {
                synchronized (BFWorkManager.this.workObjs) {
                    boolean waited = false;
                    while (BFWorkManager.this.workObjs.isEmpty()) {
                        // If a thread has waited and not gotten any work then there
                        // should only be minThreads waiting, so if freeThreads is equal
                        // to minThreads then this thread should be stopped
                        if ((waited && BFWorkManager.this.freeThreads >= BFWorkManager.this.minThreads) || this.shutdown) {
                            if (BPELTrace.isBFDebugEnabled())
                                BPELTrace.traceBF(getName() + " ending");
                            return;
                        }
                        try {
                            BFWorkManager.this.freeThreads++;
                            BFWorkManager.this.workObjs.wait(BFWorkManager.this.waitTimeout);
                            BFWorkManager.this.freeThreads--;
                            waited = true;
                        } catch (InterruptedException ignored) {
                            BFWorkManager.this.freeThreads--;
                            return;
                        }
                    }
                    this.wo = (WorkObject) BFWorkManager.this.workObjs.removeFirst();
                }

                if (this.wo != null) {
                    try {
                        checkTime();
                        if (this.wo.rState == WorkEvent.WORK_STARTED) {
                            if (BPELTrace.isBFDebugEnabled())
                                BPELTrace.traceBF("Start notification for " + this.wo);
                            synchronized (this.wo.lock) {
                                this.wo.lock.notify();  // If waiting for execution start
                            }
                        }
                        if (this.wo.wListen != null) {
                            WorkEvent we = new WorkEvent(this.wo.work, WorkEvent.WORK_STARTED,
                                    this.wo.work, null, this.wo.startTime);
                            this.wo.wListen.workStarted(we);
                        }
                        // Need to setup up ExecutionContext if necessary
                        if (this.wo.eCtx != null) {
                            //long timeout = wo.eCtx.getTransactionTimeout();
                            //Xid xid = wo.eCtx.getXid();
                            /*
                             if (xid != null) {
                                 try {
                                     if (timeout != WorkManager.UNKNOWN) {
//                                         ((Current) tm).begin(xid, timeout);
                                     } else {
//                                         ((Current) tm).begin(xid);
                                     }
                                 } catch (NotSupportedException nse) {
                                     if (BPELTrace.isBFDebugEnabled())
                                        BPELTrace.traceBF("cannot start a transaction: NotSupportedException");
                                     WorkException wex = new WorkException(
                                           "Error starting a new transaction", nse);
                                     throw wex;
                                 } catch (SystemException ex) {
                                     if (BPELTrace.isBFDebugEnabled())
                                        BPELTrace.traceBF("cannot start a transaction:\n" + ex);
                                     WorkException wex = new WorkException(
                                           "Error starting a new transaction", ex);
                                     throw wex;
                                 }
                             }
                             */
                        }
                        if (BPELTrace.isBFDebugEnabled())
                            BPELTrace.traceBF("Thread: " + getName() + " Running " + this.wo.work);
                        this.wo.work.run();
                        if (BPELTrace.isBFDebugEnabled())
                            BPELTrace.traceBF("Completed " + this.wo.work);
                        this.wo.wException = null;
                    } catch (WorkException wex) {
                        this.wo.wException = wex;
                    } catch (Throwable th) {
                        // Need to report issue and wrap with WorkException
                        WorkCompletedException wex = new WorkCompletedException(
                                "Error from work object", th);
                        this.wo.wException = wex;
                        if (BPELTrace.isBFDebugEnabled())
                            BPELTrace.traceBF("" + wex);
                    } finally {
                        if (this.wo.eCtx != null && this.wo.eCtx.getXid() != null) {
                            wrkXids.remove(this.wo.eCtx.getXid());
                        }
                        if (this.wo.wListen != null && !this.wo.rejected) {
                            WorkEvent we = new WorkEvent(this.wo.work, WorkEvent.WORK_COMPLETED,
                                    this.wo.work, this.wo.wException, this.wo.startTime);
                            this.wo.wListen.workCompleted(we);
                        }
                        if (this.wo.rState == WorkEvent.WORK_COMPLETED) {
                            synchronized (this.wo.lock) {
                                this.wo.lock.notify();  // If waiting for execution completion
                            }
                        }
                        // Make sure that our ThreadLocal storage from JOTM is cleared
                        // Another thread could have completed the transaction for us.
//                      ((Current) tm).clearThreadTx();
                        this.wo.work = null;
                    }
                }
            }
        }

        private void checkTime() throws WorkException{

            long curTime = System.currentTimeMillis();

            if (curTime > this.wo.timeout) {
                WorkException wex = new WorkException(
                        "Error starting work in specified time", WorkException.START_TIMED_OUT);
                if (this.wo.wListen != null) {
                    WorkEvent we = new WorkEvent(this.wo.work, WorkEvent.WORK_REJECTED,
                            this.wo.work, wex);
                    this.wo.wListen.workRejected(we);
                }
                this.wo.rejected = true;
                if (BPELTrace.isBFDebugEnabled())
                    BPELTrace.traceBF("" + wex);
                throw wex;
            }
            this.wo.startTime = curTime;
        }
    }


    // ResourceWorkManager Monitor thread to check for timeouts of work objects
    private class BFManagerMonitor extends Thread {

        public BFManagerMonitor() {
            super("BFWorkManagerMonitor");
            setDaemon(true);
        }

        public void run() {
            int  cnt = 0;
            long curTime;
            while (true) {
                try {
                    sleep(100); // 100 ms
                    cnt++;
                    if (!BFWorkManager.this.workObjs.isEmpty()) {
                        for (int i = 0; i < BFWorkManager.this.workObjs.size(); i++) {
                            WorkObject wo = (WorkObject) BFWorkManager.this.workObjs.get(i);
                            curTime = System.currentTimeMillis();
                            if (curTime > wo.timeout) {
                                synchronized (BFWorkManager.this.workObjs) {
                                    if (BFWorkManager.this.workObjs.remove(wo)) {
                                        WorkException wex = new WorkException(
                                                "Error starting work in specified time",
                                                WorkException.START_TIMED_OUT);
                                        if (wo.wListen != null) {
                                            WorkEvent we = new WorkEvent(wo.work, WorkEvent.WORK_REJECTED,
                                                    wo.work, wex);
                                            wo.wListen.workRejected(we);
                                        }
                                        wo.wException = wex;
                                        throw wex;
                                    }
                                }
                            }
                        }
                    }

                    // Check threads for execution timeout or thread issue every second
                    if (cnt >= 10) {
                        cnt = 0;
                        curTime = System.currentTimeMillis();
                        for (int i = 0; i < BFWorkManager.this.vThreads.size(); i++) {
                            WorkerPool wp = (WorkerPool) BFWorkManager.this.vThreads.get(i);
                            if (!wp.isAlive() && !wp.shutdown) {
                                BFWorkManager.this.vThreads.remove(i);
                                if (BFWorkManager.this.vThreads.size() < BFWorkManager.this.minThreads) {
                                    WorkerPool nwp = new WorkerPool();
                                    BFWorkManager.this.vThreads.add(nwp);
                                    nwp.start();
                                }
                            } else if (wp.wo != null && wp.wo.executionTimeout > 0
                                    && curTime > wp.wo.executionTimeout) {
                                if (wp.wo.work != null) {
                                    wp.wo.work.release();
                                }
                            }
                        }
                    }
                } catch (Exception ex) {
                    if (BPELTrace.isBFDebugEnabled())
                        BPELTrace.traceBF(ex.toString());
                }
            }
        }
    }

}
