/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.objectweb.orchestra.bpem.controller;

import graph.GraphConstants;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.Collection;
import java.util.Iterator;

import javax.security.auth.login.LoginException;
import javax.swing.ButtonGroup;
import javax.swing.JButton;
import javax.swing.JCheckBoxMenuItem;
import javax.swing.JComboBox;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JPanel;
import javax.swing.JRadioButtonMenuItem;
import javax.swing.JScrollPane;
import javax.swing.JSlider;
import javax.swing.JSplitPane;
import javax.swing.JTabbedPane;
import javax.swing.SpringLayout;
import javax.swing.Timer;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import org.objectweb.orchestra.bpem.model.ExecutionModel;
import org.objectweb.orchestra.bpem.model.ModelListener;
import org.objectweb.orchestra.bpem.model.Options;
import org.objectweb.orchestra.bpem.view.BPIListInfo;
import org.objectweb.orchestra.bpem.view.ViewAsGraphic;
import org.objectweb.orchestra.bpem.view.ViewAsProperties;
import org.objectweb.orchestra.bpem.view.ViewAsText;
import org.objectweb.orchestra.bpem.view.ViewAsTree;
import org.objectweb.orchestra.bpem.view.ViewAsVariables;
import org.objectweb.orchestra.data.BPIData;
import org.objectweb.orchestra.exception.BPELBpemException;


/**
 * BPEMContent.java
 *
 * @author Bill Murray
 * @version $Id: BPEMContent.java 5 2006-10-20 14:37:44Z lejeuneg $
 *          <p>
 *          This class extends JPanel and is the main controller for the
 *          display.
 */
public class BPEMContent extends JPanel implements ActionListener, ChangeListener {

    private JComboBox idList = null;

    private final ExecutionModel model;

    private ViewAsText textView;

    private ViewAsTree treeView;

    private ViewAsGraphic graphicView;

    private ViewAsProperties propertiesView;

    private ViewAsVariables variablesView;

    private JScrollPane scrollText;

    private JScrollPane scrollTree;

    private JScrollPane scrollGraphic;

    private JScrollPane scrollProperties;

    private JScrollPane scrollVariables;

    private JTabbedPane tabPane;

    private JScrollPane currentScrollView;

    //    private JSplitPane splitLeftRight;
    //
    private JSplitPane splitTopBottom;

    private JRadioButtonMenuItem[] viewMenuItems;

    private JRadioButtonMenuItem[] refreshMenuItems;

    private JCheckBoxMenuItem displayNbrsMenuItem;

    private JCheckBoxMenuItem displayLoopCountMenuItem;

    private JRadioButtonMenuItem[] timeFormatMenuItems;

    private JButton refreshButton;

    private Timer refreshTimer;

    /**
     * BPEMContent - constructor - layout the main display
     *
     * @param int - the initial PBID from the command line
     * @throws LoginException
     */
    public BPEMContent(int initialBPID) throws BPELBpemException {
        super(new BorderLayout());

        // Access to the BPEL engine is indirect thru our
        // internal engine class. Create it and login.
        //
        this.model = new ExecutionModel(new Engine(null));
        try {
            this.model.getEngine().login();
        } catch (LoginException e) {
            System.out.println("Unable to login to BPEL...");
            e.printStackTrace();
        }
        //
        // Create top panel to hold menu and controls
        //
        SpringLayout layout = new SpringLayout();
        JPanel topPanel = new JPanel(layout);
        //
        // Create main menu
        //
        JMenuBar menu = createMenu();
        topPanel.add(menu);
        layout.putConstraint(SpringLayout.WEST, menu, 1, SpringLayout.WEST, topPanel);
        //
        // Create BPI combo list
        //
        this.idList = setupCombo();
        topPanel.add(this.idList);
        layout.putConstraint(SpringLayout.WEST, this.idList, 3, SpringLayout.EAST, menu);
        //
        // Create refresh button & timer
        //
        this.refreshButton = new JButton("Refresh");
        this.refreshButton
        .setToolTipText("Updates instance list and currently viewed process");
        this.refreshButton.addActionListener(this);
        topPanel.add(this.refreshButton);
        layout.putConstraint(SpringLayout.WEST, this.refreshButton, 1, SpringLayout.EAST, this.idList);
        this.refreshButton.setEnabled(false);
        this.refreshTimer = new Timer(0, this);
        //
        // Create the views
        //
        this.textView = new ViewAsText(this.model);
        this.treeView = new ViewAsTree(this.model);
        this.graphicView = new ViewAsGraphic(this.model);
        this.propertiesView = new ViewAsProperties(this.model);
        this.variablesView = new ViewAsVariables(this.model);
        //
        // Get the graphic zoom slider and add to our toolbar
        //
        JSlider zoom = this.graphicView.createSlider();
        topPanel.add(zoom);
        layout.putConstraint(SpringLayout.WEST, zoom, 1, SpringLayout.EAST, this.refreshButton);
        //
        // Finish Spring Layout for top panel
        //
        layout.putConstraint(SpringLayout.NORTH, topPanel, 1, SpringLayout.NORTH, zoom);
        layout.putConstraint(SpringLayout.EAST, topPanel, 1, SpringLayout.EAST, zoom);
        layout.putConstraint(SpringLayout.SOUTH, topPanel, 1, SpringLayout.SOUTH, zoom);
        topPanel.doLayout();
        //
        // Add views to scroll panes
        //
        this.scrollText = new JScrollPane(this.textView);
        this.scrollText.setPreferredSize(new Dimension(750, 450));
        this.scrollTree = new JScrollPane(this.treeView);
        this.scrollTree.setPreferredSize(new Dimension(750, 450));
        this.scrollGraphic = new JScrollPane(this.graphicView);
        this.scrollGraphic.setPreferredSize(new Dimension(750, 450));
        this.scrollProperties = new JScrollPane(this.propertiesView);
        this.scrollProperties.setPreferredSize(new Dimension(750, 100));
        this.scrollVariables = new JScrollPane(this.variablesView);
        this.scrollVariables.setPreferredSize(new Dimension(750, 100));
        //
        // Create tab pane and add views
        //
        this.tabPane = new JTabbedPane();
        this.tabPane.addTab("Variables", this.scrollVariables);
        this.tabPane.addTab("Properties", this.scrollProperties);
        this.tabPane.addChangeListener(this);
        //
        // Create splitter panes
        //
        //        splitLeftRight = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT,
        //                scrollGraphic, scrollTree);
        this.currentScrollView = this.scrollGraphic;
        this.splitTopBottom = new JSplitPane(JSplitPane.VERTICAL_SPLIT,
                this.currentScrollView, this.tabPane);
        this.splitTopBottom.setDividerSize(3);
        //
        // Add it all to the content
        //
        add(topPanel, BorderLayout.NORTH);
        add(this.splitTopBottom, BorderLayout.CENTER);
        //        add(currentScrollView, BorderLayout.CENTER);
        //        add(scrollProperties, BorderLayout.SOUTH);

        //
        // If a instance ID was passed on the command line, then we
        // select it in the process combo list. This will trigger
        // the callbacks to each model listener to display it.
        //
        if (initialBPID != 0) {
            selectBPID(this.idList, initialBPID);
        }
        // Default is manual refresh
        //
        restartRefreshTimer(0 * 1000);
    }

    /**
     * setupCombo - Create and return a JComboBox that lists all current ID's
     * retrieved from the BPEL engine.
     *
     * @return JComboBox
     */
    private JComboBox setupCombo() throws BPELBpemException  {
        JComboBox combo = new JComboBox();
        refreshCombo(combo);
        return combo;
    }

    /**
     * refreshCombo - Gets a new set of process instances from the BPEL engine
     * and updates the list, preserving any currently selection. During the
     * update, we disable the action listener.
     *
     * @param comboList
     */
    private BPIListInfo refreshCombo(JComboBox comboList) throws BPELBpemException {
        BPIListInfo current = null;

        // Save currently selected entry
        // (If selected entry is item 0, skip it)
        //
        if (comboList.getSelectedIndex() > 0) {
            current = (BPIListInfo) comboList.getSelectedItem();
        }
        // We can not refresh the list if it is currently visible.
        //
        if (comboList.isPopupVisible()) {
            return current;
        }
        // Suspend events and clear list
        //
        comboList.removeActionListener(this);
        comboList.removeAllItems();
        //
        // Fill with new data from engine.
        //
        comboList.addItem("Select a process instance...");
        Collection running = this.model.getEngine().getIDs("running");
        if (running != null) {
            Iterator iBPI = running.iterator();
            while (iBPI.hasNext()) {
                comboList.addItem(new BPIListInfo((BPIData) iBPI.next()));
            }
        }
        Collection finished = this.model.getEngine().getIDs("finished");
        if (finished != null) {
            Iterator iBPI = finished.iterator();
            while (iBPI.hasNext()) {
                comboList.addItem(new BPIListInfo((BPIData) iBPI.next()));
            }
        }
        // Select the entry that was previously selected.
        // We can not really use the "setSelectedItem" in the
        // combo even if we put "equals" in BPIListInfo. The combo
        // display gets confused because the items really are not
        // the same. So we walk thru the combo list manually.
        //
        if (current != null) {
            if (selectBPID(comboList, current.getID()) == false) {
                current = null;
            }
        }
        comboList.addActionListener(this);
        return current;
    }

    /**
     * selectBPID - selects a BPEL proces ID by instance # in the comboList
     * provided. Returns true if the ID was found.
     *
     * Note that this function does NOT add or remove the combo action
     * listener. It can be called with the actionListener disabled or
     * enabled depending on if the selection event is to be generated.
     *
     * @param comboList
     * @param ID
     * @return boolean
     */
    private boolean selectBPID(JComboBox comboList, int ID) {
        if (comboList.getItemCount() > 1) {
            for (int i = 1; i < comboList.getItemCount(); i++) {
                if (((BPIListInfo) comboList.getItemAt(i)).getID() == ID) {
                    comboList.setSelectedIndex(i);
                    break;
                }
            }
        }
        // If we did not find the desired instance, then
        // it means that it has been removed from BPEL (unDeployed).
        //
        if (comboList.getSelectedIndex() <= 0) {
            comboList.setSelectedIndex(0);
            return false;
        }
        return true;
    }

    /**
     * actionPerformed - callback for menu and comboBox.
     */
    public void actionPerformed(ActionEvent e) {
        Object source = e.getSource();
        try {
            if (source == this.idList) {
                //
                // User selects process ID, load model
                //
                JComboBox cb = (JComboBox) source;
                if (cb.getSelectedIndex() == 0)
                    return;
                BPIListInfo listInfo = (BPIListInfo) cb.getSelectedItem();
                this.model.load(listInfo.getProcessInstance());
            } else if (source == this.viewMenuItems[0]) {
                //
                // View/Graphic menu
                //
                replaceMainView(this.scrollGraphic);
            } else if (source == this.viewMenuItems[1]) {
                //
                // View/Tree menu
                //
                replaceMainView(this.scrollTree);
            } else if (source == this.viewMenuItems[2]) {
                //
                // View/XML menu
                //
                replaceMainView(this.scrollText);
                this.textView.refresh();
            } else if (source == this.refreshButton) {
                //
                // Refresh button pressed
                //
                BPIListInfo prev = refreshCombo(this.idList);
                refreshModel(prev);
            } else if (source == this.refreshMenuItems[0]) {
                //
                // "Manual" refresh
                //
                restartRefreshTimer(0 * 1000);
            } else if (source == this.refreshMenuItems[1]) {
                //
                // 2.5 second refresh
                //
                restartRefreshTimer((int)(2.5 * 1000.0));
            } else if (source == this.refreshMenuItems[2]) {
                //
                // 5 second refresh
                //
                restartRefreshTimer(5 * 1000);
            } else if (source == this.refreshMenuItems[3]) {
                //
                // 10 second refresh
                //
                restartRefreshTimer(10 * 1000);
            } else if (source == this.refreshTimer) {
                //
                // Refresh timer
                //
                this.refreshTimer.stop();
                BPIListInfo prev = refreshCombo(this.idList);
                refreshModel(prev);
                this.refreshTimer.restart();
            } else if (source == this.displayNbrsMenuItem) {
                //
                // Display statement numbers
                // OptionListeners will update their displays
                //
                Options.setDisplayStmtNbrs(this.displayNbrsMenuItem.getState());
            } else if (source == this.displayLoopCountMenuItem) {
                //
                // Display loop count numbers
                // OptionListeners will update their displays
                //
                Options.setDisplayLoopCount(this.displayLoopCountMenuItem.getState());
            } else if (source == this.timeFormatMenuItems[0]) {
                //
                // Time format none
                // OptionListeners will update their displays
                //
                Options.setTimeFormat(Options.TIMEFORMAT_NONE);
            } else if (source == this.timeFormatMenuItems[1]) {
                //
                // Time format long
                // OptionListeners will update their displays
                //
                Options.setTimeFormat(Options.TIMEFORMAT_LONG);
            } else if (source == this.timeFormatMenuItems[2]) {
                //
                // Time format relative
                // OptionListeners will update their displays
                //
                Options.setTimeFormat(Options.TIMEFORMAT_RELATIVE);
            } else if (source == this.timeFormatMenuItems[3]) {
                //
                // Time format short
                // OptionListeners will update their displays
                //
                Options.setTimeFormat(Options.TIMEFORMAT_SHORT);
            }
        } catch (BPELBpemException ex) {
            ex.printStackTrace();
        }
        this.validate();
        this.repaint();
    }

    public void stateChanged(ChangeEvent e) {
        Object source = e.getSource();
        if (source == this.tabPane) {
            //ModelListener ml = (ModelListener)((JScrollPane)tabPane.getSelectedComponent()).getComponent(0);
            ModelListener ml = (ModelListener)((JScrollPane)this.tabPane.getSelectedComponent()).getViewport().getView();
            ml.refresh();
        }
    }

    private void refreshModel(BPIListInfo prevSelected) throws BPELBpemException {
        if (this.idList.getSelectedIndex() <= 0)
            return;
        BPIListInfo listInfo = (BPIListInfo) this.idList.getSelectedItem();
        int bpid = listInfo.getID();
        //
        // It "should" be the same ID as previously selected, but
        // check just to be sure
        //
        if (bpid == this.model.getBPID()) {
            //
            // If we previously saw the process as finished, then
            // there is no need to refresh it. We have to check the
            // state from the previous BPI to allow us to do a final
            // refresh for a process that just terminated.
            //
            if (prevSelected.isExecuting()) {
                this.model.refresh();
            }
        } else {
            this.model.load(listInfo.getProcessInstance());
        }
    }

    /**
     * restartRefreshTimer - Sets up the refresh start, manual or automatic.
     *
     * @param millis - Milliseconds between refreshes. If 0, then disable timer and
     *                 enable manuan refresh.
     */
    private void restartRefreshTimer(int millis) {
        this.refreshTimer.stop();
        if (millis == 0) {
            this.refreshButton.setEnabled(true);
        } else {
            this.refreshButton.setEnabled(false);
            this.refreshTimer.setDelay(millis);
            this.refreshTimer.start();
        }
    }

    private void replaceMainView(JScrollPane newView) {
        this.splitTopBottom.remove(this.currentScrollView);
        this.currentScrollView = newView;
        this.splitTopBottom.setTopComponent(this.currentScrollView);
    }

    /**
     * createMenu - create the top level menu bar.
     *
     * @return JMenuBar
     */
    private JMenuBar createMenu() {
        JMenuBar menuBar = new JMenuBar();
        menuBar.setMaximumSize(new Dimension(45, 25));
        menuBar.setMinimumSize(new Dimension(45, 25));
        menuBar.setPreferredSize(new Dimension(45, 25));
        JMenu optionsMenu = new JMenu("Options");
        optionsMenu.setFont(GraphConstants.GUIFont);
        menuBar.add(optionsMenu);

        // Create "View" menu items and make "Graphic" the default
        //
        this.viewMenuItems = createMenuRadioGroup(optionsMenu, new String[] {
                "Graphic", "Tree", "XML" });
        this.viewMenuItems[0].setSelected(true);
        //
        // Create "Refresh" menu items
        //
        optionsMenu.addSeparator();
        this.refreshMenuItems = createMenuRadioGroup(optionsMenu, new String[] {
                "Manual refresh", "2.5 second refresh", "5 second refresh",
        "10 second refresh" });
        this.refreshMenuItems[0].setSelected(true);
        //
        // Create "Display statement nbrs" menu
        //
        optionsMenu.addSeparator();
        this.displayNbrsMenuItem = new JCheckBoxMenuItem("Display statement #'s");
        this.displayNbrsMenuItem.setSelected(Options.getDisplayStmtNbrs());
        this.displayNbrsMenuItem.setFont(GraphConstants.GUIFont);
        this.displayNbrsMenuItem.addActionListener(this);
        optionsMenu.add(this.displayNbrsMenuItem);
        //
        // Create "Display loop count" menu
        //
        this.displayLoopCountMenuItem = new JCheckBoxMenuItem("Display loop counts");
        this.displayLoopCountMenuItem.setSelected(Options.getDisplayLoopCount());
        this.displayLoopCountMenuItem.setFont(GraphConstants.GUIFont);
        this.displayLoopCountMenuItem.addActionListener(this);
        optionsMenu.add(this.displayLoopCountMenuItem);
        //
        // Create "Time format" menu items
        //
        optionsMenu.addSeparator();
        this.timeFormatMenuItems = createMenuRadioGroup(optionsMenu, new String[] {
                "Time format: none", "Time format: full", "Time format: relative",
        "Time format: short" });
        this.timeFormatMenuItems[Options.getTimeFormat()].setSelected(true);

        return menuBar;
    }

    /**
     * createMenuRadioGroup - create a set of related menu items with the
     * display text specified.
     *
     * @param menu -
     *            the menu the new items are to be added to.
     * @param aMenuText -
     *            String array containing the desired menu text.
     * @return - JRadioButtonMenuItem[]
     */
    private JRadioButtonMenuItem[] createMenuRadioGroup(JMenu menu,
            String[] aMenuText) {
        JRadioButtonMenuItem[] items = new JRadioButtonMenuItem[aMenuText.length];

        ButtonGroup layoutGroup = new ButtonGroup();
        for (int i = 0; i < aMenuText.length; i++) {
            items[i] = new JRadioButtonMenuItem(aMenuText[i]);
            items[i].setFont(GraphConstants.GUIFont);
            items[i].addActionListener(this);
            layoutGroup.add(items[i]);
            menu.add(items[i]);
        }
        return items;
    }
}