/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.objectweb.orchestra.bpem.controller;

import java.rmi.RemoteException;
import java.util.Collection;

import javax.ejb.RemoveException;
import javax.naming.NamingException;
import javax.security.auth.login.LoginContext;
import javax.security.auth.login.LoginException;

import org.objectweb.orchestra.data.BPIData;
import org.objectweb.orchestra.exception.BPELAdminException;
import org.objectweb.orchestra.exception.BPELBpemException;
import org.objectweb.orchestra.exception.BPELMonitoringException;
import org.objectweb.orchestra.exception.BPELMonitoringObjectNotFoundException;
import org.objectweb.orchestra.monitoring.Variable;
import org.objectweb.orchestra.util.SimpleCallbackHandler;
import org.w3c.dom.Document;

import org.objectweb.orchestra.admin.BpelAdminSession;
import org.objectweb.orchestra.admin.BpelAdminSessionHome;
import org.objectweb.orchestra.admin.BpelAdminSessionUtil;
import org.objectweb.orchestra.monitoring.BpelMonitoringSession;
import org.objectweb.orchestra.monitoring.BpelMonitoringSessionHome;
import org.objectweb.orchestra.monitoring.BpelMonitoringSessionUtil;

/**
 * Engine.java.
 *
 * @author Bill Murray
 * @version $Id: Engine.java 74 2006-12-06 12:16:19Z souillac $
 *          <p>
 *          This class provides all in interraction with the actual BPEL engine.
 */
public class Engine {

    private static BPIData pi;

    /**
     * Engine - constructor.
     */
    public Engine(BPIData processInstance) {
        pi = processInstance;
    }

    /**
     * getProcessInstance - returns the current processInstance.
     *
     * @return
     */
    public static BPIData getProcessInstance() {
        return pi;
    }

    /**
     * getID - gets the currently selected instance ID.
     *
     * @return - int
     */
    public int getID() {
        if (pi == null) {
            return 0;
        }
        return pi.getBPIId();
    }

    /**
     * login - login into the BPEL with the default user/pwd.
     *
     * @throws LoginException
     */
    public void login() throws LoginException {
        char[] password = {'b', 'p', 'e', 'l', 'a', 'd', 'm', 'i', 'n'};
        SimpleCallbackHandler handler = new SimpleCallbackHandler("bpeladmin", password);
        LoginContext lc = new LoginContext("BPEL", handler);
        lc.login();
    }

    /**
     * getIDs - returns a collection of all known BPI from the engine.
     *
     * @return Collection - collection of ProcessInstance
     */
    public Collection getIDs(String type) throws BPELBpemException {
        Collection allInstances = null;

        BpelAdminSessionHome adminOpHome = null;
        BpelAdminSession adminOp = null;
        try {
            adminOpHome = BpelAdminSessionUtil.getHome();
            adminOp = adminOpHome.create();
            while (allInstances == null) {
                try {
                    if (type.equals("running")) {
                        allInstances = adminOp.getRunningInstances();
                    } else if (type.equals("finished")) {
                        allInstances = adminOp.getFinishedInstances();
                    }
                } catch (javax.ejb.NoSuchObjectLocalException e) {
                    //NOTHING : we were trying to access a bean which was removed
                }
            }
            adminOp.remove();
        } catch (javax.ejb.CreateException e) {
            throw new BPELBpemException("CreateException in controller.Engine getIDs :\n", e);
        } catch (NamingException e) {
            throw new BPELBpemException("NamingException in controller.Engine getIDs :\n", e);
        } catch (RemoteException e) {
            throw new BPELBpemException("RemoteException in controller.Engine getIDs :\n", e);
        } catch (BPELAdminException e) {
            throw new BPELBpemException("BPELAdminException in controller.Engine getIDs :\n", e);
        } catch (RemoveException e) {
            throw new BPELBpemException("RemoveException in controller.Engine getIDs :\n", e);
        }

        return allInstances;
    }

    /**
     * getExecutionTree - returns the DOM for the current process instance.
     *
     * @return Document
     */
    public Document getExecutionTree() throws BPELBpemException  {
        return getExecutionTree(pi, 0);
    }

    /**
     * refreshExecutionTree - We give the BPEL engine the timestamp
     * from the DOM of our previous call. If the engine has new data
     * for the process, it is returned, otherwise, null is returned.
     *
     * @param lastMaxTime
     * @return Document
     */
    public Document refreshExecutionTree(long lastStmtTime) throws BPELBpemException {
        return getExecutionTree(pi, lastStmtTime);
    }

    /**
     * getExecutionTree - returns the DOM for the specified process instance.
     *
     * @param bpiId
     * @return Document
     */
    public Document getExecutionTree(BPIData processInstance, long lastStmtTime) throws BPELBpemException {
        Document result = null;
        BpelMonitoringSessionHome mHome = null;
        BpelMonitoringSession monitor = null;
        try {
            mHome = BpelMonitoringSessionUtil.getHome();
            monitor = mHome.create();
            if (lastStmtTime == 0) {
                result = monitor.getStatementsTree(processInstance.getBPIId());
            } else {
                result = monitor.getStatementsTree(processInstance.getBPIId(), lastStmtTime);
            }
            monitor.remove();
        } catch (javax.ejb.CreateException e) {
            throw new BPELBpemException("CreateException in controller.Engine getExecutionTree :\n", e);
        } catch (NamingException e) {
            throw new BPELBpemException("NamingException in controller.Engine getExecutionTree :\n", e);
        } catch (RemoteException e) {
            throw new BPELBpemException("RemoteException in controller.Engine getExecutionTree :\n", e);
        } catch (RemoveException e) {
            throw new BPELBpemException("RemoveException in controller.Engine getExecutionTree :\n", e);
        } catch (BPELMonitoringObjectNotFoundException e) {
            //NOTHING
            //it means the statement tree we were trying to get was called on
            //a running instance and this instance is finished
            //in this case, the monitoring was not set to all
        } catch (BPELMonitoringException e) {
            throw new BPELBpemException("BPELMonitoringException in controller.Engine getExecutionTree :\n", e);
        }
        pi = processInstance;
        return result;
    }

    /**
     * getVariable - returns the variable value for the current process
     * instance.
     *
     * @param name -
     *            the name of the variable
     * @param statementNb -
     *            the reference statement number
     * @return Variable - the variable value
     */
    public Variable getVariable(String name, int statementNb) throws BPELBpemException {
        return getVariable(getID(), name, statementNb);
    }

    /**
     * getVariable - returns the variable value for the specified process
     * instance.
     *
     * @param name -
     *            the name of the variable
     * @param statementNb -
     *            the reference statement number
     * @return Variable - the variable value
     */
    public Variable getVariable(int bpiId, String name, int statementNb) throws BPELBpemException {
        Variable result = null;
        BpelMonitoringSessionHome mHome = null;
        BpelMonitoringSession monitor = null;
        try {
            mHome = BpelMonitoringSessionUtil.getHome();
            monitor = mHome.create();
            monitor.getProcessName(bpiId);
            result = monitor.getVar(bpiId, name, statementNb);
            monitor.remove();
        } catch (javax.ejb.CreateException e) {
            throw new BPELBpemException("CreateException in controller.Engine getVariable :\n", e);
        } catch (NamingException e) {
            throw new BPELBpemException("NamingException in controller.Engine getVariable :\n", e);
        } catch (RemoteException e) {
            throw new BPELBpemException("RemoteException in controller.Engine getVariable :\n", e);
        } catch (RemoveException e) {
            throw new BPELBpemException("RemoveException in controller.Engine getVariable :\n", e);
        } catch (BPELMonitoringObjectNotFoundException e) {
            //NOTHING
            //it means the variable we were trying to get was called on
            //a running instance and this instance is finished
            //in this case, the monitoring was not set to all
        } catch (BPELMonitoringException e) {
            throw new BPELBpemException("BPELMonitoringException in controller.Engine getVariable :\n", e);
        }
        return result;
    }

    /**
     * isIDValid - determin if the ID property is valid.
     *
     * @return boolean
     */
    public boolean isIDValid() {
        try {
            return isIDValid(getID());
        } catch (NumberFormatException ex) {
            // Nothing
        }
        return false;
    }

    public boolean isIDValid(int id) {
        return (id != 0);
    }
}
