/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.objectweb.orchestra.bpem.model;

import java.util.ArrayList;

import org.objectweb.orchestra.bpem.controller.BPEMUtils;
import org.objectweb.orchestra.bpem.controller.Engine;
import org.objectweb.orchestra.data.BPIData;
import org.objectweb.orchestra.exception.BPELBpemException;
import org.objectweb.orchestra.util.BPELConstants;
import org.w3c.dom.Document;



/**
 * ExecutionModel.java
 *
 * @author Bill Murray
 * @version $Id: ExecutionModel.java 80 2006-12-06 21:24:57Z bmurray $
 *          <p>
 *          This class holds the execution data model for the current process
 *          instance. The model consists of nested "ModelNodes's" each of which
 *          contains a reference to the cooresponding domNode.
 *
 * Here we also maintain a list of "ModelListeners". These are objects that want
 * to be informed when the data model changes.
 */
public class ExecutionModel {

    private final Engine engine;

    private Document dom;

    private ModelNode mnRoot = null;

    private ArrayList listeners = new ArrayList();

    public ExecutionModel(Engine engine) {
        this.engine = engine;
    }

    /**
     * getBPID - gets the currently loaded process ID.
     *
     * @return int
     */
    public int getBPID() {
        return this.engine.getID();
    }

    /**
     * getProcessInstance - returns the current processInstance
     *
     * @return
     */
    public BPIData getProcessInstance() {
        return Engine.getProcessInstance();
    }

    /**
     * getEngine
     *
     * @return Returns the engine. This is an abstraction for the BPEL engine.
     */
    public Engine getEngine() {
        return this.engine;
    }

    /**
     * getDom
     *
     * @return Returns the dom.
     */
    public Document getDom() {
        return this.dom;
    }

    /**
     * getModelRoot
     *
     * @return ModelNode cooresponding to the root of the data model.
     */
    public ModelNode getModelRoot() {
        return this.mnRoot;
    }

    /**
     * getLastStmtTime - returns the time of the last executed statement
     * in the model or 0 if there is no model.
     *
     * @return long
     */
    public long getLastStmtTime() {
        if (this.mnRoot == null) return 0;
        String time = this.mnRoot.getAttrValue(BPELConstants.MONITORING_MAXTIMESTAMP);
        if (time == null) return 0;
        try {
            return Long.parseLong(time);
        } catch (NumberFormatException ex) {
            return 0;
        }
    }

    /**
     * load - loads the process instance previously specified in the engine.
     *
     * @return boolean
     */
    public boolean load() throws BPELBpemException {
        this.dom = this.engine.getExecutionTree();
        if (this.dom == null) {
            //it means the tree we were trying to get was called on
            //a running instance and this instance is finished
            //in this case, the monitoring was not set to all
            //We will display an error message
            String errMsg = "The requested instance "+this.engine.getID() + "\n";
            errMsg += "is not currently running and the monitoring mode does not allow to see finished instances.\n";
            errMsg += "Please, perform a refresh.";
            BPEMUtils.showError("", errMsg);
            return false;
        }
        createModel();
        return true;
    }

    /**
     * load - load the data model for the specified process instance
     *
     * @param bpid
     * @return boolean
     */
    public boolean load(BPIData pi) throws BPELBpemException {
        this.dom = this.engine.getExecutionTree(pi, 0);
        if (this.dom == null) {
            //it means the tree we were trying to get was called on
            //a running instance and this instance is finished
            //in this case, the monitoring was not set to all
            //We will display an error message
            String errMsg = "The requested instance "+this.engine.getID() + "\n";
            errMsg += "is not currently running and the monitoring mode does not allow to see finished instances.\n";
            errMsg += "Please, perform a refresh.";
            BPEMUtils.showError("", errMsg);
            return false;
        }
        createModel();
        displayAllListeners();
        return true;
    }

    /**
     * displayAllListeners - For each registered ModelListener, call it to
     * inform it that data has changed.
     */
    public void displayAllListeners() {
        for (int i = 0; i < this.listeners.size(); i++) {
            ((ModelListener) this.listeners.get(i)).clear();
            ((ModelListener) this.listeners.get(i)).displayAll();
        }
    }

    /**
     * refresh - get updated info for the current instance from the
     * BPEL engine. If there are registered listeners, then they
     * get a "refresh" callback.
     *
     * @return boolean - true if new data was found.
     */
    public boolean refresh() throws BPELBpemException {
        Document newDom = this.engine.refreshExecutionTree(getLastStmtTime());
        if (newDom == null) {
            //it means the tree we were trying to get was called on
            //a running instance and this instance is finished
            //in this case, the monitoring was not set to all
            //We will display an error message

            //in another case, it means the execution tree has not changed since the last request
            String errMsg = "The requested instance "+this.engine.getID() + "\n";
            errMsg = "  - either is not currently running and the monitoring mode does not allow to see finished instances.\n";
            errMsg = "  - either does not have any change since the last request\n";
            errMsg = "=> Please, perform a refresh.";
            BPEMUtils.showError("Unable to retrieve instance", errMsg);
            return false;
        }
        this.dom = newDom;
        refreshModel(newDom);
        refreshAllListeners();
        return true;
    }

    /**
     * refreshAllListeners - For each registered ModelListener, call it to
     * inform it that data has changed.
     */
    public void refreshAllListeners() {
        for (int i = 0; i < this.listeners.size(); i++) {
            ((ModelListener) this.listeners.get(i)).refresh();
        }
    }

    /**
     * selectModelNode - For each registered ModelListener, call it to say that
     * the specified node has been selected.
     *
     * @param mn -
     *            the node that has been selected.
     */
    public void selectModelNode(ModelNode mn) {
        selectModelNode(mn, -1);
    }

    /**
     * selectModelNode - For each registered ModelListener EXCEPT the one
     * specified to skip, call it to say that the specified node has been
     * selected. The "skip" listener allows the caller to say that they
     * do not want the callback themselves.
     *
     * @param mn
     * @param skipListenerIndex
     */
    public void selectModelNode(ModelNode mn, int skipListenerIndex) {
        for (int i = 0; i < this.listeners.size(); i++) {
            if (i != skipListenerIndex) {
                ((ModelListener) this.listeners.get(i)).nodeSelected(mn);
            }
        }
    }


    /**
     * addListener - adds the specified ModelListener to the list and
     * returns the listener "ID". The can be used on subsequent calls.
     *
     * @param listener
     * @return int - the "ID" (index) assigned to this listener
     */
    public int addListener(ModelListener listener) {
        this.listeners.add(listener);
        return this.listeners.indexOf(listener);
    }

    /**
     * createModel - creates the data model from the DOM.
     */
    private void createModel() {
        this.mnRoot = ModelNodeFactory.createModelNode(this.dom.getFirstChild(), null);
    }

    /**
     * refreshModel - updates the current model with data from the
     * new model. Note that the current model object remain in place.
     * Only their internal attributes are updated.
     *
     * @param newDom
     */
    private void refreshModel(Document newDom) {
        if (this.mnRoot == null) return;
        ModelNode newMNRoot = ModelNodeFactory.createModelNode(newDom.getFirstChild(), null);
        if (newMNRoot == null) return;
        this.mnRoot.refresh(newMNRoot);
    }
}