/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.objectweb.orchestra.bpem.model;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;

import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;

/**
 * ModelNodeFactory.java
 *
 * @author Bill Murray
 * @version $Id: ModelNodeFactory.java 5 2006-10-20 14:37:44Z lejeuneg $
 *          <p>
 *          This class builds ModelNode derived objects based on the node type
 *          of the BPEL statement.
 */
public class ModelNodeFactory {

    private static String classNamePrefix = ModelNodeFactory.class.getName().substring(0, ModelNodeFactory.class.getName().lastIndexOf('.'));

    /**
     * createModelNode - Creates the specific ModelNode
     * @param node
     * @param mnParent
     * @return - ModelNode derivedclass
     */
    public static ModelNode createModelNode(Node node, ModelNode mnParent) {
        Class mnClass = getClass(node);
        Constructor mnConstructor;
        //
        // Get the constructor for the ModelNode derived class
        //
        try {
            mnConstructor = mnClass.getConstructor(new Class[] {Node.class , ModelNode.class});
        } catch (SecurityException e) {
            return new ModelNodeDefault(node, mnParent);
        } catch (NoSuchMethodException e) {
            return new ModelNodeDefault(node, mnParent);
        }
        // Create an instance of the class
        //
        try {
            return (ModelNode)mnConstructor.newInstance(new Object[] {node, mnParent});
        } catch (IllegalArgumentException e1) {
            return new ModelNodeDefault(node, mnParent);
        } catch (InstantiationException e1) {
            return new ModelNodeDefault(node, mnParent);
        } catch (IllegalAccessException e1) {
            return new ModelNodeDefault(node, mnParent);
        } catch (InvocationTargetException e1) {
            return new ModelNodeDefault(node, mnParent);
        }
    }

    /**
     * getClass - gets the class to be used for the node. The class name is
     * built as "org.objectweb.orchestra.bpem.model.ModelNode" + "typevalue".
     * If the class can not be found, then the default ModelNode class is
     * used.
     *
     * @param node
     * @return - class
     */
    private static Class getClass(Node node) {
        String className = getNodeType(node).toLowerCase();
        className = className.substring(0, 1).toUpperCase() + className.substring(1);
        String fullClassName = classNamePrefix + ".ModelNode" + className;
        try {
            return Class.forName(fullClassName);
        } catch (ClassNotFoundException e) {
            return ModelNodeDefault.class;
        }
    }

    /**
     * getNodeType - gets the value of the "type" attribute of this node
     *
     * @param node
     * @return - String - either the value or "Default"
     */
    private static String getNodeType(Node node) {
        String type = "Default";
        NamedNodeMap attrs = node.getAttributes();
        Node n = attrs.getNamedItem("type");
        if (n != null) {
            type = n.getNodeValue();
        }
        return type;
    }
}
