/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.objectweb.orchestra.bpem.view;

import graph.BusMultiEdgeDisplay;
import graph.GraphMultiEdge;

import java.awt.geom.Line2D;
import java.awt.geom.Rectangle2D;
import java.util.Vector;

import org.objectweb.orchestra.bpem.model.ModelEdge;



/**
 * BPEMGraphitBusMultiEdgeDisplay.java
 *
 * @author Bill Murray
 * @version $Id: BPEMGraphitBusMultiEdgeDisplay.java,v 1.1 2005/03/14 21:31:21
 *          bmurray Exp $
 *          <p>
 *          This class extends BusMultiEdgeDisplay and is the display class for
 *          bus edges. The main reason we have to extend the default
 *          "BusMutliEdgeDisplay" class is to be able to possition the "bus" bar
 *          at the appropriate location.
 */
public class BPEMGraphitBusMultiEdgeDisplay extends BusMultiEdgeDisplay {

    /**
     * @param graphMultiEdge
     */
    public BPEMGraphitBusMultiEdgeDisplay(GraphMultiEdge graphMultiEdge) {
        super(graphMultiEdge);
    }

    /**
     * Compute the curve according to connected ports and control points.
     */
    public void updateCurve() {
        // store the coordinates of ports
        processPorts();
        this.vCurve = new Vector();
        this.vShape = new Vector();
        if (this.ports.length < 2)
            return;

        // compute the coordinates of the center if not fixed
        if (this.pinned) {
            // let pinned value unchanged
            if ((this.orientation == NORTH) || (this.orientation == SOUTH)) {
                // X is static
                this.y = (int) averageY();
            } else {
                this.x = (int) averageX();
                // Y is static
            }
        } else {
            // set the center at barycentre of ports
            this.x = (int) averageX();
            this.y = (int) determineBarY();
        }

        if ((this.orientation == NORTH) || (this.orientation == SOUTH)) {
            // VERTICAL bus
            this.busStartX = this.x;
            this.busStartY = minY() - OverEdge;
            this.busEndX = this.x;
            this.busEndY = maxY() + OverEdge;

            // compute the lines representing the edge
            // the bus itself
            this.vCurve.add(new Rectangle2D.Double(this.busStartX, this.busStartY, 2, this.busEndY
                    - this.busStartY));
            // the edges to bus
            for (int i = 0; i < this.ports.length; i++) {
                this.vCurve.add(new Line2D.Double(this.ports[i].x, this.ports[i].y, this.busStartX,
                        this.ports[i].y));
            }

            // create the Vector of forms forming the shape
            // the bus itsef
            this.vShape.add(new Rectangle2D.Double(this.busStartX - ShapeMargin,
                    this.busStartY, 2 * ShapeMargin, this.busEndY - this.busStartY));
            // the edges to bus
            for (int i = 0; i < this.ports.length; i++) {
                if (this.ports[i].x < this.busStartX) {
                    this.vShape.add(new Rectangle2D.Double(this.ports[i].x, this.ports[i].y
                            - ShapeMargin, this.busStartX - this.ports[i].x,
                            2 * ShapeMargin));
                } else {
                    this.vShape.add(new Rectangle2D.Double(this.busStartX, this.ports[i].y
                            - ShapeMargin, this.ports[i].x - this.busStartX,
                            2 * ShapeMargin));
                }
            }
        } else {
            // HORIZONTAL bus
            this.busStartX = minX() - OverEdge;
            this.busStartY = this.y;
            this.busEndX = maxX() + OverEdge;
            this.busEndY = this.y;

            // compute the lines representing the edge
            // the bus itsef
            this.vCurve.add(new Rectangle2D.Double(this.busStartX, this.busStartY, this.busEndX
                    - this.busStartX, 2));
            // the edges to bus
            for (int i = 0; i < this.ports.length; i++) {
                this.vCurve.add(new Line2D.Double(this.ports[i].x, this.busStartY, this.ports[i].x,
                        this.ports[i].y));
            }

            // create the Vector of rectangles forming the shape
            // the bus itsef
            this.vShape.add(new Rectangle2D.Double(this.busStartX, this.busStartY
                    - ShapeMargin, this.busEndX - this.busStartX, 2 * ShapeMargin));
            // the edges to bus
            for (int i = 0; i < this.ports.length; i++) {
                if (this.ports[i].y < this.busStartY) {
                    this.vShape
                    .add(new Rectangle2D.Double(this.ports[i].x
                            - ShapeMargin, this.ports[i].y, 2 * ShapeMargin,
                            this.busStartY - this.ports[i].y));
                } else {
                    this.vShape
                    .add(new Rectangle2D.Double(this.ports[i].x
                            - ShapeMargin, this.busStartY, 2 * ShapeMargin,
                            this.ports[i].y - this.busStartY));
                }
            }
        }

        // update xMax and yMax
        this.xMax = Math.max((int) this.busEndX, (int) maxX());
        this.yMax = Math.max((int) this.busEndY, (int) maxY());

        // set core at center of the connected nodes
        this.coreDX = this.x - this.coreW / 2;
        this.coreDY = this.y - this.coreH / 2;
        this.xMax = Math.max(this.xMax, this.coreDX + this.coreW);
        this.yMax = Math.max(this.yMax, this.coreDY + this.coreH);
    }

    private double determineBarY() {
        if (this.ports.length == 0)
            return (0);
        if (multipleDestPorts()) {
            return averageYOfTop2Ports() + 3;
        }
        return averageYOfBottom2Ports() - 7;
    }

    private boolean multipleDestPorts() {
        int countDestPorts = 0;

        for (int i = 0; i < this.graphMultiEdge.nbOfEdges(); i++) {
            ModelEdge edge = (ModelEdge) this.graphMultiEdge.getEdge(i)
            .getUserObject();
            if (edge.getDestPort().isInbound()) {
                countDestPorts++;
                if (countDestPorts > 1)
                    return true;
            }
        }
        return false;
    }

    private double averageYOfTop2Ports() {
        double y1 = 0;
        double y2 = 0;
        //
        // Get the "top" 2 port y values
        //
        for (int i = 0; i < this.ports.length; i++) {
            double y = this.ports[i].y;
            //
            // If < both value, replace the highest
            //
            if ((y < y1) && (y < y2)) {
                if (y1 > y2) {
                    y1 = y;
                } else {
                    y2 = y;
                }
            } else {
                //
                // If < only 1 value, replace that value
                //
                if ((y < y1) || (y1 == 0)) {
                    y1 = y;
                } else if ((y < y2) || (y2 == 0)) {
                    y2 = y;
                }
            }
        }
        return (y1 + y2) / 2.0;
    }

    private double averageYOfBottom2Ports() {
        double y1 = 0;
        double y2 = 0;
        //
        // Get the "bottom" 2 port y values
        //
        for (int i = 0; i < this.ports.length; i++) {
            double y = this.ports[i].y;
            //
            // If > both value, replace the lowest
            //
            if ((y > y1) && (y > y2)) {
                if (y1 < y2) {
                    y1 = y;
                } else {
                    y2 = y;
                }
            } else {
                //
                // If > only 1 value, replace that value
                //
                if ((y > y1) || (y1 == 0)) {
                    y1 = y;
                } else if ((y > y2) || (y2 == 0)) {
                    y2 = y;
                }
            }
        }
        return (y1 + y2) / 2.0;
    }
}