/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.objectweb.orchestra.bpem.view;

import java.util.Enumeration;
import java.util.Hashtable;

import javax.swing.JTree;
import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.TreePath;

/**
 * TVExpansionState.java
 *
 * @author Bill Murray
 * @version $Id: TVExpansionState.java 5 2006-10-20 14:37:44Z lejeuneg $
 *
 * This class allows to save the current state of expansion in a tree view and
 * subsequently restore that state. This allows updating the tree view and then
 * making it look like it did previously.
 */

public class TVExpansionState {

    private final JTree tree;

    private Hashtable expandedPaths = new Hashtable();

    /**
     * TVExpansionState - constructor - Saves the expansion state for the
     * specified tree.
     *
     * @param tree -
     *            The tree whose state is to be saved and restored.
     */
    public TVExpansionState(JTree tree) {
        this.tree = tree;

        DefaultMutableTreeNode rootNode = (DefaultMutableTreeNode) tree
        .getModel().getRoot();
        TreePath rootPath = new TreePath(rootNode.getPath());
        saveState(rootPath);
    }

    /**
     * saveState - Recursive routine to save the expansion state of the current
     * path and then its children.
     *
     * @param path
     */
    private void saveState(TreePath path) {
        Enumeration expandedChildren = this.tree.getExpandedDescendants(path);
        if (expandedChildren != null) {
            //
            // The current path has expanded children, so it itself
            // must be expanded.
            //
            String key = keyForPath(path);
            this.expandedPaths.put(key, path);
            //
            // See if child nodes are expanded
            //
            while (expandedChildren.hasMoreElements()) {
                TreePath nextElement = (TreePath) expandedChildren
                .nextElement();
                saveState(nextElement);
            }
        }
    }

    /**
     * restoreState - restore the previously saved expansion state.
     */
    public void restoreState() {
        DefaultMutableTreeNode rootNode = (DefaultMutableTreeNode) this.tree
        .getModel().getRoot();
        restoreRecursive(rootNode);
    }

    /**
     * restoreRecursive - restore the previously saved expansion state.
     *
     * @param node
     */
    private void restoreRecursive(DefaultMutableTreeNode node) {
        // Walk down to lowest nodes
        //
        for (int i = 0; i < node.getChildCount(); i++) {
            restoreRecursive((DefaultMutableTreeNode) node.getChildAt(i));
        }
        // If this node was previously expanded, then do so now
        //
        TreePath current = new TreePath(node.getPath());
        String key = keyForPath(current);
        if (this.expandedPaths.get(key) != null) {
            this.tree.expandPath(current);
        }
    }

    /**
     * keyForPath - Build a string suitable for a hastable "key". We just walk
     * the path and build a key using "." seperators.
     *
     * @param path
     * @return String
     */
    private String keyForPath(TreePath path) {
        String key = "";
        for (int i = 0; i < path.getPathCount(); i++) {
            if (i > 0) {
                key += ".";
            }
            key += path.getPathComponent(i).toString();
        }
        return key;
    }
}