/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.objectweb.orchestra.bpem.view;

import java.util.ArrayList;
import java.util.Enumeration;

import javax.swing.JTree;
import javax.swing.ToolTipManager;
import javax.swing.event.TreeSelectionEvent;
import javax.swing.event.TreeSelectionListener;
import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.DefaultTreeModel;
import javax.swing.tree.TreePath;
import javax.swing.tree.TreeSelectionModel;

import org.objectweb.orchestra.bpem.model.ExecutionModel;
import org.objectweb.orchestra.bpem.model.ModelListener;
import org.objectweb.orchestra.bpem.model.ModelNode;
import org.objectweb.orchestra.bpem.model.Options;
import org.objectweb.orchestra.bpem.model.OptionsListener;


/**
 * ViewAsTree.java
 *
 * @author Bill Murray
 * @version $Id: ViewAsTree.java 5 2006-10-20 14:37:44Z lejeuneg $
 */
public class ViewAsTree extends JTree implements ModelListener,
TreeSelectionListener, OptionsListener {

    private final ExecutionModel model;

    private DefaultMutableTreeNode rootNode;

    private DefaultTreeModel treeModel;

    private final int myListenerID;

    public ViewAsTree(ExecutionModel model) {
        super();
        this.model = model;
        //
        // Listen for events from the model and Options
        //
        this.myListenerID = model.addListener(this);
        Options.addListener(this);

        this.rootNode = new DefaultMutableTreeNode("Model");
        this.treeModel = new DefaultTreeModel(this.rootNode);
        this.setModel(this.treeModel);

        setEditable(false);
        this.setRootVisible(false);
        getSelectionModel().setSelectionMode(
                TreeSelectionModel.SINGLE_TREE_SELECTION);
        setShowsRootHandles(true);

        setCellRenderer(new TVCellRenderer());
        ToolTipManager.sharedInstance().registerComponent(this);
        //
        // Listen for when the selection changes.
        //
        this.addTreeSelectionListener(this);
    }

    public DefaultMutableTreeNode addObject(DefaultMutableTreeNode parent,
            Object child) {
        return addObject(parent, child, false);
    }

    public DefaultMutableTreeNode addObject(DefaultMutableTreeNode parent,
            Object child, boolean shouldBeVisible) {
        DefaultMutableTreeNode childNode = new DefaultMutableTreeNode(child);

        if (parent == null) {
            parent = this.rootNode;
        }

        this.treeModel.insertNodeInto(childNode, parent, parent.getChildCount());

        //Make sure the user can see the lovely new node.
        if (shouldBeVisible) {
            scrollPathToVisible(new TreePath(childNode.getPath()));
        }
        return childNode;
    }

    public void clear() {
        this.rootNode.removeAllChildren();
        this.treeModel.reload();
    }

    public void displayAll() {
        dispNode(this.model.getModelRoot(), null);
    }

    /**
     * nodeSelected - This is called by the model when a node has been selected.
     */
    public void nodeSelected(ModelNode mn) {
        // Nothing
    }

    /**
     * optionChanged - Callback received when a program option changes.
     * This could affect the labels of the nodes
     */
    public void optionChanged() {
        resizeAllLabels(this.rootNode);
    }

    /**
     * resizeAllLabels - (recursive) force "nodeChanged" for each node in the
     * tree to force label widths to be recalculated.
     *
     * @param node - the node to start at.
     */
    private void resizeAllLabels(DefaultMutableTreeNode node) {
        this.treeModel.nodeChanged(node);
        Enumeration childNodes = node.children();
        if (childNodes != null) {
            while (childNodes.hasMoreElements()) {
                DefaultMutableTreeNode nextChild = (DefaultMutableTreeNode) childNodes.nextElement();
                resizeAllLabels(nextChild);
            }
        }
    }

    /**
     * refresh - The underlying attributes in the data mode have been changed.
     * We do not actually have to do anything here because the node names
     * do not change. What changes is the time stamps (tooltip) and since that
     * always comes dynamically from TVNodeInfo, it is always refreshed.
     * If a statement changes execution state (new icon) we do need to force
     * label resizing...
     */
    public void refresh() {
        if (isShowing() == true) {
            resizeAllLabels(this.rootNode);
        }
    }

    private void dispNode(ModelNode mn, DefaultMutableTreeNode parent) {
        ArrayList children = mn.getChildModelNodes();
        for (int i = 0; i < children.size(); i++) {
            ModelNode child = (ModelNode) children.get(i);
            DefaultMutableTreeNode n1 = null;
            n1 = addObject(parent, new TVNodeInfo(child));
            if (child.getChildModelNodes() != null) {
                dispNode(child, n1);
            }
        }
        this.treeModel.reload();
    }

    /**
     * valueChanged - called by the treeView when a node is selected. We tell
     * the model about the selection.
     *
     * @param e
     */
    public void valueChanged(TreeSelectionEvent e) {
        DefaultMutableTreeNode node = (DefaultMutableTreeNode) this
        .getLastSelectedPathComponent();
        if (node == null)
            return;
        Object nodeInfo = node.getUserObject();
        if ((nodeInfo instanceof TVNodeInfo) == false)
            return;
        //
        // Tell the model that a node was selected, and we do not want the callback
        //
        this.model.selectModelNode(((TVNodeInfo) nodeInfo).getModelNode(), this.myListenerID);
    }
}