/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.objectweb.orchestra.bpem.view;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.util.ArrayList;
import java.util.Calendar;

import javax.swing.JTree;
import javax.swing.ToolTipManager;
import javax.swing.event.TreeSelectionEvent;
import javax.swing.event.TreeSelectionListener;
import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.DefaultTreeModel;
import javax.swing.tree.TreePath;
import javax.swing.tree.TreeSelectionModel;
import javax.xml.namespace.QName;

import org.objectweb.orchestra.bpem.controller.BPEMUtils;
import org.objectweb.orchestra.bpem.model.ExecutionModel;
import org.objectweb.orchestra.bpem.model.ModelListener;
import org.objectweb.orchestra.bpem.model.ModelNode;
import org.objectweb.orchestra.engine.BPELMessage;
import org.objectweb.orchestra.engine.BPELVariable;
import org.objectweb.orchestra.engine.MessageStructure;
import org.objectweb.orchestra.engine.PartStructure;
import org.objectweb.orchestra.exception.BPELBpemException;
import org.objectweb.orchestra.monitoring.Variable;
import org.objectweb.orchestra.util.BPELConstants;



/**
 * ViewAsTree.java.
 *
 * @author Bill Murray
 * @version $Id: ViewAsVariables.java 80 2006-12-06 21:24:57Z bmurray $
 */
public class ViewAsVariables extends JTree  implements ModelListener,
TreeSelectionListener {

    private final ExecutionModel model;

    private DefaultMutableTreeNode rootNode;

    private DefaultTreeModel treeModel;

    private ModelNode currentVariables = null;

    private final String javalang = "java.lang.";

    private final String bpelPrefix = "org.objectweb.orchestra.";

    public ViewAsVariables(ExecutionModel model) {
        super();

        this.model = model;
        model.addListener(this);

        this.rootNode = new DefaultMutableTreeNode("Variables");
        this.treeModel = new DefaultTreeModel(this.rootNode);
        this.setModel(this.treeModel);

        setEditable(false);
        this.setRootVisible(false);
        getSelectionModel().setSelectionMode(
                TreeSelectionModel.SINGLE_TREE_SELECTION);
        setShowsRootHandles(true);

        setCellRenderer(new TVCellRenderer());
        ToolTipManager.sharedInstance().registerComponent(this);
        //
        // Listen for when the selection changes.
        //
        this.addTreeSelectionListener(this);
    }

    public void clear() {
        this.rootNode.removeAllChildren();
        this.treeModel.reload();
        this.currentVariables = null;
    }

    /**
     * displayAll - This is called by the data model when a new BPEL
     * process has been selected (The "clear" funciton is called first
     * by the model) We want to display data for the first node as the
     * initial display.
     */
    public void displayAll() {
        nodeSelected(this.model.getModelRoot().getFirstChildModelNode());
    }

    /**
     * nodeSelected - This is called by the model when a node has been selected.
     * Get the variables for the parent scope of the selected node and put them
     * in the tree. If it is the same scope as previously selected, then no
     * need to repeat.
     */
    public void nodeSelected(ModelNode mn) {
        ModelNode variables = findVariables(mn);
        if (variables == this.currentVariables) {
            //refresh();
            return;
        }

        this.currentVariables = variables;
        try {
            refreshPreserve(false);
        } catch (BPELBpemException e) {
            e.printStackTrace();
        }
    }


    public void refresh() {
        try {
            refreshPreserve(true);
        } catch (BPELBpemException e) {
            e.printStackTrace();
        }
    }


    private void refreshPreserve(boolean preserve) throws BPELBpemException {
        if ((this.currentVariables != null) && isShowing()) {
            TVExpansionState tvState = null;
            if (preserve) {
                //
                // Save current TV expansion state
                //
                new TreePath(this.rootNode.getPath());
                tvState = new TVExpansionState(this);
            }
            // Rebuild TV variable display
            //
            displayVariables(this.currentVariables);
            //
            // Restore TV expansion state
            //
            if (preserve) {
                tvState.restoreState();
            }
        }
    }

    /**
     * displayVariables - Walk thru the variables and put them in the tree.
     *
     * @param vars - the ModelNode for the BPEL "Variables".
     */
    private void displayVariables(ModelNode vars) throws BPELBpemException {
        this.rootNode.removeAllChildren();
        if (vars != null) {
            if (!vars.hasChildModelNodes()) {
                return;
            }
            ArrayList children = vars.getChildModelNodes();
            for (int i = 0; i < children.size(); i++) {
                ModelNode v1 = (ModelNode) children.get(i);
                if (v1.getType().compareToIgnoreCase("Variable") == 0) {
                    displayVariable(v1.getAttrValue("name"), v1.getStmtNumber());
                }
            }
            this.treeModel.reload();
        }
    }

    /**
     * displayVariable - Go to the BPEL engine (via our model) and get the
     * variable specified. Then display it based on its type.
     *
     * @param name - the variable name.
     * @param statementNb - the statement number of the variable.
     */
    private void displayVariable(String name, int statementNb) throws BPELBpemException {
        Variable v = this.model.getEngine().getVariable(name, statementNb);
        if (v != null) {
            String type = v.getType();

            if (BPELConstants.BPELMESSAGE.equals(type)) {
                displayBPELMessageVar(name, v);
            } else if (BPELConstants.BPELVARIABLE.equals(type)) {
                displayBPELVariableVar(name, v);
            }
        } else {
            //it means the variable we were trying to get was called on
            //a running instance and this instance is finished
            //in this case, the monitoring was not set to all
            //We will display an error message
        	String errMsg = "The requested variable : \n";
            errMsg += "  Instance : "+this.model.getEngine().getID() + "\n";
            errMsg += "  Variable name : "+name + "\n";
            errMsg += "is not currently running and the moitoring mode does not allow to see finished instances.\n";
            errMsg += "Please, perform a refresh.";
            BPEMUtils.showError("Unable to retrieve variable", errMsg);
        }
    }

    /**
     * displayBPELMessageVar - Display a variable that is a BPELMessage.
     * We have to pull apart the essage structure and display each part
     * of the message.
     *
     * @param name
     * @param v
     */
    private void displayBPELMessageVar(String name, Variable v) throws BPELBpemException {
        String type = v.getType();

        DefaultMutableTreeNode tn = addObject(null, name + " : " + type);
        BPELMessage jm = (BPELMessage) v.getValue();
        MessageStructure ms = jm.getStructure();
        PartStructure[] ps = ms.getParts();
        for (int i = 0; i < ps.length; i++) {
            String partName = ps[i].getName();
            Object partValue = jm.getPart(partName);
            if (partValue == null) {
                addObject(tn, partName + " = null");
            } else {
                //String pt = partValue.getClass().getName();
                displayUserValue(partName, partValue, tn);
            }
        }
    }

    /**
     * DisplayUserValue - Display the "value" for the specified object. The
     * object wil have "getter" methods for each field that may be a simple
     * property or a comples property that has to further be decomposed.
     * This method is recursive.
     *
     * @param name
     * @param userValue
     * @param parent
     */
    private void displayUserValue(String name, Object userValue, DefaultMutableTreeNode parent) throws BPELBpemException {
        if (userValue == null) {
            addObject(parent, name + " = null");
        } else {
            Class userClass = userValue.getClass();
            String pt = userClass.getName();
            String dispType = stripTypeName(pt);
            if (isSimpleClass(userClass) || QName.class.isAssignableFrom(userClass)) {
                addObject(parent, name + " (" + dispType + ") = "  + userValue.toString());
            } else if (Calendar.class.isAssignableFrom(userClass)) {
                String calSt = "";
                Calendar cal = (Calendar) userValue;
                calSt = calSt + "YEAR = " + cal.get(Calendar.YEAR) + " ; ";
                calSt = calSt + "MONTH = " + cal.get(Calendar.MONTH) + " ; ";
                calSt = calSt + "DATE = " + cal.get(Calendar.DATE) + " ; ";
                calSt = calSt + "HOUR_OF_DAY = " + cal.get(Calendar.HOUR_OF_DAY) + " ; ";
                calSt = calSt + "MINUTE = " + cal.get(Calendar.MINUTE) + " ; ";
                calSt = calSt + "SECOND = " + cal.get(Calendar.SECOND);
                addObject(parent, name + " (" + dispType + ") = "  + calSt);
            } else {
                DefaultMutableTreeNode userParent = addObject(parent, name + " (" + dispType + ")");
                //
                // Now get and display the "properties" for this object.
                // We walk the method list and find the "get"ters which
                // coorespond to the properties.
                //
                Method m[] = userClass.getDeclaredMethods();
                for (int i = 0; i < m.length; i++) {
                    String methodName = m[i].getName();
                    if ((methodName.startsWith("get"))
                            && ((m[i].getModifiers() & Modifier.PUBLIC) != 0)
                            && (m[i].getParameterTypes().length == 0)) {
                        String propName = methodName.substring(3);
                        try {
                            Object value = m[i].invoke(userValue, (Object[]) null);
                            displayUserValue(propName, value, userParent);
                        } catch (IllegalArgumentException e) {
                            throw new BPELBpemException("ViewAsVariables.displayUserValue, IllegalArgumentException caught :", e);
                        } catch (IllegalAccessException e) {
                            throw new BPELBpemException("ViewAsVariables.displayUserValue, IllegalAccessException caught :", e);
                        } catch (InvocationTargetException e) {
                            throw new BPELBpemException("ViewAsVariables.displayUserValue, InvocationTargetException caught :", e);
                        }
                    }
                }
            }
        }
    }

    /**
     * isSimpleClass - determines if the class specified is part of
     * java runtime.
     *
     * @param c - The class to test.
     * @return boolean
     */
    private boolean isSimpleClass(Class c) {
        if (c.isPrimitive()) {
            return true;
        }
        if (c.getName().startsWith("java.lang.")) {
            return true;
        }
        //if (c.getName().compareToIgnoreCase(String.class.getName()) == 0) return true;
        return false;
    }

    /**
     * displayBPELVariableVar - Display a variable that is a BPELVariable
     * type. These are basic variables (ints, strings...)
     *
     * @param name - The variable name.
     * @param v -  the variable data received from the BPEL engine.
     */
    private void displayBPELVariableVar(String name, Variable v) {
        String type = v.getType();
        String dispType = stripTypeName(type);
        BPELVariable jv = (BPELVariable) v.getValue();
        if (jv == null) {
            addObject(null, name + " (" + dispType + ") = null");
        } else {
            Object value = jv.getValue();
            String s = "null";
            if (value != null) {
                s =  value.toString();
            }
            addObject(null, name + " (" + dispType + ") = " + s);
        }
    }

    /**
     * stripTypeName - remove the prefix package names from the type
     * for display purposes.
     *
     * @param type
     * @return
     */
    private String stripTypeName(String type) {
        if (type.startsWith(this.javalang)) {
            return type.substring(this.javalang.length());
        }
        if (type.startsWith(this.bpelPrefix)) {
            return type.substring(this.bpelPrefix.length());
        }
        return type;
    }

    public DefaultMutableTreeNode addObject(DefaultMutableTreeNode parent,
            Object child) {
        return addObject(parent, child, false);
    }

    public DefaultMutableTreeNode addObject(DefaultMutableTreeNode parent,
            Object child, boolean shouldBeVisible) {
        DefaultMutableTreeNode childNode = new DefaultMutableTreeNode(child);

        if (parent == null) {
            parent = this.rootNode;
        }
        this.treeModel.insertNodeInto(childNode, parent, parent.getChildCount());

        //Make sure the user can see the lovely new node.
        if (shouldBeVisible) {
            scrollPathToVisible(new TreePath(childNode.getPath()));
        }
        return childNode;
    }

    private ModelNode findVariables(ModelNode mn) {
        ModelNode scope = findParentNodeType(mn, "Scope");
        ModelNode variables = findChildNodeType(scope, "Variables");
        return variables;
    }

    private ModelNode findParentNodeType(ModelNode mn, String type) {
        if (mn == null) {
            return null;
        }
        if (mn.getType().compareToIgnoreCase(type) == 0) {
            return mn;
        }
        return findParentNodeType(mn.getParent(), type);
    }

    private ModelNode findChildNodeType(ModelNode mn, String type) {
        if (mn == null) {
            return null;
        }
        if (mn.getType().compareToIgnoreCase(type) == 0) {
            return mn;
        }
        if (mn.hasChildModelNodes()) {
            ArrayList children = mn.getChildModelNodes();
            for (int i = 0; i < children.size(); i++) {
                ModelNode child = (ModelNode) children.get(i);
                ModelNode found = findChildNodeType(child, type);
                if (found != null) {
                    return found;
                }
            }
        }
        return null;
    }
    /**
     * valueChanged - called by the treeView when a node is selected.
     *
     * @param e
     */
    public void valueChanged(TreeSelectionEvent e) {
        // Nothing
    }
}
