/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.objectweb.orchestra.data;

import java.util.Collection;
import java.util.Iterator;
import java.util.Vector;

import javax.ejb.CreateException;
import javax.ejb.FinderException;
import javax.naming.NamingException;

import org.objectweb.orchestra.engine.BPELEngine;
import org.objectweb.orchestra.exception.BPELEngineException;
import org.objectweb.orchestra.util.BPELConstants;
import org.objectweb.orchestra.util.BPELTrace;


/**
 * BPUCache.java.
 *
 * @author Goulven Le Jeune & Charles Souillard
 * @version $Id: BPUCache.java 39 2006-11-21 10:09:54Z lejeuneg $
 *
 */
public class BPUCache {

    private static Vector bpuList = new Vector();
    private static int bpuNb = 0;
    private static Object mutex = new Object();

    public static int createBPU(String processName, int bpiId, String state, int currentScopeId) throws BPELEngineException {
        if (BPELTrace.isDataDebugEnabled()) {
            BPELTrace.traceData("BPUCache createBPU");
        }
        int bpuId = -1;
        if (BPELEngine.isWithEjb()) {
            try {
                BPUEntityHomeLocal bpuhl = BPUEntityBeanUtil.getLocalHome();
                BPUEntityLocal bel = bpuhl.create(processName, bpiId, state, currentScopeId);
                bpuId = bel.getId().intValue();
            } catch (NamingException e) {
                throw new BPELEngineException("NamingException BPUCache createBPU processName = " + processName + " :\n", e);
            } catch (CreateException e) {
                throw new BPELEngineException("CreateException BPUCache createBPU processName = " + processName + " :\n", e);
            }
        } else {
            synchronized (mutex) {
                bpuId = bpuNb++;
            }
            BPUData bpud  = new BPUData(bpuId, processName, bpiId, state, currentScopeId);
            bpuList.add(bpud);
        }
        if (BPELTrace.isDataDebugEnabled()) {
            BPELTrace.traceData("BPUCache end createBPU");
        }
        return bpuId;
    }

    /**
     * Find a BPU from its bpuId.
     *
     */
    public static BPUData findBPU(int bpuId) throws BPELEngineException {
        if (BPELTrace.isDataDebugEnabled()) {
            BPELTrace.traceData("BPUCache findBPU");
        }
        BPUData result = null;
        if (BPELEngine.isWithEjb()) {
            BPUEntityHomeLocal bpuhl = null;
            BPUEntityLocal bpul = null;
            try {
                bpuhl = BPUEntityBeanUtil.getLocalHome();
                bpul = bpuhl.findByPrimaryKey(new Integer(bpuId));
                result = convert(bpul);
            } catch (javax.ejb.FinderException fe) {
                //Nothing to do here. BPU not found the method will return null
                //throw new BPELEngineException("FinderException BPUCache findBPU : \n", fe);
            } catch (NamingException e) {
                throw new BPELEngineException("NamingException BPUCache findBPU :\n", e);
            }
        } else {
            synchronized (bpuList) {
                for (Iterator it = bpuList.iterator(); it.hasNext();) {
                    BPUData bpud = (BPUData) it.next();
                    if (bpud.getBPUId() == bpuId) {
                        result = bpud;
                        break;
                    }
                }
            }
        }
        if (BPELTrace.isDataDebugEnabled()) {
            BPELTrace.traceData("BPUCache end findBPU");
        }
        return result;
    }

    public static void addMonitor(int bpuId, String m) throws BPELEngineException {
        if (BPELTrace.isDataDebugEnabled()) {
            BPELTrace.traceData("BPUCache addMonitor");
        }
        if (BPELEngine.isWithEjb()) {
            BPUEntityHomeLocal bpuhl = null;
            BPUEntityLocal bpul = null;
            try {
                bpuhl = BPUEntityBeanUtil.getLocalHome();
                bpul = bpuhl.findByPrimaryKey(new Integer(bpuId));
                Vector monitor = bpul.getMonitor();
                monitor.add(m);
                bpul.setMonitor(monitor);
            } catch (javax.ejb.FinderException fe) {
                throw new BPELEngineException("FinderException BPUCache addMonitor monitor = " + m + " : \n", fe);
            } catch (NamingException e) {
                throw new BPELEngineException("NamingException BPUCache addMonitor monitor = " + m + " : \n", e);
            }
        } else {
            synchronized (bpuList) {
                for (Iterator it = bpuList.iterator(); it.hasNext();) {
                    BPUData bpud = (BPUData) it.next();
                    if (bpud.getBPUId() == bpuId) {
                        Vector monitor = bpud.getMonitor();
                        monitor.add(m);
                        bpud.setMonitor(monitor);
                        break;
                    }
                }
            }
        }
        if (BPELTrace.isDataDebugEnabled()) {
            BPELTrace.traceData("BPUCache end addMonitor");
        }
    }

    // Only called when working without EJB
    // With EJB relations are used
    public static Vector getBPIBPUs(int bpiId) {
        if (BPELTrace.isDataDebugEnabled()) {
            BPELTrace.traceData("BPUCache getBPIBPUs");
        }
        Vector result = new Vector();
        synchronized (bpuList) {
            for (Iterator it = bpuList.iterator(); it.hasNext();) {
                BPUData bpud = (BPUData) it.next();
                if (bpud.getBPIId() == bpiId) {
                    result.add(bpud);
                }
            }
        }
        if (BPELTrace.isDataDebugEnabled()) {
            BPELTrace.traceData("BPUCache end getBPIBPUs");
        }
        return result;
    }

    public static void setBPUState(int bpuId, String state) throws BPELEngineException {
        if (BPELTrace.isDataDebugEnabled()) {
            BPELTrace.traceData("BPUCache setBPUState, bpuId = "+bpuId+", state = "+state);
        }
        if (BPELEngine.isWithEjb()) {
            try {
                BPUEntityHomeLocal bpuhl = BPUEntityBeanUtil.getLocalHome();
                BPUEntityLocal bpu = bpuhl.findByPrimaryKey(new Integer(bpuId));
                bpu.setState(state);
            } catch (NamingException e) {
                throw new BPELEngineException("NamingException BPUCache setBPUState : ", e);
            } catch (FinderException e) {
                throw new BPELEngineException("FinderException BPUCache setBPUState : ", e);
            }
        } else {
            synchronized (bpuList) {
                for (Iterator it = bpuList.iterator(); it.hasNext();) {
                    BPUData bpud = (BPUData) it.next();
                    if (bpud.getBPUId() == bpuId) {
                        bpud.setState(state);
                        break;
                    }
                }
            }
        }
        if (BPELTrace.isDataDebugEnabled()) {
            BPELTrace.traceData("BPUCache end setBPUState");
        }
    }

    public static void setBPUStatementNb(int bpuId, int nb) throws BPELEngineException {
        if (BPELTrace.isDataDebugEnabled()) {
            BPELTrace.traceData("BPUCache setBPUStatementNb");
        }
        if (BPELEngine.isWithEjb()) {
            try {
                BPUEntityHomeLocal bpuhl = BPUEntityBeanUtil.getLocalHome();
                BPUEntityLocal bpu = bpuhl.findByPrimaryKey(new Integer(bpuId));
                bpu.setStatementNb(nb);
            } catch (NamingException e) {
                throw new BPELEngineException("NamingException BPUCache setBPUStatementNb : ", e);
            } catch (FinderException e) {
                throw new BPELEngineException("FinderException BPUCache setBPUStatementNb : ", e);
            }
        } else {
            synchronized (bpuList) {
                for (Iterator it = bpuList.iterator(); it.hasNext();) {
                    BPUData bpud = (BPUData) it.next();
                    if (bpud.getBPUId() == bpuId) {
                        bpud.setStatementNb(nb);
                        break;
                    }
                }
            }
        }
        if (BPELTrace.isDataDebugEnabled()) {
            BPELTrace.traceData("BPUCache end setBPUStatementNb");
        }
    }

    public static void setCurrentScopeId(int bpuId, int scopeId) throws BPELEngineException {
        if (BPELTrace.isDataDebugEnabled()) {
            BPELTrace.traceData("BPUCache setCurrentScopeId bpuId=" + bpuId + ", scopeId=" + scopeId);
        }
        if (BPELEngine.isWithEjb()) {
            try {
                BPUEntityHomeLocal bpuhl = BPUEntityBeanUtil.getLocalHome();
                BPUEntityLocal bpu = bpuhl.findByPrimaryKey(new Integer(bpuId));
                bpu.setCurrentScopeId(scopeId);
            } catch (NamingException e) {
                throw new BPELEngineException("NamingException BPUCache setBPUStatementNb : ", e);
            } catch (FinderException e) {
                throw new BPELEngineException("FinderException BPUCache setBPUStatementNb : ", e);
            }
        } else {
            synchronized (bpuList) {
                for (Iterator it = bpuList.iterator(); it.hasNext();) {
                    BPUData bpud = (BPUData) it.next();
                    if (bpud.getBPUId() == bpuId) {
                        bpud.setCurrentScopeId(scopeId);
                        break;
                    }
                }
            }
        }
        if (BPELTrace.isDataDebugEnabled()) {
            BPELTrace.traceData("BPUCache end setCurrentScopeId");
        }
    }

    public static Vector getActiveBpus(int bpiId) throws BPELEngineException {
        if (BPELTrace.isDataDebugEnabled()) {
            BPELTrace.traceData("BPUCache getActiveBpus");
        }
        Vector result = new Vector();
        if (BPELEngine.isWithEjb()) {
            try {
                BPUEntityHomeLocal bpuhl = BPUEntityBeanUtil.getLocalHome();
                Collection bpus = bpuhl.findByBpiIdAndState(new Integer(bpiId), BPELConstants.BPU_ACTIVE);
                result = convert(bpus);
            } catch (NamingException e) {
                throw new BPELEngineException("NamingException BPUCache setBPUStatementNb : ", e);
            } catch (FinderException e) {
                //no bpu found, nothing to do
            }
        } else {
            synchronized (bpuList) {
                for (Iterator it = bpuList.iterator(); it.hasNext();) {
                    BPUData bpud = (BPUData) it.next();
                    if (bpud.getBPIId() == bpiId && bpud.getState().equals(BPELConstants.BPU_ACTIVE)) {
                        result.add(bpud);
                        break;
                    }
                }
            }
        }
        if (BPELTrace.isDataDebugEnabled()) {
            BPELTrace.traceData("BPUCache end getActiveBpus");
        }
        return result;
    }

    // Only called when working without EJB
    // With EJB relations are used
    public static void removeBPIBPUs(int bpiId) {
        if (BPELTrace.isDataDebugEnabled()) {
            BPELTrace.traceData("BPUCache removeBPIBPUs");
        }
        Vector result = new Vector();
        synchronized (bpuList) {
            for (Iterator it = bpuList.iterator(); it.hasNext();) {
                BPUData bd = (BPUData) it.next();
                if (bd.getBPIId() == bpiId) {
                    result.add(bd);
                }
            }
            bpuList.removeAll(result);
        }
        if (BPELTrace.isDataDebugEnabled()) {
            BPELTrace.traceData("BPUCache end removeBPIBPUs");
        }
    }

    public static Vector convert(Collection bpuLocalCol) {
        Vector result = new Vector();
        for (Iterator it = bpuLocalCol.iterator(); it.hasNext();) {
            BPUEntityLocal bpul = (BPUEntityLocal) it.next();
            result.add(convert(bpul));
        }
        return result;
    }

    private static BPUData convert(BPUEntityLocal bpul) {
        if (bpul != null) {
            BPIEntityLocal bpi = bpul.getBPI();
            int bpiId;
            if (bpi == null) {
                bpiId = -1;
            } else {
                bpiId = bpi.getId().intValue();
            }
            return new BPUData(bpul.getId().intValue(), bpul.getProcess(), bpiId, bpul.getState(), bpul.getStatementNb(),
                    bpul.getMonitor(), bpul.getCurrentScopeId());
        }
        return null;
    }

}
