/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.objectweb.orchestra.data;

import java.util.Collection;
import java.util.Iterator;
import java.util.Vector;

import javax.ejb.CreateException;
import javax.ejb.FinderException;
import javax.naming.NamingException;

import org.objectweb.orchestra.engine.BPELEngine;
import org.objectweb.orchestra.exception.BPELEngineException;
import org.objectweb.orchestra.exception.BPELUtilException;
import org.objectweb.orchestra.util.BPELTrace;
import org.objectweb.orchestra.util.Util;


/**
 * MessageCache.java.
 *
 * @author Goulven Le Jeune & Charles Souillard
 * @version $Id: MessageCache.java 39 2006-11-21 10:09:54Z lejeuneg $
 *
 */
public class MessageCache {

    private static Vector messageList = new Vector();
    private static int messageNb = 0;
    private static Object mutex = new Object();

    public static int createMessage(String processName, String targetNamespace, String portType, String operation,
            String refPartner, Object value) throws BPELEngineException {
        return createMessage(processName, targetNamespace, portType, operation, refPartner, value, -1, false, null, null, -1);
    }

    public static int createMessage(String processName, String targetNamespace, String portType, String operation,
            String refPartner, Object value, int statementNb, boolean isFault, String faultName, String faultMessage,
            int bpiId) throws BPELEngineException {
        if (BPELTrace.isDataDebugEnabled()) {
            BPELTrace.traceData("MessageCache createMessage");
        }
        Object val;
        try {
            val = Util.putValueObject(value);
        } catch (BPELUtilException e) {
            throw new BPELEngineException("BPELUtilException MessageCache createMessage : ", e);
        }
        int messageId = -1;
        long date = System.currentTimeMillis();
        if (BPELEngine.isWithEjb()) {
            try {

                MessageHomeLocal mhl = MessageBeanUtil.getLocalHome();
                MessageLocal ml = mhl.create(processName, targetNamespace, portType, operation, refPartner, val,
                        statementNb, isFault, date, faultName, faultMessage);
                if (bpiId != -1) {
                    BPIEntityHomeLocal bpihl = BPIEntityBeanUtil.getLocalHome();
                    BPIEntityLocal bpil = bpihl.findByPrimaryKey(new Integer(bpiId));
                    ml.setBPI(bpil);
                }
                messageId = ml.getId().intValue();
            } catch (NamingException e) {
                throw new BPELEngineException("NamingException MessageCache createMessage processName = "
                        + processName + " :\n", e);
            } catch (CreateException e) {
                throw new BPELEngineException("CreateException MessageCache createMessage processName = "
                        + processName + " :\n", e);
            } catch (FinderException e) {
                throw new BPELEngineException("FinderException MessageCache createMessage processName = "
                        + processName + " :\n", e);
            }
        } else {
            synchronized (mutex) {
                messageId = messageNb++;
            }
            MessageData md  = new MessageData(messageId, processName, targetNamespace, portType, operation,
                    refPartner, val, statementNb, isFault, date, faultName, faultMessage, bpiId);
            md.setBPIId(bpiId);
            messageList.add(md);
        }
        if (BPELTrace.isDataDebugEnabled()) {
            BPELTrace.traceData("MessageCache end createMessage");
        }
        return messageId;
    }

    public static int createFaultMessage(String processName, String targetNamespace, Object value, int statementNb,
            String faultName, String faultMessage, int bpiId) throws BPELEngineException {
        if (BPELTrace.isDataDebugEnabled()) {
            BPELTrace.traceData("MessageCache createFaultMessage");
        }
        String portType = null;
        String operation = null;
        String refPartner = null;
        return createMessage(processName, targetNamespace, portType, operation, refPartner, value, statementNb, true,
                faultName, faultMessage, bpiId);
    }

    public static MessageData findMessage(int messageId) throws BPELEngineException {
        if (BPELTrace.isDataDebugEnabled()) {
            BPELTrace.traceData("MessageCache findMessage");
        }
        MessageData result = null;
        if (BPELEngine.isWithEjb()) {
            MessageLocal ml;
            try {
                MessageHomeLocal mh = MessageBeanUtil.getLocalHome();
                ml = mh.findByPrimaryKey(new Integer(messageId));
                result = convert(ml);
            } catch (NamingException ex) {
                throw new BPELEngineException("NamingException MessageCache findMessage messageId = " + messageId + " :\n", ex);
            } catch (FinderException ex) {
                // Nothing
            }
        } else {
            synchronized (messageList) {
                for (Iterator it = messageList.iterator(); it.hasNext();) {
                    MessageData md = (MessageData) it.next();
                    if (md.getMessageId() == messageId) {
                        result = md;
                        break;
                    }
                }
            }
        }
        if (BPELTrace.isDataDebugEnabled()) {
            BPELTrace.traceData("MessageCache end findMessage");
        }
        return result;
    }

    public static Vector findByBPIAndStatementNb(int bpiId, int statementNb) throws BPELEngineException {
        if (BPELTrace.isDataDebugEnabled()) {
            BPELTrace.traceData("MessageCache findByBPIAndStatementNb");
        }
        Vector result = new Vector();
        if (BPELEngine.isWithEjb()) {
            Collection messages;
            try {
                MessageHomeLocal mh = MessageBeanUtil.getLocalHome();
                messages = mh.findByBPIAndStatementNb(bpiId, statementNb);
                result = convert(messages);
            } catch (NamingException ex) {
                throw new BPELEngineException("NamingException MessageCache findByBPIAndStatementNb bpiId = "
                        + bpiId + " :\n", ex);
            } catch (FinderException ex) {
                // Nothing
                //throw new BPELEngineException("FinderException MessageCache findByBPIAndStatementNb bpiId = "
                //+ bpiId + " :\n" , ex);
            }
        } else {
            synchronized (messageList) {
                for (Iterator it = messageList.iterator(); it.hasNext();) {
                    MessageData md = (MessageData) it.next();
                    if ((md.getBPIId() == bpiId) && (md.getStatementNb() == statementNb)) {
                        result.add(md);
                    }
                }
            }
        }
        if (BPELTrace.isDataDebugEnabled()) {
            BPELTrace.traceData("MessageCache end findByBPIAndStatementNb");
        }
        return result;
    }

    public static Collection findByNoBpi() throws BPELEngineException {
        if (BPELTrace.isDataDebugEnabled()) {
            BPELTrace.traceData("MessageCache findByNoBpi");
        }
        Vector result = new Vector();
        if (BPELEngine.isWithEjb()) {
            Collection messages;
            try {
                MessageHomeLocal mh = MessageBeanUtil.getLocalHome();
                messages = mh.findByNoBpi();
                result = convert(messages);
            } catch (NamingException ex) {
                throw new BPELEngineException("NamingException MessageCache findByNoBpi :\n", ex);
            } catch (FinderException ex) {
                // Nothing
                //throw new BPELEngineException("FinderException MessageCache findByNoBpi :\n" , ex);
            }
        } else {
            synchronized (messageList) {
                for (Iterator it = messageList.iterator(); it.hasNext();) {
                    MessageData md = (MessageData) it.next();
                    if (md.getBPIId() == -1) {
                        result.add(md);
                    }
                }
            }
        }
        if (BPELTrace.isDataDebugEnabled()) {
            BPELTrace.traceData("MessageCache end findByNoBpi");
        }
        return result;
    }

    public static void setMessageStatementNb(int messageId, int nb) throws BPELEngineException {
        if (BPELTrace.isDataDebugEnabled()) {
            BPELTrace.traceData("MessageCache setMessageStatementNb");
        }
        if (BPELEngine.isWithEjb()) {
            try {
                MessageHomeLocal mhl = MessageBeanUtil.getLocalHome();
                MessageLocal ml = mhl.findByPrimaryKey(new Integer(messageId));
                ml.setStatementNb(nb);
            } catch (NamingException e) {
                throw new BPELEngineException("NamingException BPUCache setBPUStatementNb : ", e);
            } catch (FinderException e) {
                throw new BPELEngineException("FinderException BPUCache setBPUStatementNb : ", e);
            }
        } else {
            synchronized (messageList) {
                for (Iterator it = messageList.iterator(); it.hasNext();) {
                    MessageData md = (MessageData) it.next();
                    if (md.getMessageId() == messageId) {
                        md.setStatementNb(nb);
                        break;
                    }
                }
            }
        }
        if (BPELTrace.isDataDebugEnabled()) {
            BPELTrace.traceData("MessageCache setMessageStatementNb");
        }
    }

    public static void setMessageBPI(int messageId, int bpiId) throws BPELEngineException {
        if (BPELTrace.isDataDebugEnabled()) {
            BPELTrace.traceData("MessageCache setMessageBPI");
        }
        if (BPELEngine.isWithEjb()) {
            try {
                MessageHomeLocal mhl = MessageBeanUtil.getLocalHome();
                MessageLocal ml = mhl.findByPrimaryKey(new Integer(messageId));
                BPIEntityHomeLocal bpihl = BPIEntityBeanUtil.getLocalHome();
                BPIEntityLocal bpil = bpihl.findByPrimaryKey(new Integer(bpiId));
                ml.setBPI(bpil);
            } catch (NamingException e) {
                throw new BPELEngineException("NamingException BPUCache setBPUStatementNb : ", e);
            } catch (FinderException e) {
                throw new BPELEngineException("FinderException BPUCache setBPUStatementNb : ", e);
            }
        } else {
            synchronized (messageList) {
                for (Iterator it = messageList.iterator(); it.hasNext();) {
                    MessageData md = (MessageData) it.next();
                    if (md.getMessageId() == messageId) {
                        md.setBPIId(bpiId);
                        break;
                    }
                }
            }
        }
        if (BPELTrace.isDataDebugEnabled()) {
            BPELTrace.traceData("MessageCache end setMessageBPI");
        }
    }



    public static void setMessagePartner(int messageId, String partner) throws BPELEngineException {
        if (BPELTrace.isDataDebugEnabled()) {
            BPELTrace.traceData("MessageCache setMessagePartner");
        }
        if (BPELEngine.isWithEjb()) {
            try {
                MessageHomeLocal mhl = MessageBeanUtil.getLocalHome();
                MessageLocal ml = mhl.findByPrimaryKey(new Integer(messageId));
                ml.setPartner(partner);
            } catch (NamingException e) {
                throw new BPELEngineException("NamingException BPUCache setBPUStatementNb : ", e);
            } catch (FinderException e) {
                throw new BPELEngineException("FinderException BPUCache setBPUStatementNb : ", e);
            }
        } else {
            synchronized (messageList) {
                for (Iterator it = messageList.iterator(); it.hasNext();) {
                    MessageData md = (MessageData) it.next();
                    if (md.getMessageId() == messageId) {
                        md.setPartner(partner);
                        break;
                    }
                }
            }
        }
        if (BPELTrace.isDataDebugEnabled()) {
            BPELTrace.traceData("MessageCache end setMessagePartner");
        }
    }

    //Only called when working without EJB
    // With EJB relations are used
    public static Vector getBPIMessages(int bpiId) {
        if (BPELTrace.isDataDebugEnabled()) {
            BPELTrace.traceData("MessageCache getBPIMessages");
        }
        Vector result = new Vector();
        synchronized (messageList) {
            for (Iterator it = messageList.iterator(); it.hasNext();) {
                MessageData md = (MessageData) it.next();
                if (md.getBPIId() == bpiId) {
                    result.add(md);
                }
            }
        }
        if (BPELTrace.isDataDebugEnabled()) {
            BPELTrace.traceData("VariableCache end getBPIMessages");
        }
        return result;
    }

    // Only called when working without EJB
    // With EJB relations are used
    public static void removeBPIMessages(int bpiId) {
        if (BPELTrace.isDataDebugEnabled()) {
            BPELTrace.traceData("MessageCache removeBPIMessages");
        }
        Vector result = new Vector();
        synchronized (messageList) {
            for (Iterator it = messageList.iterator(); it.hasNext();) {
                MessageData md = (MessageData) it.next();
                if (md.getBPIId() == bpiId) {
                    result.add(md);
                }
            }
            messageList.removeAll(result);
        }
        if (BPELTrace.isDataDebugEnabled()) {
            BPELTrace.traceData("MessageCache end removeBPIMessages");
        }
    }

    public static Vector convert(Collection messageLocalCol) {
        Vector result = new Vector();
        for (Iterator it = messageLocalCol.iterator(); it.hasNext();) {
            MessageLocal ml = (MessageLocal) it.next();
            result.add(convert(ml));
        }
        return result;
    }

    private static MessageData convert(MessageLocal ml) {
        if (ml != null) {
            BPIEntityLocal bpi = ml.getBPI();
            int bpiId;
            if (bpi == null) {
                bpiId = -1;
            } else {
                bpiId = bpi.getId().intValue();
            }
            return new MessageData(ml.getId().intValue(), ml.getProcessName(), ml.getNamespace(), ml.getPortType(),
                    ml.getOperation(), ml.getPartner(), ml.getValue(), ml.getStatementNb(), ml.getFault(), ml.getDate(),
                    ml.getFaultName(), ml.getFaultMessage(), bpiId, ml.getPartner());
        }
        return null;
    }
}
