/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.objectweb.orchestra.data;

import java.util.Collection;
import java.util.Iterator;
import java.util.Vector;

import javax.ejb.CreateException;
import javax.ejb.FinderException;
import javax.naming.NamingException;

import org.objectweb.orchestra.engine.BPELEngine;
import org.objectweb.orchestra.exception.BPELEngineException;
import org.objectweb.orchestra.exception.BPELUtilException;
import org.objectweb.orchestra.util.BPELTrace;
import org.objectweb.orchestra.util.Util;


/**
 * VariableCache.java.
 *
 * @author Goulven Le Jeune & Charles Souillard
 * @version $Id: VariableCache.java 39 2006-11-21 10:09:54Z lejeuneg $
 *
 */
public class VariableCache {

    private static Vector variableList = new Vector();
    private static int variableNb = 0;
    private static Object mutex = new Object();

    public static int createVariable(String varName, String messageType, String varType, String element,
            String bpeljType, int bpiId, int scopeId, Object value) throws BPELEngineException {
        if (BPELTrace.isDataDebugEnabled()) {
            BPELTrace.traceData("VariableCache createVariable");
        }
        Object val;
        try {
            val = Util.putValueObject(value);
        } catch (BPELUtilException e) {
            throw new BPELEngineException("BPELUtilException VariableCache createVariable : ", e);
        }
        int variableId = -1;
        if (BPELEngine.isWithEjb()) {
            try {
                // Retrieve BPI and Scope
                BPIEntityHomeLocal bpihl = BPIEntityBeanUtil.getLocalHome();
                BPIEntityLocal bpil = bpihl.findByPrimaryKey(new Integer(bpiId));
                ScopeHomeLocal shl = ScopeBeanUtil.getLocalHome();
                ScopeLocal sl = shl.findByPrimaryKey(new Integer(scopeId));

                VariableHomeLocal vh = VariableBeanUtil.getLocalHome();
                // Create the variable
                VariableLocal v = vh.create(varName, messageType, varType, element, bpeljType);
                // Associate the variable with the right scope and BPI
                v.setBPI(bpil);
                v.setScope(sl);
                // Set the value of the variable
                v.setValue(val);
            } catch (NamingException e) {
                throw new BPELEngineException("RemoveException VariableCache createVariable varName = " + varName + " : \n", e);
            } catch (FinderException e) {
                throw new BPELEngineException("FinderException VariableCache createVariable varName = " + varName + " : \n", e);
            } catch (CreateException e) {
                throw new BPELEngineException("CreateException VariableCache createVariable varName = " + varName + " : \n", e);
            }
        } else {
            synchronized (mutex) {
                variableId = variableNb++;
            }
            VariableData vd  = new VariableData(variableId, varName, messageType,
                    varType, element, bpeljType, bpiId, scopeId, val);
            variableList.add(vd);
        }
        if (BPELTrace.isDataDebugEnabled()) {
            BPELTrace.traceData("VariableCache end createVariable");
        }
        return variableId;
    }

    public static VariableData setVariableValue(int variableId, Object value, int bpiId) throws BPELEngineException {
        if (BPELTrace.isDataDebugEnabled()) {
            BPELTrace.traceData("VariableCache setVariableValue");
        }
        VariableData result = null;
        Object val;
        try {
            val = Util.putValueObject(value);
        } catch (BPELUtilException e) {
            throw new BPELEngineException("BPELUtilException VariableCache setVariableValue : ", e);
        }
        if (BPELEngine.isWithEjb()) {
            try {
                VariableHomeLocal vhl = VariableBeanUtil.getLocalHome();
                VariableLocal vs = vhl.findByPrimaryKey(new Integer(variableId));
                vs.setValue(val);
                result = convert(vs, bpiId);
            } catch (NamingException e) {
                throw new BPELEngineException("NamingException VariableCache setVariableValue : ", e);
            } catch (FinderException e) {
                throw new BPELEngineException("FinderException VariableCache setVariableValue : ", e);
            }
        } else {
            synchronized (variableList) {
                for (Iterator it = variableList.iterator(); it.hasNext();) {
                    VariableData vd = (VariableData) it.next();
                    if (vd.getVariableId() == variableId) {
                        vd.setValue(val);
                        result = vd;
                        break;
                    }
                }
            }
        }
        if (BPELTrace.isDataDebugEnabled()) {
            BPELTrace.traceData("VariableCache end setVariableValue");
        }
        return result;
    }

    public static Collection getVariables(int bpiId, String name, int scopeId) throws BPELEngineException {
        if (BPELTrace.isDataDebugEnabled()) {
            BPELTrace.traceData("VariableCache getVariables");
        }
        Vector result = new Vector();
        if (BPELEngine.isWithEjb()) {
            Collection variables;
            try {
                VariableHomeLocal vhl = VariableBeanUtil.getLocalHome();
                variables = vhl.findVariablesByBpiByNameAndByScope(new Integer(bpiId), name, new Integer(scopeId));
                result = convert(variables, bpiId);
            } catch (NamingException e) {
                throw new BPELEngineException("NamingException VariableCache getVariables : \n" , e);
            } catch (FinderException e) {
                // Nothing
                //throw new BPELEngineException("FinderException VariableCache getVariables : \n" , e);
            }
        } else {
            synchronized (variableList) {
                for (Iterator it = variableList.iterator(); it.hasNext();) {
                    VariableData vd = (VariableData) it.next();
                    if ((vd.getBPIId() == bpiId)
                            && (vd.getVariableName().equalsIgnoreCase(name))
                            && (vd.getScopeId() == scopeId)) {
                        result.add(vd);
                    }
                }
            }
        }
        if (BPELTrace.isDataDebugEnabled()) {
            BPELTrace.traceData("VariableCache end getVariables");
        }
        return result;
    }

    public static Collection getScopeVariables(int scopeId, int bpiId) throws BPELEngineException {
        if (BPELTrace.isDataDebugEnabled()) {
            BPELTrace.traceData("VariableCache getScopeVariables");
        }
        Vector result = new Vector();
        if (BPELEngine.isWithEjb()) {
            Collection variables;
            try {
                VariableHomeLocal vhl = VariableBeanUtil.getLocalHome();
                variables = vhl.findVariablesByScope(new Integer(scopeId));
                result = VariableCache.convert(variables, bpiId);
            } catch (NamingException e) {
                throw new BPELEngineException("NamingException VariableCache getScopeVariables : \n" , e);
            } catch (FinderException e) {
                // Nothing
                //throw new BPELEngineException("FinderException VariableCache getScopeVariables : \n" , e);
            }
        } else {
            synchronized (variableList) {
                for (Iterator it = variableList.iterator(); it.hasNext();) {
                    VariableData vd = (VariableData) it.next();
                    if (vd.getScopeId() == scopeId) {
                        result.add(vd);
                    }
                }
            }
        }
        if (BPELTrace.isDataDebugEnabled()) {
            BPELTrace.traceData("VariableCache end getScopeVariables");
        }
        return result;
    }

    // Only called when working without EJB
    // With EJB relations are used
    public static void removeBPIVariables(int bpiId) {
        if (BPELTrace.isDataDebugEnabled()) {
            BPELTrace.traceData("VariableCache removeBPIVariables");
        }
        Vector result = new Vector();
        synchronized (variableList) {
            for (Iterator it = variableList.iterator(); it.hasNext();) {
                VariableData vd = (VariableData) it.next();
                if (vd.getBPIId() == bpiId) {
                    result.add(vd);
                }
            }
            variableList.removeAll(result);
        }
        if (BPELTrace.isDataDebugEnabled()) {
            BPELTrace.traceData("VariableCache end removeBPIVariables");
        }
    }

    public static Vector convert(Collection varLocalCol, int bpiId) {
        Vector result = new Vector();
        for (Iterator it = varLocalCol.iterator(); it.hasNext();) {
            VariableLocal vl = (VariableLocal) it.next();
            result.add(convert(vl, bpiId));
        }
        return result;
    }

    private static VariableData convert(VariableLocal vl, int bpiId) {
        if (vl != null) {
            return new VariableData(
                    vl.getId().intValue(),
                    vl.getName(),
                    vl.getMessageType(),
                    vl.getType(),
                    vl.getElement(),
                    vl.getBPELJType(),
                    bpiId,
                    vl.getScope().getId().intValue(),
                    vl.getValue());
        }
        return null;
    }

}
