/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.objectweb.orchestra.engine;

import java.io.Serializable;
import java.lang.reflect.Method;

/**
 * BPELVariable.java Represents the value of a BPEL variable.
 * It contains its name, Java type (according to JAX-RPC), value and the class
 * of its value.
 * This class is serialized into the database.
 *
 * @author Goulven Le Jeune & Charles Souillard
 * @version $Id: BPELVariable.java 5 2006-10-20 14:37:44Z lejeuneg $
 *
 */
public class BPELVariable implements Serializable, java.lang.Cloneable {

    /**
     * The name of this BPEL variable.
     */
    private String name = null;

    /**
     * The Java type of this BPEL variable according to the JAX-RPC 1.1 mapping.
     */
    private String type = null;

    /**
     * The class of the value attribute (useful when the value object is null).
     * The class differs from the type attribute when type is a Java primitive type
     * (it needs to be wrapped), or when the type is byte[]
     * (Then valueClass.getName() retruns "[B)").
     */
    private Class valueClass = null;

    /**
     * The value of the BPEL variable.
     */
    private Object value = null;

    /**
     * Default Constructor
     *
     * @param name The name of this BPEL variable
     * @param type The JAX-RPC 1.1 Java type of this BPEL variable
     * @param valueClass The value object class of this BPEL variable
     */
    public BPELVariable(String name, String type, Class valueClass) {
        this.name = name;
        this.type = type;
        this.valueClass = valueClass;
    }

    /**
     * Sets the value of this BPEL variable
     * @param value The value of this BPEL variable
     */
    public void setValue(Object value) {
        this.value = value;
    }

    /**
     * Retruns the name of this BPEL variable
     * @return String The name of this BPEL variable
     */
    public String getName() {
        return this.name;
    }

    /**
     * @return
     */
    public String getType() {
        return this.type;
    }

    /**
     * @return
     */
    public Class getValueClass() {
        return this.valueClass;
    }

    /**
     * Returns the value of this BPEL variable
     * @return Object The value of this BPEL variable
     */
    public Object getValue() {
        return this.value;
    }

    /*
     * (non-Javadoc)
     *
     * @see java.lang.Object#clone()
     */
    public Object clone() {
        Object result = null;
        try {
            if (this.value != null) {
                Class cl = this.value.getClass();
                try {
                    Method met = cl.getDeclaredMethod("clone", (Class[])null);
                    BPELVariable jv = new BPELVariable(getName(), getType(), getValueClass());
                    jv.setValue(met.invoke(this.value, (Object[])null));
                    result = jv;
                } catch (Exception e) {
                    result = super.clone();
                }
            } else {
                result = super.clone();
            }
        } catch (CloneNotSupportedException cnse) {
            // This never happens
            return null;
        }
        return result;
    }
}