/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.objectweb.orchestra.engine;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.Collection;
import java.util.Date;
import java.util.Enumeration;
import java.util.GregorianCalendar;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.TimeZone;
import java.util.Vector;

import javax.ejb.CreateException;
import javax.ejb.FinderException;
import javax.ejb.RemoveException;
import javax.jms.JMSException;
import javax.jms.MapMessage;
import javax.jms.QueueConnection;
import javax.jms.QueueSender;
import javax.jms.QueueSession;
import javax.jms.Session;

import javax.naming.NamingException;

import org.objectweb.orchestra.data.BPICache;
import org.objectweb.orchestra.data.BPIData;
import org.objectweb.orchestra.data.BPUCache;
import org.objectweb.orchestra.data.BPUData;
import org.objectweb.orchestra.data.CSCache;
import org.objectweb.orchestra.data.MessageCache;
import org.objectweb.orchestra.data.MessageData;
import org.objectweb.orchestra.data.ScopeCache;
import org.objectweb.orchestra.data.ScopeData;
import org.objectweb.orchestra.data.VariableData;
import org.objectweb.orchestra.exception.BPELEngineException;
import org.objectweb.orchestra.exception.BPELException;
import org.objectweb.orchestra.exception.BPELMonitoringException;
import org.objectweb.orchestra.exception.BPELUtilException;
import org.objectweb.orchestra.tree.Activity;
import org.objectweb.orchestra.tree.BPELProcess;
import org.objectweb.orchestra.tree.Catch;
import org.objectweb.orchestra.tree.CatchAll;
import org.objectweb.orchestra.tree.Flow;
import org.objectweb.orchestra.tree.Invoke;
import org.objectweb.orchestra.tree.OnAlarm;
import org.objectweb.orchestra.tree.OnMessage;
import org.objectweb.orchestra.tree.Receive;
import org.objectweb.orchestra.tree.Source;
import org.objectweb.orchestra.tree.Statement;
import org.objectweb.orchestra.tree.While;
import org.objectweb.orchestra.util.BPELConstants;
import org.objectweb.orchestra.util.BPELTrace;
import org.objectweb.orchestra.util.BPELTransaction;
import org.objectweb.orchestra.util.ClassLoaderMap;
import org.objectweb.orchestra.util.DateTime;
import org.objectweb.orchestra.util.JmsUtil;
import org.objectweb.orchestra.util.Util;
import org.w3c.dom.Document;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import org.objectweb.orchestra.data.BPUEntityBeanUtil;
import org.objectweb.orchestra.data.BPUEntityHomeLocal;
import org.objectweb.orchestra.data.BPUEntityLocal;
import org.objectweb.orchestra.data.MessageBeanUtil;
import org.objectweb.orchestra.data.MessageHomeLocal;
import org.objectweb.orchestra.data.MessageLocal;
import org.objectweb.orchestra.data.ScopeBeanUtil;
import org.objectweb.orchestra.data.ScopeHomeLocal;
import org.objectweb.orchestra.data.ScopeLocal;
import org.objectweb.orchestra.monitoring.BpelMonitoringSessionHomeLocal;
import org.objectweb.orchestra.monitoring.BpelMonitoringSessionLocal;
import org.objectweb.orchestra.monitoring.BpelMonitoringSessionUtil;

/**
 * BPI.java
 *
 * @author Goulven Le Jeune & Charles Souillard
 * @version $Id: BPI.java 39 2006-11-21 10:09:54Z lejeuneg $
 *
 */
public class BPI {

    private String process;
    private int bpiId = 0;
    private boolean stopAll = false;
    private static int pickNb = 0;
    private static int eventNb = 0;
    private BPELProcess bpelProcess = null;
    private EngineLocal engine = null;
    private BpelMonitoringSessionLocal monitor = null;
    private String targetNamespace = null;
    private int scopeId = -1;
    private int runningBPUs = 0;
    private boolean finishing = false;
    private String finishingReason = null;
    private Throwable ex = null;
    private String terminateCause = null;
    Vector pendingReceives;
    Vector pendingInvokes;
    Vector runningFlows;
    Vector runningForEach ;
    Vector pendingSources;
    Vector pendingTargets;
    Vector pendingPicks;
    Vector pendingEventHandlers;
    Vector pendingFaultHandlers;
    Vector pendingCompensationHandlers;
    Vector invokeInbounds;
    private boolean end = false;
    private String currentType = "nothing";
    private String lastType = "nothing";
    private HashMap stoppingBpusByScope = null;
    private boolean needStart = false;
    private boolean alreadyTryToTerminate = false;

    /**
     * Constructor
     *
     * @param bpiId : the Id of the BPI
     * @throws BPELEngineException
     */
    public BPI(int bpiId) throws BPELEngineException {
        this.bpiId = bpiId;
        BPIMap.addBPI(this);
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI constructor");
        this.pendingReceives = new Vector();
        this.pendingInvokes = new Vector();
        this.pendingSources = new Vector();
        this.pendingTargets = new Vector();
        this.pendingPicks = new Vector();
        this.pendingEventHandlers = new Vector();
        this.pendingFaultHandlers = new Vector();
        this.pendingCompensationHandlers = new Vector();
        this.runningFlows = new Vector();
        this.runningForEach = new Vector();
        this.invokeInbounds = new Vector();
        this.bpelProcess = null;
        this.engine = null;
        this.scopeId = -1;
        this.stoppingBpusByScope = new HashMap();
        initEngine();
        initMonitoring();
    }

    /**
     * Initializes the BPI
     *
     * @param processName : the name of the process
     * @throws BPELEngineException
     */
    public synchronized void init(String processName) throws BPELEngineException {
        if (BPELTrace.isBPIDebugEnabled()) {
            BPELTrace.traceBPI("BPI init");
        }
        this.process = processName;
        if (BPELTrace.isBPIDebugEnabled()) {
            BPELTrace.traceBPI("BPI " + this.bpiId + " onMessage, process = " + this.process + ", bpiId = " + this.bpiId);
        }

        BPIData bpid = BPICache.findBPI(this.bpiId);
        if (bpid.getBPIId() != this.bpiId) {
            throw new BPELEngineException("The BPIData is not the correct one : perhaps there is a synchrnisation problem on create/remove BPIs. BPIData.bpiId = "+bpid.getBPIId()+", BPIData.processName = "+bpid.getProcessName()+", currentBpiId = "+this.bpiId+", currentProcessName = "+this.process, null);
        }
        this.targetNamespace = bpid.getTargetNamespace();
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " onMessage BPI trouve : " + bpid.getProcessName());

        try {
            String packag = Util.getPackageFromNamespace(this.targetNamespace);
            ClassLoader classL = ClassLoaderMap.getCL(this.process);
            Class cl = Class.forName((packag + ".BPEL" + this.process), true, classL);
            Method met = cl.getMethod("getProcess",(Class[])null);
            this.bpelProcess = (BPELProcess) met.invoke(null, (Object[]) null);
        } catch (ClassNotFoundException ex) {
            throw new BPELEngineException("ClassNotFoundException BPI init process = "+this.process + ", targetNamespace = " + this.targetNamespace + ", bpiData =" + bpid + " : " ,ex);
        } catch (IllegalAccessException ex) {
            throw new BPELEngineException("IllegalAccessException BPI init process = "+this.process + " : " ,ex);
        } catch (NoSuchMethodException ex) {
            throw new BPELEngineException("NoSuchMethodException BPI init process = "+this.process + " : " ,ex);
        } catch (InvocationTargetException ex) {
            throw new BPELEngineException("InvocationTargetException BPI init process = "+this.process + " : " ,ex);
        } catch (BPELUtilException ex) {
            throw new BPELEngineException("BPELUtilException BPI init process = "+this.process + " : " ,ex);
        }
        this.needStart = true;
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " end init");
    }

    /**
     * @return
     */
    public synchronized boolean needStart() {
        return this.needStart;
    }

    /**
     * @throws BPELEngineException
     */
    public synchronized void start() throws BPELEngineException {
        if (BPELTrace.isBPIDebugEnabled()) {
            BPELTrace.traceBPI("BPI start");
            BPELTrace.traceBPI("BPI " + this.bpiId + " start start BPI");
        }
        createBPU(1, "start", null, null);
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " end start");
        this.needStart = false;
    }

    /**
     * @throws BPELEngineException
     */
    public synchronized void restartBPI() throws BPELEngineException {
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI restart");
        restart();
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " end restart");
    }

    /**
     * @return
     */
    public boolean stopAll() {
        return this.stopAll;
    }

    /**
     * @param bpuId
     * @return
     */
    public boolean stopBpuByScope(int bpuId) {
        return this.stoppingBpusByScope.containsKey(new Integer(bpuId));
    }

    /**
     * @return
     */
    public int getBpiId() {
        return this.bpiId;
    }

    /**
     * @return
     */
    public BPELProcess getBPELProcess() {
        return this.bpelProcess;
    }

    /**
     * @return
     */
    public int getCurrentScopeId() {
        return this.scopeId;
    }

    /**
     * @param e
     */
    public void setException(Throwable e) {
        this.ex = e;
    }

    /**
     * @param cause
     */
    public void setTerminateCause(String cause) {
        this.terminateCause = cause;
    }

    /**
     * @param message
     * @throws BPELEngineException
     */
    public void onMessage(MapMessage message) throws BPELEngineException {
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " onMessage");
        String type = "unset";
        String source = "unset";
        try{
            type = message.getString("type");
            this.currentType = type;
            source = message.getString("source");
            if (this.end) {
                System.out.println("**** BPI.onMessage, end=true, bpiId="+this.bpiId+" source="+source+", type="+type);
            }
            if (BPELTrace.isBPIDebugEnabled())
                BPELTrace.traceBPI("BPI " + this.bpiId + " onMessage source = " + source + ", type = " + type);
            if (source.equals("manager")) {
                if (type.equals("invokeInbound")) {
                    String operation = message.getString("operation");
                    String portType = message.getString("portType");
                    String partner = message.getString("partner");
                    String refPartner = message.getString("refPartner");
                    int messageId = message.getInt("messageId");
                    boolean oneWay = new Boolean(message.getString("oneWay")).booleanValue();
                    invokeInbound(operation, portType, partner, refPartner, messageId,oneWay);
                }
            } else if (source.equals("Admin")) {
                if (type.equals("terminate")) {
                    this.setTerminateCause("Terminated by Operator");
                    terminate();
                }
            } else if (source.equals("BPU")) {
                int bpuId = message.getInt("bpuId");
                //String name = message.getString("name");
                String joinCondition = message.getString("joinCondition");
                //String suppressJoinFailure = message.getString("suppressJoinFailure");
                this.scopeId = message.getInt("scopeId");
                if (type.equals("receive")) {
                    String operation = message.getString("operation");
                    String portType = message.getString("portType");
                    String partner = message.getString("partner");
                    int brancheNb = message.getInt("brancheNb");
                    receive(operation, portType, partner, brancheNb, bpuId);
                } else if (type.equals("flow")) {
                    int brancheNb = message.getInt("brancheNb");
                    int childNb = message.getInt("childNb");
                    int[] childs = new int[childNb];
                    for (int i = 0; i < childNb; i++) {
                        childs[i] = message.getInt("branche" + Integer.toString(i));
                    }
                    flow(brancheNb, childNb, childs);
                } else if (type.equals("forEach")) {
                    int brancheNb = message.getInt("brancheNb");
                    int vChildNb = message.getInt("vChildNb");
                    int startCounterValue = message.getInt("startCounterValue") ;
                    String counterName = message.getString("counterName") ;
                    int childBrancheNb = message.getInt("branche");
                    forEach(brancheNb, startCounterValue, vChildNb, counterName, childBrancheNb) ;
                } else if (type.equals("endFlow")) {
                    int brancheNb = message.getInt("brancheNb");
                    endFlow(bpuId, brancheNb);
                }  else if (type.equals("endForEach")) {
                    int brancheNb = message.getInt("brancheNb");
                    int sId = message.getInt("scopeId") ;
                    endForEach(sId, bpuId, brancheNb);
                } else if (type.equals("endScope")) {
                    int oldScopeNb = message.getInt("oldScopeId");
                    endScope(oldScopeNb);
                } else if (type.equals("source")) {
                    String linkName = message.getString("linkName");
                    String transitionCondition = message.getString("transitionCondition");
                    source(linkName, transitionCondition);
                } else if (type.equals("targets")) {
                    int brancheNb = message.getInt("brancheNb");
                    int childNb = message.getInt("childNb");
                    String[] targetLinkNames = new String[childNb];
                    int[] branchesNb = new int[childNb];
                    for (int i = 1; i <= childNb; i++) {
                        targetLinkNames[i - 1] = message.getString("linkName" + i);
                        branchesNb[i - 1] = message.getInt("branche" + i);
                    }
                    targets(brancheNb, childNb, targetLinkNames, bpuId, joinCondition,branchesNb);
                } else if (type.equals("wait")) {
                    String wUntil = message.getString("until");
                    String wFor = message.getString("for");
                    int brancheNb = message.getInt("brancheNb");
                    if (wUntil != null) waitUntil(bpuId, wUntil, brancheNb);
                    if (wFor != null) waitFor(bpuId, wFor, brancheNb);
                } else if (type.equals("pick")) {
                    int brancheNb = message.getInt("brancheNb");
                    event(brancheNb, bpuId,"pick");
                } else if (type.equals("reply")) {
                    int brancheNb = message.getInt("brancheNb");
                    String operation = message.getString("operation");
                    String portType = message.getString("portType");
                    String partner = message.getString("partner");
                    reply(operation, portType, partner, brancheNb);
                } else if (type.equals("eventHandler")) {
                    int brancheNb = message.getInt("brancheNb");
                    event(brancheNb, bpuId,"event");
                } else if (type.equals("faultHandler")) {
                    int brancheNb = message.getInt("brancheNb");
                    faultHandler(brancheNb);
                } else if (type.equals("compensationHandler")) {
                    int brancheNb = message.getInt("brancheNb");
                    compensationHandler(brancheNb);
                } else if (type.equals("throw")) {
                    String faultName = message.getString("faultName");
                    String variable = message.getString("faultVariable");
                    fault(this.scopeId, faultName, variable);
                } else if (type.equals("BPELJfault")) {
                    String faultName = message.getString("faultName");
                    Object faultData = null;
                    byte[] msgFaultData = message.getBytes("faultData");
                    if (msgFaultData != null) {
                        try {
                            faultData = Util.getValueObject(msgFaultData, this.process);
                        } catch (BPELUtilException e) {
                            throw new BPELEngineException("BPI.onMessage: "
                                    + "Exception occured during the deserialization of the fault data", e);
                        }
                    }
                    BPELJfault(this.scopeId, faultName, faultData);
                } else if (type.equals("fault")) {
                    String faultName = message.getString("faultName");
                    String faultMessage = message.getString("faultMessage");
                    int messageId = message.getInt("messageId");
                    fault(this.scopeId, faultName, faultMessage, messageId);
                } else if (type.equals("compensate")) {
                    String scopeName = message.getString("scopeName");
                    compensate(scopeName);
                } else if (type.equals("terminate")) {
                    this.setTerminateCause("Terminated by Process");
                    terminate();
                } else if (type.equals("stopped")) {
                    //int brancheNb = message.getInt("brancheNb");
                    // TODO bpuStopped(bpuId);
                } else if (type.equals("endBPI")) {
                    //endBPI(BPELConstants.BPI_FINISHED);
                    this.finishing = true;
                } else if (type.equals("endBPU")) {
                    endBPU();
                } else if (type.equals("invokeReturn")) {
                    InvokerObject io = (InvokerObject) Util.getValueObject(message.getBytes("invokerObject"),this.process);
                    manageInvokeReturn(io);
                }

            } else if (source.equals("timerReceiver")) {
                int brancheNb = message.getInt("brancheNb");
                this.scopeId = message.getInt("scopeId");
                //int bpuId = message.getInt("bpuId");
                if (type.equals("wait")) {
                    // Launch a new BPU on the wait statement
                    createBPU( brancheNb,"endWait", null, null);
                } else if (type.equals("pickOnAlarm")) {
                    int pickId = message.getInt("Id");
                    // Remove Pick from PendingPicks if it was not already
                    //removes (received another timer or message)
                    if (removeEvent(pickId,this.pendingPicks)) {
                        // Launch a new BPU on the wait statement
                        createBPU(brancheNb, "pickOnAlarm", null, null);
                    }
                } else if (type.equals("eventOnAlarm")) {
                    int eventId = message.getInt("Id");
                    if (removeEvent(eventId,this.pendingEventHandlers)) {
                        createBPU(brancheNb, "eventOnAlarm", null, null);
                    }
                }
            }
            /*
             if (this.end) {
             System.out.println("BPI.onMessage end of method, end = true, bpiId = "+this.bpiId+", type = "+type+", runningBpus = "+this.runningBPUs);
             }
             */
        } catch (JMSException e) {
            throw new BPELEngineException("BPI.onMessage, JMSException caught : ",e);
        } catch (BPELException e) {
            System.out.println("BPI catch ex, bpiId = "+this.bpiId);
            System.out.println("BPI catch ex, currentType = "+this.currentType);
            System.out.println("BPI catch ex, lastType = "+this.lastType);
            System.out.println("BPI catch ex, runningBpus = "+this.runningBPUs);
            e.printStackTrace();
            throw new BPELEngineException("BPI.onMessage, BPELException caught : ",e);
        } catch (NullPointerException e) {
            System.out.println("BPI.onMessage : NPE caught : cause = "+e.getCause());
            System.out.println("type = "+type+", source = "+source+", message = "+message);
            e.printStackTrace();
            throw new BPELEngineException("BPI.onMessage, NullPointerException caught : ",e);
        }
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " end onMessage");
    }

    /**
     * @param io
     * @throws BPELEngineException
     */
    private void manageInvokeReturn(InvokerObject io) throws BPELEngineException {
        createBPU(io.getInvoke().getNb().intValue(), "invokeReturn", new String[]{"invokerObject"}, new Object[]{io});
    }

    /**
     * @throws BPELEngineException
     */
    private void throwJoinFailure() throws BPELEngineException {
        String faultName = "bpws:joinFailure";
        String variable = null;
        fault(this.scopeId, faultName, variable);
    }

    /**
     * @param bpuId
     * @param wUntil
     * @param brancheNb
     */
    private void waitUntil(int bpuId, String wUntil, int brancheNb) throws BPELEngineException {
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " waitUntil");
        Calendar lNext;
        lNext = analyseDate(wUntil, "Until");
        Date now = new Date();
        if (lNext.before(now)) {
            if (BPELTrace.isBPIDebugEnabled())
                BPELTrace.traceBPI("BPI " + this.bpiId + " waitUntil : timer not launched, specified date ("+lNext.getTimeInMillis()+") before now, now = "+now.getTime());
            createBPU(brancheNb, "endWait", null, null);
        } else {
            if (BPELTrace.isBPIDebugEnabled())
                BPELTrace.traceBPI("BPI " + this.bpiId + " waitUntil : timer launched, specified date ("+lNext.getTimeInMillis()+") after now, now = "+now.getTime());
            BPELTimerManager.getInstance().launchTimer(this.bpiId,lNext, brancheNb, "wait",bpuId,this.scopeId);
        }
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " End waitUntil");
    }

    /**
     * @param bpuId
     * @param wFor
     * @param brancheNb
     */
    private void waitFor(int bpuId, String wFor, int brancheNb) throws BPELEngineException {
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " waitFor");
        Calendar lNext = null;
        lNext = analyseDate(wFor, "For");
        Date now = new Date();
        if (lNext.before(now)) {
            createBPU(brancheNb, "endWait", null, null);
        } else {
            BPELTimerManager.getInstance().launchTimer(this.bpiId,lNext, brancheNb, "wait",bpuId,this.scopeId);
        }
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " End waitFor");
    }

    /**
     * @param d
     * @param type
     * @return
     * @throws BPELEngineException
     */
    private Calendar analyseDate(String d, String type) throws BPELEngineException {
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " analyseDate Date : " + d+", type = "+type);
        Calendar result = null;

        Object param1 = Expression.analyseExpression(d, this.bpiId, this.scopeId, this.process,BPELConstants.DATE_EXPR);
        if (param1 == null) {
            if (BPELTrace.isBPIDebugEnabled())
                BPELTrace.traceBPI("BPI " + this.bpiId + " param null");
            throw new BPELEngineException("Error BPIMdb analyseDate : param1 null",null);
        }

        if (type.equals("For")) {
            result = DateTime.addDuration((String) param1);
        } else if (type.equals("Until")) {
            if (param1 instanceof Calendar) {
                Calendar cal =  (Calendar) param1;
                if ((cal.get(Calendar.DAY_OF_MONTH) == 31) && (cal.get(Calendar.MONTH) == Calendar.DECEMBER) && (cal.get(Calendar.YEAR) == 2)) {
                    // Test not clean but I don't any other solution for the moment...
                    if (BPELTrace.isBPIDebugEnabled())
                        BPELTrace.traceBPI("BPI : this Calendar is a time");
                    cal = correctCal(cal);
                    cal.getFirstDayOfWeek();
                }
                result = (Calendar)cal.clone();
                if (BPELTrace.isBPIDebugEnabled())
                    BPELTrace.traceBPI("BPI " + this.bpiId + " Calendar : " + result.getTime() + " " + result.toString());
            } else if (param1 instanceof String) {
                SimpleDateFormat df = new SimpleDateFormat("dd/MM/yyyy HH:mm:ss");
                try {
                    result = new GregorianCalendar();
                    String dateTime = DateTime.getDateTime((String) param1);
                    if (BPELTrace.isBPIDebugEnabled())
                        BPELTrace.traceBPI("BPI " + this.bpiId + " dateTime : " + dateTime);
                    Date date = df.parse(dateTime);
                    if (BPELTrace.isBPIDebugEnabled())
                        BPELTrace.traceBPI("BPI " + this.bpiId + " date : " + date);
                    result.setTime(date);
                    result.get(Calendar.HOUR);
                } catch (java.text.ParseException e) {
                    throw new BPELEngineException("ParseException caught BPIMdb analyseDate : ", e);
                }
                if (BPELTrace.isBPIDebugEnabled())
                    BPELTrace.traceBPI("BPI " + this.bpiId + " String : " + result.toString());
            }
        }
        if (BPELTrace.isBPIDebugEnabled()) {
            BPELTrace.traceBPI("result = "+result.getTimeInMillis());
            BPELTrace.traceBPI("BPI " + this.bpiId + " End analyseDate : " + result.toString() + " " + result.getTime());
        }
        return result;
    }


    /**
     * @param c1
     * @return
     */
    private Calendar correctCal(Calendar c1) {
        Calendar c3 = (Calendar)c1.clone();
        c3.setTimeZone(TimeZone.getDefault());
        c3.getTime();

        Calendar c2 = new GregorianCalendar ();
        c2.set(Calendar.HOUR_OF_DAY,c3.get(Calendar.HOUR_OF_DAY));
        c2.set(Calendar.MINUTE,c3.get(Calendar.MINUTE));
        c2.set(Calendar.SECOND,c3.get(Calendar.SECOND));
        c2.getTime();
        return c2;


    }

    /**
     * @return
     * @throws BPELEngineException
     */
    private String getProcessSuppressJoinFailure() throws BPELEngineException {
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " getProcessJoinFailure");
        String result = null;
        try {
            String packag = Util.getPackageFromNamespace(this.targetNamespace);
            ClassLoader classL = ClassLoaderMap.getCL(this.process);
            if (BPELTrace.isBPIDebugEnabled())
                BPELTrace.traceBPI("BPI " + this.bpiId + " getProcessJoinFailure, classL = "+classL);
            Class cl = Class.forName((packag + ".BPEL" + this.process), true,classL);
            Object o = cl.newInstance();
            Method met = cl.getMethod("getSuppressJoinFailure",(Class[])null);
            result = (String) met.invoke(o, (Object[])null);
        } catch (java.lang.ClassNotFoundException ex) {
            throw new BPELEngineException("ClassNotFoundException caught BPIMdb initBPELProcess : ", ex);
        } catch (java.lang.InstantiationException ex) {
            throw new BPELEngineException("InstantiationException caught BPIMdb initBPELProcess : " , ex);
        } catch (java.lang.IllegalAccessException ex) {
            throw new BPELEngineException("IllegalAccessException caught BPIMdb initBPELProcess : " , ex);
        } catch (java.lang.NoSuchMethodException ex) {
            throw new BPELEngineException("NoSuchMethodException caught BPIMdb initBPELProcess : " , ex);
        } catch (java.lang.reflect.InvocationTargetException ex) {
            throw new BPELEngineException("InvocationTargetException caught BPIMdb initBPELProcess : " , ex);
        } catch (BPELUtilException ex) {
            throw new BPELEngineException("BPELUtilException caught BPIMdb initBPELProcess : " , ex);
        }
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " End getProcessJoinFailure, result = "+result);
        return result;
    }

    /**
     * @throws BPELEngineException
     */
    private void terminate() throws BPELEngineException {
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " Terminate");
        if (!this.alreadyTryToTerminate) {
            this.alreadyTryToTerminate = true;
            // Send a message to all BPUs on BPELFaultQueue
            if (!isInRestartMode()) {
                //sendFaultBpus();
                this.stopAll = true;
            }
            sendTerminateToWS();
            this.finishing = true;
            this.finishingReason = BPELConstants.BPI_TERMINATED;
        } else {
            String toPrint = "BPI " + this.bpiId + " Terminate, alreadyTryToTerminate=true";
            toPrint += "\n";
            toPrint += "********** INTERNAL ERROR **********";
            toPrint += "The server encountered an internal error preventing him to terminate the instance : "+this.bpiId+" of the process "+this.process+" properly.";
            toPrint += "Here is the stack trace : \n";
            toPrint += "\n";
            toPrint += Util.getExceptionAsString(this.ex);
            toPrint += "\n";
            if (BPELTrace.isBPIDebugEnabled())
                BPELTrace.traceBPI(toPrint);
            System.out.println(toPrint);
        }
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " End Terminate");
    }

    /**
     * @throws BPELEngineException
     */
    private void sendTerminateToWS() throws BPELEngineException {
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " sendTerminateToWS");
        for (int i = 0; i < this.invokeInbounds.size(); i++) {
            InvokeInbound ii = (InvokeInbound) this.invokeInbounds.get(i);
            String refPartner = ii.getRefPartner();
            String excep;
            if (this.ex != null) {
                excep = "\n bpiId = " + this.bpiId;
                excep += Util.getExceptionAsString(this.ex);
            } else {
                excep = this.terminateCause;
            }
            createBPU(-1, "replyError", new String[]{"refPartner","exception"}, new String[]{refPartner,excep},300000);
        }
    }

    /**
     * @param scopeId
     * @throws BPELEngineException
     */
    private void sendFaultBpusByScope(int scopeId) throws BPELEngineException {
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " sendFaultBpusByScope, scopeId = "+scopeId);
        Vector activeBpus = BPUCache.getActiveBpus(this.bpiId);
        for (int i = 0; i < activeBpus.size(); i++) {
            BPUData bpu = (BPUData) activeBpus.get(i);
            if (bpu.getCurrentScopeId() == scopeId) {
                this.stoppingBpusByScope.put(new Integer(bpu.getBPUId()),new Object());
                //sendMessageFaultQueue(bpu.getBPUId());
            }
        }

        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " end sendFaultBpusByScope");
    }

    /**
     * @param scopeId
     * @param faultName
     * @param variable
     * @return
     * @throws BPELEngineException
     */
    private boolean fault(int scopeId, String faultName, String variable) throws BPELEngineException {
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " Fault : faultName : " + faultName + ", variable : " + variable);
        boolean terminate = false;
        // Stop all BPUs running statements in this scope
        sendFaultBpusByScope(scopeId);
        PendingFaultHandler p = findFaultHandler(scopeId);
        if (p != null) {
            // Found a fault Handler for this scope => look for a corresponding
            // catch
            PendingCatch pc = null;
            if (variable != null) {
                VariableData vd = getVariable(variable);
                pc = p.findCatch(faultName, vd.getType(), vd.getMessageType(), vd.getElement());
            } else {
                pc = p.findCatch(faultName, null, null, null);
            }
            if (pc != null) {
                String[] argsName = new String[1];
                String[] argsValue = new String[1];
                argsName[0] = "variable";
                argsValue[0] = variable;
                // Found a Catch => Launch new BPU on brancheNb
                createBPU(pc.getBrancheNb(), "catchFault", argsName, argsValue);
            } else {
                // No corresponding Catch => implicit FaultHandler
                terminate = implicitFaultHandler(scopeId, faultName, variable);
            }
        } else {
            // No Fault Handler for this scope => implicit Fault Handler
            terminate = implicitFaultHandler(scopeId, faultName, variable);
        }
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " End Fault");
        return terminate;
    }

    /**
     * @param scopeId
     * @param faultName
     * @param faultMessage
     * @param messageId
     * @return
     * @throws BPELEngineException
     */
    private boolean fault(int scopeId, String faultName, String faultMessage, int messageId) throws BPELEngineException {
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " Fault : faultName : " + faultName + ", faultMessage : " + faultMessage + ", messageId : " + messageId);
        if (scopeId == -1) {
            terminate();
            return true;
        }
        boolean terminate = false;
        // Stop all BPUs running statements in this scope
        sendFaultBpusByScope(scopeId);
        PendingFaultHandler p = findFaultHandler(scopeId);
        if (p != null) {
            // Found a fault Handler for this scope => look for a corresponding
            // catch
            PendingCatch pc = null;
            pc = p.findCatch(faultName, null, faultMessage, null);
            if (pc != null) {
                String[] argsName = new String[1];
                String[] argsValue = new String[1];
                argsName[0] = "messageId";
                argsValue[0] = Integer.toString(messageId);
                // Found a Catch => Launch new BPU on brancheNb
                createBPU(pc.getBrancheNb(), "catchFault", argsName, argsValue);
            } else {
                // No corresponding Catch => implicit FaultHandler
                terminate = implicitFaultHandler(scopeId, faultName, faultMessage, messageId);
            }
        } else {
            // No Fault Handler for this scope => implicit Fault Handler
            terminate = implicitFaultHandler(scopeId, faultName, faultMessage, messageId);
        }
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " End Fault");
        return terminate;
    }

    /**
     * @param scopeId
     * @param faultName
     * @param faultData
     * @return
     * @throws BPELEngineException
     */
    private boolean BPELJfault(int scopeId, String faultName, Object faultData) throws BPELEngineException {
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " BPELJfault : faultName : " + faultName + ", faultData : " + faultData);

        boolean terminate = false;
        // Stop all BPUs running statements in this scope
        sendFaultBpusByScope(scopeId);
        PendingFaultHandler p = findFaultHandler(scopeId);
        if (p != null) {
            // Found a fault Handler for this scope => look for a corresponding catch
            PendingCatch pc = p.findBPELJCatch(faultName, faultData);
            if (pc != null) {
                String[] argsName = {
                        "faultData"
                };
                Object[] argsValue = {
                        faultData
                };
                // Found a Catch => Launch new BPU on brancheNb
                /* To save up some unuseful bytes transmission, the faultData should not be send into a JMS message.
                 * The faultData might be put directly into the good variable.
                 */
                createBPU(pc.getBrancheNb(), "BPELJCatchFault", argsName, argsValue);
            } else {
                // No corresponding Catch => implicit FaultHandler
                terminate = implicitBPELJFaultHandler(scopeId, faultName, faultData);
            }
        } else {
            // No Fault Handler for this scope => implicit Fault Handler
            terminate = implicitBPELJFaultHandler(scopeId, faultName, faultData);
        }
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " End Fault");
        return terminate;
    }

    /**
     * @param scopeId
     * @param faultName
     * @param variable
     * @return
     * @throws BPELEngineException
     */
    private boolean implicitFaultHandler(int scopeId, String faultName, String variable) throws BPELEngineException {
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " ImplicitFaultHandler : faultName : " + faultName + ", variable : " + variable);
        boolean terminate = false;
        int fatherId = implicitFaultHandlerCompensate(scopeId);

        if (fatherId != -1) {
            // ReLaunch the Fault to the Father scope
            terminate = fault(fatherId, faultName, variable);
        } else {
            terminate = true;
            setTerminateCause(faultName);
            terminate();
        }

        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " End ImplicitFaultHandler");
        return terminate;
    }

    /**
     * @param scopeId
     * @param faultName
     * @param faultMessage
     * @param messageId
     * @return
     * @throws BPELEngineException
     */
    private boolean implicitFaultHandler(int scopeId, String faultName, String faultMessage, int messageId) throws BPELEngineException {
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " ImplicitFaultHandler : faultName : " + faultName + ", faultMessage : " + faultMessage);
        boolean terminate = false;

        int fatherId = implicitFaultHandlerCompensate(scopeId);

        if (fatherId != -1) {
            // ReLaunch the Fault to the Father scope
            terminate = fault(fatherId, faultName, faultMessage, messageId);
        } else {
            terminate = true;
            terminate();
        }

        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " End ImplicitFaultHandler");
        return terminate;
    }

    /**
     * @param scopeId
     * @param faultName
     * @param faultData
     * @return
     * @throws BPELEngineException
     */
    private boolean implicitBPELJFaultHandler(int scopeId, String faultName, Object faultData) throws BPELEngineException {
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " ImplicitFaultHandler : faultName : " + faultName + ", faultData : " + faultData);
        boolean terminate = false;
        int fatherId = implicitFaultHandlerCompensate(scopeId);
        if (fatherId != -1) {
            // ReLaunch the Fault to the Father scope
            terminate = BPELJfault(fatherId, faultName, faultData);
        } else {
            terminate = true;
            terminate();
        }

        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " End ImplicitFaultHandler");
        return terminate;
    }

    /**
     * @param scopeId
     * @return
     * @throws BPELEngineException
     */
    private int implicitFaultHandlerCompensate(int scopeId) throws BPELEngineException {
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " ImplicitFaultHandlerCompensate");

        // Get Children Scopes
        Collection children = ScopeCache.getChildren(scopeId);

        // Get Father Scope
        ScopeData father = ScopeCache.getFather(scopeId);

        // Launch all Compensation Handlers for all children scopes
        for (java.util.Iterator it = children.iterator(); it.hasNext();) {
            ScopeData s = (ScopeData) it.next();
            int sId = s.getScopeId();
            compensate(sId);
        }
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " End ImplicitFaultHandlerCompensate");
        if (father != null)
            return father.getScopeId();
        return -1;
    }

    /**
     * @param scopeId
     * @throws BPELEngineException
     */
    private void compensate(int scopeId) throws BPELEngineException {
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " Compensate");
        PendingCompensationHandler p = findCompensationHandler(scopeId);
        if (p != null) {
            // Found a compensation Handler for this scope => launch new BPU on
            // brancheNb
            createBPU(p.getBrancheNb(), "compensate", null, null);
        } else {
            // No Compensation Handler for this scope => implicit compensation
            // Handler
            implicitCompensationHandler();
        }
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " End Compensate");
    }

    /**
     * @param scopeName
     * @throws BPELEngineException
     */
    private void compensate(String scopeName) throws BPELEngineException {
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " Compensate : scopeName : " + scopeName);
        if (scopeName != null) {
            // Find Scopes Bean corresponding to the scope Name
            Collection scopes = ScopeCache.findByNameOrderByDate(this.bpiId, scopeName);

            // Launch all Compensation Handlers for all scopes
            for (java.util.Iterator it = scopes.iterator(); it.hasNext();) {
                ScopeData s = (ScopeData) it.next();
                int sId = s.getScopeId();
                compensate(sId);
            }
        } else {
            implicitCompensationHandler();
        }
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " End Compensate");
    }

    /**
     * @throws BPELEngineException
     */
    private void implicitCompensationHandler() throws BPELEngineException {
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " ImplicitCompensationHandler");

        // Get Children Scopes
        Collection children = ScopeCache.getChildren(this.bpiId);

        // Launch all Compensation Handlers for all children scopes
        for (java.util.Iterator it = children.iterator(); it.hasNext();) {
            ScopeData s = (ScopeData) it.next();
            int sId = s.getScopeId();
            compensate(sId);
        }
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " End ImplicitCompensationHandler");
    }

    /**
     * @param brancheNb
     * @throws BPELEngineException
     */
    private void faultHandler(int brancheNb) throws BPELEngineException {
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " FaultHandler");
        Statement s = this.bpelProcess.findStatement(brancheNb);
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " FaultHandler : s.getType() : " + s.getType());
        if ((s != null) && (s.getType().equals("FaultHandler"))) {
            PendingFaultHandler pfh = new PendingFaultHandler(this.scopeId);
            Statement[] childs = s.getChilds();
            int childNb = s.getChildNb();
            for (int i = 0; i < childNb; i++) {
                if (childs[i].getType().equals("Catch")) {
                    if (BPELTrace.isBPIDebugEnabled())
                        BPELTrace.traceBPI("BPI " + this.bpiId + " FaultHandler : Catch");
                    Catch c = (Catch) childs[i];
                    String faultName = c.getFaultName();
                    String faultVariable = c.getFaultVariable();
                    int catchNb = c.getNb().intValue();
                    if (BPELTrace.isBPIDebugEnabled())
                        BPELTrace.traceBPI("BPI " + this.bpiId + " FaultHandler : add Catch : " + catchNb + " " + faultName + " " + faultVariable);
                    if (faultVariable != null) {
                        VariableData vd = getVariable(faultVariable);
                        pfh.addCatch(catchNb, faultName, vd.getType(), vd.getMessageType(), vd.getElement(), vd.getBPELJType());
                    } else {
                        pfh.addCatch(catchNb, faultName, null, null, null, null);
                    }
                } else if (childs[i].getType().equals("CatchAll")) {
                    if (BPELTrace.isBPIDebugEnabled())
                        BPELTrace.traceBPI("BPI " + this.bpiId + " EventHandler : CatchAll");
                    CatchAll ca = (CatchAll) childs[i];
                    int catchAllNb = ca.getNb().intValue();
                    if (BPELTrace.isBPIDebugEnabled())
                        BPELTrace.traceBPI("BPI " + this.bpiId + " FaultHandler : add CatchAll : " + catchAllNb);
                    pfh.addCatchAll(catchAllNb);
                }
            }
            if (BPELTrace.isBPIDebugEnabled())
                BPELTrace.traceBPI("BPI " + this.bpiId + " FaultHandler : Store faultHandler in PendingFaultHandlers");
            // Store faultHandler in PendingFaultHandlers
            this.pendingFaultHandlers.add(pfh);
        } else
            throw new BPELEngineException("Error in BPIMdb : faultHandler statement called with wrong statementNb or Statement Type not faultHandler",null);
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " End FaultHandler");
    }

    /**
     * @param brancheNb
     */
    private void compensationHandler(int brancheNb) {
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " CompensationHandler");
        PendingCompensationHandler pch = new PendingCompensationHandler(this.scopeId, brancheNb);
        // Store compensationHandler in PendingCompensationHandlers
        this.pendingCompensationHandlers.add(pch);
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " End CompensationHandler");
    }

    /**
     * @param brancheNb
     * @param bpuId
     * @throws BPELEngineException
     */
    private void event(int brancheNb, int bpuId, String type) throws BPELEngineException {
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " event, type = "+type);
        int nb = -1;
        if (type.equals("pick")) {
            nb = ++pickNb;
        } else if (type.equals("event")) {
            nb = ++eventNb;
        }
        PendingEvent pe = new PendingEvent(bpuId,nb, this.scopeId,brancheNb,type);

        Statement s = this.bpelProcess.findStatement(brancheNb);
        if (s != null) {
            Vector v = null;
            if (type.equals("pick")) {
                v = this.pendingPicks;
            } else if (type.equals("event")) {
                v = this.pendingEventHandlers;
            }
            Statement[] childs = s.getChilds();
            int childNb = s.getChildNb();
            boolean finished = false;
            int i = 0;
            while ((i < childNb) && (!finished)) {
                finished = analyseEventChild (childs[i], bpuId, pe);
                i++;
            }
            if (!finished) {
                // Store event in PendingPicks or pendingEventHandlers
                v.add(pe);
            } else {
                if (!isInRestartMode()) {
                    BPELTimerManager.getInstance().removeAlarms(this.bpiId,pe);
                }
            }
        } else
            throw new BPELEngineException("Error in BPIMdb : Event statement called with wrong statementNb or Statement Type not event",null);
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " End event");
    }

    /**
     * @param eventChild
     * @param bpuId
     * @param pe
     * @return
     * @throws BPELEngineException
     */
    private boolean analyseEventChild (Statement eventChild, int bpuId, PendingEvent pe) throws BPELEngineException {
        boolean finished = false;
        if (eventChild.getType().equals("OnMessage")) {
            OnMessage om = (OnMessage) eventChild;
            String partnerLink = om.getOnMessagePartnerLink();
            String portType = om.getOnMessagePortType();
            String operation = om.getOnMessageOperation();
            String variable = om.getOnMessageVariable();
            int onMessageNb = om.getNb().intValue();

            int id = findInvoke(operation, portType, partnerLink);
            if (id != -1) {
                if (BPELTrace.isBPIDebugEnabled())
                    BPELTrace.traceBPI("BPI " + this.bpiId + " analyseEventChild : onMessage : there is a corresponding pending invoke");
                // Remove PendingInvoke
                PendingInvoke pi = (PendingInvoke) this.pendingInvokes.remove(id);

                if (!isInRestartMode()) {
                    int messageId = pi.getMessageId();

                    // Send message to BPU to tell him to continue
                    String[] argsName = new String[1];
                    String[] argsValue = new String[1];
                    argsName[0] = "messageId";
                    argsValue[0] = Integer.toString(messageId);
                    createBPU(onMessageNb, pe.getType()+"OnMessage", argsName, argsValue);
                }
                finished = true;
            } else {
                if (BPELTrace.isBPIDebugEnabled())
                    BPELTrace.traceBPI("BPI " + this.bpiId + " analyseEventChild : onMessage : there is no corresponding invoke");
                // if no corresponding Invoke, store onMessage
                pe.addOnMessage(onMessageNb, partnerLink, portType, operation, variable);
                addBpiWaiting( portType, operation);
            }
        } else if (eventChild.getType().equals("OnAlarm")) {
            finished = launchAlarms((OnAlarm) eventChild,bpuId,finished,pe);
        }
        return finished;
    }

    /**
     * @param oa
     * @param bpuId
     * @param finished
     * @param pe
     * @return
     * @throws BPELEngineException
     */
    private boolean launchAlarms(OnAlarm oa, int bpuId,boolean finished,PendingEvent pe) throws BPELEngineException {
        if (!isInRestartMode()) {
            if (BPELTrace.isBPIDebugEnabled())
                BPELTrace.traceBPI("BPI " + this.bpiId + " launchAlarms");

            String onAlarmFor = oa.getFor();
            String onAlarmUntil = oa.getUntil();
            int onAlarmNb = oa.getNb().intValue();
            Calendar lNext = null;

            if (onAlarmFor != null) {
                lNext = analyseDate(onAlarmFor, "For");
            }
            if (onAlarmUntil != null) {
                lNext = analyseDate(onAlarmUntil, "Until");
            }
            String peType = pe.getType();
            Date now = new Date();
            if (lNext.before(now)) {
                createBPU(onAlarmNb, peType+"OnAlarm", null, null);
                finished = true;
            } else {
                String type = peType+"OnAlarm:" + pe.getId();
                pe.addAlarm(type);
                BPELTimerManager.getInstance().launchTimer(this.bpiId,lNext, onAlarmNb, type,bpuId,this.scopeId);
            }
        }
        return finished;
    }

    /**
     * @param linkName
     * @param transitionCondition
     * @throws BPELEngineException
     */
    private void source(String linkName, String transitionCondition) throws BPELEngineException {
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " source linkName = " + linkName+", transitionCondition = '"+transitionCondition+"'");
        boolean found = false;
        found = findTarget(linkName, transitionCondition);
        if (!found) {
            //Store source
            this.pendingSources.add(new PendingSource(linkName, transitionCondition));
        }
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " end source, found = " + found + " , linkName = " + linkName);
    }

    /**
     * @param bpuId
     * @param s
     * @throws BPELEngineException
     */
    private void update(int bpuId,Statement s) throws BPELEngineException {
        update(bpuId,s,new String[0]);
    }

    /**
     * @param bpuId
     * @param s
     * @param others
     * @throws BPELEngineException
     */
    private void update(int bpuId,Statement s, String[] others) throws BPELEngineException {
        initEngine();
        DbUpdater dbu = new DbUpdater(this.engine);
        dbu.updateBPUStatementNb(this.bpiId,bpuId,s,others);
    }

    /**
     * @param linkName
     * @param tc
     * @return
     * @throws BPELEngineException
     */
    private boolean findTarget(String linkName, String tc) throws BPELEngineException {
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " findTarget linkName = " + linkName);
        boolean found = false;
        for (int i=0;i<this.pendingTargets.size();i++) {
            PendingTarget pt = (PendingTarget) this.pendingTargets.get(i);
            boolean tcValue = new Boolean(tc).booleanValue();
            found = pt.newSource(linkName, tcValue);
            if (BPELTrace.isBPIDebugEnabled())
                BPELTrace.traceBPI("BPI " + this.bpiId + " findTarget PendingTarget found = "+found);
            if (found) {
                int targetNb = pt.getTargetBrancheNb(linkName);
                if (!isInRestartMode()) {
                    Statement s = this.bpelProcess.findStatement(targetNb);
                    update(pt.getBpuId(),s);
                }
                boolean finished = pt.isFinish();
                if (BPELTrace.isBPIDebugEnabled())
                    BPELTrace.traceBPI("BPI " + this.bpiId + " findTarget pendingTarget finished = "+finished);
                if (finished) {
                    this.pendingTargets.remove(i);
                    ptFinished(pt);
                    if (isInRestartMode()) {
                        if (BPELTrace.isRestartDebugEnabled())
                            BPELTrace.traceRestart("BPI " + this.bpiId + " findTarget, adding to finishedTargets = "+pt.getBrancheNb());
                        this.finishedTargets.put(new Integer(pt.getBrancheNb()),pt);
                    }
                }
                break;
            }
        }
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " end findTarget found = " + found + " , linkName = " + linkName);
        return found;
    }

    /**
     * @param pt
     * @throws BPELEngineException
     */
    private void ptFinished(PendingTarget pt) throws BPELEngineException {
        boolean join = pt.evaluateJoinCondition();
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " findTarget joinCondition = "+join);
        //if suppressjoinfailure = yes execute the next statement, else thros bpws:joinFailure
        String suppressJoinFailure = getSuppressJoinFailure(pt.getBrancheNb());
        if (BPELTrace.isRestartDebugEnabled())
            BPELTrace.traceRestart("BPI " + this.bpiId + " ptFinished, pt = "+pt.getBrancheNb());
        if (join) {
            this.scopeId = pt.getScopeId();
            if (!isInRestartMode()) {
                createBPU(pt.getBrancheNb(), "target",null, null);
            }
        } else if (suppressJoinFailure.equals("yes")) {
            if (!isInRestartMode()) {
                createBPU(pt.getBrancheNb(), "sourceAndNextStatement",null, null);
            }
        } else {
            if (BPELTrace.isBPIDebugEnabled())
                BPELTrace.traceBPI("BPI " + this.bpiId + " trow fault bpws:joinFailure");
            if (!isInRestartMode()) {
                throwJoinFailure();
            }
        }
    }

    /**
     * @param statementNb
     * @return
     * @throws BPELEngineException
     */
    private String getSuppressJoinFailure(int statementNb) throws BPELEngineException {
        String sjf = this.bpelProcess.getSuppressJoinFailure(this.bpelProcess.findStatement(statementNb));
        if (sjf == null) {
            sjf = getProcessSuppressJoinFailure();
        }
        return sjf;
    }
    /**
     * @param brancheNb
     * @param childNb
     * @param targetLinkNames
     * @param bpuId
     * @param joinCondition
     * @param bn
     * @throws BPELEngineException
     */
    private void targets(int brancheNb, int childNb, String[] targetLinkNames, int bpuId, String joinCondition, int[] bn) throws BPELEngineException {
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " Targets, childNb : " + childNb + ", activityNb = "+brancheNb+", pendingSources size : " + this.pendingSources.size());
        PendingTarget pt = new PendingTarget(this.bpiId, this.process,bpuId,brancheNb, this.scopeId, childNb, targetLinkNames, joinCondition,bn);
        Vector psTmp = (Vector) this.pendingSources.clone();
        for (int i = 0; i<psTmp.size();i++) {
            PendingSource ps = (PendingSource) psTmp.get(i);
            if (pt.newSource(ps.getLinkName(), ps.getTransitionCondition())) {
                int targetNb = pt.getTargetBrancheNb(ps.getLinkName());
                if (!isInRestartMode()) {
                    Statement s = this.bpelProcess.findStatement(targetNb);
                    update(pt.getBpuId(),s);
                }
                if (BPELTrace.isBPIDebugEnabled())
                    BPELTrace.traceBPI("BPI " + this.bpiId + " Targets, found source for : " + ps.getLinkName());
                this.pendingSources.remove(ps);
            }
        }
        if (pt.isFinish()) {
            if (BPELTrace.isBPIDebugEnabled())
                BPELTrace.traceBPI("BPI " + this.bpiId + " Targets pendingTarget finished");
            ptFinished(pt);
        } else {
            if (BPELTrace.isBPIDebugEnabled())
                BPELTrace.traceBPI("BPI " + this.bpiId + " Targets pendingTarget Not finished");
            this.pendingTargets.add(pt);
        }
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " end Targets");
    }

    /**
     * A endFlow message was send by the BPU meaning that a branche of the flow
     * is finished
     *
     * @param bpuId
     * @param brancheNb
     */
    private void endFlow(int bpuId, int brancheNb) throws BPELEngineException {
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " endFlow : BPU : " + bpuId+", brancheNb="+brancheNb);
        // Find Flow in runningFlows
        RunningFlow rf = findFlow(brancheNb);
        if (BPELTrace.isRestartDebugEnabled())
            BPELTrace.traceRestart("BPI " + this.bpiId + " endFlow normal : "+brancheNb+", bpuId = "+bpuId);
        // Remove one active child on that flow
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " endFlow, activeChildren = "+rf.getActiveChildren());
        boolean finished = rf.childFinished();
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " endFlow, finished = "+finished);
        if (finished) {
            this.runningFlows.remove(rf);
            // Launch a new BPU on the next statement after the flow if needed
            createBPU(rf.getStatementNb(), "endFlow", null, null);
        }
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " end endFlow");
    }

    /**
     * endForEach
     * A endForEach message was send by the BPU meaning that a branche of
     * the forEach is finished
     *
     * @param scopeId
     * @param bpuId
     * @param brancheNb
     * @throws BPELEngineException
     */
    private void endForEach(int scopeId, int bpuId, int brancheNb) throws BPELEngineException {
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " endForEach : BPU : " + bpuId+", brancheNb="+brancheNb);
        // Find Flow in runningFlows
        RunningForEach rf = findForEach(scopeId, brancheNb);
        boolean finished = rf.childFinished();
        if (finished) {
            this.runningForEach.remove(rf);
            // Launch a new BPU on the next statement after the foreach if needed
            createBPU(rf.getStatementNb(), "endForEach", null, null);
        }
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " end endForEach");
    }

    /**
     * A flow message was send by the BPU
     *
     * @param bpuId
     * @param brancheNb
     * @param childNb
     * @param childs
     */
    private void flow(int brancheNb, int childNb, int[] childs) throws BPELEngineException {
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " flow, brancheNb : " + brancheNb + ", childNb : " + childNb);
        // add Flow to the list of running Flows
        RunningFlow rf = new RunningFlow(brancheNb, childNb);

        this.runningFlows.add(rf);
        if (!isInRestartMode()) {
            for (int i = 0; i < childNb; i++) {
                createBPU(childs[i], "flow", null, null);
            }
        }
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " end flow");
    }

    /**
     * forEach
     * A parallel ForEach message has been sent by the BPU
     * @author Alexandre Greco
     *
     * @param brancheNb ForEach branch number
     * @param startCounterValue
     * @param vChildNb Number of branches to be created
     * @param counterName Name of the scope associated variable
     * @param childBrancheNb ForEach Scope activity branch number
     * @throws BPELEngineException
     */
    private void forEach(int brancheNb, int startCounterValue, int vChildNb, String counterName, int childBrancheNb) throws BPELEngineException {
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " forEach, brancheNb : " + brancheNb + ", vChildNb : " + vChildNb);

        // Add the ForEach to the list of running ForEach
        RunningForEach rf = new RunningForEach(this.scopeId, brancheNb, vChildNb);
        this.runningForEach.add(rf);

        int finalCounterValue = startCounterValue + vChildNb ;

        if (!isInRestartMode()) {
            for (int i = startCounterValue; i < finalCounterValue; i++) {

                String [] argsName = new String[3] ;
                String [] argsValue = new String[3] ;
                argsName[0] = new String("counterName") ; argsValue[0] = counterName ;
                argsName[1] = new String("counterValue") ; argsValue[1] = (new Integer(i)).toString() ;
                argsName[2] = new String("scopeId") ; argsValue[2] = (new Integer(this.scopeId)).toString() ;

                createBPU(childBrancheNb, "forEach", argsName, argsValue) ;
            }
        }

        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " end foreach");
    }

    /**
     * A endScope message was send by the BPU
     *
     * @param newScopeId
     * @param bpuId
     */
    private void endScope(int oldScopeNb) throws BPELEngineException {
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " endscope : " + oldScopeNb);

        ScopeCache.setScopeFinished(oldScopeNb, true);
        cleanHandlers(oldScopeNb);
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " end endscope");
    }

    /**
     * An Invoke Inbound has arrived
     *
     * @param operation
     * @param portType
     * @param partner
     * @param refPartner
     * @param messageId
     * @param oneWay
     * @throws BPELEngineException
     */
    private void invokeInbound(String operation, String portType, String partner, String refPartner, int messageId,boolean oneWay) throws BPELEngineException {
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " invokeInbound : " + operation + " " + portType + " " + partner + " " + refPartner);
        InvokeInbound ii = new InvokeInbound(operation, portType, partner, refPartner,oneWay);
        if (!oneWay) {
            this.invokeInbounds.add(ii);
        }
        // Search for corresponding Receive
        int id = findReceive(operation, portType, partner);
        if (id != -1) {
            if (BPELTrace.isBPIDebugEnabled())
                BPELTrace.traceBPI("BPI " + this.bpiId + " invokeInbound there is a corresponding receive");
            PendingReceive pr = (PendingReceive) this.pendingReceives.get(id);
            int brancheNb = pr.getBrancheNb();

            String[] argsName = new String[2];
            String[] argsValue = new String[2];
            argsName[0] = "messageId";
            argsValue[0] = Integer.toString(messageId);
            argsName[1] = "oldBpuId";
            argsValue[1] = Integer.toString(pr.getBpuId());
            this.scopeId = pr.getScopeId();
            createBPU(brancheNb, "receive",argsName, argsValue);

            // Remove PendingReceive
            this.pendingReceives.remove(pr);
        } else {
            if (BPELTrace.isBPIDebugEnabled())
                BPELTrace.traceBPI("BPI " + this.bpiId + " invokeInbound there is no corresponding receive");
            id = findEvent(operation, portType, partner,this.pendingPicks);
            if (id != -1) {
                if (BPELTrace.isBPIDebugEnabled())
                    BPELTrace.traceBPI("BPI " + this.bpiId + " invokeInbound there is a corresponding pick");
                PendingEvent pe = (PendingEvent) this.pendingPicks.get(id);
                //perhaps the pick has been removed by an alarm or another onMessage
                if (pe != null) {
                    int brancheNb = pe.getFoundOnMessageBrancheNb();
                    String[] argsName = new String[1];
                    String[] argsValue = new String[1];
                    argsName[0] = "messageId";
                    argsValue[0] = Integer.toString(messageId);

                    this.scopeId = pe.getScopeId();
                    BPELTimerManager.getInstance().removeAlarms(this.bpiId,pe);
                    createBPU(brancheNb, "pickOnMessage",argsName, argsValue);

                    // Remove PendingPick
                    this.pendingPicks.remove(pe);
                }
            } else {
                if (BPELTrace.isBPIDebugEnabled())
                    BPELTrace.traceBPI("BPI " + this.bpiId + " invokeInbound there is no corresponding pick");
                id = findEvent(operation, portType, partner,this.pendingEventHandlers);
                if (id != -1) {
                    if (BPELTrace.isBPIDebugEnabled())
                        BPELTrace.traceBPI("BPI " + this.bpiId + " invokeInbound there is a corresponding EventHandler");
                    PendingEvent pe = (PendingEvent) this.pendingEventHandlers.get(id);
                    int brancheNb = pe.getFoundOnMessageBrancheNb();

                    String[] argsName = new String[1];
                    String[] argsValue = new String[1];
                    argsName[0] = "messageId";
                    argsValue[0] = Integer.toString(messageId);

                    this.scopeId = pe.getScopeId();
                    createBPU(brancheNb, "eventOnMessage",argsName, argsValue);
                } else {
                    if (BPELTrace.isBPIDebugEnabled())
                        BPELTrace.traceBPI("BPI " + this.bpiId + " invokeInbound there is no corresponding Event");
                    // if no corresponding Receive, pick or event, store Invoke
                    PendingInvoke pi = new PendingInvoke(operation, portType, partner, messageId);
                    this.pendingInvokes.add(pi);
                }
            }
        }
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " end invokeInbound");
    }

    /**
     * The BPU sent a message telling that he is on a receive tag
     *
     * @param operation
     * @param portType
     * @param partner
     * @param brancheNb
     * @param bpuId
     * @throws BPELEngineException
     */
    private void receive(String operation, String portType, String partner, int brancheNb, int bpuId) throws BPELEngineException {
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " receive");
        // Search for an corresponding invoke
        int id = findInvoke(operation, portType, partner);
        if (id != -1) {
            if (BPELTrace.isBPIDebugEnabled())
                BPELTrace.traceBPI("BPI " + this.bpiId + " receive there is a corresponding invoke");
            PendingInvoke pi = (PendingInvoke) this.pendingInvokes.get(id);
            int messageId = pi.getMessageId();

            // Send message to BPU to tell him to continue
            String[] argsName = new String[2];
            String[] argsValue = new String[2];
            argsName[0] = "messageId";
            argsValue[0] = Integer.toString(messageId);
            argsName[1] = "oldBpuId";
            argsValue[1] = Integer.toString(bpuId);
            if (!isInRestartMode()) {
                createBPU(brancheNb, "receive", argsName, argsValue);
            }
            // Remove PendingInvoke
            this.pendingInvokes.remove(pi);
        } else {
            if (BPELTrace.isBPIDebugEnabled())
                BPELTrace.traceBPI("BPI " + this.bpiId + " receive there is no corresponding invoke");
            // if no corresponding Invoke, store Receive
            PendingReceive pr = new PendingReceive(bpuId,operation, portType, partner, brancheNb, this.scopeId);
            this.pendingReceives.add(pr);
            // Notify the Manager that a call is waited
            addBpiWaiting( portType, operation);
        }
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " end receive");
    }

    /**
     * @param operation
     * @param portType
     * @param partner
     * @param brancheNb
     * @throws BPELEngineException
     */
    private void reply(String operation, String portType, String partner, int brancheNb) throws BPELEngineException {
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " reply, operation = "+operation+", portType = "+portType+", partner = "+partner+", brancheNb = "+brancheNb);
        String refPartner = findRefPartner(operation, portType, partner);
        String[] argsName = new String[1];
        String[] argsValue = new String[1];
        argsName[0] = "refPartner";
        argsValue[0] = refPartner;
        createBPU(brancheNb, "reply", argsName, argsValue);
    }


    /**
     * @param portType
     * @param operation
     * @throws BPELEngineException
     */
    private void addBpiWaiting (String portType, String operation) throws BPELEngineException {
        try {
            ManagerHomeLocal mhl = ManagerUtil.getLocalHome();
            ManagerLocal ml = mhl.create();
            ml.bpiWaiting(this.bpiId, portType, operation);
            ml.remove();
        } catch (NamingException e) {
            throw new BPELEngineException("NamingException BPIMdb addBpiWaiting : " , e);
        } catch (CreateException e) {
            throw new BPELEngineException("CreateException BPIMdb addBpiWaiting : " , e);
        } catch (RemoveException e) {
            throw new BPELEngineException("RemoveException BPIMdb addBpiWaiting : " , e);
        }
    }

    /**
     * End of a BPU
     *
     * @throws BPELEngineException
     */
    private void endBPU() throws BPELEngineException {
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " endBPU");
        this.runningBPUs--;
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " endBPU : remaining BPUs : " + this.runningBPUs);
        if (this.finishing && this.runningBPUs == 0) {
            if (this.finishingReason == null)
                this.finishingReason = BPELConstants.BPI_FINISHED;
            endBPI();
        }
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " end endBPU");
    }

    /**
     * End of a BPI : removes the BPI Entity
     *
     * @throws BPELEngineException
     */
    private void endBPI() throws BPELEngineException {
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " endBPI");
        this.end = true;
        this.lastType = this.currentType;
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " endBPI, canFinish = true");
        notifyManagerEndBPI();
        //choose : setState or remove bpi
        try {
            BPELTransaction tx = null;
            boolean txActive = false;
            try {
                tx = new BPELTransaction();
                tx.beginTx();
                txActive = true;
                BPICache.setBPIState(this.bpiId, this.finishingReason);
                CSCache.removeCS(this.bpiId);
                tx.commitTx();
                txActive = false;
            } finally {
                if (txActive) {
                    tx.roolbackTx();
                }
            }
            flushBPI();
            if (this.engine != null) {
                this.engine.remove();
                this.engine = null;
            }
            if (this.monitor != null) {
                this.monitor.remove();
                this.monitor = null;
            }
        } catch (BPELUtilException e) {
            throw new BPELEngineException("BPELUtilException : BPI.endBPI : "+e,e);
        } catch (RemoveException e) {
            throw new BPELEngineException("RemoveException : BPI.endBPI : "+e,e);
        } finally {
            // This has to be ran whatever happens
            BPICache.removeBPI(this.bpiId);
            BPIMap.removeBPI(this.bpiId);
        }
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " end endBPI");
    }

    /**
     * @throws BPELEngineException
     */
    private void flushBPI() throws BPELEngineException {
        BPIData bpid = BPICache.findBPI(this.bpiId);
        if (BPELEngine.flushInstance()) {
            BPICache.flushBPIToXMLFile(bpid);
            BPICache.storeScopesVariables(bpid);
        }
        if (BPELEngine.flushMessages()) {
            BPICache.storeMessages(bpid);
        }
    }

    /**
     * @throws BPELEngineException
     */
    private void notifyManagerEndBPI() throws BPELEngineException {
        // Notify the Manager that the BPI is finished
        try {
            ManagerHomeLocal mhl = ManagerUtil.getLocalHome();
            ManagerLocal ml = mhl.create();
            ml.endBPI(this.bpiId);
            ml.remove();
        } catch (NamingException e) {
            throw new BPELEngineException("NamingException BPIMdb notifyManagerEndBPI : " , e);
        } catch (CreateException e) {
            throw new BPELEngineException("CreateException BPIMdb notifyManagerEndBPI : " , e);
        } catch (RemoveException e) {
            throw new BPELEngineException("RemoveException BPIMdb notifyManagerEndBPI : " , e);
        }
    }

    /**
     * Returns the RunningFlow that has the statementNb searched
     *
     * @param statementNb
     * @return
     */
    private RunningFlow findFlow(int statementNb) {
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " findFlow");
        boolean found = false;
        int i = 0;
        while (i < this.runningFlows.size() && !found) {
            found = (((RunningFlow) this.runningFlows.get(i)).getStatementNb() == statementNb);
            i++;
        }
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " end findFlow found = " + found);
        if (found)
            return (RunningFlow) this.runningFlows.get(i - 1);
        return null;
    }

    /**
     * findForEach
     * Returns the RunningForEach that has the statementNb
     * and the scopeId searched
     *
     * @param scopeId
     * @param statementNb
     */
    private RunningForEach findForEach(int scopeId, int statementNb) {
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " findForEach");
        RunningForEach result = null;

        int i = 0;
        while (i < this.runningForEach.size()) {
            RunningForEach rf= (RunningForEach) this.runningForEach.get(i) ;
            if ((rf.getStatementNb() == statementNb) && (rf.getScopeId() == scopeId)) {
                result = rf;
                break;
            }
            i++;
        }
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " end findForEach found = " + (result!=null));
        return result;
    }

    /**
     * Looks in vector receives for a Receive corresponding to the Invoke
     *
     * @param operation
     * @param portType
     * @param partner
     * @return
     */
    private int findReceive(String operation, String portType, String partner) {
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " findReceive");
        int found = -1;
        PendingReceive p = null;
        PendingReceive p2 = new PendingReceive(operation, portType, partner);
        for (int i = 0; i < this.pendingReceives.size() && found == -1; i++) {
            p = (PendingReceive) this.pendingReceives.get(i);
            if (p.equals(p2)) {
                found = i;
            }
        }
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " end findReceive found = " + found);
        return found;
    }

    /**
     * @param eId
     * @param v
     * @return
     */
    private boolean removeEvent(int eId, Vector v) {
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " removeEvent");
        boolean found = false;
        int i = 0;
        while (i < v.size()) {
            PendingEvent pe = (PendingEvent) v.get(i);
            if (pe.getId() == eId) {
                found = true;
                v.remove(i);
                break;
            }
            i++;
        }
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " end removeEvent found = " + found);
        return found;
    }

    /**
     * @param scopeId
     */
    private void cleanHandlers(int scopeId) {
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " cleanHandlers");
        //clean EventHandlers
        Vector pehTmp = (Vector) this.pendingEventHandlers.clone();
        for (Enumeration elmts = pehTmp.elements(); elmts.hasMoreElements();) {
            PendingEvent p = (PendingEvent) elmts.nextElement();
            if (p.getScopeId() == scopeId) {
                this.pendingEventHandlers.remove(p);
            }
        }
        //Clean FaultHandlers
        Vector pfhTmp = (Vector) this.pendingFaultHandlers.clone();
        for (Enumeration elmts = pfhTmp.elements(); elmts.hasMoreElements();) {
            PendingFaultHandler p = (PendingFaultHandler) elmts.nextElement();
            if (p.getScopeId() == scopeId) {
                this.pendingFaultHandlers.remove(p);
            }
        }

        // Activate corresponding Compensation Handlers
        Vector pchTmp = (Vector) this.pendingCompensationHandlers.clone();
        for (Enumeration elmts = pchTmp.elements(); elmts.hasMoreElements();) {
            PendingCompensationHandler p = (PendingCompensationHandler) elmts.nextElement();
            if (p.getScopeId() == scopeId) {
                this.pendingCompensationHandlers.remove(p);
                p.activate();
                this.pendingCompensationHandlers.add(p);
            }
        }
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " end cleanHandlers");
    }

    /**
     * Looks in vector pendingFaultHandlers for a FaultHandler for the scope
     *
     * @param scopeId
     * @return
     */
    private PendingFaultHandler findFaultHandler(int scopeId) {
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " findFaultHandler : " + scopeId);
        boolean found = false;
        PendingFaultHandler p = null;
        PendingFaultHandler result = null;
        for (int i = 0; i < this.pendingFaultHandlers.size() && !found; i++) {
            p = (PendingFaultHandler) this.pendingFaultHandlers.get(i);
            found = p.equals(scopeId);
            if (found) result = p;
        }
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " end findFaultHandler found = " + found);
        return result;
    }

    /**
     * Looks in vector pendingCompensationHandlers for a CompensationHandler
     * for the scope
     *
     * @param scopeId
     * @return
     */
    private PendingCompensationHandler findCompensationHandler(int scopeId) {
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " findCompensationHandler : " + scopeId);
        boolean found = false;
        PendingCompensationHandler p = null;
        PendingCompensationHandler result = null;
        for (int i = 0; i < this.pendingCompensationHandlers.size() && !found; i++) {
            if (BPELTrace.isBPIDebugEnabled())
                BPELTrace.traceBPI("BPI " + this.bpiId + " findCompensationHandler ici");
            p = (PendingCompensationHandler) this.pendingCompensationHandlers.get(i);
            found = p.equals(scopeId);
            if (found) result = p;
        }
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " end findCompensationHandler found = " + found);
        return result;
    }

    /**
     * @param operation
     * @param portType
     * @param partner
     * @param v
     * @return
     */
    private int findEvent(String operation, String portType, String partner, Vector v) {
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " findEvent");
        int found = -1;
        PendingEvent pe = null;
        for (int i = 0; i < v.size(); i++) {
            pe = (PendingEvent) v.get(i);
            if (pe.equals(operation, portType, partner)) {
                found = i;
                break;
            }
        }
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " end findEvent found = " + found);
        return found;
    }

    /**
     * @param eventStatementNb
     * @param scopeId
     * @param v
     * @return
     */
    private PendingEvent findEvent(int eventStatementNb, int scopeId, Vector v) {
        PendingEvent result = null;
        for (int i = 0; i < v.size() && result==null; i++) {
            PendingEvent pe = (PendingEvent) v.get(i);
            if (pe.getScopeId()==scopeId && pe.getStamentNb()==eventStatementNb) {
                result = pe;
                break;
            }
        }
        return result;
    }
    /**
     * Looks in vector invokes for a Invoke corresponding to the Receive
     *
     * @param operation : the name of the operation
     * @param portType : the name of the port type
     * @param partner : the name of the partner
     * @return
     */
    private int findInvoke(String operation, String portType, String partner) {
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " findInvoke");
        int found = -1;
        PendingInvoke p = null;
        PendingInvoke p2 = new PendingInvoke(operation, portType, partner);
        for (int i = 0; i < this.pendingInvokes.size() && found == -1; i++) {
            p = (PendingInvoke) this.pendingInvokes.get(i);
            if (p.equals(p2)) {
                found = i;
            }
        }
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " end findInvoke found = " + found);
        return found;
    }

    /**
     * Looks in vector invokeInbounds for the corresponding refPartner
     *
     * @param operation : the name of the operation
     * @param portType : the name of the port type
     * @param partner : the name of the partner
     * @return : the Reference to the partner
     */
    private String findRefPartner(String operation, String portType, String partner) {
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " findRefPartner, operation = "+operation+", portType = "+portType+", partner = "+partner);
        String found = null;
        InvokeInbound ii = null;
        InvokeInbound ii2 = new InvokeInbound(operation, portType, partner,true);
        for (int i = 0; i < this.invokeInbounds.size(); i++) {
            ii = (InvokeInbound) this.invokeInbounds.get(i);
            if (ii.equals(ii2)) {
                found = ii.getRefPartner();
                this.invokeInbounds.remove(i);
                break;
            }
        }
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " end findRefPartner found = " + found);
        return found;
    }

    /**
     * @param brancheNb : the brancheNb on which the BPU starts to work
     * @param type : the type of statement
     * @param argsName : the arguments names of the message sent to the BPU
     * @param argsValue : the argument values
     * @return the Id of the BPU
     * @throws BPELEngineException
     */
    private int createBPU(int brancheNb, String type, String[] argsName, Object[] argsValue) throws BPELEngineException {
        return createBPU(brancheNb,type,argsName,argsValue,0);
    }

    /**
     * Creates a new BPU and tell him to start analysing brancheNb
     *
     * @param brancheNb : the brancheNb on which the BPU starts to work
     * @param type : the type of statement
     * @param argsName : the arguments names of the message sent to the BPU
     * @param argsValue : the argument values
     * @param timeToLive
     * @return the Id of the BPU
     * @throws BPELEngineException
     */
    private int createBPU(int brancheNb, String type, String[] argsName, Object[] argsValue, long timeToLive) throws BPELEngineException {
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " createBPU : brancheNb : " + brancheNb+", type = "+type);
        int bpuId = BPUCache.createBPU(this.process, this.bpiId, BPELConstants.BPU_ACTIVE,this.scopeId);
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " createBPU : bpuId : " + bpuId);
        this.runningBPUs++;
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " createBPU : BPUs started : " + this.runningBPUs);
        try {
            QueueConnection connection = JmsUtil.getConnection();
            Session session = connection.createQueueSession(false, Session.AUTO_ACKNOWLEDGE);
            QueueSender prod = ((QueueSession) session).createSender(JmsUtil.getBPUQueue());
            prod.setTimeToLive(timeToLive);
            MapMessage message = session.createMapMessage();
            message.setStringProperty("destination", "BPU");
            message.setString("source", "BPI");
            message.setIntProperty("bpuId", bpuId);
            message.setInt("bpiId", this.bpiId);
            message.setString("type", type);
            message.setInt("brancheNb", brancheNb);
            message.setInt("scopeId", this.scopeId);
            // The process name can be easily found in the BPUMdbBean by doing
            // "this.bpelProcess.getName()" after "analyseMessage()" has been called
            //message.setString("process", this.process);
            message.setString("targetNamespace", this.targetNamespace);
            if (argsName != null) {
                for (int i = 0; i < argsName.length; i++) {
                    if (argsValue[i] != null &&  (argsValue[i] instanceof String) ) {
                        message.setString(argsName[i], (String)argsValue[i]);
                    } else {
                        message.setBytes(argsName[i], Util.getBytesFromObject(argsValue[i]));
                    }
                }
            }
            if (BPELTrace.isBPIDebugEnabled())
                BPELTrace.traceBPI(this.bpiId+" "+System.currentTimeMillis()+" BPI BPU "+bpuId+" send before BPUQueue Before sending a message in BPEL Engine (BPI type="+type+") : To = BPUQueue..."+message);
            if (isRestarting()) {
                if (BPELTrace.isRestartDebugEnabled())
                    BPELTrace.traceRestart("BPI " + this.bpiId + " createBPU : addBPU in messagesToSendToBpu, bpuId = "+bpuId+", brancheNb : " + brancheNb+", type = "+type);
                this.messagesToSendToBpu.addElement(message);
            } else {
                if (BPELTrace.isRestartDebugEnabled())
                    BPELTrace.traceRestart("BPI " + this.bpiId + " createBPU : bpuId = "+bpuId+", brancheNb : " + brancheNb+", type = "+type);
                prod.send(message);
            }
            if (BPELTrace.isBPIDebugEnabled())
                BPELTrace.traceBPI(this.bpiId+" "+System.currentTimeMillis()+" BPI BPU "+bpuId+" send after BPUQueue Sending a message in BPEL Engine (BPI) : To = BPUQueue..."+message);
            prod.close();
            session.close();
            connection.close();
            prod = null;
            session = null;
            connection = null;
            message = null;
        } catch (BPELUtilException e) {
            throw new BPELEngineException("BPELUtilException BPIMdb createBPU :\n" , e);
        } catch (JMSException e) {
            throw new BPELEngineException("JMSException BPIMdb sendMessageBPU :\n" , e);
        }
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " end createBPU");
        return bpuId;
    }

    /**
     * @throws BPELEngineException
     */
    private void initEngine() throws BPELEngineException {
        if (this.engine == null) {
            try {
                EngineHomeLocal ehl = EngineUtil.getLocalHome();
                this.engine = ehl.create();
            } catch (javax.naming.NamingException e) {
                throw new BPELEngineException("NamingException BPIMdb initEngine : " , e);
            } catch (javax.ejb.CreateException e) {
                throw new BPELEngineException("CreateException BPIMdb initEngine : " , e);
            }
        }
    }

    /**
     * @param varName : Name of the variable
     * @return a VariableData containing the Variable
     * @throws BPELEngineException
     */
    private VariableData getVariable(String varName) throws BPELEngineException {
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPI " + this.bpiId + " getVariable : varName : " + varName);
        return this.engine.getVar(this.bpiId, varName, this.scopeId);
    }

    /*
     * #############################################################################
     * ##################################   RESTART   ##############################
     * #############################################################################
     */

    private boolean restartMode = false;
    private boolean restarting = false;
    private Document restartDoc = null;
    private HashMap finishedTargets = null;
    private Vector messagesToSendToBpu = null;
    private HashMap insertedTargets = null;
    private boolean isRestarting() {
        return this.restarting;
    }

    private boolean isInRestartMode() {
        return this.restartMode;
    }

    private void restart() throws BPELEngineException {
        //restart an instance is composed by different steps
        // 1) place the memory as it was before the crash
        // 2) restart each bpu wich were running before crash
        if (BPELTrace.isRestartDebugEnabled()) {
            BPELTrace.traceRestart("################################################################");
            BPELTrace.traceRestart("################################################################");
            BPELTrace.traceRestart("BPI " + this.bpiId + " restart");
            BPELTrace.traceRestart("################################################################");
            BPELTrace.traceRestart("################################################################");
        }
        this.restartMode = true;
        initEngine();
        initMonitoring();
        initScopeHomeLocal();
        this.finishedTargets = new HashMap();
        this.insertedTargets = new HashMap();
        this.messagesToSendToBpu = new Vector();
        Vector endFlows = null;
        Vector targets = null;
        try {
            //monitoring statement tree
            this.restartDoc = this.monitor.getStatementsTree(this.bpiId);
            //all endflows performed by this bpi before the crash
            endFlows = this.monitor.getEndFlows(this.bpiId);
            //all targets performed by this bpi before the crash
            targets = this.monitor.getTargets(this.bpiId);
        } catch (BPELMonitoringException e) {
            throw new BPELEngineException("BPELMonitoringException BPIMdb restart : " , e);
        }
        removeUnusedBpus();
        Collection bpus = getBpusByState(this.bpiId,BPELConstants.BPU_ACTIVE);

        fillPendingInvokes();
        fillPendingTargets(targets);
        fillPendingReceives();
        boolean stop = readExecutionTree(this.restartDoc);
        if (!stop) {
            this.restartMode = false;
            this.restarting = true;
            invokeInbounds();
            removeDuplicateRunningFlows();
            putChildFinishedOnFlow(endFlows);

            for (int i = 0;i<this.pendingTargets.size();i++) {
                PendingTarget pt = (PendingTarget) this.pendingTargets.get(i);
                if (BPELTrace.isRestartDebugEnabled())
                    BPELTrace.traceRestart("Restart, pt = "+pt.getBrancheNb());
            }
            for (int i = 0;i<this.pendingSources.size();i++) {
                PendingSource ps = (PendingSource) this.pendingSources.get(i);
                if (BPELTrace.isRestartDebugEnabled())
                    BPELTrace.traceRestart("Restart, ps = "+ps.getLinkName());
            }
            restartBPUs(bpus);
            this.restarting = false;
            launchAllBpus();
            this.needStart = false;
        }
        /*
        try {
            this.engine.remove();
            this.engine = null;
            this.monitor.remove();
            this.monitor = null;
        } catch (RemoveException e) {
            throw new BPELEngineException("RemoveException BPI restart : " , e);
        }
         */
        if (BPELTrace.isRestartDebugEnabled())
            BPELTrace.traceRestart("BPI " + this.bpiId + " end restart");
    }

    private void fillPendingReceives() throws BPELEngineException {
        Collection waitingBpus = getBpusByState(this.bpiId,null);
        for (Iterator it = waitingBpus.iterator();it.hasNext();) {
            BPUEntityLocal bpu = (BPUEntityLocal) it.next();
            String state = bpu.getState();
            if (state.startsWith(BPELConstants.BPU_WAITING_RECEIVE)) {
                int receiveNb = Integer.parseInt(state.substring(BPELConstants.BPU_WAITING_RECEIVE.length()));
                Receive receive = (Receive) this.bpelProcess.findStatement(receiveNb);
                String operation = receive.getReceiveOperation();
                String portType = receive.getReceivePortType();
                String partner = receive.getReceivePartnerLink();
                Node receiveNode = getNodeFromStatementNb(this.restartDoc,receiveNb);
                int scopeNb = this.bpelProcess.findStatementOfType("Scope",receive);
                int iter = getLoopsNb(receiveNode);
                int scopeId = getScopeId(scopeNb,iter+1);
                if (scopeId == -1) {
                    scopeId = getScopeId(scopeNb,iter);
                }
                PendingReceive pr = new PendingReceive(bpu.getId().intValue(),operation, portType, partner, receiveNb, scopeId);
                this.pendingReceives.add(pr);
                // Notify the Manager that a call is waited
                addBpiWaiting( portType, operation);
            }
        }
    }

    private void launchAllBpus() throws BPELEngineException {
        if (BPELTrace.isRestartDebugEnabled())
            BPELTrace.traceRestart("BPI " + this.bpiId + " launchAllBpus : messagesToSendToBpu.size() : " + this.messagesToSendToBpu.size());
        try {
            QueueConnection connection = JmsUtil.getConnection();
            Session session = connection.createQueueSession(false, Session.AUTO_ACKNOWLEDGE);
            QueueSender prod = ((QueueSession) session).createSender(JmsUtil.getBPUQueue());

            for (int i=0;i<this.messagesToSendToBpu.size();i++) {
                MapMessage message = (MapMessage) this.messagesToSendToBpu.get(i);
                prod.send(message);
                if (BPELTrace.isRestartDebugEnabled())
                    BPELTrace.traceRestart("BPI " + this.bpiId + " launchAllBpus : brancheNb = "+message.getInt("brancheNb")+", bpuId = "+message.getIntProperty("bpuId"));
                this.runningBPUs++;
                if (BPELTrace.isRestartDebugEnabled())
                    BPELTrace.traceRestart("BPI " + this.bpiId + " launchAllBpus : BPUs started : " + this.runningBPUs);
                message = null;
            }
            prod.close();
            session.close();
            connection.close();
            prod = null;
            session = null;
            connection = null;
        } catch (BPELUtilException e) {
            throw new BPELEngineException("NamingException BPIMdb launchAllBpus :\n" , e);
        } catch (JMSException e) {
            throw new BPELEngineException("JMSException BPIMdb launchAllBpus :\n" , e);
        }

        if (BPELTrace.isRestartDebugEnabled())
            BPELTrace.traceRestart("BPI " + this.bpiId + " end launchAllBpus");
    }

    private void removeDuplicateRunningFlows() {
        HashMap hm = new HashMap();
        for (int i=0;i<this.runningFlows.size();i++) {
            RunningFlow rf = (RunningFlow) this.runningFlows.get(i);
            int flowNb = rf.getStatementNb();
            if (hm.containsKey(new Integer(flowNb))) {
                this.runningFlows.remove(i);
                i--;
            } else {
                hm.put(new Integer(flowNb),new Object());
            }
        }
    }

    private void removeUnusedBpus() throws BPELEngineException {
        if (BPELTrace.isRestartDebugEnabled())
            BPELTrace.traceRestart("BPI " + this.bpiId + " removeUnusedBpus");
        //some bpus are created, they didn't had time to execute
        //completely one statement before the crash : these one must be
        //deleted. Some new one will be restarted instead
        try {
            BPUEntityHomeLocal bpuhl = BPUEntityBeanUtil.getLocalHome();
            Collection bpus = bpuhl.findByBpiId(new Integer(this.bpiId));
            for (Iterator it = bpus.iterator();it.hasNext();) {
                BPUEntityLocal bpu = (BPUEntityLocal) it.next();
                if (bpu.getMonitor().size() == 0) {
                    if (BPELTrace.isRestartDebugEnabled())
                        BPELTrace.traceRestart("BPI " + this.bpiId + " removeUnusedBpus, remove bpuEntity : bpuId="+bpu.getId());
                    bpu.remove();
                } else if (bpu.getStatementNb() == -1) {
                    //if the bpu has one statement in its monitor vector
                    //but the statementNb was not set, we set it
                    Vector m = bpu.getMonitor();
                    String st = null;
                    for (int i=0;i<m.size();i++) {
                        st = (String) m.get(i);
                        String[] split = st.split(BPELConstants.MONITORING_INFO_SEP);
                        if (split.length >= 3) {
                            bpu.setStatementNb(Integer.parseInt(split[1]));
                        }
                    }
                    if (bpu.getStatementNb() == -1) {
                        bpu.setState(BPELConstants.BPU_FINISHED);
                    }
                }
            }
        } catch (FinderException e) {
            //no bpu for this bpi
        } catch (javax.ejb.RemoveException ex) {
            throw new BPELEngineException("RemoveException BPI removeUnusedBpus : " , ex);
        } catch (NamingException e) {
            throw new BPELEngineException("NamingException BPI removeUnusedBpus : " , e);
        }
    }

    private void fillPendingTargets(Vector targets) throws BPELEngineException {
        if (BPELTrace.isRestartDebugEnabled())
            BPELTrace.traceRestart("BPI " + this.bpiId + " fillPendingTargets");
        //Target statement is diffucult to rebuild as it is different from a source
        //a PendingSource correspond to one source statement
        //a PendingTarget is a set of many Target statements : all targets of a given activity
        //to fill this BPI pendingTargets Vector, we put in it each targets which were put
        //in it during the entire execution of this BPI before the crash
        //then some of them were finished, they will be removed by the restartAction
        //on a source statement
        //We have to manage loops and we have to find the scopeId and the bpuId associated
        //to each pendingtarget


        for (int i=0;i<targets.size();i++) {
            Integer activityNb = (Integer) targets.get(i);
            if (this.insertedTargets.containsKey(activityNb)) {
                Integer integ = (Integer) this.insertedTargets.get(activityNb);
                this.insertedTargets.put(activityNb,new Integer(integ.intValue()+1));
            } else {
                this.insertedTargets.put(activityNb,new Integer(1));
            }
        }

        Set s = this.insertedTargets.entrySet();
        for (Iterator i = s.iterator(); i.hasNext();) {
            Map.Entry e = (Map.Entry) i.next();
            int activityNb = ((Integer) e.getKey()).intValue();
            int loops = ((Integer) e.getValue()).intValue();
            Activity a = (Activity) this.bpelProcess.findStatement(activityNb);
            for (int j=1;j<=loops;j++) {
                int scopeNb = this.bpelProcess.findStatementOfType("Scope",a);
                int aScopeId = getScopeId(scopeNb,j);
                int aBpuId = getTargetBpuId(this.restartDoc,activityNb,j);
                PendingTarget pt = buildPendingTarget(a,aScopeId,aBpuId);
                if (BPELTrace.isRestartDebugEnabled())
                    BPELTrace.traceRestart("BPI " + this.bpiId + " fillPendingTargets, add pt = "+pt.getBrancheNb());
                this.pendingTargets.add(pt);
            }
        }
    }

    private int getTargetBpuId(Node n, int activityNb, int loopNb) {
        //get the bpuId which executed a given activity thanks to its brancheNband loop nb
        int result;
        Node activityN = getNodeWithNb(n,activityNb);
        int totalLoops = getLoopsNb(activityN);
        if (loopNb == totalLoops) {
            result = getBpuId(activityN);
        } else {
            Node father = activityN.getParentNode();
            result = getBpuId(father);
        }

        return result;
    }

    private PendingTarget buildPendingTarget(Activity a, int aScopeId, int aBpuId) {
        //build a pendingTarget
        PendingTarget pt = null;

        int activityNb = a.getNb().intValue();
        Vector tmp1 = new Vector();
        Vector tmp2 = new Vector();
        int childrenNb = a.getChildNb();
        int targetNb = 0;
        org.objectweb.orchestra.tree.Statement[] tab = a.getChilds();
        org.objectweb.orchestra.tree.Target t;
        for (int i = 0; i < childrenNb; i++) {
            if (tab[i].getType().equals("Target")) {
                targetNb++;
                t = (org.objectweb.orchestra.tree.Target) tab[i];
                tmp1.addElement(t.getNb());
                tmp2.addElement(t.getLinkName());
            }
        }

        int childNb = targetNb;
        String[] targetLinkNames = new String[childNb];
        int[] branchesNb = new int[childNb];
        for (int i = 0; i < targetNb; i++) {
            branchesNb[i] = ((Integer)tmp1.get(i)).intValue();
            targetLinkNames[i] = (String)tmp2.get(i);
        }

        pt = new PendingTarget(this.bpiId,this.process,aBpuId,activityNb, aScopeId, childNb, targetLinkNames, a.getJoinCondition(),branchesNb);

        return pt;
    }

    private void fillPendingInvokes() throws BPELEngineException {
        //pendingInvoke are all messages received by the manager and not already used
        //here we put in it all messages that were used, so the restartAction on
        //a receive or pick or eventhandler will remove it
        //once all restartActions will be done, we will move to !restartMode
        //and make some invokeInbounds as it is done in the normal execution
        //By doing that after restartActions, if the corresponding invokeinbound free
        //a receive,pick... it will start a new BPU....
        if (BPELTrace.isRestartDebugEnabled())
            BPELTrace.traceRestart("BPI " + this.bpiId + " fillPendingInvokes");
        try {
            MessageHomeLocal hl = MessageBeanUtil.getLocalHome();
            Collection messages = hl.findByBPIWithStatementNb(this.bpiId);

            for (Iterator i = messages.iterator(); i.hasNext();) {
                MessageLocal mess = (MessageLocal) i.next();
                if (!mess.getFault()) {
                    //it is a normal message and it was consumed
                    String operation = mess.getOperation();
                    String portType = mess.getPortType();
                    String partner = mess.getPartner();
                    String refPartner = mess.getRefPartner();
                    int messageId = mess.getId().intValue();
                    boolean oneWay = isOneWay();
                    if (BPELTrace.isRestartDebugEnabled())
                        BPELTrace.traceRestart("BPI " + this.bpiId + " fillPendingInvokes, addInvoke, operation="+operation+", portType="+portType+", partner="+partner+", refPartner="+refPartner+", messageId="+messageId);
                    InvokeInbound ii = new InvokeInbound(operation, portType, partner, refPartner,oneWay);
                    this.invokeInbounds.add(ii);
                    PendingInvoke pi = new PendingInvoke(operation, portType, partner, messageId);
                    this.pendingInvokes.add(pi);
                }
            }
        } catch (NamingException e) {
            throw new BPELEngineException("NamingException BPI fillPendingInvokes :\n" , e);
        } catch (FinderException e) {
            //no message for this BPI
        }
    }

    private void invokeInbounds() throws BPELEngineException {
        //for each message not already used, we perform an invokeinbound
        //when this method is called, we are not in restartMode, so if corresponding
        //receive,pick... is found, a new BPU will be launched
        if (BPELTrace.isRestartDebugEnabled())
            BPELTrace.traceRestart("BPI " + this.bpiId + " invokeInbounds");
        try {
            MessageHomeLocal hl = MessageBeanUtil.getLocalHome();
            Collection messages = hl.findByBPIAndStatementNb(this.bpiId,-1);

            for (Iterator i = messages.iterator(); i.hasNext();) {
                MessageLocal mess = (MessageLocal) i.next();
                if (!mess.getFault()) {
                    String operation = mess.getOperation();
                    String portType = mess.getPortType();
                    String partner = mess.getPartner();
                    String refPartner = mess.getRefPartner();
                    int messageId = mess.getId().intValue();
                    boolean oneWay = isOneWay();
                    invokeInbound(operation, portType, partner, refPartner, messageId,oneWay);
                }
            }
        } catch (NamingException e) {
            throw new BPELEngineException("NamingException ManagerBean restartEngine :\n" , e);
        } catch (FinderException e) {
            //no message for this BPI
        }
    }

    private boolean isOneWay() {
        boolean result = true;
        Statement[] allS = this.bpelProcess.getTabStatement();
        // not i = 0 as their is no statement with number = 0
        for (int i=1;i<allS.length;i++) {
            Statement s = allS[i];
            if (s.getType().equals("Reply")) {
                result = false;
                break;
            }
        }
        return result;
    }

    //if the last statement executed is one of the following, when restarting the bpu we have to
    //put downAndNext in the type
    private String[] restartDownAndNext = {"Assign",
            "Catch",
            "CatchAll",
            "Compensate",
            "CorrelationSet",
            "CorrelationSets",
            "Empty",
            "Link",
            "Links",
            "OnAlarm",
            "OnMessage",
            "Partner",
            "Partners",
            "PartnerLink",
            "PartnerLinks",
            "Scope",
            "Sequence",
            "Variable",
            "Variables",
            "Case",
            "Otherwise"
    };

    //if the last statement executed is one of the following, when restarting the bpu we have to
    //put nextStatement in the type
    private String[] restartNext = {"CompensationHandler",
            "Copy",
            "Receive",
            "Reply",
            "Wait",
            "FaultHandler"
    };

    private HashMap endScopes = null;

    private boolean contains(String[] tab, String st) {
        //return true if tab contains a string equals to st
        boolean result = false;
        for (int i=0;i<tab.length;i++) {
            if (tab[i].equals(st)) {
                result = true;
                break;
            }
        }
        return result;
    }

    private boolean readExecutionTree(Node n) throws BPELEngineException {
        //read the monitoring execution tree
        //for each statement executed, perform eventually a restartAction
        this.endScopes = new HashMap();
        boolean stop = false;
        NodeList children = n.getChildNodes();
        for (int i=0;i<children.getLength();i++) {
            Node child = children.item(i);
            boolean executed = getExecuted(child);
            if (executed) {
                String type = getType(child);
                if (type != null) {
                    stop = executeRetartAction(child,type);
                }
            }
            if (stop) {
                break;
            }
            if (child.hasChildNodes()) {
                readExecutionTree(child);
            }
        }
        return stop;
    }

    private void manageEnds(int loopsNb,int brancheNb) throws BPELEngineException{
        //perform endScope
        Integer integ = new Integer(brancheNb);
        if (this.endScopes.containsKey(integ)) {
            for (int j=0;j<loopsNb;j++) {
                int finishedScopeId = getScopeId(((Integer)this.endScopes.get(integ)).intValue(),j);
                endScope(finishedScopeId);
            }
        }
    }

    private boolean executeRetartAction(Node n, String type) throws BPELEngineException {
        //in order to use a maximum of existing code, the idea of the restartAction
        //is to do the same thing as during the normal execution but without modifying
        //the db and without launching bpus
        //with this mechanism, all the memory will be filled as if were are in normal execution
        //Some statements have no impact on the BPI memory, so nothing is done
        boolean stop = false;
        int brancheNb = getBrancheNb(n);
        int loopsNb = getLoopsNb(n);
        int bpuId = getBpuId(n);
        if (BPELTrace.isRestartDebugEnabled())
            BPELTrace.traceRestart("BPI " + this.bpiId + " executeRetartAction, type = "+type+", brancheNb = "+brancheNb+", bpuId = "+bpuId);

        manageEnds(loopsNb,brancheNb);
        Statement s = this.bpelProcess.findStatement(brancheNb);
        int scopeNb = this.bpelProcess.findStatementOfType("Scope",s);

        for (int j=0;j<loopsNb;j++) {
            this.scopeId = getScopeId(scopeNb,j);
            if (type.equals("Receive")) {
                //do the normal code without launching a new BPU
                Receive r = (Receive) s;
                String operation = r.getReceiveOperation();
                String portType = r.getReceivePortType();
                String partner = r.getReceivePartnerLink();
                receive(operation, portType, partner, brancheNb, bpuId);
            } else if (type.equals("Flow")) {
                //do the normal code without launching a new BPU
                Flow f = (Flow) s;
                //all children of the flow (excluding source/target)
                int childrenNb = f.getChildNb();
                Statement[] tab = f.getChilds();
                int childNb = 0;
                Vector tmp = new Vector();
                for (int i = 0; i < childrenNb; i++) {
                    if (!tab[i].getType().equals("Target") && !tab[i].getType().equals("Source")) {
                        tmp.addElement(tab[i].getNb());
                        childNb++;
                    }
                }
                int[] childs = new int[childNb];
                for (int i = 0; i < tmp.size(); i++) {
                    childs[i] = ((Integer)tmp.get(i)).intValue();
                }
                flow(brancheNb, childNb, childs);
            } else if (type.equals("Scope")) {
                //if the scope has a next statement we put it in endScopes
                //if the scope next statement was executed, we will perform an endScope
                //for each time it was executed (done by manageEnds)
                if (s.getNext() != null) {
                    this.endScopes.put(s.getNext().getNb(),s.getNb());
                }
            } else if (type.equals("Source")) {
                Source so = (Source)s;
                String linkName = so.getLinkName();
                String transitionCondition = getTransitionCondition(n,j+1);
                //here many cases can occurs,
                source(linkName, transitionCondition);
            } else if (type.equals("Pick")) {
                //do the normal code without launching a new BPU, and removing expired alarms
                event(brancheNb, bpuId,"pick");
            } else if (type.equals("EventHandler")) {
                //do the normal code without launching a new BPU, and removing expired alarms
                event(brancheNb, bpuId,"event");
            } else if (type.equals("Terminate")) {
                //stop this bpi,
                //do not launch messages to bpus as they are not already restarted
                terminate();
                stop = true;
                break;
            } else if (type.equals("FaultHandler")) {
                //do the normal code
                faultHandler(brancheNb);
            } else if (type.equals("CompensationHandler")) {
                //do the normal code
                compensationHandler(brancheNb);
            } else if (type.equals("Throw")) {
                //Throw t = (Throw) s;
                //String faultName = t.getFaultName();
                //String variable = t.getFaultVariable();
                //********************* TO DO ***********************
                //fault(scopeId, faultName, variable);
            } else if (type.equals("Invoke")) {
                MessageData md = getFaultMessage (brancheNb, this.bpiId, j);
                if (md != null) {
                    //String faultName = md.getFaultName();
                    //String faultMessage = md.getFaultMessage();
                    //int messageId = md.getMessageId();
                    //********************* TO DO ***********************
                    //fault(scopeId, faultName, faultMessage, messageId);
                }
            } else if (type.equals("Compensate")) {
                //Compensate c = (Compensate)s;
                //String scopeName = c.getScopeName();
                //********************* TO DO ***********************
                //compensate(scopeId, scopeName);
            }
        }
        return stop;
    }

    private String getTransitionCondition(Node n,int loopNb) {
        //get the value obtained at execution time for this transition condition
        //thanks to monitoring
        String result = getAttibuteValue(n,"transitionConditionEvaluation_loop"+loopNb);
        return result;
    }

    private boolean getSuccess(Node n,int loopNb) {
        //get the value obtained at execution time for this invoke
        //thanks to monitoring
        String result = getAttibuteValue(n,"success_loop"+loopNb);
        return result.equals("true");
    }

    private MessageData getFaultMessage (int invokeNb, int bpiId, int loopNb) throws BPELEngineException {
        //return null if no fault occured for this invoke at the given loop
        //if a fault occured, a corresponding message was associated to this invoke, return it
        MessageData md = null;
        Vector messages = null;
        try {
            messages = MessageCache.findByBPIAndStatementNb(bpiId,invokeNb);
        } catch (BPELEngineException e) {
            if (!e.getCause().getClass().equals(FinderException.class)) {
                throw e;
            }
        }
        if (messages != null && messages.size()>0) {
            for (int i = 0; i < messages.size()-1; i++) {
                int im = i;
                for (int j = i+1 ; j < messages.size(); j++) {
                    MessageData mdj = (MessageData)messages.get(j);
                    MessageData mdim = (MessageData)messages.get(im);
                    if (mdj.getDate() < mdim.getDate()) {
                        im = j;
                    }
                }
                Object tmp = messages.get(i);
                Object tmp2 = messages.get(im);
                messages.remove(i);
                messages.add(i,tmp2);
                messages.remove(im);
                messages.add(im,tmp);

            }
            /*TEST*/
            long before = 0;
            for (int i = 0; i < messages.size()-1; i++) {
                MessageData m = (MessageData)messages.get(i);
                if (m.getDate()<before) {
                    System.out.println("************** BPI getFaultMessage NOT SORTED **************");
                    if (BPELTrace.isRestartDebugEnabled())
                        BPELTrace.traceRestart("************** BPI getFaultMessage NOT SORTED **************");

                }
                before = m.getDate();
            }
            md = (MessageData) messages.get(loopNb);
            if (!md.isFault()) {
                md = null;
            }
        }
        return md;
    }

    private Collection getBpusByState(int bpiId,String state) throws BPELEngineException {
        Collection bpus = null;
        try {
            BPUEntityHomeLocal bpuhl = BPUEntityBeanUtil.getLocalHome();
            if (state != null) {
                bpus = bpuhl.findByBpiIdAndState(new Integer(bpiId),state);
            } else {
                bpus = bpuhl.findByBpiId(new Integer(bpiId));
            }
        } catch (FinderException e) {
            //no bpu for this bpi
        } catch (NamingException e) {
            throw new BPELEngineException("NamingException BPI getActiveBpus : " , e);
        }
        return bpus;
    }

    private void restartBPUs(Collection bpus) throws BPELEngineException {
        //restart all active bpus
        //restart all flow children not already launched
        //restart on the next statement of a flow for each finished runningFlow
        //restart all bpus having finished targets but no new bpu launched on the corresponding activity
        launchFlowChildrenNotLaunched();
        launchNextOfFinishedFlow();
        launchTargetNeedRestartBPUS();
        for (Iterator it = bpus.iterator();it.hasNext();) {
            BPUEntityLocal bpu = (BPUEntityLocal) it.next();
            restartBPU(bpu,this.restartDoc);
        }
    }

    private void launchNextOfFinishedFlow() throws BPELEngineException {
        if (BPELTrace.isRestartDebugEnabled())
            BPELTrace.traceRestart("BPI " + this.bpiId + " launchNextOfFinishedFlow");
        for (int i=0;i<this.runningFlows.size();i++) {
            RunningFlow rf = (RunningFlow) this.runningFlows.get(i);
            if (rf.getActiveChildren() == 0) {
                this.runningFlows.remove(rf);
                // Launch a new BPU on the next statement after the flow if needed
                createBPU(rf.getStatementNb(), "endFlow", null, null);
            }
        }
        if (BPELTrace.isRestartDebugEnabled())
            BPELTrace.traceRestart("BPI " + this.bpiId + " end launchNextOfFinishedFlow");
    }

    private void launchTargetNeedRestartBPUS() throws BPELEngineException {
        if (BPELTrace.isRestartDebugEnabled())
            BPELTrace.traceRestart("BPI " + this.bpiId + " launchTargetNeedRestartBPUS, finishedTargets.size() = "+this.finishedTargets.size());
        Set s = this.finishedTargets.entrySet();
        for (Iterator i = s.iterator(); i.hasNext();) {
            Map.Entry e = (Map.Entry) i.next();
            int activityNb = ((Integer) e.getKey()).intValue();
            PendingTarget pt = (PendingTarget) e.getValue();
            Node activityN = getNodeWithNb(this.restartDoc,activityNb);
            int activityLoops = getLoopsNb(activityN);
            int bpuId = pt.getBpuId();
            int[] targets = pt.getBranchesNb();
            int targetLoops = -1;
            for (int j=0;j<targets.length;j++) {
                int targetBrancheNb = targets[j];
                Node targetN = getNodeWithNb(activityN,targetBrancheNb);
                targetLoops = getLoopsNb(targetN);
                String targetLinkName = getAttibuteValue(targetN,"linkName");
                Node sourceN = findCorrespondingSource(this.restartDoc,targetLinkName);
                int sourceLoops = getLoopsNb(sourceN);
                int targetsInsertedTimes = -1;
                Object tmp = this.insertedTargets.get(new Integer(activityNb));
                targetsInsertedTimes = ((Integer)tmp).intValue();
                if (sourceLoops > targetLoops && targetsInsertedTimes == sourceLoops) {
                    if (BPELTrace.isRestartDebugEnabled())
                        BPELTrace.traceRestart("BPI " + this.bpiId + " launchTargetNeedRestartBPUS, executes target : "+targetBrancheNb);
                    Statement targetS = this.bpelProcess.findStatement(targetBrancheNb);
                    update(bpuId,targetS);
                    targetLoops++;
                }
            }
            int activitySourceLoops = getMaxSourceExecNb(activityN);
            //if the current activity has one or more source already executed
            //the same nb of times as targetLoops, it means, the activity was not
            //executed due to its joinCondition. So we must no execute one time more
            //these sources
            if (targetLoops > activityLoops && activitySourceLoops < targetLoops) {
                if (BPELTrace.isRestartDebugEnabled())
                    BPELTrace.traceRestart("Launch a needRestart BPU on brancheNb : "+activityNb);
                ptFinished(pt);
            }
        }
        if (BPELTrace.isRestartDebugEnabled())
            BPELTrace.traceRestart("BPI " + this.bpiId + " end launchNeedRestartBPUS");
    }

    private int getMaxSourceExecNb(Node activityN) {
        int result = 0;
        NodeList children = activityN.getChildNodes();
        for (int i=0;i<children.getLength();i++) {
            Node child = children.item(i);
            String type = getType(child);
            if (type != null && type.equals("Source")) {
                int exec = getLoopsNb(child);
                if (exec > result) {
                    result = exec;
                }
            }
        }
        return result;
    }

    private Node findCorrespondingSource(Node n,String targetLinkName) {
        Node source = null;
        NodeList children = n.getChildNodes();
        for (int i=0;i<children.getLength() && source == null;i++) {
            Node child = children.item(i);
            String type = getType(child);
            if (type != null && type.equals("Source")) {
                String sourceLinkName = getAttibuteValue(child,"linkName");
                if (sourceLinkName.equals(targetLinkName)) {
                    source = child;
                    break;
                }
            }
            if (child.hasChildNodes()) {
                source = findCorrespondingSource(child,targetLinkName);
            }
        }
        return source;
    }

    private Node getNodeFromStatementNb(Node n,int statementNb) {
        Node result = null;
        NodeList children = n.getChildNodes();
        for (int i=0;i<children.getLength();i++) {
            Node child = children.item(i);
            if (getBrancheNb(child) == statementNb) {
                result = child;
                break;
            }
            result = getNodeFromStatementNb(child,statementNb);
        }
        return result;
    }

    private void launchFlowChildrenNotLaunched() throws BPELEngineException {
        if (BPELTrace.isRestartDebugEnabled())
            BPELTrace.traceRestart("BPI " + this.bpiId + " launchFlowChildrenNotLaunched");
        for (int k=0;k<this.runningFlows.size();k++) {
            RunningFlow rf = (RunningFlow) this.runningFlows.get(k);
            int flowNb = rf.getStatementNb();
            Flow f = (Flow) this.bpelProcess.findStatement(flowNb);
            Node flowN = getNodeWithNb(this.restartDoc,flowNb);

            int flowLoops = getLoopsNb(flowN);
            org.objectweb.orchestra.tree.Statement[] tab = f.getChilds();
            for (int i = 0; i < tab.length; i++) {
                if (!tab[i].getType().equals("Target") && !tab[i].getType().equals("Source")) {
                    Statement child = tab[i];
                    int childNb = child.getNb().intValue();
                    Node childN = getNodeWithNb(flowN,childNb);
                    //if this flow child was executed the same times as the flow,
                    //it was already launched
                    if (getLoopsNb(childN)!=flowLoops) {
                        //this children was not yet launched or it has targets
                        //before launching a new BPU on this child,
                        //verify it has no target already executed
                        //here, child can be an activity or a links statement
                        //if the child is a links, we must launch it as we know it was
                        //not executed the same nb of times as it flow father
                        //if the child is an activity, perhaps it was already launched
                        //it was already launched if this activity has at least one target executed the same nb$
                        //of times as the flow
                        boolean launch = !child.isActivity() || (child.isActivity() && !hasOneTargetExecuted((Activity)child, flowLoops, childN));
                        if (launch) {
                            createBPU(childNb, "flow", null, null);
                        }
                    }
                }
            }
        }
        if (BPELTrace.isRestartDebugEnabled())
            BPELTrace.traceRestart("BPI " + this.bpiId + " end launchFlowChildrenNotLaunched");
    }

    private void restartBPU(BPUEntityLocal bpu,Node doc) throws BPELEngineException {
        //we never reused an existing bpu, so we put it to FINISHED and we create a new one
        bpu.setState(BPELConstants.BPU_FINISHED);
        int statementNb = bpu.getStatementNb();
        int bpuId = bpu.getId().intValue();
        if (BPELTrace.isRestartDebugEnabled()) {
            BPELTrace.traceRestart("BPI " + this.bpiId + " restartBPU, bpuId = "+bpuId+", statementNb = "+statementNb);
            BPELTrace.traceRestart("BPI " + this.bpiId + " restartBPU, statementNb = "+statementNb);
        }
        Statement s = this.bpelProcess.findStatement(statementNb);
        String type = s.getType();
        if (BPELTrace.isRestartDebugEnabled())
            BPELTrace.traceRestart("BPI " + this.bpiId + " restartBPU, bpuId = "+bpuId+", type = "+type);

        if (contains(this.restartDownAndNext,type)) {
            createBPU(statementNb,"downAndNext",null,null);
        } else if (contains(this.restartNext,type)) {
            createBPU(statementNb,"nextStatement",null,null);
        } else if (type.equals("Source")) {
            executeSourcesNotAlreadyExecuted(s,bpuId,doc);
            createBPU(s.getFather().getNb().intValue(),"onlyNextStatement",null,null);
        } else if (type.equals("Switch")) {
            restartSwitch(statementNb);
        } else if (type.equals("While")) {
            restartWhile(s);
        } else if (type.equals("Correlations")) {
            //here we are sure we are around an invoke
            //if the invoke was performed :  down and next
            //if the invoke was not performed : invoke, outcorrelations, down and next
            executeInvoke(s,doc);
        } else if (type.equals("Invoke")) {
            //exec correlations
            //downAndNext
            boolean success = wasInvokeSuccessful(statementNb,doc);
            String successSt = Boolean.toString(success);
            createBPU(statementNb,"outCorrelations",new String[]{"success"},new String[]{successSt});
        } else if (type.equals("Throw")) {
            //********************* TO DO ***********************
            //TO DO : depends on what is done in executeRestartAction for Throw
        } else if (type.equals("Terminate")) {
            //nothing, this case should never happen as this bpi was stopped
            //by restartAction
        } else if (type.equals("Pick")) {
            PendingEvent pe = findEvent(statementNb,this.scopeId,this.pendingPicks);
            launchAlarms(s, doc, bpuId,pe);
        } else if (type.equals("EventHandler")) {
            PendingEvent pe = findEvent(statementNb,this.scopeId,this.pendingEventHandlers);
            launchAlarms(s, doc, bpuId, pe);
        }
    }

    private boolean hasOneTargetExecuted(Activity a, int flowLoops, Node activityN) {
        //return true if this activity has a target which was executed the same
        //nb of times as flowLoops
        org.objectweb.orchestra.tree.Statement[] tab = a.getChilds();
        boolean oneTargetExecuted = false;
        for (int j = 0; j < tab.length; j++) {
            if (tab[j].getType().equals("Target")) {
                Statement target = tab[j];
                int targetNb = target.getNb().intValue();
                Node targetN = getNodeWithNb(activityN,targetNb);
                if (getLoopsNb(targetN)==flowLoops) {
                    oneTargetExecuted = true;
                    break;
                }
            }
        }
        return oneTargetExecuted;
    }

    private void launchAlarms(Statement s, Node doc,int bpuId, PendingEvent pe) throws BPELEngineException {
        //launch all alarms of a given event if no one onMessage was executed
        Node eventN = getNodeWithNb(doc,s.getNb().intValue());
        int eventLoops = getLoopsNb(eventN);
        org.objectweb.orchestra.tree.Statement[] tab = s.getChilds();
        Vector alarms = new Vector();
        boolean launchAlarms = true;
        for (int i = 0; i < tab.length; i++) {
            if (!tab[i].getType().equals("Target") && !tab[i].getType().equals("Source")) {
                Statement child = tab[i];
                int childNb = child.getNb().intValue();
                Node childN = getNodeWithNb(eventN,childNb);
                if (getLoopsNb(childN)!=eventLoops && child.getType().equals("OnAlarms")) {
                    alarms.addElement(child);
                } else if (child.getType().equals("OnMessage")) {
                    //an onMessage was executed, do not launch alarms
                    launchAlarms = false;
                }
            }
        }
        if (launchAlarms) {
            boolean finished = false;
            for (int i=0;i<alarms.size() && !finished;i++) {
                OnAlarm oa = (OnAlarm) alarms.get(i);
                finished = launchAlarms(oa,bpuId,finished,pe);
            }
            if (finished) {
                BPELTimerManager.getInstance().removeAlarms(this.bpiId,pe);
            }
        }
    }

    private void executeInvoke(Statement s, Node doc) throws BPELEngineException {
        //restart a bpu depending on if the invoke was already performed or not
        Invoke i = (Invoke) s.getFather();
        int invokeNb = i.getNb().intValue();
        Node iN = getNodeWithNb(doc,invokeNb);
        Node corN = getNodeWithNb(doc,s.getNb().intValue());
        int iLoops = getLoopsNb(iN);
        int correlationLoops = getLoopsNb(corN);

        boolean success;
        if (iLoops < correlationLoops) {
            //the invoke was not executed, so execute it
            //out correlations were not executed, so execute them
            createBPU(invokeNb,"execInvoke",null,null);
        } else {
            //here the invoke was executed
            success = wasInvokeSuccessful(invokeNb,doc);
            if (success) {
                createBPU(invokeNb,"downAndNext",null,null);
            }
        }
    }

    private boolean wasInvokeSuccessful (int invokeNb, Node doc) {
        //retur true if the invoke executed normally
        //return false if a fault occured during the invoke
        Node iN = getNodeWithNb(doc,invokeNb);
        int loops = getLoopsNb(iN);
        return getSuccess(iN,loops);
    }

    private void executeSourcesNotAlreadyExecuted(Statement s, int bpuId,Node doc) throws BPELEngineException {
        if (BPELTrace.isRestartDebugEnabled())
            BPELTrace.traceRestart("BPI " + this.bpiId + " executeSourcesNotAlreadyExecuted");
        //execute all sources not already executed
        Activity a = (Activity) s.getFather();
        Node activityN = getNodeWithNb(doc,a.getNb().intValue());
        int activityLoops = getLoopsNb(activityN);

        if (a.getChildNb() > 0) {
            int childrenNb = a.getChildNb();
            org.objectweb.orchestra.tree.Statement[] tab = a.getChilds();
            for (int i = 0; i < childrenNb; i++) {
                if (tab[i].getType().equals("Source")) {
                    Node source = getNodeWithNb(activityN,tab[i].getNb().intValue());
                    int loops = getLoopsNb(source);
                    if (loops < activityLoops) {
                        String tc = ((Source) tab[i]).getTransitionCondition();
                        String linkName = ((Source) tab[i]).getLinkName();
                        boolean tcValue = ((Boolean) Expression.analyseExpression(tc, this.bpiId, this.scopeId, this.process, BPELConstants.BOOLEAN_EXPR)).booleanValue();
                        update(bpuId,tab[i],new String[]{"transitionConditionEvaluation",Boolean.toString(tcValue)});
                        String transitionContion = Boolean.toString(tcValue)/*+"()"*/;
                        source(linkName,transitionContion);
                    }
                }
            }
        }
        if (BPELTrace.isRestartDebugEnabled())
            BPELTrace.traceRestart("BPI " + this.bpiId + " end executeSourcesNotAlreadyExecuted");

    }

    public static Node getNodeWithNb(Node n,int nb) {
        //get the node having the corresponding statementnb in the monitoring tree
        //BPELTrace.traceRestart("BPI getNodeWithNb, nb="+nb);
        NodeList children = n.getChildNodes();
        Node result = null;
        for (int i=0;i<children.getLength();i++) {
            Node child = children.item(i);
            NamedNodeMap attributes = child.getAttributes();
            Node nNb = attributes.getNamedItem("nb");
            if (nNb!=null && Integer.parseInt(nNb.getNodeValue()) == nb) {
                result = child;
            }
            if (result == null && child.hasChildNodes()) {
                result = getNodeWithNb(child,nb);
            }
        }
        //BPELTrace.traceRestart("BPI end getNodeWithNb, result="+result);
        return result;
    }

    private void restartSwitch(int switchNb) throws BPELEngineException {
        createBPU(switchNb,"casesOfSwitch",null,null);
    }

    private void restartWhile(Statement s) throws BPELEngineException {
        While w = (While)s;
        int whileNb = s.getNb().intValue();
        boolean tcValue = ((Boolean) Expression.analyseExpression(w.getCondition(), this.bpiId, this.scopeId,this.process, BPELConstants.BOOLEAN_EXPR)).booleanValue();
        if (tcValue) {
            createBPU(whileNb,"downAndNext",null,null);
        } else {
            createBPU(whileNb,"nextStatement",null,null);
        }
    }

    private void putChildFinishedOnFlow (Vector endFlows) {
        //for each endFlow executed by a bpu, say it to the corresponding flow
        if (BPELTrace.isRestartDebugEnabled())
            BPELTrace.traceRestart("BPI " + this.bpiId + " putChildFinishedOnFlow");
        HashMap flows = new HashMap();

        for (int i=0;i<endFlows.size();i++) {
            Integer flowNb = (Integer) endFlows.get(i);
            if (flows.containsKey(flowNb)) {
                Integer integ = (Integer) flows.get(flowNb);
                flows.put(flowNb,new Integer(integ.intValue()+1));
            } else {
                flows.put(flowNb,new Integer(1));
            }
        }

        Set s = flows.entrySet();
        for (Iterator i = s.iterator(); i.hasNext();) {
            Map.Entry e = (Map.Entry) i.next();
            int flowNb = ((Integer) e.getKey()).intValue();
            int ends = ((Integer) e.getValue()).intValue();
            RunningFlow rf = findFlow(flowNb);
            if (rf != null) {
                int flowChildren = rf.getActiveChildren();
                int endsOfThisLoop = 0;

                int q = ends/flowChildren;
                endsOfThisLoop = ends - q*flowChildren;

                for (int j=0;j<endsOfThisLoop;j++) {
                    if (BPELTrace.isRestartDebugEnabled())
                        BPELTrace.traceRestart("BPI " + this.bpiId + " putChildFinishedOnFlow on : "+flowNb);
                    rf.childFinished();
                }
            }
        }
    }

    public static Node getAttibute(Node n, String attributeName) {
        NamedNodeMap attributes = n.getAttributes();
        return attributes.getNamedItem(attributeName);
    }

    private String getAttibuteValue(Node n, String attributeName) {
        String result = null;
        Node attrNode = getAttibute(n,attributeName);
        if (attrNode != null) {
            result = attrNode.getNodeValue();
        }
        return result;
    }

    public static int getAttributeIntValue(Node n,String attributeName) {
        int result = -1;
        Node attr = getAttibute(n,attributeName);
        if (attr!=null) {
            result = Integer.parseInt(attr.getNodeValue());
        }
        return result;
    }

    public static int getLoopsNb(Node n) {
        int result = 0;
        int loops = getAttributeIntValue(n,BPELConstants.MONITORING_LOOPS_NODE_NAME);
        if (loops != -1) {
            result = loops;
        }
        return result;
    }

    private int getBrancheNb(Node n) {
        return getAttributeIntValue(n,"nb");
    }

    private int getBpuId(Node n) {
        return getAttributeIntValue(n,BPELConstants.MONITORING_BPUID_NODE_NAME);
    }

    private boolean getExecuted(Node n) {
        boolean ret = false;
        String executed = getAttibuteValue(n,BPELConstants.MONITORING_EXECUTED_NODE_NAME);
        if (executed != null && executed.equals("true")) {
            ret = true;
        }
        return ret;
    }

    private String getType(Node n) {
        return getAttibuteValue(n,"type");
    }

    private int getScopeId(int scopeNb, int iter) throws BPELEngineException {
        //get the scopeId corresponding to the given scopeNb and iter
        //return -1 if no one is found
        if (BPELTrace.isRestartDebugEnabled())
            BPELTrace.traceRestart("BPI " + this.bpiId + " getScopeId, scopeNb="+scopeNb+"iter="+iter);
        int scopeId = -1;
        String scopeNbName = Integer.toString(this.bpiId) + scopeNb;
        try {
            initScopeHomeLocal();
            Collection scopes = this.sh.findByNb(scopeNbName);
            for (Iterator it = scopes.iterator();it.hasNext();) {
                ScopeLocal sl = (ScopeLocal) it.next();
                int currentIter = sl.getIter().intValue();
                if (currentIter == iter) {
                    scopeId = sl.getId().intValue();
                }
            }
        } catch (FinderException ex) {
            throw new BPELEngineException("FinderException BPIMdb getScopeId :\n" , ex);
        }
        if (BPELTrace.isRestartDebugEnabled())
            BPELTrace.traceRestart("BPI " + this.bpiId + " end getScopeId, scopeId ="+scopeId);
        return scopeId;
    }

    private ScopeHomeLocal sh = null;

    private void initMonitoring() throws BPELEngineException {
        if (this.monitor == null) {
            try {
                BpelMonitoringSessionHomeLocal mHome = BpelMonitoringSessionUtil.getLocalHome();
                this.monitor = mHome.create();
            } catch (javax.ejb.CreateException e) {
                throw new BPELEngineException("CreateException BPIMdb initMonitoring : " , e);
            } catch (NamingException e) {
                throw new BPELEngineException("NamingException BPIMdb initMonitoring : " , e);
            }
        }
    }

    private void initScopeHomeLocal() throws BPELEngineException {
        if (this.sh == null) {
            try {
                this.sh = ScopeBeanUtil.getLocalHome();
            } catch (NamingException e) {
                throw new BPELEngineException("NamingException BPIMdb initScopeHomeLocal : " , e);
            }
        }
    }


}