/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.objectweb.orchestra.engine;

import javax.ejb.MessageDrivenBean;
import javax.ejb.MessageDrivenContext;
import javax.jms.MapMessage;
import javax.jms.MessageListener;
import javax.jms.QueueConnection;
import javax.jms.QueueReceiver;
import javax.jms.QueueSession;
import javax.jms.Session;

import org.objectweb.orchestra.data.BPICache;
import org.objectweb.orchestra.data.BPIData;
import org.objectweb.orchestra.data.CSCache;
import org.objectweb.orchestra.exception.BPELEngineException;
import org.objectweb.orchestra.exception.BPELUtilException;
import org.objectweb.orchestra.util.BPELConstants;
import org.objectweb.orchestra.util.BPELTrace;
import org.objectweb.orchestra.util.BPELTransaction;
import org.objectweb.orchestra.util.JmsUtil;
import org.objectweb.orchestra.util.Util;


/**
 * BPIMdbBean is an Message Driven bean
 *
 * @author Goulven Le Jeune & Charles Souillard
 * @version $Id: BPIMdbBean.java 5 2006-10-20 14:37:44Z lejeuneg $
 *
 * @ejb:message-driven
 *
 * @ejb.bean name="BPIMdb"
 *           description="Deployment descriptor for the BPIMdbBean bean"
 *           transaction-type="Container"
 *           acknowledge-mode="Auto-acknowledge"
 *           destination-type="javax.jms.Queue"
 *
 * @ejb:transaction type="NotSupported"
 *
 * @jonas.message-driven-destination jndi-name="BPIQueue"
 *
 * @jonas.min-pool-size @BPIMdbBean.min-pool-size@
 * @jonas.max-cache-size @BPIMdbBean.max-cache-size@
 * @ejb.activation-config-property name="maxNumberOfWorks" value="@BPIMdbBean.maxNumberOfWorks@"
 *
 * @ejb.ejb-ref ejb-name="BPUEntityBean"
 *              view-type="local"
 *              ref-name="ejb/BPUEntityHomeLocal"
 *
 * @ejb.ejb-ref ejb-name="BPIEntityBean"
 *              view-type="local"
 *              ref-name="ejb/BPIEntityHomeLocal"
 *
 * @ejb.security-identity run-as="admin"
 *
 */
public class BPIMdbBean implements MessageDrivenBean, MessageListener {

    private transient MessageDrivenContext mdbContext;

    public MessageDrivenContext getContext() {
        return this.mdbContext;
    }

    public void onMessage(javax.jms.Message message) {
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPIMdbBean onMessage");
        MapMessage m = (MapMessage) message;
        int bpiId = -1;
        BPI bpi = null;
        try {
            bpiId = m.getIntProperty("bpiId");
            if (BPELTrace.isBPIDebugEnabled())
                BPELTrace.traceBPI("BPIMdbBean onMessage, bpiId = "+bpiId);
            //System.out.println("BPIMdbBean onMessage, bpiId = "+bpiId);
            bpi = BPIMap.getBPI(bpiId);
            //System.out.println("BPIMdbBean onMessage, bpi = "+bpi);
            if (bpi != null) {
                synchronized (bpi) {
                    //System.out.println("Executing bpi : "+bpiId);
                    //can not use the message received on BPIQ as many messages
                    //from different bpus having the same destination bpi can
                    //be received in parallel in this onMessage method.
                    //we want to consumer messages in the order they arrive
                    MapMessage mess = getMessageOnFromBPELQueue(bpiId);
                    bpi.onMessage(mess);
                    //bpi.onMessage(m);
                    //System.out.println("end bpi : "+bpiId);
                }
            } else {
                if (BPELTrace.isBPIDebugEnabled())
                    BPELTrace.traceBPI("BPIMdbBean onMessage, no BPI matching in BPIMap with id : "+bpiId);
                MapMessage mess = getMessageOnFromBPELQueue(bpiId);
                String type = mess.getString("type");
                String source = mess.getString("source");
                // Even if the BPI is not found the instance has to be terminated
                if ((source.equals("Admin")) && (type.equals("terminate"))) {
                    if (BPELTrace.isBPIDebugEnabled())
                        BPELTrace.traceBPI("BPIMdbBean onMessage, Terminating instance : "+bpiId);
                    try {
                        BPELTransaction tx = null;
                        boolean txActive = false;
                        try {
                            tx = new BPELTransaction();
                            tx.beginTx();
                            txActive = true;
                            BPICache.setBPIState(bpiId, BPELConstants.BPI_TERMINATED);
                            CSCache.removeCS(bpiId);
                            tx.commitTx();
                            txActive = false;
                        } finally {
                            if (txActive) {
                                tx.roolbackTx();
                            }
                        }
                        BPIData bpid = BPICache.findBPI(bpiId);
                        if (BPELEngine.flushInstance()) {
                            BPICache.flushBPIToXMLFile(bpid);
                            BPICache.storeScopesVariables(bpid);
                        }
                        if (BPELEngine.flushMessages()) {
                            BPICache.storeMessages(bpid);
                        }
                    } catch (BPELUtilException e) {
                        throw new BPELEngineException("BPELUtilException : BPI.endBPI : "+e,e);
                    } finally {
                        // This has to be ran whatever happens
                        BPICache.removeBPI(bpiId);
                        BPIMap.removeBPI(bpiId);
                    }
                }
            }
        } catch (Throwable e) {
            System.out.println("ERROR : BPIMdb OnMessage Exception caught BPI : " + bpiId);
            e.printStackTrace();
            if (BPELTrace.isBPIDebugEnabled())
                BPELTrace.traceBPI("BPI Exception, Thread = "+Thread.currentThread()+" : \n"+Util.getExceptionAsString(e));
            try {
                EngineHomeLocal ehl = EngineUtil.getLocalHome();
                EngineLocal engine = ehl.create();
                bpi.setException(e);
                engine.sendMessageBPI(bpiId, -1, bpi.getCurrentScopeId(), "fault", new String[]{"faultName","faultMessage","messageId"}, new Object[]{"OrchestraInternalError","",new Integer(-1)});
                engine.remove();
            } catch (Exception ex) {
                ex.printStackTrace();
            }
        }
        if (BPELTrace.isBPIDebugEnabled())
            BPELTrace.traceBPI("BPIMdbBean " + bpiId + " end onMessage");
    }

    private MapMessage getMessageOnFromBPELQueue(int bpiId) throws Exception {
        QueueConnection connection = JmsUtil.getConnection();
        Session session = connection.createQueueSession(false, Session.AUTO_ACKNOWLEDGE);
        QueueReceiver prod = ((QueueSession) session).createReceiver(JmsUtil.getBPELQueue(), "destination = 'BPI' and bpiId = " + bpiId);
        connection.start();
        MapMessage message = (MapMessage) prod.receive();
        prod.close();
        session.close();
        connection.close();
        prod = null;
        session = null;
        connection = null;
        return message;
    }

    public BPIMdbBean() {
        // Constructor
    }
    public void setMessageDrivenContext(MessageDrivenContext ctx) {
        this.mdbContext = ctx;
    }
    public void ejbRemove() {
        // Mandatory method
    }
    public void ejbCreate() {
        // Mandatory method

    }

}