/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.objectweb.orchestra.engine;


import javax.ejb.MessageDrivenBean;
import javax.ejb.MessageDrivenContext;
import javax.jms.JMSException;
import javax.jms.MapMessage;
import javax.jms.MessageListener;

import org.objectweb.orchestra.data.BPUCache;
import org.objectweb.orchestra.data.BPUData;
import org.objectweb.orchestra.exception.BPELEngineException;
import org.objectweb.orchestra.exception.BPELUtilException;
import org.objectweb.orchestra.tree.BPELProcess;
import org.objectweb.orchestra.tree.Invoke;
import org.objectweb.orchestra.tree.Statement;
import org.objectweb.orchestra.tree.Switch;
import org.objectweb.orchestra.util.BPELConstants;
import org.objectweb.orchestra.util.BPELTrace;
import org.objectweb.orchestra.util.Util;


/**
 * BPUMdb is an Message Driven bean
 *
 * @author Goulven Le Jeune & Charles Souillard
 * @version $Id: BPUMdbBean.java 5 2006-10-20 14:37:44Z lejeuneg $
 *
 * @ejb:message-driven
 * @ejb.bean name="BPUMdb"
 *           description="Deployment descriptor for the BPUMdbBean bean"
 *           transaction-type="Container"
 *           acknowledge-mode="Auto-acknowledge"
 *           destination-type="javax.jms.Queue"
 *
 * @ejb:transaction type="NotSupported"
 *
 * @jonas.message-driven-destination jndi-name="BPUQueue"
 *
 * @jonas.min-pool-size @BPUMdbBean.min-pool-size@
 * @jonas.max-cache-size @BPUMdbBean.max-cache-size@
 *
 * @ejb.activation-config-property name="maxNumberOfWorks" value="@BPUMdbBean.maxNumberOfWorks@"
 *
 * @ejb.security-identity run-as="admin"
 *
 */
public class BPUMdbBean implements MessageDrivenBean, MessageListener {

    private transient MessageDrivenContext mdbContext;

    //private String process;
    private int bpiId;
    private int bpuId;
    private StatementManager sm;
    private BPELProcess bpelProcess;
    private String targetNamespace;
    private long messagesReceivedFromBPI = 0;

    public MessageDrivenContext getContext() {
        return this.mdbContext;
    }

    /**
     * onMessage method
     *
     * @param Message The message to analyse
     */
    public void onMessage(javax.jms.Message message) {
        if (BPELTrace.isBPUDebugEnabled())
            BPELTrace.traceBPU("BPUMdbBean " + this.bpuId + " onMessage");
        this.sm = null;
        MapMessage m = (MapMessage) message;
        BPI bpi = null;
        try {
            this.bpiId = m.getInt("bpiId");
            this.bpuId = m.getIntProperty("bpuId");
            this.targetNamespace = m.getString("targetNamespace");

            bpi = BPIMap.getBPI(this.bpiId);
            if (bpi != null) {
                this.bpelProcess = bpi.getBPELProcess();
                this.sm = new StatementManager(this.bpelProcess, this.bpiId, this.bpuId, this.targetNamespace);

                analyseMessage(m);
            }

        } catch (Throwable e) {
            System.out.println("ERROR : BPUMdbBean OnMessage Exception caught BPI : " + this.bpiId + ", BPU : " + this.bpuId);
            e.printStackTrace();
            if (BPELTrace.isBPUDebugEnabled())
                BPELTrace.traceBPU("BPUMdbBean Exception, : \n"+Util.getExceptionAsString(e));
            try {
                EngineHomeLocal ehl = EngineUtil.getLocalHome();
                EngineLocal engine = ehl.create();
                bpi.setException(e);
                engine.sendMessageBPI(this.bpiId, this.bpuId, this.sm.getScopeId(), "fault", new String[]{"faultName","faultMessage","messageId"}, new Object[]{"OrchestraInternalError","",new Integer(-1)});
                engine.remove();
            } catch (Exception ex) {
                ex.printStackTrace();
            }
        } finally {
            try {
                if (this.sm !=null) {
                    this.sm.end();
                    this.sm = null;
                }

                endBPU();
            } catch (BPELEngineException e) {
                System.out.println("ERROR : BPUMdb OnMessage Impossible to end BPU. BPI : " + this.bpiId + ", BPU : " + this.bpuId);
                e.printStackTrace();
            }
        }
        if (BPELTrace.isBPUDebugEnabled())
            BPELTrace.traceBPU("BPUMdbBean " + this.bpuId + " end onMessage");
    }

    public void analyseMessage(MapMessage m) throws BPELEngineException {
        if (BPELTrace.isBPUDebugEnabled())
            BPELTrace.traceBPU("BPUMdbBean " + this.bpuId + " analyseMessage, bpiId = " + this.bpiId);
        try {
            String type = m.getString("type");

            //find the statement to execute. It is not necessary the first as
            // there are communications with the BPI.
            //There could also have a restart
            this.messagesReceivedFromBPI++;
            int statementNb = m.getInt("brancheNb");

            if (BPELTrace.isBPUDebugEnabled())
                BPELTrace.traceBPU("BPUMdbBean " + this.bpuId + " analyseMessage, type = " + type+", statementNb = "+statementNb);

            //System.out.println("STATEMENT MANAGER, statementNb = "+statementNb);
            Statement s = this.bpelProcess.findStatement(statementNb);
            if (s == null && !type.equals("replyError")) {
                System.out.println("BPUMdbBean.analyseMessage, bpiId = "+this.bpiId+", bpuId = "+this.bpuId+", processName = "+this.bpelProcess.getName()+", type = "+type+", statementNb = "+statementNb);
                throw new BPELEngineException("BPUMdbBean.analyseMessage, no statement found with nb : "+statementNb,null);
            }

            //System.out.println("STATEMENT MANAGER, s = "+s);
            if (BPELTrace.isBPUDebugEnabled())
                BPELTrace.traceBPU("BPUMdbBean " + this.bpuId + " analyseMessage, bpiId = " + this.bpiId + " type = " + type + ", brancheNb = " + statementNb + ", statement = " + s);
            int scopeId = m.getInt("scopeId");
            //System.out.println("STATEMENT MANAGER, ScopeId = "+scopeId);
            this.sm.setScopeId(scopeId);
            //in some cases, specific actions have to be run
            if (type.equals("receive")) {
                Integer messageId = new Integer(m.getString("messageId"));
                int oldBpuId = Integer.parseInt(m.getString("oldBpuId"));
                this.sm.executeReceiveReceive(s, messageId,oldBpuId);
            } else if (type.equals("target")) {
                this.sm.executeReceiveTarget(s);
            } else if (type.equals("endFlow") || type.equals("endForEach")) {
                this.sm.executeNextOfFlow(s);
            } else if (type.equals("endWait")) {
                this.sm.executeNextOfWait(s);
            } else if (type.equals("forEach")) {
                 String counterName = m.getString("counterName") ;
                 int counterValue = m.getInt("counterValue") ;
                 scopeId = m.getInt("scopeId") ;
                 this.sm.setScopeId(scopeId) ;
                 this.sm.executeParallelForEach(s, counterName, counterValue);
            } else if (type.equals("pickOnMessage")) {
                Integer messageId = new Integer(m.getString("messageId"));
                this.sm.executeReceiveOnMessage(s, messageId);
            } else if (type.equals("pickOnAlarm")) {
                this.sm.executeReceiveOnAlarm(s);
            } else if (type.equals("eventOnMessage")) {
                Integer messageId = new Integer(m.getString("messageId"));
                this.sm.executeEventOnMessage(s, messageId);
            } else if (type.equals("eventOnAlarm")) {
                this.sm.executeEventOnAlarm(s);
            } else if (type.equals("catchFault")) {
                String messageId = m.getString("messageId");
                if (messageId !=null) {
                    Integer mId = new Integer(messageId);
                    this.sm.executeCatchFault(s, mId);
                } else {
                    String variable = m.getString("variable");
                    this.sm.executeCatchFault(s, variable);
                }
            } else if (type.equals("BPELJCatchFault")) {
                String processName = this.bpelProcess.getName();
                Object faultData = Util.getValueObject(m.getBytes("faultData"), processName);
                this.sm.executeBPELJCatchFault(s, faultData);

            } else if (type.equals("compensate")) {
                this.sm.executeCompensation(s);
            } else if (type.equals("reply")) {
                String refPartner = m.getString("refPartner");
                this.sm.executeReceiveReply(s, refPartner);
            } else if (type.equals("replyError")) {
                String refPartner = m.getString("refPartner");
                String ex = m.getString("exception");
                this.sm.executeReceiveReplyError(refPartner,ex);
            } else if (type.equals("nextStatement")) {
                this.sm.executeNextStatement(s);
            } else if (type.equals("downAndNext")) {
                this.sm.executeDownAndNext(s);
            } else if (type.equals("casesOfSwitch")) {
                this.sm.executeCasesOfSwitch((Switch)s);
            } else if (type.equals("sourceAndNextStatement")) {
                this.sm.executeSourceAndNextStatement(s);
            } else if (type.equals("execInvoke")) {
                this.sm.executeInvokeAfterInCorrelations(s);
            } else if (type.equals("outCorrelations")) {
                boolean success = new Boolean(m.getString("success")).booleanValue();
                this.sm.executeOutCorrelations(s,success);
            } else if (type.equals("onlyNextStatement")) {
                this.sm.executeOnlyNextStatement(s);
            } else if (type.equals("invokeReturn")) {
                String processName = this.bpelProcess.getName();
                InvokerObject io = (InvokerObject)Util.getValueObject(m.getBytes("invokerObject"),processName);
                this.sm.executeInvokeReturn((Invoke)s,io);
            } else {
                this.sm.execute(s);
            }

            while (this.sm.needWhile() != null) {
                this.sm.executeBreakWhile(this.sm.needWhile());
            }
        } catch (JMSException e) {
            throw new BPELEngineException("JMSException BPUMdbBean " + this.bpuId + " analyseTree : " , e);
        } catch (BPELUtilException e) {
            throw new BPELEngineException("BPELUtilException BPUMdbBean " + this.bpuId + " analyseTree : " , e);
        }

        if (BPELTrace.isBPUDebugEnabled())
            BPELTrace.traceBPU("BPUMdbBean " + this.bpuId + " end analyseMessage");
    }

    private void endBPU() throws BPELEngineException {
        if (BPELTrace.isBPUDebugEnabled())
            BPELTrace.traceBPU("BPUMdbBean " + this.bpuId + " endBPU");
        try {
            BPUData bpud = BPUCache.findBPU(this.bpuId);
            if (bpud != null) {
                if (!bpud.getState().startsWith(BPELConstants.BPU_WAITING_RECEIVE)) {
                    BPUCache.setBPUState(this.bpuId, BPELConstants.BPU_FINISHED);
                }
            } else {
                // Is this normal ????? Probably not. What should be done ?
                System.out.println("No BPU found for this bpuId : " + this.bpuId);
            }
            EngineHomeLocal ehl = EngineUtil.getLocalHome();
            EngineLocal engine = ehl.create();
            engine.sendMessageBPI(this.bpiId, this.bpuId, -1, "endBPU", new String[0] , new Object[0]);
            engine.remove();
        } catch (javax.naming.NamingException e) {
            throw new BPELEngineException("NamingException BPUMdb endBPU : " , e);
        } catch (javax.ejb.CreateException e) {
            throw new BPELEngineException("CreateException BPUMdb endBPU : " , e);
        } catch (javax.ejb.RemoveException e) {
            throw new BPELEngineException("RemoveException BPUMdb endBPU : " , e);
        }
        if (BPELTrace.isBPUDebugEnabled())
            BPELTrace.traceBPU("BPUMdbBean " + this.bpuId + " end endBPU");
    }

    public BPUMdbBean() {
        // Mandatory method
    }
    public void setMessageDrivenContext(MessageDrivenContext ctx) {
        this.mdbContext = ctx;
    }
    public void ejbRemove() {
        // Mandatory method
    }
    public void ejbCreate() {
        // Mandatory method
    }

}