/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.objectweb.orchestra.engine;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.Collection;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.Vector;

import javax.ejb.CreateException;
import javax.ejb.SessionBean;
import javax.ejb.SessionContext;
import javax.xml.namespace.QName;

import org.apache.axis.AxisFault;
import org.apache.axis.utils.XMLUtils;
import org.apache.axis.wsdl.toJava.Utils;
import org.objectweb.orchestra.bindingframework.ServiceData;
import org.objectweb.orchestra.bindingframework.ServiceEvent;
import org.objectweb.orchestra.bindingframework.ServiceImpl;
import org.objectweb.orchestra.data.BPICache;
import org.objectweb.orchestra.data.BPIData;
import org.objectweb.orchestra.data.BPUCache;
import org.objectweb.orchestra.data.CSCache;
import org.objectweb.orchestra.data.CSData;
import org.objectweb.orchestra.data.MessageCache;
import org.objectweb.orchestra.data.MessageData;
import org.objectweb.orchestra.data.PartnerLinkCache;
import org.objectweb.orchestra.data.PartnerLinkData;
import org.objectweb.orchestra.data.ScopeCache;
import org.objectweb.orchestra.data.ScopeData;
import org.objectweb.orchestra.data.VariableCache;
import org.objectweb.orchestra.data.VariableData;
import org.objectweb.orchestra.exception.BPELBindingFrameworkException;
import org.objectweb.orchestra.exception.BPELEngineException;
import org.objectweb.orchestra.exception.BPELUtilException;
import org.objectweb.orchestra.tree.From;
import org.objectweb.orchestra.tree.Statement;
import org.objectweb.orchestra.tree.To;
import org.objectweb.orchestra.util.BPELConstants;
import org.objectweb.orchestra.util.BPELString;
import org.objectweb.orchestra.util.BPELTrace;
import org.objectweb.orchestra.util.ClassLoaderMap;
import org.objectweb.orchestra.util.JmsUtil;
import org.objectweb.orchestra.util.UDDIDirectory;
import org.objectweb.orchestra.util.Util;
import org.objectweb.orchestra.util.WSDLJavaMatching;
import org.w3c.dom.Element;


/**
 * EngineBean.java
 *
 * @author Goulven Le Jeune & Charles Souillard
 * @version $Id: EngineBean.java 235 2007-02-19 12:58:32Z souillac $
 *
 * @ejb:bean name="Engine"
 *           display-name="Session Engine Bean"
 *           type="Stateless"
 *           transaction-type="Container"
 *           local-jndi-name="org/objectweb/orchestra/engine/EngineHome_L"
 *           view-type="local"
 *
 * @ejb.home local-class="org.objectweb.orchestra.engine.EngineHomeLocal"
 *
 * @ejb.interface local-class="org.objectweb.orchestra.engine.EngineLocal"
 *
 *
 * @jonas.bean ejb-name="Engine" jndi-name="org/objectweb/orchestra/engine/EngineHome"
 *
 * @ejb:transaction type="Supports"
 *
 * @ejb.permission role-name="admin"
 *
 * @jonas.max-cache-size @EngineBean.max-cache-size@
 *
 */
public class EngineBean implements SessionBean {

    // -------------------------------------------------------------------------
    // Members
    // -------------------------------------------------------------------------
    private SessionContext mContext;


    // -------------------------------------------------------------------------
    // Methods - SessionBean implementation
    // -------------------------------------------------------------------------

    public SessionContext getContext() {
        return this.mContext;
    }

    /**
     * Create the Session Bean
     *
     * @throws CreateException
     *
     * @ejb:create-method
     */

    public void ejbCreate() throws CreateException {
        // Mandatory method
    }

    public void setSessionContext(final SessionContext context) {
        this.mContext = context;
    }

    public void ejbRemove() {
        // Mandatory method
    }

    public void ejbActivate() {
        // Mandatory method
    }

    public void ejbPassivate() {
        // Mandatory method
    }

    private void monitor (String toInsert, int bpuId) throws BPELEngineException {
        if (BPELEngine.monitorRunning()) {
            BPUCache.addMonitor(bpuId, toInsert);
        }
    }

    /**
     *
     * @ejb:interface-method
     */
    public void insertEndFlow(int bpiId, int bpuId, int flowNb) throws BPELEngineException {
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("EngineBean.insertEndFlow, bpiId= "+bpiId+", bpuId = "+bpuId+", flowNb = "+flowNb);
        if (BPELTrace.isRestartDebugEnabled())
            BPELTrace.traceRestart("EngineBean.insertEndFlow, bpiId= "+bpiId+", bpuId = "+bpuId+", flowNb = "+flowNb);

        monitor(BPELConstants.MONITORING_ENDFLOW+flowNb,bpuId);
    }

    /**
    *
    * @ejb:interface-method
    */
   public void insertEndForEach(int bpiId, int bpuId, int forEachNb) throws BPELEngineException {
       BPELTrace.traceEngine("EngineBean.insertEndForEach, bpiId= "+bpiId+", bpuId = "+bpuId+", forEachNb = "+forEachNb);
       BPELTrace.traceRestart("EngineBean.insertEndForEach, bpiId= "+bpiId+", bpuId = "+bpuId+", forEachNb = "+forEachNb);
       monitor(BPELConstants.MONITORING_ENDFOREACH+forEachNb,bpuId);
   }

   /**
     *
     * @ejb:interface-method
     */
    public void insertTargets(int bpiId, int bpuId, int activityNb) throws BPELEngineException {
        if (BPELTrace.isEngineDebugEnabled()) {
            BPELTrace.traceEngine("EngineBean.insertTargets, bpiId= "+bpiId+", bpuId = "+bpuId+", activityNb = "+activityNb);
            if (BPELTrace.isRestartDebugEnabled())
                BPELTrace.traceRestart("EngineBean.insertTargets, bpiId= "+bpiId+", bpuId = "+bpuId+", activityNb = "+activityNb);
        }
        monitor(BPELConstants.MONITORING_TARGETS+activityNb,bpuId);
    }

    /**
     * update the stamentNb in the BPUEntity to know what is the last statement
     * executed
     *
     * @ejb:interface-method
     * @ejb:transaction type="Required"
     */

    public void updateBPUStatementNb(int bpiId, int bpuId, Statement s, String[] others) throws BPELEngineException {
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("EngineBean updateBPUStatementNb, s = " + s+", bpuId = "+bpuId+", bpiId = "+bpiId);
        int nb = s.getNb().intValue();
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("EngineBean updateBPUStatementNb, nb = " + nb+", bpuId = "+bpuId+", bpiId = "+bpiId);
        if (BPELTrace.isRestartDebugEnabled())
            BPELTrace.traceRestart("EngineBean updateBPUStatementNb, nb = " + nb+", bpuId = "+bpuId+", bpiId = "+bpiId);
        String type = s.getType();

        String st = type+BPELConstants.MONITORING_INFO_SEP+nb+BPELConstants.MONITORING_INFO_SEP+System.currentTimeMillis();
        for (int i=0;i<others.length;i++) {
            st += BPELConstants.MONITORING_INFO_SEP+others[i];
        }

        monitor(st,bpuId);
        BPUCache.setBPUStatementNb(bpuId,nb);
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("EngineBean end updateBPUStatementNb, nb = " + nb);
    }

    /**
     * @ejb:interface-method
     */
    public void exit() throws Exception {
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("EngineBean exiting...");
        while(true) {
            try {
                System.exit(2);
                break;
            } catch (Exception e) {
                // Nothing
            }
        }
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("EngineBean exited");
    }
    /**
     * create a scope
     *
     * @param bpiId The BPI Id
     * @param s The Scope
     * @param currentScopeId Current Scope Id
     * @return ScopeId
     * @ejb:interface-method
     */
    public int createScope(int bpiId, org.objectweb.orchestra.tree.Scope s, int currentScopeId) throws BPELEngineException {
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("EngineBean createScope called : " + bpiId + s.getNb().toString() + "  in BPI : " + bpiId + ",  currentScope = " + currentScopeId);
        int scopeId = -1;
        String scopeNbName = Integer.toString(bpiId) + s.getNb().toString();
        //System.out.println("createScope fatherId : " + currentScopeId);
        scopeId = ScopeCache.createScope(s.getName(),scopeNbName,bpiId,currentScopeId);
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("EngineBean end createScope created scope : " + scopeId);
        return scopeId;
    }

    /**
     * get the father of a scope
     *
     * @param scopeId The ScopeId
     * @return The Scope Id of the scope father of the given scope
     * @ejb:interface-method
     */
    public int getScopeFather(int scopeId) throws BPELEngineException {
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("EngineBean getScopeFather, scopeId = " + scopeId);
        int father = -1;
        ScopeData sd = ScopeCache.getFather(scopeId);
        if (sd != null)
            father = sd.getScopeId();
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("EngineBean end getScopeFather, father = " + father);
        return father;
    }

    /**
     * create a variable
     *
     * @param var Variable
     * @param p Process
     * @param scopeId The scope Id
     * @param bpiId The BPIId
     * @ejb:interface-method
     */
    public void createVariable(org.objectweb.orchestra.tree.Variable var, int scopeId, int bpiId) throws BPELEngineException {
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("EngineBean createVariable = " + var.getVariableName() + "  in scope = " + scopeId + "  in BPI = " + bpiId);
        try {
            // Find the right BPI
            BPIData bpid = BPICache.findBPI(bpiId);

            // If this BPEL variable is a messageType, the value of the Variable
            // Entity Bean should be a BPELMessage, else a BPELVariable
            String processName = bpid.getProcessName();
            Object val = null;
            String t = null;
            if (var.getMessageType() != null) {
                t = var.getMessageType();
                val = Util.buildJM(processName,Util.buildQname(t));
            } else {
                if (var.getVarType() != null) {
                    t = var.getVarType();
                } else if (var.getElement() != null) {
                    t = var.getElement();
                } else if (var.getBPELJType() != null) {
                    t = var.getBPELJType();
                }
                String valueClassName = t;
                if (WSDLJavaMatching.isJavaPrimitiveType(t)) {
                    valueClassName = WSDLJavaMatching.getWrapper(t);
                } else if ("byte[]".equals(t)) {
                    valueClassName = "[B";
                }
                // ClassLoader classL = Thread.currentThread().getContextClassLoader();
                ClassLoader classL = ClassLoaderMap.getCL(processName);
                Class cl = Class.forName(valueClassName, true, classL);
                val = new BPELVariable(var.getVariableName(), t, cl);
            }
            // Creates the variable with the associated bpiId, scopeId and value
            VariableCache.createVariable(var.getVariableName(), var.getMessageType(), var.getVarType(), var.getElement(), var.getBPELJType(), bpiId, scopeId, val);

        } catch (ClassNotFoundException e) {
            throw new BPELEngineException("ClassNotFoundException EngineBean createVariable :\n" , e);
        } catch (BPELUtilException e) {
            throw new BPELEngineException("BPELUtilException EngineBean createVariable :\n" , e);
        }
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("EngineBean end createVariable");
    }

    /**
     * send a message to BPIMdb to say I am on a flow
     *
     * @param bpiId
     * @param bpuId
     * @param f
     * @param scopeId
     * @ejb:interface-method
     */
    public void sendFlowBPI(int bpiId, int bpuId, org.objectweb.orchestra.tree.Flow f, int scopeId) throws BPELEngineException {
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("EngineBean sendFlowBPI, BPI = " + bpiId + "  BPU = " + bpuId);
        Vector v = new Vector();
        v.addElement("brancheNb"); v.addElement(f.getNb());
        v.addElement("name"); v.addElement(f.getName());
        v.addElement("joinCondition"); v.addElement(f.getJoinCondition());
        v.addElement("suppressJoinFailure"); v.addElement(f.getSuppressJoinFailure());

        int childrenNb = f.getChildNb();
        org.objectweb.orchestra.tree.Statement[] tab = f.getChilds();
        int child;
        int childNb = 0;
        //give to BPI all children which must be executed
        for (int i = 0; i < childrenNb; i++) {
            if (!tab[i].getType().equals("Target") && !tab[i].getType().equals("Source")) {
                child = tab[i].getNb().intValue();
                v.addElement("branche" + childNb); v.addElement(new Integer(child));
                childNb++;
            }
        }
        v.addElement("childNb"); v.addElement(new Integer(childNb));

        sendMessageBPI(bpiId, bpuId, scopeId, "flow", v);
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("EngineBean end sendFlowBPI");
    }

    /**
     * sendForEachBPI
     * Send a message to BPIMdb to say I am on a parallel ForEach
     * @author Alexandre Greco
     *
     * @param bpiId Current BPI ID
     * @param bpuId Current BPU ID
     * @param f ForEach statement
     * @param scopeId Current Scope ID
     *
     * @throws BPELEngineException
     * @ejb:interface-method
     */
    public void sendForEachBPI(int bpiId, int bpuId, org.objectweb.orchestra.tree.ForEach f, int scopeId) throws BPELEngineException {

    		BPELTrace.traceEngine("EngineBean sendForEachBPI, BPI = " + bpiId + "  BPU = " + bpuId);
        Vector v = new Vector();
        v.addElement("brancheNb"); v.addElement(f.getNb());
        v.addElement("name"); v.addElement(f.getName());
        v.addElement("joinCondition"); v.addElement(f.getJoinCondition());
        v.addElement("suppressJoinFailure"); v.addElement(f.getSuppressJoinFailure());

        int childrenNb = f.getChildNb();
        org.objectweb.orchestra.tree.Statement[] tab = f.getChilds();
        int child;

        // Retrieving scope branch number
        for (int i = 0; i < childrenNb; i++) {
            if ( tab[i].getType().equals("Scope") ) {
                child = tab[i].getNb().intValue();
                v.addElement("branche"); v.addElement(new Integer(child));
            }
        }
        // Scope branch number retrieved

        v.addElement("startCounterValue"); v.addElement(new Integer(f.getStartCounterValue())) ;
        v.addElement("counterName"); v.addElement(f.getCounterName()) ;
        v.addElement("vChildNb"); v.addElement(new Integer(f.getFinalCounterValue() - f.getStartCounterValue() + 1)) ; // Number of branches to launch

        sendMessageBPI(bpiId, bpuId, scopeId, "forEach", v);
        BPELTrace.traceEngine("EngineBean end sendForEachBPI");
    }

    public void sendMessageBPI(int bpiId, int bpuId, int scopeId, String type, Vector v) throws BPELEngineException {
        int size = v.size();
        String[] argsName = new String[size/2];
        Object[] argsValue = new Object[size/2];
        int j=0;
        for (int i = 0; i < size-1 ; i = i+2) {
            argsName[j] = (String)v.get(i);
            argsValue[j] = v.get(i+1);
            j++;
        }
        sendMessageBPI(bpiId,bpuId,scopeId,type,argsName,argsValue);
    }
    /**
     * send a message to BPIMdb
     *
     * @param bpiId
     * @param bpuId
     * @param c
     * @param scopeId
     * @ejb:interface-method
     */
    public void sendMessageBPI(int bpiId, int bpuId, int scopeId, String type, String[] argsName, Object[] argsValue) throws BPELEngineException {
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("EngineBean sendMessageBPI, BPI = " + bpiId + "  BPU = " + bpuId+", scopeId = "+scopeId+", type = "+type);
        if (argsName.length != argsValue.length) {
            throw new BPELEngineException("BPELEngineException EngineBean sendMessageBPI argsName and argsValue have not the same number of elements",null);
        }

        try {
            int size = argsName.length;
            String newNames[] = new String[size+2];
            Object newValues[] = new Object[size+2];

            for (int i=0;i<size;i++) {
                newNames[i] = argsName[i];
                newValues[i] = argsValue[i];
            }
            newNames[size] = "bpuId";
            newValues[size] = new Integer(bpuId);
            newNames[size+1] = "scopeId";
            newValues[size+1] = new Integer(scopeId);
            JmsUtil.sendMessageBPI("BPU",bpiId,type,newNames,newValues);
        } catch (BPELUtilException ex) {
            throw new BPELEngineException("JMSException EngineBean sendMessageBPI with type : "+type+" :\n" , ex);
        }
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("EngineBean end sendMessageBPI");
    }

    /**
     * send a message to BPIMdb to say I an on a target statement
     *
     * @ejb:interface-method
     */
    public void sendTargetsBPI(int bpiId, int bpuId, org.objectweb.orchestra.tree.Activity a, int scopeId) throws BPELEngineException {
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("EngineBean sendTargetsBPI, BPI = " + bpiId + "  BPU = " + bpuId);

        Vector v = new Vector();
        v.addElement("brancheNb"); v.addElement(a.getNb());

        int childrenNb = a.getChildNb();
        int targetNb = 0;
        org.objectweb.orchestra.tree.Statement[] tab = a.getChilds();
        org.objectweb.orchestra.tree.Target t;
        for (int i = 0; i < childrenNb; i++) {
            if (tab[i].getType().equals("Target")) {
                targetNb++;
                t = (org.objectweb.orchestra.tree.Target) tab[i];
                v.addElement("branche" + targetNb); v.addElement(t.getNb());
                v.addElement("linkName" + targetNb); v.addElement(t.getLinkName());
            }
        }

        v.addElement("childNb"); v.addElement(new Integer(targetNb));
        v.addElement("joinCondition"); v.addElement(a.getJoinCondition());
        v.addElement("suppressJoinFailure"); v.addElement(a.getSuppressJoinFailure());


        sendMessageBPI(bpiId, bpuId, scopeId, "targets", v);
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("EngineBean end sendTargetsBPI");
    }

    /**
     * get the range of the first var with the closest scope
     *
     * @param vars Vector of Variables
     * @param scopeId Scope Id
     * @return the rank of the first Variable by Scope
     * @ejb:interface-method
     */
    public VariableData getFirstVarByScope(int bpiId, String name, int scopeId) throws BPELEngineException {
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("EngineBean getFirstVarByScope, scopeId = " + scopeId );
        Vector vars = (Vector) VariableCache.getVariables(bpiId, name, scopeId);
        VariableData result = null;
        if (vars.size() != 0)
            result = (VariableData) vars.get(0);
        else {
            ScopeData sFather = ScopeCache.getFather(scopeId);
            if (sFather != null) {
                int idScope = sFather.getScopeId();
                result = getFirstVarByScope(bpiId, name, idScope);
            } else {
                throw new BPELEngineException("Error : variable not found in the scope and its fathers, variableName = "+name, null);
            }
        }
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("EngineBean end getFirstVarByScope");
        return result;
    }

    /**
     * get the rank of the first cs with the closest scope
     *
     * @param css Vector of CS
     * @param scopeId The Scope Id
     * @return Rank of the first CS by scope
     * @ejb:interface-method
     */
    public CSData getFirstCSByScope(int bpiId, String csName, int scopeId) throws BPELEngineException {
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("EngineBean getFirstCSByScope, scopeId = " + scopeId);
        Vector css = CSCache.findByBpiNameAndScope(bpiId, csName, scopeId);
        CSData result = null;
        if (css.size() != 0)
            result = (CSData) css.get(0);
        else {
            ScopeData sFather = ScopeCache.getFather(scopeId);
            if (sFather != null) {
                int idScope = sFather.getScopeId();
                result = getFirstCSByScope(bpiId, csName,idScope);
            } else {
                throw new BPELEngineException("Error : CS not found in the scope and its fathers", null);
            }
        }
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("EngineBean end getFirstCSByScope");
        return result;
    }

    /**
     * set a message to a variable
     *
     * @ejb:interface-method
     */
    public void setMessageToVariable(int bpiId, String varName, Integer messageId, int scopeId, String processName, int statementNb) throws BPELEngineException {
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("EngineBean setMessageToVariable, BPI = " + bpiId + " scopeId = " + scopeId + "  varName = " + varName + "  messageId = " + messageId);
        try {
            VariableData v = getVar(bpiId, varName, scopeId);
            MessageData md = MessageCache.findMessage(messageId.intValue());
            MessageCache.setMessageStatementNb(messageId.intValue(), statementNb);
            BPELMessage jmsg = (BPELMessage) Util.getValueObject(md.getValue(),processName);
            Object value = jmsg.clone();
            VariableCache.setVariableValue(v.getVariableId(), value, bpiId);
        } catch (BPELUtilException ex) {
            throw new BPELEngineException("BPELUtilException EngineBean setMessageToVariable :\n" , ex);
        }
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("EngineBean end setMessageToVariable");

    }

    /**
     * get a var thanks to its bpi, name and scope in local mode
     *
     * @param bpiId The BPI Id
     * @param name The variable name
     * @param scopeId The Scope Id
     * @return the variable entity bean
     * @ejb:interface-method
     */
    public VariableData getVar(int bpiId, String name, int scopeId) throws BPELEngineException {
        //System.out.println("EngineBean getVar, BPI = " + bpiId + "  name = " + name + "  scopeId = " + scopeId);
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("EngineBean getVar, BPI = " + bpiId + "  name = " + name + "  scopeId = " + scopeId);
        VariableData  v = getFirstVarByScope(bpiId, name, scopeId);
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("EngineBean end getVar, v = " + v);
        return v;
    }

    /**
     * Sets the variable value. If the variable is a BPELMessage, then its value is directly set.
     * Otherwise, this is a BPELVariable and this is the BPELVariable value  that is set.
     * @param bpiId the BPI identifier
     * @param variableName the variable name
     * @param scopeId the scope identifier
     * @param processName the process name
     * @param value the new value of the variable
     * @throws BPELEngineException
     * @ejb:interface-method
     */
    public void setVar(int bpiId, String variableName, int scopeId, String processName, Object value) throws BPELEngineException {
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("EngineBean setVar, bpiId=" + bpiId + ", variableName=" + variableName
                    + ", scopeId=" + scopeId + ", processName=" + processName + ", value=" + value
                    + ((value != null)? ", valueClass=" + value.getClass().getName() : "") );
        try {
            VariableData varData = getVar(bpiId, variableName, scopeId);
            Object oldValue = org.objectweb.orchestra.util.Util.getValueObject( (byte[]) varData.getValue(), processName);
            if (oldValue instanceof BPELMessage) {
                VariableCache.setVariableValue(varData.getVariableId(), value, bpiId);
            } else if (oldValue instanceof BPELVariable) {
                BPELVariable bv = (BPELVariable) oldValue;
                bv.setValue(value);
                VariableCache.setVariableValue(varData.getVariableId(), bv, bpiId);
            } else if (oldValue != null) {
                throw new BPELEngineException("The variable " + variableName
                        + " as a value with the unexpected following type :"
                        + oldValue.getClass().getName() + ".", null);
            } else {
                throw new BPELEngineException("The variable " + variableName + " is null, which is unexpected.", null);
            }
        } catch (BPELUtilException e) {
            throw new BPELEngineException("EngineBean.setVar: "
                    + "Exception occured during value saving for a variable"
                    + ", process=" + processName + ", bpiId=" + bpiId + ", scopeId=" + scopeId
                    + ", variableName=" + variableName + ", value=" + value +
                    ((value != null)? ", valueClass=" + value.getClass().getName() : "")
                    , e);
        }
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("EngineBean end setVar");
    }

    /**
     * @param bpiId
     * @param fromVariable
     * @param toVariable
     * @param scopeId
     * @param proc
     * @throws BPELEngineException
     * @ejb:interface-method
     */
    public void copyVariableToVariable(int bpiId, String fromVariable, String toVariable, int scopeId, String proc) throws BPELEngineException {
        Object fromVal = null;
        VariableData fromVar = null;
        VariableData toVar = null;

        fromVar = getVar(bpiId, fromVariable, scopeId);
        toVar = getVar(bpiId, toVariable, scopeId);
        try {
            fromVal = ((BPELMessage) Util.getValueObject(fromVar.getValue(),proc)).clone();
            VariableCache.setVariableValue(toVar.getVariableId(), fromVal, bpiId);
        } catch (BPELUtilException ex) {
            throw new BPELEngineException("BPELUtilException EngineBean copyVariableToVariable :\n" , ex);
        }

    }

    /**
     * copy a var or a part or... to another
     *
     * @param bpiId The BPI Id
     * @param fromType The type of the "From" value
     * @param toType The type of the "To" value
     * @param f From
     * @param t To
     * @param scopeId The Scope Id
     * @param proc Process
     * @ejb:interface-method
     */
    public void copy(int bpiId, int fromType, int toType, org.objectweb.orchestra.tree.From f, org.objectweb.orchestra.tree.To t, int scopeId, String proc) throws BPELEngineException {
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("EngineBean copy, BPI = " + bpiId + "  scopeId = " + scopeId);
        try {
            Object fromVal = null;
            Object toVal = null;
            VariableData fromVar = null;
            VariableData toVar = null;

            //if there is a variable (it is not a partnerlink, literal,
            // expression, opaque
            if (fromType != 4 && fromType != 6 && fromType != 7 && fromType != 8) {
                fromVar = getVar(bpiId, f.getVariable(), scopeId);
            }
            if (toType != 4) {
                toVar = getVar(bpiId, t.getVariable(), scopeId);
            }
            BPELVariable jv = null;

            printCopy(f,t);

            //get the from value (complex)
            if (fromType == 1) {
                //var (message)
                fromVal = ((BPELMessage) Util.getValueObject(fromVar.getValue(),proc)).clone();
            } else if (fromType == 2) {
                //var (!message)
                fromVal = ((BPELVariable) ((BPELVariable) Util.getValueObject(fromVar.getValue(),proc)).clone()).getValue();
            } else if (fromType == 3) {
                //var part
                fromVal = getMsgPart(fromVar, f.getPart(),proc);
            } else if (fromType == 4) {
                //partnerlink
                fromVal = getPartnerLinkEndpoint(bpiId, f.getPartnerLink(), f.getEndpointReference());
            } else if (fromType == 5) {
                //var property
                fromVal = ((BPELMessage) Util.getValueObject(fromVar.getValue(),proc)).clone();
                fromVal = getProperty(fromVal, f.getProperty(),proc);
            } else if (fromType == 6) {
                //expression
                fromVal = getObjectFromExpression(f.getExpression(), bpiId, scopeId, proc);
            } else if (fromType == 7) {
                //literal
            } else if (fromType == 8) {
                //opaque

            } else if (fromType == 9) {
                //var part query
                Object partValue = getMsgPart(fromVar, f.getPart(),proc);
                fromVal = getQuery(f.getQuery(), partValue, proc,bpiId,scopeId);
            }

            //get the to value (complex)
            if (toType == 1) {
                //var (message)
                VariableCache.setVariableValue(toVar.getVariableId(), fromVal, bpiId);
            } else if (toType == 2) {
                //var (!message)
                jv = (BPELVariable) Util.getValueObject(toVar.getValue(),proc);
                fromVal = Util.convertNumber(jv.getValueClass(),fromVal);
                jv.setValue(fromVal);
                VariableCache.setVariableValue(toVar.getVariableId(), jv, bpiId);
            } else if (toType == 3) {
                //var part
                setMsgPart(toVar, t.getPart(), fromVal,proc, bpiId);
            } else if (toType == 4) {
                //partnerlink
                setPartnerLinkEndpoint(bpiId, t.getPartnerLink(), (String)fromVal);
            } else if (toType == 5) {
                //var property
                toVal = setProperty(toVar, t.getProperty(), fromVal, proc);
                VariableCache.setVariableValue(toVar.getVariableId(), toVal, bpiId);
            } else if (toType == 6) {
                //var part query
                Object partValue = getMsgPart(toVar, t.getPart(),proc);
                if (partValue == null) {
                    toVar = setPartToDefault(toVar, t.getPart(),proc, bpiId);
                    partValue = getMsgPart(toVar, t.getPart(),proc);
                }
                String query = t.getQuery();
                //setQuery(query, partValue, fromVal);
                BPELMessage msg = getBPELMessage(toVar,proc);
                String messageName = msg.getMessageName();
                String messagePack = msg.getClass().getPackage().getName();
                setQuery(bpiId, scopeId,query, partValue, fromVal, proc, messageName, messagePack);
                setMsgPart(toVar, t.getPart(), partValue,proc, bpiId);
            }
        } catch (BPELUtilException ex) {
            throw new BPELEngineException("BPELUtilException EngineBean copy :\n" , ex);
        }
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("EngineBean end copy");
    }

    private void printCopy(From f, To t) {
        String st = "copy from";
        if (f.getVariable() != null) {
            st += " var=" + f.getVariable();
        }
        if (f.getPart() != null) {
            st += " part=" + f.getPart();
        }
        if (f.getQuery() != null) {
            st += " query=" + f.getQuery();
        }
        if (f.getProperty() != null) {
            st += " property=" + f.getProperty();
        }
        if (f.getExpression() != null) {
            st += " expression=" + f.getExpression();
        }
        if (f.getPartnerLink() != null) {
            st += " partnerLink=" + f.getPartnerLink();
        }
        if (f.getEndpointReference() != null) {
            st += " endpointReference=" + f.getEndpointReference();
        }
        st += "  to";
        if (t.getVariable() != null) {
            st += " var=" + t.getVariable();
        }
        if (t.getPart() != null) {
            st += " part=" + t.getPart();
        }
        if (t.getQuery() != null) {
            st += " query=" + t.getQuery();
        }
        if (t.getProperty() != null) {
            st += " property=" + t.getProperty();
        }
        if (t.getPartnerLink() != null) {
            st += " partnerLink=" + t.getPartnerLink();
        }
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("EngineBean copy : " + st);
    }

    private VariableData setPartToDefault(VariableData var, String partname, String processName, int bpiId) throws BPELEngineException {
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("EngineBean setPartToDefault, partName = " + partname);

        VariableData result = null;
        //set a part with a default value
        //it allows to create the tree to access a field under the part
        try {
            BPELMessage bpelmsg = (BPELMessage) Util.getValueObject(var.getValue(),processName);
            MessageStructure ms = bpelmsg.getStructure();
            PartStructure[] ps = ms.getParts();
            for (int i = 0; i < ps.length; i++) {
                if (ps[i].getName().equals(partname)) {
                    //Class cl
                    // =Class.forName(WSDLJavaMatching.getJavaType(ps[i].getType(),proc));
                    ClassLoader classL = ClassLoaderMap.getCL(processName);
                    Class cl = Class.forName(WSDLJavaMatching.getWrapperType(ps[i].getType()),true,classL);
                    try {
                        Constructor constr = cl.getDeclaredConstructor((Class[])null);
                        Object tmp = constr.newInstance((Object[])null);
                        bpelmsg.setPart(partname, tmp);
                        result = VariableCache.setVariableValue(var.getVariableId(), bpelmsg, bpiId);
                    } catch (NoSuchMethodException e) {
                        if (BPELTrace.isEngineDebugEnabled())
                            BPELTrace.traceEngine("NoSuchMethodException EngineBean setPartToDefault :\n");
                        result = var;
                    }
                    break;
                }
            }
        } catch (ClassNotFoundException ex) {
            throw new BPELEngineException("ClassNotFoundException EngineBean setPartToDefault :\n" , ex);
        } catch (InstantiationException ex) {
            throw new BPELEngineException("InstantiationException EngineBean setPartToDefault :\n" , ex);
        } catch (IllegalArgumentException ex) {
            throw new BPELEngineException("IllegalArgumentException EngineBean setPartToDefault :\n" , ex);
        } catch (IllegalAccessException ex) {
            throw new BPELEngineException("IllegalAccessException EngineBean setPartToDefault :\n" , ex);
        } catch (InvocationTargetException ex) {
            throw new BPELEngineException("InvocationTargetException EngineBean setPartToDefault :\n" , ex);
        } catch (BPELUtilException ex) {
            throw new BPELEngineException("BPELUtilException EngineBean setPartToDefault :\n" , ex);
        }
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("EngineBean end setPartToDefault");
        return result;
    }

    private BPELMessage getBPELMessage(VariableData var,String processName) throws BPELEngineException {
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("EngineBean getBPELMessage");
        BPELMessage result = null;
        try {
            result = (BPELMessage) ((BPELMessage) Util.getValueObject(var.getValue(),processName)).clone();
        } catch (BPELUtilException ex) {
            throw new BPELEngineException("BPELUtilException EngineBean getBPELMessage :\n" , ex);
        }
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("EngineBean end getBPELMessage, result = " + result);
        return result;
    }

    private Object getMsgPart(VariableData var, String partname,String processName) throws BPELEngineException {
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("EngineBean getMsgPart, partName = " + partname);
        Object result = null;
        try {
            //get the part of a message
            BPELMessage bpelmsg = (BPELMessage) ((BPELMessage) Util.getValueObject(var.getValue(),processName)).clone();
            result = bpelmsg.getPart(partname);
        } catch (BPELUtilException ex) {
            throw new BPELEngineException("BPELUtilException EngineBean getMsgPart :\n" , ex);
        }
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("EngineBean end getMsgPart, result = " + result);
        return result;
    }

    private void setMsgPart(VariableData var, String partname, Object partvalue,String processName, int bpiId) throws BPELEngineException {
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("EngineBean setMsgPart, partName = " + partname);
        //set the part of a message
        try {
            BPELMessage bpelmsg = (BPELMessage) Util.getValueObject(var.getValue(),processName);

            if (partvalue != null && WSDLJavaMatching.getPrimitiveClass(partvalue.getClass()) != partvalue.getClass()) {
                partvalue = Util.convertNumber(bpelmsg,partname,partvalue);
            }
            bpelmsg.setPart(partname, partvalue);
            VariableCache.setVariableValue(var.getVariableId(), bpelmsg, bpiId);
        } catch (BPELUtilException ex) {
            throw new BPELEngineException("BPELUtilException EngineBean setMsgPart :\n" , ex);
        }
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("EngineBean end setMsgPart");
    }

    private Object getQuery(String query, Object partValue, String processName,int bpiId, int scopeId) throws BPELEngineException {
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("EngineBean getQuery, query = " + query);
        Object result = Expression.executeGetQuery(partValue,query,bpiId,scopeId,processName);
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("EngineBean end getQuery, result = " + result);
        return result;
    }


    private void setQuery(int bpiId, int scopeId,String query, Object partValue, Object val, String processName, String messageName, String messagePack) throws BPELEngineException {
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("EngineBean setQuery, query = " + query+", val = "+val);
        Expression.executeSetQuery(bpiId,scopeId,partValue,query,val,processName,messageName, messagePack);
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("EngineBean end setQuery");
    }

    private Object getProperty(Object messValue, String property, String processName) throws BPELEngineException {
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("EngineBean getProperty, property = " + property);
        BPELMessage jm = (BPELMessage) messValue;

        // TODO : Should be performed at compilation time
        // Replace the prefix with actual namespace
        String[] parts = property.split(":");
        String prefix;
        String prop;
        if (parts.length == 2) {
            prefix = parts[0];
            prop = parts[1];
        } else {
            prefix = "targetNamespace";
            prop = property;
        }
        String ns = Expression.getNS(processName, prefix);
        if (ns != null)
            property = ns + "," + prop;
        // End replace prefix

        Object o = jm.getPropertyValue(property);
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("EngineBean end getProperty, result = " + o);
        return o;
    }

    private Object setProperty(VariableData var, String property, Object val, String processName) throws BPELEngineException {
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("EngineBean setProperty, property = " + property);

        BPELMessage jm = null;
        try {
            jm = (BPELMessage) Util.getValueObject(var.getValue(),processName);

            // TODO : Should be performed at compilation time
            // Replace the prefix with actual namespace
            String[] parts = property.split(":");
            String prefix;
            String prop;
            if (parts.length == 2) {
                prefix = parts[0];
                prop = parts[1];
            } else {
                prefix = "targetNamespace";
                prop = property;
            }
            String ns = Expression.getNS(processName, prefix);
            if (ns != null)
                property = ns + "," + prop;
            // End replace prefix

            jm.setPropertyValue(property,val);

        } catch (BPELUtilException ex) {
            throw new BPELEngineException("BPELUtilException EngineBean setProperty :\n" , ex);
        }
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("EngineBean end setProperty, jm = " + jm);
        return jm;
    }

    /**
     * copy a var or a part or... to another
     *
     * @param bpiId the BPI Id
     * @param scopeId The Scope Id
     * @param varName Variable Name
     * @return
     * @ejb:interface-method
     */
    public Object getVariableData(int bpiId, int scopeId, String varName,String processName) throws BPELEngineException {
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("EngineBean getVariableData, varName = " + varName);
        //get the value of a variable (BPELVariable)
        Object result = null;
        try {
            VariableData var = getVar(bpiId, varName, scopeId);
            result = ((BPELVariable) Util.getValueObject(var.getValue(),processName)).getValue();
        } catch (BPELUtilException ex) {
            throw new BPELEngineException("BPELUtilException EngineBean getVariableData :\n" , ex);
        }
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("EngineBean end getVariableData, result = " + result);
        return result;
    }

    /**
     * copy a var or a part or... to another
     *
     * @param bpiId
     * @param scopeId
     * @param varName
     * @param partName
     * @return
     * @ejb:interface-method
     */
    public Object getVariableData(int bpiId, int scopeId, String varName, String partName,String processName) throws BPELEngineException {
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("EngineBean getVariableData, varName = " + varName + "  partName = " + partName);
        //get the value of a part of a variable (BPELMessage)
        Object result = null;
        VariableData var = getVar(bpiId, varName, scopeId);
        result = getMsgPart(var, partName,processName);
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("EngineBean end getVariableData, result = " + result);
        return result;
    }

    /**
     * copy a var or a part or... to another
     *
     * @param bpiId
     * @param scopeId
     * @param varName
     * @param partName
     * @param query
     * @return
     * @ejb:interface-method
     */
    public Object getVariableData(int bpiId, int scopeId, String varName, String partName, String query, String processName) throws BPELEngineException {
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("EngineBean getVariableData, varName = " + varName + "  partName = " + partName + "  query = " + query);
        //get the value of a part of a variable (BPELMessage) after executing
        // the query
        Object result = null;
        VariableData var = getVar(bpiId, varName, scopeId);
        Object partValue = getMsgPart(var, partName,processName);
        result = getQuery(query, partValue,processName,bpiId, scopeId);
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("EngineBean end getVariableData, result = " + result);
        return result;
    }

    /**
     * copy a var or a part or... to another
     *
     * @param bpiId the BPIId
     * @param scopeId the ScopeId
     * @param p the Property
     * @param varName Name of Variable
     * @param property  Property Name
     * @param proc
     * @return
     * @ejb:interface-method
     */
    public Object getVariableProperty(int bpiId, int scopeId, String varName, String property,String processName) throws BPELEngineException {
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("EngineBean getVariableProperty, varName = " + varName + "   property = " + property);
        //get the value of a variable (BPELMessage) thanks to a property
        Object result = null;
        try {
            VariableData var = getVar(bpiId, varName, scopeId);
            result = Util.getValueObject(var.getValue(),processName);
            result = getProperty(result, property,processName);
        } catch (BPELUtilException ex) {
            throw new BPELEngineException("BPELUtilException EngineBean getVariableProperty :\n" , ex);
        }
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("EngineBean end getVariableProperty, result = " + result);
        return result;
    }

    /**
     * Invoke a Web Service
     *
     * @param bpiId
     * @param scopeId
     * @param process
     * @param wsdlBinding
     * @param i
     * @ejb:interface-method
     */
    public boolean invoke(int bpiId, int bpuId, int scopeId, String processName, WsdlBinding wsdlBinding, org.objectweb.orchestra.tree.Invoke i, String targetNamespace) throws BPELEngineException {
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("EngineBean invoke");
        String suffix = "";
        if (BPELEngine.isJbi()) {
            suffix = BPELConstants.MESSAGE_PORT_SUFFIX;
        }
        //execute the invoke statement
        boolean result = true;
        String qnamePortType = i.getInvokePortType();
        String[] service = null;
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("EngineBean invoke qnamePortType = "+qnamePortType);
        try {
            service = UDDIDirectory.getService(processName,qnamePortType);
        } catch (BPELUtilException e) {
            throw new BPELEngineException("BPELUtilException caught in EngineBean.invoke : ",e);
        }

        QName qnameService = new QName(BPELString.getNS(service[0]), BPELString.removeNS(service[0]));
        String endpoint = service[1];
        //String qnameService = service[0]+suffix;
        String port = service[2]+suffix;

        //WARNING !!
        //Here UDDI is coded to return the first port having the correct portType
        //but don't forget, there can have many bindings for one portType
        //and so there can have many ports for a portType
        //In addition there can have many ports for one binding !!
        //To be taken in charge when writing the binding framework
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("EngineBean invoke endpoint/port = "+endpoint+"/"+port);
        if (qnameService == null) {
            throw new BPELEngineException("Service for process : "+processName+", and qnamePortType : "+qnamePortType+" not found in UDDI Directory. Please verify the partner links definitions (local names and namespaces).\n" , null);
        }


        try {
            //get the service class
            String namespace = qnameService.getNamespaceURI();
            //System.out.println("EngineBean, invoke, qnameService="+qnameService+",   namespace="+namespace);
            String packag = Util.getPackageFromNamespace(namespace);
            String serviceLocatorClName = packag + "." + Utils.xmlNameToJavaClass(qnameService.getLocalPart()) + "Locator";

            String address = null;
            Collection plCol = BPICache.getPartnerLinks(bpiId);
            String plName = null;
            for (Iterator iter = plCol.iterator();iter.hasNext();) {
                PartnerLinkData pld = (PartnerLinkData) iter.next();
                if (pld.getPlName().equals(i.getInvokePartnerLink())) {
                    plName = pld.getPlName();
                    address = pld.getPlPartnerRoleEndpoint()+suffix;
                    if (BPELTrace.isDebugDebugEnabled())
                        BPELTrace.traceDebug("******************* EngineBean invoke : plName = "+plName+", address = "+address);

                    break;
                }
            }

            if (BPELTrace.isEngineDebugEnabled())
                BPELTrace.traceEngine("EngineBean invoke : address : " + address + ", port = "+port+", serviceLocatorClName = "+serviceLocatorClName+", service = "+service);

            if (i.getInvokePort() == null) {
                result = executeBFInvoke(i, bpiId, bpuId, processName, scopeId, plName, qnamePortType, port, serviceLocatorClName, address, targetNamespace, wsdlBinding, packag);
            } else {
                // TODO : ?????????
            }
        } catch (Exception ex) {
            throw new BPELEngineException("Exception EngineBean invoke :\n" , ex);
        }
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("EngineBean end invoke");
        return result;
    }

    private boolean executeBFInvoke(org.objectweb.orchestra.tree.Invoke i, int bpiId, int bpuId, String processName, int scopeId, String plName, String qnamePortType, String port, String serviceLocatorClassName, String address, String targetNamespace, WsdlBinding wsdlBinding, String packag) throws BPELEngineException {
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("EngineBean executeBFInvoke");
        try {
            Method met = null;
            ClassLoader classL = ClassLoaderMap.getCL(processName);
            if (BPELTrace.isEngineDebugEnabled())
                BPELTrace.traceEngine("EngineBean executeBFInovoke  classL = " + classL);
            Class cl = Class.forName(serviceLocatorClassName, true, classL);
            Constructor constr = cl.getConstructor((Class [])null);
            //service is an instance of the service
            Object service = constr.newInstance((Object [])null);

            String inputVar = i.getInputVariable();
            String outputVar = i.getOutputVariable();

            met = service.getClass().getMethod("get"+port,(Class [])null);
            Object stub = met.invoke(service, (Object [])null);
            cl = stub.getClass();
            //TEMPORARY SET THE USERNAME AND THE PASSWORD
            /*
             met = cl.getMethod("setUsername",new Class[]{String.class});
             met.invoke(stub,new Object[]{"jonas"});
             met = cl.getMethod("setPassword",new Class[]{String.class});
             met.invoke(stub,new Object[]{"jonas"});
             */
            //END OF TEMP CODE

            // construct the input parameters tab
            met = Util.getMethodByName(cl, Utils.xmlNameToJava(i.getInvokeOperation()));
            boolean requestResponse = false;
            BPELMessage jmsgin = null;
            BPELMessage jmsgout = null;
            if (inputVar != null) {
                String input = inputVar;
                String output = null;
                if (outputVar != null) {
                    output = outputVar;
                    requestResponse = true;
                }
                VariableData in = getVar(bpiId, input, scopeId);
                Object inValue = Util.getValueObject(in.getValue(),processName);
                jmsgin = (BPELMessage) inValue;

                if (output != null) {
                    VariableData out = getVar(bpiId, output, scopeId);
                    Object outValue = Util.getValueObject(out.getValue(),processName);
                    jmsgout = (BPELMessage) outValue;
                }
            }

            // The following is to just illustrate the listener
            OrchestraServiceListener osl = new OrchestraServiceListener();
            //boolean retData = false;
            boolean ret = false;
            if (BPELTrace.isEngineDebugEnabled())
                BPELTrace.traceEngine("EngineBean executeBFInvoke : before calling method : "+met);
            if (BPELTrace.isDebugDebugEnabled())
                BPELTrace.traceDebug("******************* EngineBean executeBFInvoke : before call, requestResponse = "+requestResponse);
            if (requestResponse) {
                //request-response WS, we are waiting for a response so, an InvokerMdbBean will be created
                InvokerObject io = new InvokerObject(plName, qnamePortType,port, serviceLocatorClassName,i.getInvokeOperation(),jmsgin,jmsgout,address,requestResponse,processName,bpiId,scopeId,wsdlBinding,i,packag,targetNamespace,osl);
                if (BPELTrace.isDebugDebugEnabled())
                    BPELTrace.traceDebug("******************* EngineBean executeBFInvoke : before call, io = "+io+", processName = "+processName);
                JmsUtil.sendMessageInvoker(io,processName);
            } else {
                //oneway WS, we do not desynchronize
                ret = ServiceImpl.invoke(classL, plName, qnamePortType, port, serviceLocatorClassName,i.getInvokeOperation(), jmsgin, jmsgout, address, requestResponse, processName, bpiId, osl);
                if (osl.getBfResult().getException() != null) {
                    manageFault(osl.getBfResult().getException(),wsdlBinding,qnamePortType,i,processName,packag,targetNamespace,bpiId,bpuId,scopeId);
                }
            }
            if (BPELTrace.isEngineDebugEnabled())
                BPELTrace.traceEngine("EngineBean executeBFInvoke : after call: " + ret);
            if (BPELTrace.isDebugDebugEnabled())
                BPELTrace.traceDebug("******************* EngineBean executeBFInvoke : after call");
            /*
                 while (!retData) {
                 synchronized (invokeLock) {
                 try {
                 invokeLock.wait();
                 retData = true;
                 } catch (InterruptedException ie) { }
                 }
                 }
             */
        } catch (BPELBindingFrameworkException ex) {
            throw new BPELEngineException("BPELBindingFrameworkException EngineBean executeBFInvoke :\n" , ex);
        } catch (BPELUtilException ex) {
            throw new BPELEngineException("BPELUtilException EngineBean executeBFInvoke :\n" , ex);
        } catch (ClassNotFoundException ex) {
            throw new BPELEngineException("ClassNotFoundException EngineBean executeBFInvoke :\n" , ex);
        } catch (NoSuchMethodException ex) {
            throw new BPELEngineException("NoSuchMethodException EngineBean executeBFInvoke :\n" , ex);
        } catch (InvocationTargetException ex) {
            throw new BPELEngineException("InvocationTargetException EngineBean executeBFInvoke :\n" , ex);
        } catch (InstantiationException ex) {
            throw new BPELEngineException("InstantiationException EngineBean executeBFInvoke :\n" , ex);
        } catch (IllegalAccessException ex) {
            throw new BPELEngineException("IllegalAccessException EngineBean executeBFInvoke :\n" , ex);
        }
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("EngineBean end executeBFInvoke");
        return true;
    }

    /**
     * @ejb:interface-method
     */
    public void manageInvokeResponse(OrchestraServiceListener osl, int bpiId, int scopeId, String outputVar, String processName) throws BPELEngineException {
        try {
            ServiceEvent se = osl.getBfResult();
            ServiceData sd = se.getServiceData();
            Object result = sd.getResult();
            Map outputMap = sd.getOutputMap();
            Set set = outputMap.entrySet();

            VariableData var = getVar(bpiId, outputVar, scopeId);
            Object value = Util.getValueObject(var.getValue(),processName);
            //if the output var (bpel) is a BPELMessage message
            if (value.getClass().getSuperclass().getName().equals("org.objectweb.orchestra.engine.BPELMessage")) {
                //each part : for each part put the new value
                BPELMessage jm = (BPELMessage) value;
                Class cl = jm.getClass();
                Constructor constr = cl.getDeclaredConstructor((Class [])null);
                BPELMessage jmNew = (BPELMessage) constr.newInstance((Object [])null);
                MessageStructure ms = jmNew.getStructure();
                PartStructure[] ps = ms.getParts();
                int[] usedParts = new int[ps.length];
                for (int k = 0; k < ps.length; k++) {
                    usedParts[k] = -1;
                    String partName = ps[k].getName();
                    String partElement = ps[k].getElement();
                    String key = partName;
                    if (partElement != null) {
                        key = partElement;
                    }
                    Object tmp = getOutputParam(set, key, usedParts, k);
                    if (tmp != null) {
                        jmNew.setPart(partName, tmp);
                    }
                }
                if (result != null) {
                    for (int k = 0; k < ps.length; k++) {
                        if (usedParts[k] == -1) {
                            jmNew.setPart(ps[k].getName(), result);
                        }
                    }
                }
                VariableCache.setVariableValue(var.getVariableId(), jmNew, bpiId);
            } else {
                //it is a BPELVariable
                if (result == null && set != null && set.size() > 0) {
                    Map.Entry entry = (Map.Entry) set.iterator().next();
                    result = entry.getValue();
                }
                BPELVariable jv = (BPELVariable) value;
                jv.setValue(result);
                VariableCache.setVariableValue(var.getVariableId(), jv, bpiId);
            }
        } catch(BPELUtilException e) {
            throw new BPELEngineException("BPELUtilException EngineBean manageInvokeResponse :\n" , e);
        } catch(NoSuchMethodException e) {
            throw new BPELEngineException("NoSuchMethodExceptionilException EngineBean manageInvokeResponse :\n" , e);
        } catch(InstantiationException e) {
            throw new BPELEngineException("InstantiationException EngineBean manageInvokeResponse :\n" , e);
        } catch(IllegalAccessException e) {
            throw new BPELEngineException("IllegalAccessException EngineBean manageInvokeResponse :\n" , e);
        } catch(InvocationTargetException e) {
            throw new BPELEngineException("InvocationTargetException EngineBean manageInvokeResponse :\n" , e);
        }
    }

    /**
     * @ejb:interface-method
     */
    public void manageFault(Exception e, WsdlBinding wsdlBinding, String qnamePortType, org.objectweb.orchestra.tree.Invoke i, String processName, String packag, String targetNamespace, int bpiId, int bpuId, int scopeId) throws BPELEngineException {
        if (BPELTrace.isEngineDebugEnabled()) {
            BPELTrace.traceEngine("EngineBean manageFault : Fault : "+e);
            BPELTrace.traceEngine("EngineBean manageFault : FaultCause : "+e.getCause());
            BPELTrace.traceEngine("EngineBean manageFault : FaultMessage : "+e.getMessage());
        }
        AxisFault af = (AxisFault) e.getCause();
        Element[] elems = af.getFaultDetails();
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("EngineBean manageFault : elems : " + elems.length + ", " + elems[0]);
        String exceptionData = XMLUtils.getInnerXMLString(elems[0]);
        String exceptionName = XMLUtils.getInnerXMLString(elems[1]);

        String faultMessage = exceptionName;
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("EngineBean : manageFault : faultMessage = " + faultMessage);
        String faultMessageNS = Util.replacePackageByNS(faultMessage);
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("EngineBean : manageFault : faultMessageNS = " + faultMessageNS);
        String faultName = wsdlBinding.getFaultName(qnamePortType, i.getInvokeOperation(), faultMessage);

        String bmclassName = packag + "." + BPELConstants.BM + processName + BPELString.toLower(BPELString.removePackage(faultMessage));
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("EngineBean : manageFault : bmclassName : " + bmclassName);
        Class clFault = null;
        try {
            ClassLoader classL = ClassLoaderMap.getCL(processName);
            clFault = Class.forName(bmclassName,true,classL);
        } catch (ClassNotFoundException ex) {
            throw new BPELEngineException("ClassNotFoundException EngineBean manageFault :\n" , ex);
        } catch (BPELUtilException ex) {
            throw new BPELEngineException("BPELUtilException EngineBean manageFault :\n" , ex);
        }
        BPELMessage bmFault = null;
        try {
            bmFault = (BPELMessage) clFault.newInstance();
        } catch (IllegalAccessException ex) {
            throw new BPELEngineException("IllegalAccessException EngineBean manageFault :\n" , ex);
        } catch (InstantiationException ex) {
            throw new BPELEngineException("InstantiationException EngineBean manageFault :\n" , ex);
        }

        MessageStructure msFault = bmFault.getStructure();
        PartStructure[] psFault = msFault.getParts();
        for (int k = 0; k < psFault.length; k++) {
            bmFault.setPart(psFault[k].getName(), exceptionData);
        }

        int messageId = -1;
        // create a Message from the value contained in the BPELMessage
        messageId = MessageCache.createFaultMessage(processName,targetNamespace, bmFault, i.getNb().intValue(), faultName, faultMessageNS, bpiId);
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("EngineBean : manageFault, Fault messageId: " + messageId);
        sendMessageBPI(bpiId, bpuId, scopeId, "fault", new String[]{"faultName","faultMessage","messageId"}, new Object[]{faultName,faultMessageNS,new Integer(messageId)});
        //end Fault Management
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("EngineBean manageFault : End Fault");
    }
    private Object getOutputParam(Set set, String key, int[] usedParts, int index) {
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("EngineBean getOutputParam key = " + key);
        //get the output param corresponding to a key
        Object value = null;
        Map.Entry entry = null;
        for (Iterator n = set.iterator(); n.hasNext();) {
            entry = (Map.Entry) n.next();
            if (entry.getKey().toString().equals(key)) {
                value = entry.getValue();
                usedParts[index] = 1;
                break;
            }
        }
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("EngineBean end getOutputParam value = " + value);
        return value;
    }

    private Object getObjectFromExpression(String expr, int bpiId, int scopeId, String processName) throws BPELEngineException {
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("EngineBean getObjectFromExpression  expr = " + expr);
        //get an object value based on a xpath expression
        //System.out.println("**** getObjectFromExpression : "+expr);
        Object result = Expression.analyseExpression(expr, bpiId, scopeId, processName, BPELConstants.GENERAL_EXPR);
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("EngineBean end getObjectFromExpression  result = " + result);
        if (result != null) {
            if (BPELTrace.isEngineDebugEnabled())
                BPELTrace.traceEngine("EngineBean end getObjectFromExpression  result.class=" + result.getClass());
        }
        return result;
    }

    /**
     * @ejb:interface-method
     */
    /*
     public String getExprValue(String exp, int bpiId, int scopeId, Property p,String processName) throws BPELEngineException {
     if (BPELTrace.isEngineDebugEnabled())
     BPELTrace.traceEngine("EngineBean getExprValue  expr = " + exp);
     //get an object value based on an expression starting by
      // getVariableData or getVariableProperty
       String expr = new String (exp);
       if (expr.indexOf("getVariableData") != -1) {
       expr = expr.substring(expr.indexOf("(") + 1, expr.indexOf(")"));
       String[] param = expr.split(",");
       for (int i = 0; i < param.length; i++) {
       param[i] = param[i].substring(param[i].indexOf("'") + 1);
       param[i] = param[i].substring(0, param[i].indexOf("'"));
       }
       if (param.length == 1) {
       expr = (String) getVariableData(bpiId, scopeId, param[0],processName);
       } else if (param.length == 2) {
       expr = (String) getVariableData(bpiId, scopeId, param[0], param[1],processName);
       } else {
       expr = (String) getVariableData(bpiId, scopeId, param[0], param[1], Util.getQueryFromGetVariableData(exp),processName);
       }
       } else if (expr.indexOf("getVariableProperty") != -1) {
       String[] param = expr.split(",");
       for (int i = 0; i < param.length; i++) {
       param[i] = param[i].substring(param[i].indexOf("'") + 1);
       param[i] = param[i].substring(0, param[i].indexOf("'"));
       }
       expr = (String) getVariableProperty(bpiId, scopeId, p, param[0], param[1],processName);
       }
       if (BPELTrace.isEngineDebugEnabled())
       BPELTrace.traceEngine("EngineBean end getExprValue  expr = " + new String(expr));
       return new String(expr);
       }
     */
    /**
     * return the value of the expr getVariableData(...)
     *
     * @param expr
     * @param bpiId The BPI Id
     * @param scopeId The Scope Id
     * @return The result
     * @ejb:interface-method
     */
    /*
     public Object getDataValue(String exp, int bpiId, int scopeId,String processName) throws BPELEngineException {
     if (BPELTrace.isEngineDebugEnabled())
     BPELTrace.traceEngine("EngineBean getDataValue  exp = " + exp);
     String expr = new String(exp);
     expr = expr.substring(expr.indexOf("(") + 1, expr.lastIndexOf(")"));
     String[] param = expr.split(",");
     Object result = null;
     for (int i = 0; i < param.length; i++) {
     param[i] = param[i].substring(param[i].indexOf("'") + 1);
     param[i] = param[i].substring(0, param[i].indexOf("'"));
     }
     if (param.length == 1) {
     result = getVariableData(bpiId, scopeId, param[0],processName);
     } else if (param.length == 2) {
     result = getVariableData(bpiId, scopeId, param[0], param[1],processName);
     } else {
     result = getVariableData(bpiId, scopeId, param[0], param[1], Util.getQueryFromGetVariableData(exp),processName);
     }
     if (BPELTrace.isEngineDebugEnabled())
     BPELTrace.traceEngine("EngineBean end getDataValue  result = " + result);
     return result;
     }
     */
    /**
     * return the value of the expr getVariableProperty(var,prop)
     *
     * @param expr
     * @param bpiId The BPI Id
     * @param scopeId The Scope Id
     * @param p Property
     * @param proc The Process
     * @return Value of the Property
     * @ejb:interface-method
     */
    public Object getPropertyValue(String expr, int bpiId, int scopeId, String processName) throws BPELEngineException {
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("EngineBean getPropertyValue  expr = " + expr);
        Object result = null;
        String[] param = expr.split(",");
        for (int i = 0; i < param.length; i++) {
            param[i] = param[i].substring(param[i].indexOf("'") + 1);
            param[i] = param[i].substring(0, param[i].indexOf("'"));
        }
        result = getVariableProperty(bpiId, scopeId, param[0], param[1],processName);
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("EngineBean end getPropertyValue result = " + result);
        return result;
    }

    /**
     * assign a scope to a correlationSet if it exists, creates it if not
     *
     * @param bpiId The BPI Id
     * @param scopeId The Scope Id
     * @param c The Correlation Set
     * @ejb:interface-method
     */
    public void assignScopeToCorrelationSet(int bpiId, int scopeId, org.objectweb.orchestra.tree.CorrelationSet c) throws BPELEngineException {
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("EngineBean assignScopeToCorrelationSet  bpiId = " + bpiId + "  scopeId = " + scopeId + "  CSName = " + c.getName());
        String csName = c.getName();
        CSData csd = null;
        CSData tmp = null;
        Collection csList = null;

        csList = CSCache.findByNameAndBpi(csName, bpiId);
        for (Iterator it = csList.iterator(); it.hasNext();) {
            tmp = (CSData) it.next();
            if (tmp.getScopeId() == -1) {
                csd = tmp;
            }
        }
        if (csd == null) {
            String[] props = c.getProperties().split(" ");
            CSProperties CSVal = new CSProperties();
            for (int i = 0; i < props.length; i++) {
                CSVal.addProperty(props[i]);
            }
            CSCache.createCS(bpiId, csName, CSVal,scopeId);
        } else
            CSCache.setCSScope(csd.getCSId(), scopeId);
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("EngineBean end assignScopeToCorrelationSet");
    }

    /**
     * assign a value to a correlation
     *
     * @ejb:interface-method
     */
    public void setCorrelationValue(int bpiId, int scopeId, org.objectweb.orchestra.tree.Correlation c, String inputVar, String outputVar, String processName) throws BPELEngineException {
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("EngineBean setCorrelationValue  bpiId = " + bpiId + "  scopeId = " + scopeId + "  set = " + c.getCorrelationSet() + "  initiate = " + c.getInitiate() + "  pattern = " + c.getPattern());
        try {
            if (c.getInitiate().equals("yes")) {
                CSData csd = null;
                String csName = c.getCorrelationSet();

                csd = getFirstCSByScope(bpiId, csName, scopeId);
                Object csLocalValue = csd.getCSValue();
                CSProperties CSVal = (CSProperties) Util.getValueObject(csLocalValue,processName);
                if (CSVal.getPropertiesSize() > 0 && CSVal.getValue(0) == null) {
                    VariableData vin = getVar(bpiId, inputVar, scopeId);
                    BPELMessage jmin = (BPELMessage) Util.getValueObject(vin.getValue(),processName);
                    if (outputVar == null) {
                        //reply or invoke one-way
                        setCSValuesFromBPELMessage(CSVal, jmin);
                    } else {
                        //invoke request-response
                        if (c.getPattern().equals("out") || c.getPattern().equals("out-in")) {
                            setCSValuesFromBPELMessage(CSVal, jmin);
                        } else {
                            VariableData vout = getVar(bpiId, outputVar, scopeId);
                            BPELMessage jmout = (BPELMessage) Util.getValueObject(vout.getValue(),processName);
                            setCSValuesFromBPELMessage(CSVal, jmout);
                        }
                    }
                    CSCache.setCSValue(csd.getCSId(), CSVal);
                }
            }
        } catch (BPELUtilException ex) {
            throw new BPELEngineException("BPELUtilException EngineBean setCorrelationValue :\n" , ex);
        }
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("EngineBean end setCorrelationValue");
    }

    private void setCSValuesFromBPELMessage(CSProperties CSVal, BPELMessage jm) {
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("EngineBean setCSValuesFromBPELMessage, CSVal = "+CSVal+", jm = "+CSVal);
        String prop = null;
        for (int j = 0; j < CSVal.getPropertiesSize(); j++) {
            prop = CSVal.getProperty(j);
            CSVal.setValue(prop, jm.getPropertyValue(prop));
        }
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("EngineBean end setCSValuesFromBPELMessage");
    }

    /**
     * create a partnerLink
     *
     * @ejb:interface-method
     */
    public void createPartnerLink(int bpiId, String processName,String plName, QName plType, String plMyRole, String plPartnerRole, QName plMyRolePT, QName plPartnerRolePT, String plMyRoleEndpoint, String plPartnerRoleEndpoint) throws BPELEngineException {
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("EngineBean createPartnerLink, bpiId="+bpiId+", processName="+processName+", plName="+plName+", plType="+plType+", plMyRole="+plMyRole+", plPartnerRole="+plPartnerRole+", plMyRolePT="+plMyRolePT+", plPartnerRolePT="+plPartnerRolePT+", plMyRoleEndpoint="+plMyRoleEndpoint+", plPartnerRoleEndpoint="+plPartnerRoleEndpoint);
        PartnerLinkCache.createPartnerLink(plName, processName, plType, plMyRole, plPartnerRole, plMyRolePT, plPartnerRolePT, plMyRoleEndpoint, plPartnerRoleEndpoint, bpiId);
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("EngineBean end createPartnerLink");
    }

    private String getPartnerLinkEndpoint(int bpiId, String plName, String role) throws BPELEngineException{
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("EngineBean getPartnerLinkEndpoint, bpiId="+bpiId+", plName="+plName+", role="+role);
        String endPoint = null;
        Collection plCol = BPICache.getPartnerLinks(bpiId);
        for (Iterator iter = plCol.iterator();endPoint == null && iter.hasNext();) {
            PartnerLinkData pl = (PartnerLinkData) iter.next();
            if (pl.getPlName().equals(plName)) {
                if (role.equals("partnerRole")) {
                    endPoint = pl.getPlPartnerRoleEndpoint();
                } else {
                    endPoint = pl.getPlMyRoleEndpoint();
                }
            }
        }
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("EngineBean getPartnerLinkEndpoint, endPoint="+endPoint);
        return endPoint;
    }

    private void setPartnerLinkEndpoint(int bpiId, String plName, String newEndpoint) throws BPELEngineException{
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("EngineBean setPartnerLinkEndpoint, bpiId="+bpiId+", plName="+plName+", newEndpoint="+newEndpoint);
        Collection plCol = BPICache.getPartnerLinks(bpiId);
        for (Iterator iter = plCol.iterator();iter.hasNext();) {
            PartnerLinkData pl = (PartnerLinkData) iter.next();
            if (pl.getPlName().equals(plName)) {
                PartnerLinkCache.setPartnerRoleEndpoint(pl.getPlId(), newEndpoint);
                break;
            }
        }
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("EngineBean end setPartnerLinkEndpoint");
    }

    //class end
}