/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.objectweb.orchestra.engine;

import javax.ejb.CreateException;
import javax.ejb.MessageDrivenBean;
import javax.ejb.MessageDrivenContext;
import javax.jms.MapMessage;
import javax.jms.MessageListener;
import javax.naming.NamingException;

import org.objectweb.orchestra.bindingframework.ServiceEvent;
import org.objectweb.orchestra.bindingframework.ServiceImpl;
import org.objectweb.orchestra.util.BPELTrace;
import org.objectweb.orchestra.util.ClassLoaderMap;
import org.objectweb.orchestra.util.Util;


/**
 * InvokerMdbBean is an Message Driven bean
 *
 * @author Goulven Le Jeune & Charles Souillard
 * @version $Id: InvokerMdbBean.java 5 2006-10-20 14:37:44Z lejeuneg $
 *
 * @ejb:message-driven
 *
 * @ejb.bean name="InvokerMdb"
 *           description="Deployment descriptor for the InvokerMdbBean bean"
 *           transaction-type="Container"
 *           acknowledge-mode="Auto-acknowledge"
 *           destination-type="javax.jms.Queue"
 *
 * @ejb:transaction type="NotSupported"
 *
 * @jonas.message-driven-destination jndi-name="InvokerQueue"
 *
 * @jonas.min-pool-size @InvokerMdbBean.min-pool-size@
 * @jonas.max-cache-size @InvokerMdbBean.max-cache-size@
 * @ejb.activation-config-property name="maxNumberOfWorks" value="@InvokerMdbBean.maxNumberOfWorks@"
 *
 *
 * @ejb.security-identity run-as="admin"
 *
 */
public class InvokerMdbBean implements MessageDrivenBean, MessageListener {

    private transient MessageDrivenContext mdbContext;

    public MessageDrivenContext getContext() {
        return this.mdbContext;
    }

    public void onMessage(javax.jms.Message message) {
        if (BPELTrace.isInvokerDebugEnabled())
            BPELTrace.traceInvoker("InvokerMdbBean onMessage");
        MapMessage m = (MapMessage) message;
        org.objectweb.orchestra.engine.EngineLocal engine = null;
        boolean ret = false;
        InvokerObject io = null;
        try {
            byte[] bytes = m.getBytes("invokerObject");
            if (BPELTrace.isInvokerDebugEnabled()) {
                BPELTrace.traceInvoker("InvokerMdbBean bytes = "+bytes);
                BPELTrace.traceInvoker("InvokerMdbBean bytes.length = "+bytes.length);
            }
            String processName = m.getString("processName");
            io = (InvokerObject)Util.getValueObject(bytes,processName);
            engine = getEngine();
            ClassLoader classL = ClassLoaderMap.getCL(io.getProcessName());
            ret = ServiceImpl.invoke(classL, io.getPartnerLinkName(), io.getQnamePortType(),io.getPort(),io.getServiceLocatorClassName(), io.getInvokeOperation(),io.getInBM(),io.getOutBM(), io.getAddress(), io.isRequestResponse(),io.getProcessName(), io.getBpiId(),io.getServiceListener());
            ServiceEvent se = io.getServiceListener().getBfResult();
            if (se.getException() != null) {
                engine.manageFault(se.getException(), io.getWsdlBinding(), io.getQnamePortType(), io.getInvoke(), io.getProcessName(), io.getPackag(), io.getTargetNamespace(), io.getBpiId(), -1, io.getScopeId());
            } else {
                io.setRet(ret);
                engine.sendMessageBPI(io.getBpiId(), -1, io.getScopeId(), "invokeReturn", new String[]{"invokerObject"}, new Object[]{io});
            }
        } catch (Exception e) {
            System.out.println("ERROR : InvokerMdbBean OnMessage Exception caught : bpiId = " + io.getBpiId());
            e.printStackTrace();
            if (BPELTrace.isInvokerDebugEnabled())
                BPELTrace.traceInvoker("InvokerMdbBean Exception, : \n"+Util.getExceptionAsString(e));
            try {
                if (engine == null) {
                    engine = getEngine();
                }
                BPIMap.getBPI(io.getBpiId()).setException(e);
                engine.sendMessageBPI(io.getBpiId(), -1, io.getScopeId(), "fault", new String[]{"faultName","faultMessage","messageId"}, new Object[]{"OrchestraInternalError","",new Integer(-1)});
                engine.remove();
            } catch (Exception ex) {
                ex.printStackTrace();
            }
        }
        if (BPELTrace.isInvokerDebugEnabled())
            BPELTrace.traceInvoker("InvokerMdbBean end onMessage");
    }

    private EngineLocal getEngine() throws NamingException, CreateException {
        EngineHomeLocal ehl = EngineUtil.getLocalHome();
        EngineLocal engine = ehl.create();
        return engine;
    }
    public InvokerMdbBean() {
        // Constructor
    }
    public void setMessageDrivenContext(MessageDrivenContext ctx) {
        this.mdbContext = ctx;
    }
    public void ejbRemove() {
        // Mandatory method
    }
    public void ejbCreate() {
        // Mandatory method

    }

}