/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.objectweb.orchestra.engine;

import java.util.Collection;
import java.util.Iterator;
import java.util.Vector;

import javax.ejb.CreateException;
import javax.ejb.FinderException;
import javax.ejb.RemoveException;
import javax.ejb.SessionBean;
import javax.ejb.SessionContext;
import javax.naming.NamingException;

import org.objectweb.orchestra.data.BPICache;
import org.objectweb.orchestra.data.CSCache;
import org.objectweb.orchestra.data.CSData;
import org.objectweb.orchestra.data.MessageCache;
import org.objectweb.orchestra.data.MessageData;
import org.objectweb.orchestra.exception.BPELEngineException;
import org.objectweb.orchestra.exception.BPELUtilException;
import org.objectweb.orchestra.util.BPELConstants;
import org.objectweb.orchestra.util.BPELTrace;
import org.objectweb.orchestra.util.JmsUtil;
import org.objectweb.orchestra.util.Util;

import org.objectweb.orchestra.data.BPIEntityBeanUtil;
import org.objectweb.orchestra.data.BPIEntityHomeLocal;
import org.objectweb.orchestra.data.BPIEntityLocal;

/**
 * ManagerBean.java.
 *
 * @author Steve Krupanszky, Goulven Le Jeune & Charles Souillard
 * @version $Id: ManagerBean.java 108 2007-01-09 15:55:22Z lejeuneg $
 *
 * @ejb:bean name="Manager"
 *           display-name="Session Manager Bean"
 *           type="Stateless"
 *           transaction-type="Container"
 *           jndi-name="org/objectweb/orchestra/engine/ManagerHome"
 *           local-jndi-name="org/objectweb/orchestra/engine/ManagerHome_L"
 *           view-type="both"
 *
 * @ejb.home local-class="org.objectweb.orchestra.engine.ManagerHomeLocal"
 *           remote-class="org.objectweb.orchestra.engine.ManagerHome"
 * @ejb.interface local-class="org.objectweb.orchestra.engine.ManagerLocal"
 *                remote-class="org.objectweb.orchestra.engine.Manager"
 *
 * @ejb:transaction type="NotSupported"
 *
 * @ejb.ejb-ref ejb-name="BPIEntityBean" view-type="local"
 *              ref-name="ejb/BPIEntityHomeLocal"
 *
 * @ejb.permission role-name="admin"
 *
 * @jonas.max-cache-size @ManagerBean.max-cache-size@
 */
public class ManagerBean implements SessionBean {

    // -------------------------------------------------------------------------
    // Members
    // -------------------------------------------------------------------------
    private SessionContext mContext;
    private static Vector runningBPIs = new Vector();

    // -------------------------------------------------------------------------
    // Methods - SessionBean implementation
    // -------------------------------------------------------------------------

    /**
     * @return SessionContext
     */
    public SessionContext getContext() {
        return this.mContext;
    }

    /**
     * Create the All Projects Session Bean.
     *
     * @throws CreateException : ejb exeception
     *
     * @ejb:create-method
     */
    public void ejbCreate() throws CreateException {
        // Mandatory method
    }

    public void setSessionContext(final SessionContext context) {
        this.mContext = context;
    }

    public void ejbRemove() {
        // Mandatory method
    }

    public void ejbActivate() {
        // Mandatory method
    }

    public void ejbPassivate() {
        // Mandatory method
    }

    // -------------------------------------------------------------------------
    // Methods - public
    // -------------------------------------------------------------------------

    /**
     * Invoke.
     *
     * @param process : Name of the process
     * @param refPartner : Reference of the partner
     * @param portType : Name of the PortType
     * @param operation : Name of the Operation
     * @param jmsg : Message of the invoke
     * @param targetNamespace : targetnamespace
     * @param oneWay : true if invoke is one way
     * @throws BPELEngineException : Engine exception
     *
     * @ejb:interface-method
     *
     */
    public void invoke(final String process, final String refPartner, final String portType, final String operation,
            final BPELMessage jmsg, final String targetNamespace, final boolean oneWay) throws BPELEngineException {
        if (BPELTrace.isManagerDebugEnabled())
            BPELTrace.traceManager("Manager invoke with no message, process: " + process + ", portType: " +
                    portType + ", operation: " + operation+", refPartner = "+refPartner+", targetNamespace = " +
                    targetNamespace+", oneWay = "+oneWay);
        invoke(process, refPartner, portType, operation, jmsg, targetNamespace,oneWay,-1);
        if (BPELTrace.isManagerDebugEnabled())
            BPELTrace.traceManager("Manager end invoke with no message");
    }

    private void invoke(String process, String refPartner, String portType, String operation, BPELMessage jmsg,
            String targetNamespace, boolean oneWay, int messageNb) throws BPELEngineException {
        if (BPELTrace.isManagerDebugEnabled())
            BPELTrace.traceManager("Manager invoke, process: " + process + ", portType: " + portType + ", operation: " +
                    operation+", refPartner = "+refPartner+", targetNamespace = "+targetNamespace+", oneWay = " +
                    oneWay+", messageNb = "+messageNb);
        int messageId = -1;
        Vector pmrs = null;
        ProcessMapReceive pmr = null;
        String csName = null;
        int bpiId = -1;
        String partner = null;
        //Vector[] csBPIIds = null;
        if (BPELTrace.isManagerDebugEnabled())
            BPELTrace.traceManager("Manager invoke, messageNb: " + messageNb);

        if (messageNb == -1) {
            messageId = analyseMessage(process,targetNamespace,portType, operation, refPartner, jmsg);
        } else {
            messageId = messageNb;
        }
        if (BPELTrace.isManagerDebugEnabled())
            BPELTrace.traceManager("Manager invoke, messageId: " + messageId);

        //instantiate a ProcessMap
        ProcessMap.loadProcess(process);

        //retrieve list of the corresponding CS from the ProcessMap
        pmrs = ProcessMap.getPMR(process, portType, operation);
        //getPMR should always return something
        if (pmrs.size() == 0) {
            if (BPELTrace.isManagerDebugEnabled())
                BPELTrace.traceManager("Manager invoke, ProcessMap.getPMR() returned empty list for " + process + " " +
                        portType + " " + operation);
            throw new BPELEngineException("Manager invoke, ProcessMap.getPMR() returned empty list",null);
        }

        //find the exact CS, and corresponding BPI
        pmr = (ProcessMapReceive) pmrs.get(pmrs.size()-1);
        csName = pmr.getCS();
        partner = pmr.getPartner();
        MessageCache.setMessagePartner(messageId, partner);
        if (csName != null) {
            if (BPELTrace.isManagerDebugEnabled())
                BPELTrace.traceManager("Manager invoke, csnames!=null, csName="+csName);
            bpiId = analyseCS(pmrs,process,jmsg,operation,portType,targetNamespace);
        } else { // csnames==null
            if (BPELTrace.isManagerDebugEnabled())
                BPELTrace.traceManager("Manager invoke, csname==null, createInstance = " + pmr.getCreateInstance());
            if (pmr.getCreateInstance().equals("yes")) {
                bpiId = createBPI(process, targetNamespace);
            } else {
                bpiId = findBPI(process, portType, operation);
                if (BPELTrace.isManagerDebugEnabled())
                    BPELTrace.traceManager("Manager invoke, csname==null, bpiId = " + bpiId);
            }
        } //if(csname)

        if (messageNb == -1) {
            linkMessBpiAndSendMessBpi(bpiId,messageId,oneWay,operation,partner,refPartner,portType,process);
        } else {
            linkMessBpi(messageId,bpiId);
        }

        BPI bpi = BPIMap.getBPI(bpiId);
        if (bpi!=null && bpi.needStart()) {
            bpi.start();
        }
        if (BPELTrace.isManagerDebugEnabled())
            BPELTrace.traceManager("Manager end invoke");
    }

    private int analyseCS(Vector pmrs,String process,BPELMessage jmsg,String operation,String portType,String targetNamespace)
    throws BPELEngineException {
        int bpiId = -1;
        synchronized (getClass()) {
            //gate the (potential) creation of a new BPI
            if (BPELTrace.isManagerDebugEnabled())
                BPELTrace.traceManager("Manager analyseCS");
            String[] csPropertiesNames = null;
            Object csPropertyValue = null;
            CSData cs = null;
            Vector[] csBPIIds = new Vector[pmrs.size()];
            ProcessMapReceive pmr = null;
            String csName = null;
            if (BPELTrace.isManagerDebugEnabled())
                BPELTrace.traceManager("Manager analyseCS, csBPIIds.length = " + csBPIIds.length);
            for (int i = 0; i < pmrs.size(); i++) {
                csBPIIds[i] = new Vector();
                pmr = (ProcessMapReceive) pmrs.get(i);
                csName = pmr.getCS();
                csPropertiesNames = pmr.getCSProperties();
                Collection csCol = null;

                csCol = CSCache.findCS(csName);
                if (BPELTrace.isManagerDebugEnabled())
                    BPELTrace.traceManager("Manager analyseCS, csName = " + csName + ", csCol.size() = " + csCol.size());
                for (Iterator it = csCol.iterator(); it.hasNext();) {
                    cs = (CSData) it.next();
                    CSProperties csProperties;
                    try {
                        csProperties = (CSProperties) Util.getValueObject(cs.getCSValue(),process);
                    } catch (BPELUtilException e) {
                        throw new BPELEngineException("BPELUtilException ManagerBean analyseCS :\n" , e);
                    }
                    boolean allmatch = true;
                    for (int j = 0; j < csPropertiesNames.length; j++) {
                        String csPropertyName = csPropertiesNames[j];
                        if (BPELTrace.isManagerDebugEnabled())
                            BPELTrace.traceManager("Manager analyseCS, csPropertyName = " + csPropertyName);
                        if (jmsg.getPropertyValue(csPropertyName) != null) {
                            if (BPELTrace.isManagerDebugEnabled())
                                BPELTrace.traceManager("Manager analyseCS, jmsg.getPropertyValue(csPropertyName) != null");
                            csPropertyValue = csProperties.getValue(csPropertyName);
                            if (csPropertyValue == null) {
                                if (BPELTrace.isManagerDebugEnabled())
                                    BPELTrace.traceManager("Manager analyseCS, csPropertyValue == null");
                                allmatch = false;
                            } else {
                                if (BPELTrace.isManagerDebugEnabled())
                                    BPELTrace.traceManager("Manager analyseCS, csPropertyValue != null");
                                allmatch = allmatch && jmsg.compare(csPropertyName, csPropertyValue);
                            }
                        } else {
                            if (BPELTrace.isManagerDebugEnabled())
                                BPELTrace.traceManager("Manager analyseCS, jmsg.getPropertyValue(csPropertyName) == null");
                            allmatch = false;
                        }
                    }
                    if (BPELTrace.isManagerDebugEnabled())
                        BPELTrace.traceManager("Manager analyseCS, end of for on csPropertiesNames, allmatch = "+allmatch);
                    if (allmatch) {
                        csBPIIds[i].add(new Integer(cs.getBPIId()));
                    }
                }
            }//end for
            //search all bpi having their cs (there can be many cs) corresponding to the desired cs
            Vector commonBPIIds = csBPIIds[0];
            for (int i = 1; i < pmrs.size(); i++) {
                Vector bpiIds = csBPIIds[i];
                commonBPIIds.retainAll(bpiIds);
            }
            /*
            for (int i = 0; i < commonBPIIds.size(); i++) {
                if (BPELTrace.isManagerDebugEnabled())
                BPELTrace.traceManager("Manager analyseCS, commonBPIIds["+i+"] = "+((Integer)commonBPIIds.get(i)).intValue());
            }
             */
            //if there is at least one bpi found (corresponding to cs)
            if (commonBPIIds.size() >= 1) {
                bpiId = ((Integer) commonBPIIds.get(0)).intValue();
                if (BPELTrace.isManagerDebugEnabled())
                    BPELTrace.traceManager("Manager analyseCS, allMatch bpiId = " + bpiId);
                // found the desired CS
            } else {
                if (pmr.getCreateInstance().equals("yes")) {
                    if (BPELTrace.isManagerDebugEnabled())
                        BPELTrace.traceManager("Manager analyseCS, createInstance=yes");
                    bpiId = createBPI(process, targetNamespace);
                } else {
                    if (BPELTrace.isManagerDebugEnabled())
                        BPELTrace.traceManager("Manager analyseCS, createInstance=no");
                    // No existing BPI having all desired CS and createInstance = No => find any process
                    // and initiate CS value
                    bpiId = findBPI(process, portType, operation);
                }

                // Initiate CS Value
                for (int i = 0; i < pmrs.size(); i++) {
                    pmr = (ProcessMapReceive) pmrs.get(i);
                    csPropertiesNames = pmr.getCSProperties();
                    CSProperties csp = new CSProperties();
                    for (int j = 0; j < csPropertiesNames.length; j++) {
                        String csPropertyName = csPropertiesNames[j];
                        csPropertyValue = jmsg.getPropertyValue(csPropertyName);
                        if (csPropertyValue != null) csp.addProperty(csPropertyName, csPropertyValue);
                    }
                    createCS(bpiId, csName, csp);
                }
            }
            if (BPELTrace.isManagerDebugEnabled())
                BPELTrace.traceManager("Manager end analyseCS, bpiId = "+bpiId);
        }
        return bpiId;
    }

    private void invokeWithExistingMessage(MessageData message) throws BPELEngineException {
        if (BPELTrace.isManagerDebugEnabled())
            BPELTrace.traceManager("Manager invokeWithExistingMessage, messageId : " + message.getMessageId());
        String processName = message.getProcessName();
        BPELMessage BPELMess = null;
        try {
            BPELMess = (BPELMessage) Util.getValueObject(message.getValue(),processName);
        } catch (BPELUtilException e) {
            throw new BPELEngineException("BPELUtilException ManagerBean invokeWithExistingMessage :\n" , e);
        }
        invoke(processName, message.getRefPartner(), message.getPortType(), message.getOperation(), BPELMess,
                message.getTargetNamespace(),true,message.getMessageId());
        if (BPELTrace.isManagerDebugEnabled())
            BPELTrace.traceManager("Manager end invokeWithExistingMessage");
    }

    /**
     * analyseMessage
     * @ejb:transaction type="Required"
     */
    private int analyseMessage(String process,String targetNamespace, String portType, String operation, String refPartner,
            BPELMessage jmsg) throws BPELEngineException {
        //      create a Message from the value contained in the BPELMessage
        int messageId;
        messageId = MessageCache.createMessage(process,targetNamespace,portType,operation,refPartner,jmsg);
        return messageId;
    }

    /**
     * linkMessBpi
     */
    private void linkMessBpi(int messageId,int bpiId) throws BPELEngineException{
        if (BPELTrace.isManagerDebugEnabled())
            BPELTrace.traceManager("Manager linkMessBpi, setting message to BPI : " + bpiId);
        //set the relation between the Message and the corresponding BPI
        MessageCache.setMessageBPI(messageId, bpiId);
    }

    /**
     * linkMessBpiAndSendMessBpi
     */
    private void linkMessBpiAndSendMessBpi(int bpiId,int messageId, boolean oneWay,String operation,String partner,
            String refPartner,String portType,String process) throws BPELEngineException{
        if (BPELTrace.isManagerDebugEnabled())
            BPELTrace.traceManager("Manager linkMessBpiAndSendMessBpi, setting message to BPI : " + bpiId+
                    ", messageId = "+messageId);
        linkMessBpi(messageId,bpiId);
        try {
            String[] argsName = new String[]{"messageId","operation","partner","refPartner","portType","process","oneWay"};
            Object[] argsValue = new Object[]{new Integer(messageId),operation,partner,refPartner,
                    portType,process,Boolean.toString(oneWay)};
            JmsUtil.sendMessageBPI("manager",bpiId,"invokeInbound",argsName,argsValue);
        } catch (BPELUtilException ex) {
            throw new BPELEngineException("JMSException Manager linkMessBpiAndSendMessBpi :\n" , ex);
        }
    }

    private void analyseMessagesWithoutBPI() throws BPELEngineException {
        Object[] noBpiMessages = MessageCache.findByNoBpi().toArray();
        for (int i = 0; i < noBpiMessages.length; i++) {
            MessageData message = (MessageData) noBpiMessages[i];
            if (!message.isFault()) {
                //it is not a fault message created by EngineBean
                invokeWithExistingMessage(message);
            }
        }
    }

    /**
     * restart - must be called only once, ever.
     *
     * @ejb:interface-method
     */
    public void restart(String processName, int bpiId) throws BPELEngineException {
        synchronized (getClass()) {
            BPI bpi_ = new BPI(bpiId);
            bpi_.init(processName);
            bpi_.restartBPI();
        }

    }

    /**
     * restart - must be called only once, ever.
     *
     * @ejb:interface-method
     */
    public void restartEngine() throws BPELEngineException {
        if (BPELTrace.isManagerDebugEnabled())
            BPELTrace.traceManager("Manager restartEngine");
        /*
        System.out.println("**************************************");
        System.out.println("Restart not available in this release.");
        System.out.println("**************************************");
         */

        if (runningBPIs.size() == 0) {
            //get all messages having no bpi, analyse them to associate them a bpi
            analyseMessagesWithoutBPI();
            try {
                Object[] allBPI = null;
                BPIEntityHomeLocal bpihl = BPIEntityBeanUtil.getLocalHome();
                allBPI = bpihl.findAllBPI().toArray();
                BPIEntityLocal bpi = null;
                for (int i = 0; i < allBPI.length; i++) {
                    bpi = (BPIEntityLocal) allBPI[i];
                    String state = bpi.getState();
                    if (!state.equals(BPELConstants.BPI_FINISHED)) {
                        String pn = bpi.getProcessName();
                        int bpiId = bpi.getId().intValue();
                        RunningBPI rbpi = new RunningBPI(bpiId, pn);
                        runningBPIs.add(rbpi);
                        if (BPELTrace.isManagerDebugEnabled())
                            BPELTrace.traceManager("Manager restartEngine, add runningBpi : "+bpiId+", with processName : "+pn);
                        if (state.equals(BPELConstants.BPI_ACTIVE)) {
                            restart(bpi.getProcessName(),bpi.getId().intValue());
                        }
                    } else {
                        bpi.remove();
                    }
                }
            } catch (NamingException e) {
                throw new BPELEngineException("NamingException ManagerBean restartEngine :\n" , e);
            } catch (FinderException e) {
                throw new BPELEngineException("FinderException ManagerBean restartEngine :\n" , e);
            } catch (RemoveException e) {
                throw new BPELEngineException("RemoveException ManagerBean restartEngine :\n" , e);
            }

        } else {
            if (BPELTrace.isManagerDebugEnabled())
                BPELTrace.traceManager("the engine is already running. Don't need to restart it");
        }
        if (BPELTrace.isManagerDebugEnabled())
            BPELTrace.traceManager("Manager end restartEngine");
    }

    /**
     * Notify that the BPI is waiting for a call
     *
     * @param bpiId the bpiId
     * @param pt the portType
     * @param op the Operation
     * @ejb:interface-method
     */
    public void bpiWaiting(int bpiId, String pt, String op) {
        if (BPELTrace.isManagerDebugEnabled())
            BPELTrace.traceManager("Manager bpiWaiting : bpiId :" + bpiId + ", portType : " + pt + ", operation : " + op);
        boolean found = false;
        for (int i = 0; i < runningBPIs.size(); i++) {
            RunningBPI rbpi = (RunningBPI) runningBPIs.get(i);
            if (bpiId == rbpi.getBPIId()) {
                rbpi.addCallWaiting(pt, op);
                found = true;
                break;
            }
        }
        if (BPELTrace.isManagerDebugEnabled())
            BPELTrace.traceManager("Manager end bpiWaiting, found = "+found);
    }

    /**
     * Notification sent by the BPIMdb that a BPI has ended
     *
     * @param bpiId the BPIId
     * @ejb:interface-method
     */
    public void endBPI(int bpiId) {
        if (BPELTrace.isManagerDebugEnabled())
            BPELTrace.traceManager("Manager endBPI : bpiId :" + bpiId);
        synchronized (runningBPIs) {
            for (int i = 0; i < runningBPIs.size(); i++) {
                RunningBPI rbpi = (RunningBPI) runningBPIs.get(i);
                if (bpiId == rbpi.getBPIId()) {
                    runningBPIs.remove(i);
                    break;
                }
            }
        }
        if (BPELTrace.isManagerDebugEnabled())
            BPELTrace.traceManager("Manager end endBPI");
    }

    private int findBPI(String processName, String pt, String op) {
        if (BPELTrace.isManagerDebugEnabled())
            BPELTrace.traceManager("Manager findBPI : processName :" + processName +
                    ", portType : " + pt + ", operation : " + op);
        int bpiId = -1;
        for (int i = 0; i < runningBPIs.size() ; i++) {
            RunningBPI rbpi = (RunningBPI) runningBPIs.get(i);
            if (processName.equals(rbpi.getProcessName())) {
                bpiId = rbpi.getBPIId();
                break;
            }
        }
        if (BPELTrace.isManagerDebugEnabled())
            BPELTrace.traceManager("Manager end findBPI : found : " + (bpiId!=-1) + ", bpiId :" + bpiId);
        return bpiId;
    }

    /**
     * createBPI
     * @ejb:transaction type="Supports"
     */
    private int createBPI(String processName, String targetNamespace) throws BPELEngineException {
        if (BPELTrace.isManagerDebugEnabled())
            BPELTrace.traceManager("Manager createBPI, processName = " + processName);

        int bpiId = BPICache.createBPI(processName, targetNamespace);

        RunningBPI rbpi = new RunningBPI(bpiId, processName);
        runningBPIs.add(rbpi);

        BPI bpi = new BPI(bpiId);
        bpi.init(processName);

        //System.out.println("***************** createBPI, bpiId = "+this.bpiId+", processName = "+this.process+", this = "+this);
        if (BPELTrace.isManagerDebugEnabled())
            BPELTrace.traceManager("Manager end createBPI, bpiId: " + bpiId);
        return bpiId;
    }

    private void createCS(int bpiId, String csname, CSProperties csvalue) throws BPELEngineException {
        if (BPELTrace.isManagerDebugEnabled())
            BPELTrace.traceManager("Manager createCS bpiId: " + bpiId + ", csname: " + csname +
                    ", csvalue : " + csvalue.toString());
        CSCache.createCS(bpiId,csname,csvalue);
        if (BPELTrace.isManagerDebugEnabled())
            BPELTrace.traceManager("Manager end createCS");
    }

}
