/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.objectweb.orchestra.engine;

import javax.xml.namespace.QName;

import org.objectweb.orchestra.exception.BPELEngineException;
import org.objectweb.orchestra.util.BPELTrace;
import org.objectweb.orchestra.util.Util;

/**
 * PendingCatch.java
 *
 * @author Goulven Le Jeune & Charles Souillard
 * @version $Id: PendingCatch.java 5 2006-10-20 14:37:44Z lejeuneg $
 *
 */
public class PendingCatch {

    private int brancheNb;
    private String faultName;
    private String faultVariableType;
    private String faultVariableMessageType;
    private String faultVariableElement;
    private String faultVariableBPELJType;

    /**
     * Constructor for a pending <code>catch</code>
     *
     * @param nb brancheNb
     * @param fn FaultName
     * @param fv FaultVariable
     */
    public PendingCatch(int nb, String fn, String fvt, String fvmt, String fve, String fvbt) {
        this.brancheNb = nb;
        this.faultName = fn;
        this.faultVariableType = fvt;
        this.faultVariableMessageType = fvmt;
        this.faultVariableElement = fve;
        this.faultVariableBPELJType = fvbt;
    }

    /**
     * Constructor for a pending <code>catchAll</code>
     * @param nb
     */
    public PendingCatch(int nb) {
        this.brancheNb = nb;
        this.faultName = null;
        this.faultVariableType = null;
        this.faultVariableMessageType = null;
        this.faultVariableElement = null;
        this.faultVariableBPELJType = null;
    }

    /**
     * @return
     */
    public int getBrancheNb() {
        return this.brancheNb;
    }

    /**
     * @return
     */
    public String getFaultName() {
        return this.faultName;
    }

    /**
     * @return
     */
    public String getFaultVariableType() {
        return this.faultVariableType;
    }

    private String getType() {
        if (this.faultVariableType != null)
            return this.faultVariableType;
        else if (this.faultVariableMessageType != null)
            return this.faultVariableMessageType;
        else if (this.faultVariableElement != null)
            return this.faultVariableElement;
        else if (this.faultVariableBPELJType != null)
            return this.faultVariableBPELJType;
        else
            return null;
    }

    /**
     * Returns the correctness of this Catch according to the following
     * parameters.
     * @param fn the fault name
     * @param fvt the fault variable type
     * @param fvmt the fault variable message type
     * @param fve the fault variable element
     * @return 0 if not correct 1 if correct 2 if partially correct
     */
    public int correctCatch(String fn, String fvt, String fvmt, String fve) {
        if (BPELTrace.isEngineDebugEnabled()) {
            BPELTrace.traceEngine("PendingCatch correctCatch : fn : " + fn + " fvt : " + fvt + " fvmt : " + fvmt + " fve : " + fve);
            BPELTrace.traceEngine("PendingCatch Fault Name : " + this.faultName + ", faultVariableType :" + this.faultVariableType + ", faultVariableMessageType : " + this.faultVariableMessageType + ", faultVariableElement : " + this.faultVariableElement);
        }
        String fv = null;
        String faultVariable = null;
        int result = 0;
        if (fvt != null) {
            fv = fvt;
            faultVariable = this.faultVariableType;
        } else if (fvmt != null) {
            fv = fvmt;
            faultVariable = this.faultVariableMessageType;
        } else if (fve != null) {
            fv = fve;
            faultVariable = this.faultVariableElement;
        } else
            faultVariable = getType();

        if (fv == null) {
            if (faultVariable == null) {
                if (this.faultName.equals(fn))
                    result = 1;
                else
                    result = 0;
            } else {
                result = 0;
            }
        } else {
            if (faultVariable == null) {
                result = 0;
            } else {
                if (this.faultName == null) {
                    if (faultVariableEquals(faultVariable,fv))
                        result = 2;
                    else
                        result = 0;
                } else {
                    if (this.faultName.equals(fn) && faultVariableEquals(faultVariable,fv))
                        result = 1;
                    else
                        result = 0;
                }
            }
        }
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("PendingCatch End correctCatch : result : " + result);
        return result;
    }

    private boolean faultVariableEquals(String fv1, String fv2) {
        QName qn1 = Util.buildQname(fv1);
        QName qn2 = Util.buildQname(fv2);
        return qn1.getLocalPart().equals(qn2.getLocalPart()) && Util.namespacesEquals(qn1.getNamespaceURI(), qn2.getNamespaceURI());
    }

    /**
     * Returns the correctness of this Catch according to the following
     * parameters (specific to BPELJ).
     * @param faultName the fault name
     * @param faultData the fault data object
     * @return boolean True if this Catch is correct, false otherwise
     * @throws BPELEngineException
     */
    public boolean correctBPELJCatch(String faultName, Object faultData) throws BPELEngineException {
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("PendingCatch correctBPELJCatch : faultName : " + faultName + " faultData : " + faultData);

        boolean result = false;
        if (faultData == null) {
            // the fault has no data
            if (this.faultName != null) {
                // the catch specifies a fault name
                result = this.faultName.equals(faultName);
            } else {
                // the catch does not specify a fault name
                result = true;// this is the catchAll handler
            }
        } else {
            // the fault has data
            boolean isTypeCompatible = false;
            if (this.faultVariableBPELJType != null) {
                try {
                    // is the catch faultVariable class a superclass of the fault data class ?
                    Class catchFaultVariableClass = Class.forName(this.faultVariableBPELJType);
                    Class faultDataClass = faultData.getClass();
                    isTypeCompatible = catchFaultVariableClass.isAssignableFrom(faultDataClass);
                } catch (ClassNotFoundException e) {
                    throw new BPELEngineException("PendingCatch.correctBPELJCatch :\n"
                            + "Impossible to load the class named " + this.faultVariableBPELJType, e);
                }
            } else {
                isTypeCompatible = true;
            }
            if (this.faultName != null) {
                // the catch specifies a fault name
                if (this.faultName.equals(faultName) ) {
                    // same names
                    result = isTypeCompatible;
                } else {
                    // different names;
                    result = false;
                }
            } else {
                // the catch does not specify a fault name
                result = isTypeCompatible;
            }
        }

        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("PendingCatch End correctBPELJCatch : result : " + result);
        return result;
    }
}