/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.objectweb.orchestra.engine;

import java.util.Vector;

import org.objectweb.orchestra.exception.BPELEngineException;
import org.objectweb.orchestra.util.BPELTrace;


/**
 * PendingFaultHandler.java
 *
 * @author Goulven Le Jeune & Charles Souillard
 * @version $Id: PendingFaultHandler.java 5 2006-10-20 14:37:44Z lejeuneg $
 *
 */
public class PendingFaultHandler {

    private Vector catches;
    private PendingCatch catchAll;
    private int scopeId;

    /**
     * Default Constructor
     *
     * @param sId Scope Id
     */
    public PendingFaultHandler(int sId) {
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("PendingFaultHandler Constructor : " + sId);
        this.catches = new Vector();
        this.catchAll = null;
        this.scopeId = sId;
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("PendingFaultHandler End Constructor");
    }

    public boolean equals(int sId) {
        return (sId == this.scopeId);
    }

    /**
     * @param nb BrancheNb
     * @param fn faultName
     * @param fv FaultVariable
     */
    public void addCatch(int nb, String fn, String fvt, String fvmt, String fve, String fvbt) {
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("PendingFaultHandler addCatch");
        this.catches.add(new PendingCatch(nb, fn, fvt, fvmt, fve, fvbt));
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("PendingFaultHandler End addCatch");
    }

    /**
     * @param nb BrancheNb
     */
    public void addCatchAll(int nb) {
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("PendingFaultHandler addCatchAll");
        this.catchAll = new PendingCatch(nb);
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("PendingFaultHandler End addCatchAll");
    }

    /**
     * Finds the Catch corresponding to the given parameters.
     * @param fn the fault name
     * @param fvt the fault variable type
     * @param fvmt the fault variable message type
     * @param fve the fault variable element
     * @return PendingCatch the pending catch that corresponds to the given parameters.
     */
    public PendingCatch findCatch(String fn, String fvt, String fvmt, String fve) {
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("PendingFaultHandler findCatch");
        PendingCatch result = null;
        PendingCatch pc = null;
        for (int i = 0; i < this.catches.size(); i++) {
            pc = (PendingCatch) this.catches.get(i);
            int isCorrect = pc.correctCatch(fn, fvt, fvmt, fve);
            if (isCorrect == 1) {
                result = pc;
                break;
            } else if (isCorrect == 2) {
                result = pc;
            }
        }
        if (result == null) result = this.catchAll;
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("PendingFaultHandler End findCatch");
        return result;
    }

    /**
     * Finds the Catch corresponding to the given parameters (specific to BPELJ).
     * @param faultName the fault name
     * @param faultData the fault data object
     * @return PendingCatch the pending catch that corresponds to the given parameters.
     */
    public PendingCatch findBPELJCatch(String faultName, Object faultData) throws BPELEngineException {
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("PendingFaultHandler findBPELJCatch");
        PendingCatch result = null;
        for (int i = 0; i < this.catches.size(); i++) {
            PendingCatch pc = (PendingCatch) this.catches.get(i);
            if (pc.correctBPELJCatch(faultName, faultData)) {
                result = pc;
                break;
            }
        }
        if (result == null) {
            result = this.catchAll;
        }
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("PendingFaultHandler End findBPELJCatch");
        return result;
    }

    /**
     * @return Scope Id
     */
    public int getScopeId() {
        return this.scopeId;
    }

}