/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.objectweb.orchestra.engine;

import org.objectweb.orchestra.exception.BPELEngineException;
import org.objectweb.orchestra.util.BPELTrace;


/**
 * PendingTarget.java
 *
 * @author Goulven Le Jeune & Charles Souillard
 * @version $Id: PendingTarget.java 5 2006-10-20 14:37:44Z lejeuneg $
 *
 */
public class PendingTarget extends Object {

    private int brancheNb;
    private int scopeId;
    private int bpiId;
    private String processName;
    private int childNb;
    private String[] targetLinkNames;
    private boolean[] solved;
    private boolean[] transitionCondition;
    private int[] branchesNb;
    private String joinCondition;
    private int bpuId;
    private boolean needToStart = false;

    /**
     * Default Constructor
     * @param nb the BrancheNb
     * @param sId the ScopeId
     * @param tlnNb the Number of targets
     * @param tln an array containing the name of the target
     * @param jc the joinCondition
     */
    public PendingTarget(int bpiId, String processName,int bpuId,int nb, int sId, int tlnNb, String[] tln, String jc,int[] bn) {
        this.bpiId = bpiId;
        this.processName = processName;
        this.brancheNb = nb;
        this.scopeId = sId;
        this.targetLinkNames = tln;
        this.joinCondition = jc;
        this.childNb = tlnNb;
        this.solved = new boolean[tlnNb];
        this.transitionCondition = new boolean[tlnNb];
        for (int i = 0; i < tlnNb; i++) {
            this.solved[i] = false;
        }
        this.branchesNb = bn;
        this.bpuId = bpuId;
    }

    /**
     * @return true if all of the targets have received a corresponding source
     */
    public boolean isFinish() {
        boolean result = true;
        int i = 0;
        while ((i < this.childNb) && (result)) {
            result = this.solved[i++];
        }
        return result;
    }

    /**
     * @return the BrancheNb
     */
    public int getBrancheNb() {
        return this.brancheNb;
    }

    /**
     * @return the ScopeId
     */
    public int getScopeId() {
        return this.scopeId;
    }

    /**
     * @return the bpuId
     */
    public int getBpuId() {
        return this.bpuId;
    }

    /**
     * @return a string containing the joinCondition
     */
    public String getJoinCondition() {
        return this.joinCondition;
    }

    /**
     * @param linkName the Name of the Link
     * @param tc the Transition Condition
     * @return true if the linkName correspond to a target not allready solved
     */
    public boolean newSource(String linkName, boolean tc) {
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("PendingTarget newSource : linkName = " + linkName);
        boolean found = false;
        int index = getIndexOfTarget(linkName);
        if (index != -1) {
            if (BPELTrace.isEngineDebugEnabled())
                BPELTrace.traceEngine("PendingTarget newSource, Corresponding target !, linkName = "+linkName+", tc = "+tc);
            this.solved[index] = true; // Target corresponding to the source found
            this.transitionCondition[index] = tc;
            found = true;
        }
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("PendingTarget end newSource : found = " + found);
        return found;

    }

    private int getIndexOfTarget(String linkName) {
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("PendingTarget getIndexOfTarget : linkName = " + linkName);
        int index = -1;
        int i = 0;
        while (i < this.childNb && index == -1) {
            if (linkName.equals(this.targetLinkNames[i]) && !this.solved[i]) {
                index = i;
            }
            i++;
        }
        if (BPELTrace.isEngineDebugEnabled())
            BPELTrace.traceEngine("PendingTarget end getIndexOfTarget : index = " + index);
        return index;
    }

    /**
     * @return true if the joinCondition is realised
     */
    public boolean evaluateJoinCondition() throws BPELEngineException {
        boolean result = false;
        if (this.joinCondition == null) {
            int i = 0;
            //System.out.println("PendingTarget.evalJC");
            while ((i < this.childNb) && (!result)) {
                result = this.transitionCondition[i];
                //System.out.println("PendingTarget.evalJC, link = "+targetLinkNames[i]+", tc = "+result);
                i++;
            }
        } else {
            // ********analyse JoinCondition*******
            //System.out.println("PendingTarget.evalJC defined, jc = "+joinCondition);
            result = ((Boolean) Expression.analyseJoinCondition(this.joinCondition,this,this.bpiId,this.scopeId,this.processName)).booleanValue();
        }
        return result;
    }

    /**
     * return the value of the expr getLinkStatus('linkName')
     *
     * @param expr
     * @param bpiId The BPI Id
     * @param scopeId The Scope Id
     * @return The result
     * @ejb:interface-method
     */
    public Boolean getTransitionCondition(String linkName) throws BPELEngineException {
        Boolean result = null;
        boolean found = false;
        int i = 0;
        while (i < this.childNb && !found) {
            found = linkName.equals(this.targetLinkNames[i++]);
        }
        if (found) {
            found = this.transitionCondition[i - 1];
        } else {
            throw new BPELEngineException("BPEL Error : No transition condition for linkName : " + linkName,null);
        }
        result = new Boolean(found);

        return result;
    }

    public int getTargetBrancheNb(String linkName) {
        int i = 0;
        int result = -1;
        while (i < this.childNb) {
            if (linkName.equals(this.targetLinkNames[i])) {
                result = this.branchesNb[i];
            }
            i++;
        }
        return result;
    }

    public String toString() {
        String st = "----------\n";
        st += "PendingTarget - brancheNb = "+this.brancheNb+"\n";
        st += "PendingTarget - bpuId = "+this.bpuId+"\n";
        st += "PendingTarget - scopeId = "+this.scopeId+"\n";
        st += "PendingTarget - childNb = "+this.childNb+"\n";
        st += "PendingTarget - joinCondition = "+this.joinCondition+"\n";
        st += "PendingTarget - needToStart = "+this.needToStart+"\n";
        for (int i=0;i<this.targetLinkNames.length;i++) {
            st += "PendingTarget - targetLinkNames["+i+"] = "+this.targetLinkNames[i]+"\n";
            st += "PendingTarget - solved["+i+"] = "+this.solved[i]+"\n";
            st += "PendingTarget - branchesNb["+i+"] = "+this.branchesNb[i]+"\n";
            st += "PendingTarget - transitionCondition["+i+"] = "+this.transitionCondition[i]+"\n";
        }
        st += "----------\n";
        return st;
    }

    public int[] getBranchesNb() {
        return this.branchesNb;
    }
}
