/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.objectweb.orchestra.generator;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.PrintWriter;
import java.util.HashMap;
import java.util.Map;
import java.util.Random;
import java.util.Vector;

import javax.xml.namespace.QName;
import javax.xml.parsers.ParserConfigurationException;

import org.apache.axis.wsdl.symbolTable.SymbolTable;
import org.objectweb.orchestra.data.PartnerLinkData;
import org.objectweb.orchestra.exception.BPELGeneratorException;
import org.objectweb.orchestra.exception.BPELUtilException;
import org.objectweb.orchestra.tree.Variable;
import org.objectweb.orchestra.util.BPELConstants;
import org.objectweb.orchestra.util.BPELString;
import org.objectweb.orchestra.util.Util;
import org.objectweb.orchestra.util.WSDLJavaMatching;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;


/**
 * BPELCompilo.java, generates a java representation of the xml file (bpel file)
 *
 * @author Goulven Le Jeune & Charles Souillard
 * @version $Id: BPELCompilo.java 126 2007-01-30 08:45:59Z souillac $
 *
 */
public class BPELCompilo {

    // -------------------------------------------------------------------------
    // Private Variables
    // -------------------------------------------------------------------------
    private String BPELFileName = null;
    private String processName = null;
    private int statementNb;
    private BPELProcessFile f = null;
    private BPELJProcessFile BPELJFile = null;
    private ProcessMapFile processMapFile = null;
    private boolean text_node = false;
    private String inoutType = "";
    private boolean receivePrinted = true;
    private String[] receiveInfo = new String[6];
    private String node_type = null;
    private int scopeId = 0;
    private Document document ;
    /**
     * Represents the correlationSet of the process
     */
    private Vector CS = null;
    private String mCreateInstance = "no";
    private Process proc = null;
    private String targetNS = null;
    private String BPELVersion = null ;
    private SymbolTable symbolTable = null;
    private Map plTypes = null;
    private Wsdl wsdl = null;

    /**
     * default constructor
     *
     * @throws BPELGeneratorException
     */
    public BPELCompilo() {
        this.CS = new Vector();
    }

    /**
     * setBPELVersion
     * Assign the BPELVersion attribute
     * @author Alexandre Greco
     *
     * @param s A string containing the BPEL namespace
     * @throws BPELGeneratorException
     */
    private void setBPELVersion (String s) throws BPELGeneratorException {
        if ( !(s.equals(BPELConstants.NS_BPEL_V1_1) || s.equals(BPELConstants.NS_BPEL_V2_0)) ) {
            throw new BPELGeneratorException("BPELCompilo.setBPELVersion : Invalid BPEL version", null) ;
        }
        this.BPELVersion = s ;
    }

    /**
     * analyse the process document
     *
     * @param document the DOM document of the BPEL file
     * @param symbolTable the symbol table of the WSDL file correcponding to this BPEL document
     * @param plTypes the partner link types of the WSDL file correcponding to this BPEL document
     * @param wsdl the wsdl NamespaceDef corresponding to this BPEL document
     * @return the processName
     * @throws BPELGeneratorException
     */
    public String analyseDocument(Document document, SymbolTable symbolTable, Map plTypes, Wsdl wsdl) throws BPELGeneratorException {
        this.symbolTable = symbolTable;
        this.plTypes = plTypes;
        this.wsdl = wsdl;

        //go through each node of the dom tree and analyse it
        for (Node child = document.getFirstChild(); child != null; child = child.getNextSibling()) {
            analyseNode(child);
        }
        this.proc.writeProperties();
        return this.processName;
    }

    /**
     * returns the DOM document of the specified BPEL file
     * @param bpelFileURI the bpel file uri
     * @return the document corresponding to this BPEL file
     * @throws BPELGeneratorException
     */
    public Document getDocumentFromBpelFile(String bpelFileURI) throws BPELGeneratorException {
        this.BPELFileName = bpelFileURI;

        File originalBPELFile = new File(this.BPELFileName);
        File originalBPELFileBackup = new File(originalBPELFile.getAbsolutePath() + new Random().nextInt());

        rename(originalBPELFile,originalBPELFileBackup);
        try {
            originalBPELFile.createNewFile();
        } catch (Exception e) {
            rename(originalBPELFileBackup,originalBPELFile);
            throw new BPELGeneratorException("BPELCompilo.analyseFile :", e);
        }

        File SAXCompliantBPELFile = originalBPELFile;
        boolean isFileTransformed = transformBPELJFileToSAXCompliantFile(originalBPELFileBackup, SAXCompliantBPELFile);
        if (!isFileTransformed) {
            makeSomeCleanUp(originalBPELFile, originalBPELFileBackup, SAXCompliantBPELFile);
        }

        // build a DOM document
        try {
            this.document = DocumentParser.getDocument(new File(this.BPELFileName).toURL(), true, new BpelEntityResolver());
        } catch (Exception e) {
            try {
                makeSomeCleanUp(originalBPELFile, originalBPELFileBackup, SAXCompliantBPELFile);
            } catch (Exception e2) {
                e.printStackTrace();
                throw new BPELGeneratorException("BPELCompilo.analyseFile : unable to do some cleanup after having caught : "+e, e);
            }
            throw new BPELGeneratorException("BPELCompilo.analyseFile :", e);
        }

        if (isFileTransformed) {
            makeSomeCleanUp(originalBPELFile, originalBPELFileBackup, SAXCompliantBPELFile);
        }

        return this.document;
    }

    private void rename(File from,File to) throws BPELGeneratorException {
        try {
            boolean success = from.renameTo(to);
            if (!success) {
                Util.copyFile(from, to);
                from.delete();
            }
        } catch (BPELUtilException e) {
            throw new BPELGeneratorException("BPELCompilo.rename : error while trying to rename "+from+" to "+to,e);
        }
    }
    /**
     * Cleans up the created files.
     * @param originalBPELFile
     * @param originalBPELFileBackup
     * @param SAXCompliantBPELFile
     */
    private final void makeSomeCleanUp(File originalBPELFile, File originalBPELFileBackup, File SAXCompliantBPELFile) throws BPELGeneratorException {
        SAXCompliantBPELFile.delete();
        rename(originalBPELFileBackup,originalBPELFile);
    }

    /**
     * Transforms the BPELJ source file into the SAXCompliant destination file.
     * This method enables BPELJ files to be parsed by the SAX Parser.
     * If the source file is not a BPELJ source file, then this methods does nothing.
     * @param fileSrc the BPELJ source file to transform
     * @param fileDest the SAXCompliant file result of the transformation
     * @return boolean true if the fileSrc is a BPELJ source file, false otherwise
     */
    private boolean transformBPELJFileToSAXCompliantFile(File fileSrc, File fileDest) throws BPELGeneratorException {
        boolean result = false;
        try {
            long srcLength = fileSrc.length();
            char[] buffer = new char[(int)srcLength]; // We assume l < Integer.MAX_VALUE i.e. file is less than 2 GB
            BufferedReader sourceFile = new BufferedReader(new FileReader(fileSrc));
            PrintWriter resultFile = new PrintWriter(new BufferedWriter(new FileWriter(fileDest)));
            sourceFile.read(buffer, 0, (int)srcLength); // We assume we read l bytes
            String BPELFileContent = new String(buffer);

            // If there is a prefix for the BPELJ namespace, then this file is BPELJ aware
            String[] prefixes = getBPELJNSPrefixes(BPELFileContent);
            if ( (prefixes != null) && (prefixes.length > 0) ) {
                // Replaces into Java snippets the valid Java characters that are also invalid XML characters
                // by valid XML entities to allow parsing of this file
                result = writeSAXCompliantFile(resultFile, BPELFileContent, prefixes);
                sourceFile.close();
                resultFile.close();
            } else {
                result = false;
            }
        } catch (Exception e) {
            throw new BPELGeneratorException("BPELCompilo.transformBPELJFileToSAXCompliantFile :", e);
        }
        return result;
    }

    /**
     * Write the BPELJ File into a SAXCompliant file.
     * Basically, this method replaces '&' and '<' character within the content of the BPELFile
     * with the corresponding XML entities "&amp;" and "&lt;" into the <code>PrintWriter</code>.
     * This enables the SAX parser to perform parsing of the output file.
     * @param resultFile the <code>Printwriter</code> where to write the SAX compliant result file
     * @param BPELFileContent the string containing the BPELJ File to transform
     * @param prefixes the prefixes of the BPELJ namespace in this document
     * @return boolean true if the file has been changed, false otherwise
     */
    private boolean writeSAXCompliantFile(PrintWriter resultFile, String BPELFileContent, String[] prefixes) {
        // Those arrays are only used has out parameters of the method findNext
        int[] index = new int[2];
        int[] length = new int[2];
        int lastIndexEnd = 0;
        boolean stillHasMarkups = findNextBPELJTags(index, length, BPELFileContent, prefixes);
        if (!stillHasMarkups) {
            return false;
        }
        // writes the content of the file from the beginning to the first <prefix:code> markup included
        resultFile.write(BPELFileContent, 0, index[BEGIN] + length[BEGIN]);
        // Replaces into Java snippets the valid Java characters that are also invalid XML characters
        // by valid XML entities to allow parsing of this file
        while (stillHasMarkups) {
            // gets the Java code of the next Java snippet and replaces the invalid XML characters
            // by the XML entities equivalent and writes the modified Java code
            String javaCode = BPELFileContent.substring(index[BEGIN] + length[BEGIN], index[END]);
            resultFile.write(javaCode.replaceAll("&", "&amp;").replaceAll("<", "&lt;"));
            // searches for the next <prefix:code> and </prefix:code> markups
            //indexBegin = BPELFileContent.indexOf(begin, indexEnd + end.length());
            lastIndexEnd = index[END];
            stillHasMarkups = findNextBPELJTags(index, length, BPELFileContent, prefixes);
            if (stillHasMarkups) {
                resultFile.write(BPELFileContent, lastIndexEnd, index[BEGIN] + length[BEGIN] - lastIndexEnd);
                stillHasMarkups = stillHasMarkups && (index[END] != -1);
            }
        }
        // writes the content of the file from the last </prefix:code> markup included to the end of the file
        resultFile.write(BPELFileContent, lastIndexEnd, BPELFileContent.length() - lastIndexEnd);
        return true;
    }

    /**
     * Only useful for the methods writeSAXCompliantFile() and findNextTags()
     */
    private static final int BEGIN = 0;
    /**
     * Only useful for the methods writeSAXCompliantFile() and findNextTags()
     */
    private static final int END = 1;

    /**
     * Finds the next BPELJ tags
     * @param index the indexes of the begin and end tags found
     * @param length the lengthes of the begin and end tags found
     * @param BPELFileContent the string containing the BPELJ File to transform
     * @param prefixes the prefixes of the BPELJ namespace in this document
     * @return boolean true if the file has been changed, false otherwise
     */
    private boolean findNextBPELJTags(int[] index, int[] length, String BPELFileContent, String[] prefixes) {
        // these local parts are to be used with the bpelj prefixes
        String[] localParts1 = {"code", "value"};
        // these local parts are not to be used with the bpelj prefixes
        String[] localParts2 = {"condition", "joinCondition", "transitionCondition", "until", "for"};
        boolean result = false;

        int lastIndexEnd = (length[BEGIN] == 0)? 0 : index[END] + length[END] - 1;
        index[BEGIN] = Integer.MAX_VALUE;
        index[END] = Integer.MAX_VALUE;

        for (int i=0; i<prefixes.length; ++i) {
            for (int j=0; j<localParts1.length; ++j) {
                String begin = "<" + prefixes[i] + ":" + localParts1[j];
                String end = "</" + prefixes[i] + ":" + localParts1[j] + ">";
                int indexBeginTmp = BPELFileContent.indexOf(begin, lastIndexEnd);
                int indexEndTmp = BPELFileContent.indexOf(end, indexBeginTmp);
                // We also verify indexEndTmp to be sure we aught the right markup
                // ex: in the case of <forEach>, indexBeginTmp matches, but not indexEndTmp
                if (indexBeginTmp != -1 && indexBeginTmp<index[BEGIN] && indexEndTmp != -1) {
                    index[BEGIN] = indexBeginTmp;
                    index[END] = indexEndTmp;
                    length[END] = end.length();
                    result = true;
                }
            }
        }
        for (int j=0; j<localParts2.length; ++j) {
            String begin = "<" + localParts2[j];
            String end = "</" + localParts2[j] + ">";
            int indexBeginTmp = BPELFileContent.indexOf(begin, lastIndexEnd);
            int indexEndTmp = BPELFileContent.indexOf(end, indexBeginTmp);
            // We also verify indexEndTmp to be sure we aught the right markup
            // ex: in the case of <forEach>, indexBeginTmp matches, but not indexEndTmp
            if (indexBeginTmp != -1 && indexBeginTmp<index[BEGIN] && indexEndTmp != -1) {
                index[BEGIN] = indexBeginTmp;
                index[END] = indexEndTmp;
                length[END] = end.length();
                result = true;
            }
        }
        if (result) {
            int indexGt = BPELFileContent.indexOf(">", index[BEGIN]);
            length[BEGIN] = indexGt - index[BEGIN] + 1;
        }
        return result;
    }

    /**
     * Get the BPELJ namespace prefixes of this BPELJ file
     * @param BPELFileContent the content of the BPEL file where to search for the BPELJ namespace prefixes
     * @return String[] the BPELJ namespace prefixes of this BPELJ file
     */
    private String[] getBPELJNSPrefixes(String BPELFileContent) {
        int lastIndexBPELJNS = -BPELConstants.NS_BPELJ.length();
        Vector prefixes = new Vector();

        int indexProcessElement = BPELFileContent.indexOf("<process");
        if (indexProcessElement == -1) {
            indexProcessElement = BPELFileContent.indexOf(":process");
        }
        int indexProcessElement2 = BPELFileContent.indexOf(">", indexProcessElement);
        String processElement = BPELFileContent.substring(indexProcessElement, indexProcessElement2);

        while (true) {
            int indexBPELJNS = processElement.indexOf(BPELConstants.NS_BPELJ, lastIndexBPELJNS + BPELConstants.NS_BPELJ.length());
            if (indexBPELJNS==-1) {
                break;
            }
            String substring = processElement.substring(lastIndexBPELJNS + BPELConstants.NS_BPELJ.length(), indexBPELJNS);
            int indexXMLNS = substring.lastIndexOf("xmlns:");
            int indexEqual = processElement.indexOf("=", indexXMLNS);
            if (indexXMLNS==-1 || indexEqual==-1) {
                break;
            }
            String prefix = processElement.substring(indexXMLNS + "xmlns:".length(), indexEqual);
            prefixes.add(prefix);
            lastIndexBPELJNS = indexBPELJNS;
        }

        // Prepares the result
        String[] result = new String[prefixes.size()];
        for (int i=0; i<prefixes.size(); ++i) {
            result[i] = (String) prefixes.get(i);
        }
        return result;
    }

    /**
     * get the targetNamespace of the bpel file
     *
     * @return the targetNamespace of the bpel file
     */
    public String getTargetNamespace() {
        return this.targetNS;
    }

    /**
     * Recursive method to analyse the DOM tree
     *
     * @param n
     *            the node to analyse
     * @return int the statement number of the node
     */
    private int analyseNode(Node n) throws BPELGeneratorException {

        int nodeStatementNb = this.statementNb + 1;

        int type = n.getNodeType();
        this.text_node = false;

        switch (type) {
        case Node.ELEMENT_NODE:
            String elementName = n.getLocalName();
            if (elementName == null) {
                elementName = n.getNodeName();
            }
            //System.out.println("BPELCompilo, analyseNode, elementName = "+elementName);
            if (elementName.equals("process")) {
                //analyse the process tag
                NamedNodeMap atts = n.getAttributes();
                Node att = null;
                for (int i = 0; i < atts.getLength(); i++) {
                    att = atts.item(i);
                    if (att.getNodeName().equals("name")) {
                        this.processName = new String(att.getNodeValue());
                        break;
                    }
                }
                //creates a ProcessMapFile which will contains all mapping
                // between processes and correlationSets
                this.processMapFile = new ProcessMapFile(this.processName);
                //creates a process file which will contains the mapping
                // between namespaces and prefix,
                //the processName and process tag attributes
                this.proc = new org.objectweb.orchestra.generator.Process(this.processName);
                //add to process file a very useful mapping to manage cases
                // when WSDLJavaMatching class return a xsd type without prefix
                this.proc.addProperty("xmlns:xsd", BPELConstants.NS_XSD);
                //for each mapping namespace mapping in process tag, add it
                // into process file
                for (int i = 0; i < atts.getLength(); i++) {
                    att = atts.item(i);
                    this.proc.addProperty(att.getNodeName(), att.getNodeValue());
                }
                this.targetNS = this.proc.getTargetNamespace();

                // Set the BPEL version used in this process
                setBPELVersion(this.proc.getXmlNS()) ;

                //creates a BPELProcessFile which will be the java
                // representation of the dom tree
                this.f = new BPELProcessFile(this.processName, this.proc);
                //init this class, java.lang.reflect package is used to find
                // classes in classLoader tree
                /*
                 * f.insertLine(" java.lang.ClassLoader classL =
                 * java.lang.Thread.currentThread().getContextClassLoader();");
                 * f.insertLine(" java.lang.Class cl = null;"); f.insertLine("
                 * java.lang.reflect.Constructor constr = null;");
                 * f.insertLine(" java.util.ArrayList tabObj = new
                 * java.util.ArrayList();"); f.insertLine(" java.lang.Class cl2 =
                 * null;"); f.insertLine(" java.lang.Class[] tabcl = null;");
                 * f.insertLine(" java.lang.reflect.Method met = null;");
                 * f.insertLine(" java.lang.Object[] tabo = null;");
                 * f.insertLine(" java.lang.Object result = null;");
                 * f.insertLine(" java.lang.Class clstatement =
                 * java.lang.Class.forName(\"org.objectweb.orchestra.tree.Statement\",true,classL);");
                 * f.insertLine(" java.lang.Class[] tabclstatement = new
                 * java.lang.Class [1];"); f.insertLine(" tabclstatement[0] =
                 * clstatement;"); f.insertLine(" tabo = new java.lang.Object
                 * [1];");
                 */
                this.f.insertLine("      p = new org.objectweb.orchestra.tree.BPELProcess(\"" + this.processName + "\");");
                this.f.insertLine("      tab.add(\"This case will not be used as statements numbers start from 1\");");
                /*
                 * f.insertLine(""); //f.insertLine(" //p = new
                 * org.objectweb.orchestra.tree.BPELProcess(processName);"); f.insertLine(" cl =
                 * java.lang.Class.forName(\"org.objectweb.orchestra.tree.BPELProcess\",true,classL);");
                 * f.insertLine(" constr = cl.getConstructor(new
                 * Class[]{java.lang.String.class});"); f.insertLine("
                 * tabObj.add(\"" + processName + "\");"); f.insertLine(" p =
                 * constr.newInstance(tabObj.toArray());"); f.insertLine("
                 * tabObj.clear();"); f.insertLine("");
                 */
                if (this.proc.isBPELJProcess()) {
                    this.BPELJFile = new BPELJProcessFile(this.processName, this.proc);
                }

                this.statementNb = 0;
                analyseScope(n);

                if (this.proc.isBPELJProcess()) {
                    this.BPELJFile.finish();
                }
                //f.insertLine(" p.addChild(s1);");
                this.f.insertLine("      p.addChild((org.objectweb.orchestra.tree.Statement)tab.get(1));");
                /*
                 * f.insertLine(" cl =
                 * java.lang.Class.forName(\"org.objectweb.orchestra.tree.BPELProcess\",true,classL);");
                 * f.insertLine(" met =
                 * cl.getMethod(\"addChild\",tabclstatement);"); f.insertLine("
                 * tabo[0] = s1;"); f.insertLine(" result =
                 * met.invoke(p,tabo);"); f.insertLine("");
                 */
                this.f.insertLine("      p.setTotalStatementNb("+this.statementNb+");");
                this.f.finish();
                this.processMapFile.finish();
            } else if (elementName.equals("sequence")) {
                analyseSequence(n);
            } else if (elementName.equals("flow")) {
                analyseFlow(n);
            } else if (elementName.equals("receive")) {
                analyseReceive(n);
            } else if (elementName.equals("reply")) {
                analyseReply(n);
            } else if (elementName.equals("invoke")) {
                analyseInvoke(n);
            } else if (elementName.equals("switch")) {
                if ( !this.BPELVersion.equals(BPELConstants.NS_BPEL_V1_1) )
                    throw new BPELGeneratorException ("BPELCompilo.analyseSwitch : Switch activity is a BPEL 1.1 activity only", null) ;
                analyseSwitch(n);
            } else if (elementName.equals("if")) {
                analyseIf(n);
            } else if (elementName.equals("case")) {
                analyseCase(n);
            } else if (elementName.equals("otherwise")) {
                analyseOtherwise(n);
            } else if (elementName.equals("assign")) {
                analyseAssign(n);
            } else if (elementName.equals("copy")) {
                analyseCopy(n);
            } else if (elementName.equals("from")) {
                analyseFrom(n);
            } else if (elementName.equals("to")) {
                analyseTo(n);
            } else if (elementName.equals("links")) {
                analyseLinks(n);
            } else if (elementName.equals("link")) {
                analyseLink(n);
            } else if (elementName.equals("correlations")) {
                analyseCorrelations(n);
            } else if (elementName.equals("correlation")) {
                analyseCorrelation(n);
            } else if (elementName.equals("source")) {
                analyseSource(n);
            } else if (elementName.equals("target")) {
                analyseTarget(n);
            } else if (elementName.equals("while") || elementName.equals("repeatUntil")) {
                analyseWhileOrRepeatUntil(n);
            } else if (elementName.equals("pick")) {
                analysePick(n);
            } else if (elementName.equals("onMessage")) {
                analyseOnMessage(n);
            } else if (elementName.equals("onAlarm")) {
                analyseOnAlarm(n);
            } else if (elementName.equals("throw")) {
                analyseThrow(n);
            } else if (elementName.equals("catch")) {
                analyseCatch(n);
            } else if (elementName.equals("catchAll")) {
                analyseCatchAll(n);
            } else if (elementName.equals("terminate")) {
                analyseTerminate(n);
            } else if (elementName.equals("wait")) {
                analyseWait(n);
            } else if (elementName.equals("empty")) {
                analyseEmpty(n);
            } else if (elementName.equals("scope")) {
                analyseScope(n);
            } else if (elementName.equals("variables")) {
                analyseVariables(n);
            } else if (elementName.equals("variable")) {
                analyseVariable(n);
            } else if (elementName.equals("correlationSets")) {
                analyseCorrelationSets(n);
            } else if (elementName.equals("correlationSet")) {
                analyseCorrelationSet(n);
            } else if (elementName.equals("faultHandlers")) {
                analyseFH(n);
            } else if (elementName.equals("compensationHandler")) {
                analyseCH(n);
            } else if (elementName.equals("eventHandlers")) {
                analyseEH(n);
            } else if (elementName.equals("partnerLinks")) {
                analysePartnerLinks(n);
            } else if (elementName.equals("partnerLink")) {
                analysePartnerLink(n);
            } else if (elementName.equals("partners")) {
                analysePartners(n);
            } else if (elementName.equals("partner")) {
                analysePartner(n);
            } else if (elementName.equals("compensate")) {
                analyseCompensate(n);
            } else if (elementName.equals("forEach")) {
                analyseForEach(n);
            } else {
                if (hasPrefixOfBPELJNS(n.getNodeName())) {
                    if (elementName.equals("snippet")) {
                        analyseBPELJSnippet(n);
                    } else if (elementName.equals("code")) {
                        analyseBPELJCode(n);
                    } else if (elementName.equals("value")) {
                        analyseBPELJValue(n);
                    } else {
                        // TODO remove when BPELJ implementation is finished
                        throw new BPELGeneratorException("Statement " + elementName + " is not yet implemented");
                    }
                } else {
                    // TODO remove when BPELJ implementation is finished
                    throw new BPELGeneratorException("Statement " + elementName + " is not yet implemented");
                }
            }
            this.text_node = false;
            this.node_type = new String(elementName);
            this.node_type = BPELString.toUpper(this.node_type);
            if (this.node_type.equals("EventHandlers") || this.node_type.equals("FaultHandlers")) {
                this.node_type = this.node_type.substring(0, this.node_type.length() - 1);
            }

            break;
        case Node.TEXT_NODE:
            this.text_node = true;
            nodeStatementNb = 0;
            //System.out.println("TEXT_NODE" + n.getNodeValue());
            break;
        case Node.COMMENT_NODE:
            this.text_node = true;
            nodeStatementNb = 0;
            //System.out.println("TEXT_NODE" + n.getNodeValue());
            break;
        default:
            nodeStatementNb = 0;
        //analyseNode(n);
        break;
        }
        return nodeStatementNb;
    }

    /**
     * create an instance of a java class by the way of reflection (use to
     * create org.objectweb.orchestra.tree classes)
     *
     * @param className
     *            the className of the class to instanciate
     * @param tabObj
     *            the parameters needed to build the instance
     * @param statementName
     *            the name of the java object which will reference thos new
     *            instance
     */
    private void createInstance(String className, String[] tabObj/*, String statementName*/) throws BPELGeneratorException {
        /*
         * f.insertLine(" cl = java.lang.Class.forName(\"" + className +
         * "\",true,classL);"); f.insertLine(" constr =
         * (cl.getConstructors())[0];"); for (int i=0;i <tabObj.length;i++){
         * f.insertLine(" tabObj.add(" + tabObj[i] + ");"); } f.insertLine("
         * java.lang.Object " + statementName + " =
         * constr.newInstance(tabObj.toArray());"); f.insertLine("
         * tabObj.clear();"); f.insertLine("");
         */
        /*
         * String l = " "+className+" " + statementName + " = new "+className
         * +"("; for (int i=0;i <tabObj.length;i++){ if (i>0) { l += ","; } l +=
         * tabObj[i]; } l+=");";
         */

        //String l = " tab["+sNb + "] = (org.objectweb.orchestra.tree.Statement) new "+className
        // +"(";
        String l = "      tab.add(new " + className + "(";
        for (int i = 0; i < tabObj.length; i++) {
            if (i > 0) {
                l += ",";
            }
            l += tabObj[i];
        }
        l += "));";

        this.f.insertLine(l);
    }

    /**
     * invoke a method on an object (a org.objectweb.orchestra.tree object) by the way of
     * reflection, for example for example : s1.addChild(s2);
     *
     * @param className
     *            the className of the class to where the method to call is
     *            defined
     * @param methodName
     *            the method name to call
     * @param caller
     *            the name of the java object on which the call must be done
     * @param statementName
     *            the parameter name to give to the method called, this
     *            parameter name references an object which is a statement
     */
    private void statementCall(String methodName, String caller, String statementName) throws BPELGeneratorException {
        /*
         * f.insertLine(" cl = java.lang.Class.forName(\"" + className +
         * "\",true,classL);"); f.insertLine(" met = cl.getMethod(\"" +
         * methodName + "\",tabclstatement);"); f.insertLine(" tabObj.add(" +
         * statementName + ");"); f.insertLine(" result = met.invoke(" + caller
         * +", tabObj.toArray());"); f.insertLine(" tabObj.clear();");
         *
         * f.insertLine("");
         */

        //String l = " "+caller +
        // "."+methodName+"((org.objectweb.orchestra.tree.Statement)"+statementName+");";
        int callerNb = Integer.parseInt(caller.substring(1));
        int paramNb = Integer.parseInt(statementName.substring(1));
        String l = "      ((org.objectweb.orchestra.tree.Statement)tab.get(" + callerNb + "))." + methodName + "((org.objectweb.orchestra.tree.Statement)tab.get(" + paramNb + "));";

        this.f.insertLine(l);

    }

    /**
     * invoke a method on an object (a org.objectweb.orchestra.tree object) by the way of
     * reflection, for example for example : s1.setChildNd(10);
     *
     * @param className
     *            the className of the class to where the method to call is
     *            defined
     * @param methodName
     *            the method name to call
     * @param caller
     *            the name of the java object on which the call must be done
     * @param param
     *            the parameter name to give to the method called, this
     *            parameter name references an int
     */
    private void intCall(String className, String methodName, String caller, int param) throws BPELGeneratorException {
        /*
         * f.insertLine(" cl = java.lang.Class.forName(\"" + className +
         * "\",true,classL);"); f.insertLine(" tabcl = new java.lang.Class
         * [1];"); f.insertLine(" tabcl[0] =
         * java.lang.Class.forName(\"java.lang.Integer\");"); f.insertLine(" met =
         * cl.getMethod(\"" + methodName + "\",tabcl);"); f.insertLine(" tabo[0] =
         * new java.lang.Integer(" + param + ");"); f.insertLine(" result =
         * met.invoke(" + caller + ", tabo);"); f.insertLine("");
         */
        //String l = " "+caller + "."+methodName+"(new
        // java.lang.Integer("+param+"));";
        int callerNb = Integer.parseInt(caller.substring(1));
        String l = "      ((" + className + ")tab.get(" + callerNb + "))." + methodName + "(new java.lang.Integer(" + param + "));";
        this.f.insertLine(l);
    }

    /**
     * invoke a method on an object (a org.objectweb.orchestra.tree object) by the way of
     * reflection, for example for example : s1.setFor("'PT10S'");
     *
     * @param className
     *            the className of the class to where the method to call is
     *            defined
     * @param methodName
     *            the method name to call
     * @param caller
     *            the name of the java object on which the call must be done
     * @param param
     *            the parameter name to give to the method called, this
     *            parameter name references a string
     */
    private void stringCall(String className, String methodName, String caller, String param) throws BPELGeneratorException {
        /*
         * f.insertLine(" cl = java.lang.Class.forName(\"" + className +
         * "\",true,classL);"); f.insertLine(" tabcl = new java.lang.Class
         * [1];"); f.insertLine(" tabcl[0] =
         * java.lang.Class.forName(\"java.lang.String\");"); f.insertLine(" met =
         * cl.getMethod(\"" + methodName + "\",tabcl);"); f.insertLine(" tabo[0] =
         * new java.lang.String(\"" + param + "\");"); f.insertLine(" result =
         * met.invoke(" + caller + ", tabo);"); f.insertLine("");
         */
        //String l = " "+caller + "."+methodName+"(new
        // java.lang.String(\""+param+"\"));";
        int callerNb = Integer.parseInt(caller.substring(1));
        String l = "      ((" + className + ")tab.get(" + callerNb + "))." + methodName + "(new java.lang.String(\"" + param + "\"));";
        this.f.insertLine(l);
    }

    /**
     * creates the code to instanciate a statement which takes a parameter
     * @param className the className of the statement object to instantiate
     * @param stmtNb the statement number of the created statement
     * @param parameterClassName the class name of the parameter
     */
    private void createStatementInstance(String className, int stmtNb, String parameterClassName) throws BPELGeneratorException {
        String l = "      tab.add(new " + className + "("
        + "new java.lang.Integer(" + stmtNb + "), (" + parameterClassName + ")tab.remove(" + stmtNb + ")"
        + "));";
        this.f.insertLine(l);
    }

    /**
     * creates the code to instanciate a statement which takes a parameter
     * @param className the className of the statement object to instantiate
     * @param stmtNb the statement number of the created statement
     * @param parameterClassName the class name of the parameter
     * @param isBlock whether the statement is a "block" statement
     */
    private void createStatementInstance(String className, int stmtNb, String parameterClassName, boolean isBlock) throws BPELGeneratorException {
        String l = "      tab.add(new " + className + "("
        + "new java.lang.Integer(" + stmtNb + "), (" + parameterClassName + ")tab.remove(" + stmtNb + ")," + isBlock
        + "));";
        this.f.insertLine(l);
    }

    /**
     * analyse node throw
     *
     * @param n
     *            the xml node representing this throw
     */
    private void analyseThrow(Node n) throws BPELGeneratorException {
        //System.out.println("BPELCompilo : AnalyseThrow");
        int throwNb = ++this.statementNb;
        String throwFaultName = null;
        String throwFaultVariable = null;
        String name = null;
        String joinCondition = null;
        String suppressJoinFailure = null;

        NamedNodeMap atts = n.getAttributes();
        for (int i = 0; i < atts.getLength(); i++) {
            Node att = atts.item(i);
            if (att.getNodeName().equals("faultName")) {
                throwFaultName = new String(att.getNodeValue());
                QName tfnQN = Util.buildQname(throwFaultName, this.proc);
                throwFaultName = Util.buildStringFromQName(tfnQN);
                //System.out.println("Throw FaultName : " + throwFaultName);
            } else if (att.getNodeName().equals("faultVariable")) {
                throwFaultVariable = new String(att.getNodeValue());
                //System.out.println("Throw FaultVariable : " +
                // throwFaultVariable);
            } else if (att.getNodeName().equals("name")) {
                name = new String(att.getNodeValue());
            } else if (att.getNodeName().equals("joinCondition")) {
                joinCondition = new String(att.getNodeValue());
            } else if (att.getNodeName().equals("suppressJoinFailure")) {
                suppressJoinFailure = new String(att.getNodeValue());
            }
        }

        //f.insertLine(" //org.objectweb.orchestra.tree.Throw s" + throwNb + " = new
        // org.objectweb.orchestra.tree.Throw(" + throwNb + ", \"" + throwFaultName + "\", \"" +
        // throwFaultVariable + ", \"" + name + "\", \"" + joinCondition + "\",
        // \"" + suppressJoinFailure + "\");");

        String[] tabParam = new String[6];
        tabParam[0] = "new java.lang.Integer(" + throwNb + ")";
        tabParam[1] = "\"" + throwFaultName + "\"";
        tabParam[2] = "\"" + throwFaultVariable + "\"";
        tabParam[3] = "\"" + name + "\"";
        tabParam[4] = "\"" + joinCondition + "\"";
        tabParam[5] = "\"" + suppressJoinFailure + "\"";
        createInstance("org.objectweb.orchestra.tree.Throw", tabParam);
        analyseChilds(n, "Throw");
        //System.out.println("BPELCompilo : End AnalyseThrow");
    }

    /**
     * analyse node FaultHandler
     *
     * @param n
     *            the xml node representing this FaultHandler
     */
    private void analyseFH(Node n) throws BPELGeneratorException {
        //System.out.println("BPELCompilo : AnalyseFaultHandler");
        int faultHandlerNb = ++this.statementNb;

        //f.insertLine(" //org.objectweb.orchestra.tree.FaultHandler s" + faultHandlerNb + " = new
        // org.objectweb.orchestra.tree.FaultHandler(" + faultHandlerNb + ");");

        String[] tabParam = new String[1];
        tabParam[0] = "new java.lang.Integer(" + faultHandlerNb + ")";
        createInstance("org.objectweb.orchestra.tree.FaultHandler", tabParam);

        analyseChilds(n, "FaultHandler");

    }

    /**
     * analyse node EventHandler
     *
     * @param n
     *            the xml node representing this EventHandler
     */
    private void analyseEH(Node n) throws BPELGeneratorException {
        //System.out.println("BPELCompilo : AnalyseEventHandler");
        int eventHandlerNb = ++this.statementNb;

        //f.insertLine(" //org.objectweb.orchestra.tree.EventHandler s" + eventHandlerNb + " = new
        // org.objectweb.orchestra.tree.EventHandler(" + eventHandlerNb + ");");

        String[] tabParam = new String[1];
        tabParam[0] = "new java.lang.Integer(" + eventHandlerNb + ")";
        createInstance("org.objectweb.orchestra.tree.EventHandler", tabParam);

        analyseChilds(n, "EventHandler");
        //System.out.println("BPELCompilo : End AnalyseEventHandler");
    }

    /**
     * analyse node CompensationHandler
     *
     * @param n
     *            the xml node representing this CompensationHandler
     */
    private void analyseCH(Node n) throws BPELGeneratorException {
        //System.out.println("BPELCompilo : AnalyseCompensationHandler");
        int compensationHandlerNb = ++this.statementNb;

        //f.insertLine(" //org.objectweb.orchestra.tree.CompensationHandler s" +
        // compensationHandlerNb + " = new org.objectweb.orchestra.tree.CompensationHandler(" +
        // compensationHandlerNb + ");");

        String[] tabParam = new String[1];
        tabParam[0] = "new java.lang.Integer(" + compensationHandlerNb + ")";
        createInstance("org.objectweb.orchestra.tree.CompensationHandler", tabParam);

        analyseChilds(n, "CompensationHandler");

        //System.out.println("BPELCompilo : End AnalyseCompensationHandler");
    }

    /**
     * analyse node Catch
     *
     * @param n
     *            the xml node representing this Catch
     */
    private void analyseCatch(Node n) throws BPELGeneratorException {
        //System.out.println("BPELCompilo : AnalyseCatch");
        int catchNb = ++this.statementNb;
        String catchFaultName = null;
        String catchFaultVariable = null;

        NamedNodeMap atts = n.getAttributes();
        for (int i = 0; i < atts.getLength(); i++) {
            Node att = atts.item(i);
            if (att.getNodeName().equals("faultName")) {
                catchFaultName = new String(att.getNodeValue());
                QName cfnQN = Util.buildQname(catchFaultName, this.proc);
                catchFaultName = Util.buildStringFromQName(cfnQN);
                //System.out.println("BPELCompilo : Catch FaultName : " +
                // catchFaultName);
            } else if (att.getNodeName().equals("faultVariable")) {
                catchFaultVariable = new String(att.getNodeValue());
                //System.out.println("BPELCompilo : Catch FaultVariable : " +
                // catchFaultVariable);
            }
        }

        //f.insertLine(" //org.objectweb.orchestra.tree.Catch s" + catchNb + " = new
        // org.objectweb.orchestra.tree.Catch(" + catchNb + ", \"" + catchFaultName + "\", \"" +
        // catchFaultVariable + "\");");

        String[] tabParam = new String[3];
        tabParam[0] = "new java.lang.Integer(" + catchNb + ")";
        tabParam[1] = "\"" + catchFaultName + "\"";
        tabParam[2] = "\"" + catchFaultVariable + "\"";
        createInstance("org.objectweb.orchestra.tree.Catch", tabParam);

        analyseChilds(n, "Catch");

        //System.out.println("BPELCompilo : End AnalyseCatch");
    }

    /**
     * analyse node CatchAll
     *
     * @param n
     *            the xml node representing this CatchAll
     */
    private void analyseCatchAll(Node n) throws BPELGeneratorException {
        //System.out.println("BPELCompilo : AnalyseCatchAll");
        int catchAllNb = ++this.statementNb;

        //f.insertLine(" //org.objectweb.orchestra.tree.CatchAll s" + catchAllNb + " = new
        // org.objectweb.orchestra.tree.CatchAll(" + catchAllNb + ");");

        String[] tabParam = new String[1];
        tabParam[0] = "new java.lang.Integer(" + catchAllNb + ")";
        createInstance("org.objectweb.orchestra.tree.CatchAll", tabParam);

        analyseChilds(n, "CatchAll");

        //System.out.println("BPELCompilo : End AnalyseCatchAll");
    }

    /**
     * analyse node Pick
     *
     * @param n
     *            the xml node representing this Pick
     */
    private void analysePick(Node n) throws BPELGeneratorException {
        //System.out.println("BPELCompilo : AnalysePick");
        int pickNb = ++this.statementNb;
        String createInstance = "no";
        String name = null;
        String joinCondition = null;
        String suppressJoinFailure = null;

        NamedNodeMap atts = n.getAttributes();
        for (int i = 0; i < atts.getLength(); i++) {
            Node att = atts.item(i);
            if (att.getNodeName().equals("createInstance")) {
                createInstance = new String(att.getNodeValue());
            } else if (att.getNodeName().equals("name")) {
                name = new String(att.getNodeValue());
            } else if (att.getNodeName().equals("joinCondition")) {
                joinCondition = new String(att.getNodeValue());
            } else if (att.getNodeName().equals("suppressJoinFailure")) {
                suppressJoinFailure = new String(att.getNodeValue());
            }
        }
        this.mCreateInstance = createInstance;

        //f.insertLine(" //org.objectweb.orchestra.tree.Pick s" + pickNb + " = new
        // org.objectweb.orchestra.tree.Pick(" + pickNb + ", \"" + createInstance + ", \"" + name +
        // "\", \"" + joinCondition + "\", \"" + suppressJoinFailure + "\");");

        String[] tabParam = new String[5];
        tabParam[0] = "new java.lang.Integer(" + pickNb + ")";
        tabParam[1] = "\"" + createInstance + "\"";
        tabParam[2] = "\"" + name + "\"";
        tabParam[3] = "\"" + joinCondition + "\"";
        tabParam[4] = "\"" + suppressJoinFailure + "\"";
        createInstance("org.objectweb.orchestra.tree.Pick", tabParam);

        analyseChilds(n, "Pick");
        this.mCreateInstance = "no";
        //System.out.println("BPELCompilo : End AnalysePick");
    }

    /**
     * analyse node OnMessage
     *
     * @param n
     *            the xml node representing this OnMessage
     */
    private void analyseOnMessage(Node n) throws BPELGeneratorException {
        //System.out.println("BPELCompilo : AnalyseOnMessage");
        String inoutTypeSave = this.inoutType;
        this.inoutType = "OnMessage";
        int onMessageNb = ++this.statementNb;
        String partnerLink = null;
        String portType = null;
        String operation = null;
        String variable = null;
        String name = null;
        NamedNodeMap atts = n.getAttributes();
        for (int i = 0; i < atts.getLength(); i++) {
            Node att = atts.item(i);
            if (att.getNodeName().equals("partnerLink")) {
                partnerLink = new String(att.getNodeValue());
            } else if (att.getNodeName().equals("portType")) {
                portType = new String(att.getNodeValue());
                QName ptQN = Util.buildQname(portType, this.proc);
                portType = Util.buildStringFromQName(ptQN);
            } else if (att.getNodeName().equals("operation")) {
                operation = new String(att.getNodeValue());
            } else if (att.getNodeName().equals("variable")) {
                variable = new String(att.getNodeValue());
            } else if (att.getNodeName().equals("name")) {
                name = new String(att.getNodeValue());
            }
        }
        boolean receivePrintedSave = this.receivePrinted;
        setReceiveInfo("onMessage", partnerLink, portType, operation, this.mCreateInstance,name);
        this.receivePrinted = false;
        //f.insertLine(" //org.objectweb.orchestra.tree.OnMessage s" + onMessageNb + " = new
        // org.objectweb.orchestra.tree.OnMessage(" + onMessageNb + ", \"" + partnerLink + "\", \""
        // + portType + "\", \"" + operation + "\", \"" + variable + "\");");

        String[] tabParam = new String[5];
        tabParam[0] = "new java.lang.Integer(" + onMessageNb + ")";
        tabParam[1] = "\"" + partnerLink + "\"";
        tabParam[2] = "\"" + portType + "\"";
        tabParam[3] = "\"" + operation + "\"";
        tabParam[4] = "\"" + variable + "\"";

        createInstance("org.objectweb.orchestra.tree.OnMessage", tabParam);

        analyseChilds(n, "OnMessage");
        if (!this.receivePrinted) {
            setReceiveInfo("onMessage", partnerLink, portType, operation, this.mCreateInstance,name);
            this.processMapFile.insertLine(this.receiveInfo);
        }
        this.receivePrinted = receivePrintedSave;
        this.inoutType = inoutTypeSave;
        //System.out.println("BPELCompilo : End AnalyseOnMessage");
    }

    /**
     * analyse node OnAlarm
     *
     * @param n
     *            the xml node representing this OnAlarm
     */
    private void analyseOnAlarm(Node n) throws BPELGeneratorException {
        //System.out.println("BPELCompilo : AnalyseOnAlarm");
        int onAlarmNb = ++this.statementNb;
        String onAlarmFor = null;
        String onAlarmUntil = null;

        //f.insertLine(" //org.objectweb.orchestra.tree.OnAlarm s" + onAlarmNb + " = new
        // org.objectweb.orchestra.tree.OnAlarm(" + onAlarmNb + ");");
        String[] tabParam = new String[1];
        tabParam[0] = "new java.lang.Integer(" + onAlarmNb + ")";
        createInstance("org.objectweb.orchestra.tree.OnAlarm", tabParam);

        NamedNodeMap atts = n.getAttributes();
        for (int i = 0; i < atts.getLength(); i++) {
            Node att = atts.item(i);
            if (att.getNodeName().equals("for")) {
                onAlarmFor = new String(att.getNodeValue());
                //System.out.println("BPELCompilo : OnAlarm For : " +
                // onAlarmFor);
                //f.insertLine(" //s" + onAlarmNb + ".setFor(\"" + onAlarmFor +
                // "\");");
                stringCall("org.objectweb.orchestra.tree.OnAlarm", "setFor", "s" + onAlarmNb, onAlarmFor);

            } else if (att.getNodeName().equals("until")) {
                onAlarmUntil = new String(att.getNodeValue());
                //System.out.println("BPELCompilo : OnAlarm Until : " +
                // onAlarmUntil);
                //f.insertLine(" //s" + onAlarmNb + ".setUntil(\"" +
                // onAlarmUntil + "\");");
                stringCall("org.objectweb.orchestra.tree.OnAlarm", "setUntil", "s" + onAlarmNb, onAlarmUntil);

            }
        }

        analyseChilds(n, "OnAlarm");

        //System.out.println("BPELCompilo : End AnalyseOnAlarm");
    }

    /**
     * analyseWait
     * Analyse node Wait
     *
     * @param n The XML node representing this Wait
     * @throws BPELGeneratorException
     */
    private void analyseWait(Node n) throws BPELGeneratorException {
        //System.out.println("BPELCompilo : AnalyseWait");
        int waitNb = ++this.statementNb;
        String waitFor = null;
        String waitUntil = null;
        String name = null;
        String joinCondition = null;
        String suppressJoinFailure = null;

        NamedNodeMap atts = n.getAttributes();
        for (int i = 0; i < atts.getLength(); i++) {
            Node att = atts.item(i);
            if (att.getNodeName().equals("name")) {
                name = new String(att.getNodeValue());
            } else if (att.getNodeName().equals("joinCondition")) {
                joinCondition = new String(att.getNodeValue());
            } else if (att.getNodeName().equals("suppressJoinFailure")) {
                suppressJoinFailure = new String(att.getNodeValue());
            }
        }

        //f.insertLine(" //org.objectweb.orchestra.tree.Wait s" + waitNb + " = new
        // org.objectweb.orchestra.tree.Wait(" + waitNb + ", \"" + name + "\", \"" + joinCondition
        // + "\", \"" + suppressJoinFailure + "\");");
        String[] tabParam = new String[4];
        tabParam[0] = "new java.lang.Integer(" + waitNb + ")";
        tabParam[1] = "\"" + name + "\"";
        tabParam[2] = "\"" + joinCondition + "\"";
        tabParam[3] = "\"" + suppressJoinFailure + "\"";
        createInstance("org.objectweb.orchestra.tree.Wait", tabParam);

        if ( this.BPELVersion.equals(BPELConstants.NS_BPEL_V1_1) ) { // BPEL 1.1 case : the time we look for is in the attributes
            atts = n.getAttributes();
            for (int i = 0; i < atts.getLength(); i++) {
                Node att = atts.item(i);
                if (att.getNodeName().equals("for")) {
                    waitFor = new String(att.getNodeValue());
                    stringCall("org.objectweb.orchestra.tree.Wait", "setFor", "s" + waitNb, waitFor);
                } else if (att.getNodeName().equals("until")) {
                    waitUntil = new String(att.getNodeValue());
                    stringCall("org.objectweb.orchestra.tree.Wait", "setUntil", "s" + waitNb, waitUntil);
                }
            }
        }
        else if ( this.BPELVersion.equals(BPELConstants.NS_BPEL_V2_0) ) { // BPEL 2.0 case : the time we look for is in a child
            Vector children = getElementChildren(n) ;
            Node child = null ;
            int i ;
            for ( i = 0 ; i < children.size(); i++ ) {
                child = (Node) children.get(i) ;
                if ( child.getNodeName().equals("for") ) {
                    Node forchild = child.getFirstChild() ;
                    waitFor = new String( replaceEndLine(forchild.getNodeValue()) ) ;
                    stringCall("org.objectweb.orchestra.tree.Wait", "setFor", "s" + waitNb, waitFor);
                    n.removeChild(child) ;
                    break ;
                }
                else if ( child.getNodeName().equals("until") ) {
                    Node untilchild = child.getFirstChild() ;
                    waitUntil = new String( replaceEndLine(untilchild.getNodeValue()) ) ;
                    stringCall("org.objectweb.orchestra.tree.Wait", "setUntil", "s" + waitNb, waitUntil);
                    n.removeChild(child) ;
                    break ;
                }
            }
            if ( i == children.size() ) {
                throw new BPELGeneratorException("BPELCompilo.analyseWait : A BPEL 2.0 Wait statement must have a For or an Until child", null) ;
            }
        }

        analyseChilds(n, "Wait");

        //System.out.println("BPELCompilo : End AnalyseWait");
    }

    /**
     * analyse node Terminate
     *
     * @param n
     *            the xml node representing this Terminate
     */
    private void analyseTerminate(Node n) throws BPELGeneratorException {
        //System.out.println("BPELCompilo : AnalyseTerminate");
        int terminateNb = ++this.statementNb;
        String name = null;
        String joinCondition = null;
        String suppressJoinFailure = null;

        NamedNodeMap atts = n.getAttributes();
        for (int i = 0; i < atts.getLength(); i++) {
            Node att = atts.item(i);
            if (att.getNodeName().equals("name")) {
                name = new String(att.getNodeValue());
            } else if (att.getNodeName().equals("joinCondition")) {
                joinCondition = new String(att.getNodeValue());
            } else if (att.getNodeName().equals("suppressJoinFailure")) {
                suppressJoinFailure = new String(att.getNodeValue());
            }
        }

        //f.insertLine(" //org.objectweb.orchestra.tree.Terminate s" + terminateNb + " = new
        // org.objectweb.orchestra.tree.Terminate(" + terminateNb + ", \"" + name + "\", \"" +
        // joinCondition + "\", \"" + suppressJoinFailure + "\");");
        String[] tabParam = new String[4];
        tabParam[0] = "new java.lang.Integer(" + terminateNb + ")";
        tabParam[1] = "\"" + name + "\"";
        tabParam[2] = "\"" + joinCondition + "\"";
        tabParam[3] = "\"" + suppressJoinFailure + "\"";
        createInstance("org.objectweb.orchestra.tree.Terminate", tabParam);

        analyseChilds(n, "Terminate");

        //System.out.println("BPELCompilo : End AnalyseTerminate");
    }

    /**
     * analyse node Empty
     *
     * @param n
     *            the xml node representing this Empty
     */
    private void analyseEmpty(Node n) throws BPELGeneratorException {
        //System.out.println("BPELCompilo : AnalyseEmpty");
        int emptyNb = ++this.statementNb;
        String name = null;
        String joinCondition = null;
        String suppressJoinFailure = null;

        NamedNodeMap atts = n.getAttributes();
        for (int i = 0; i < atts.getLength(); i++) {
            Node att = atts.item(i);
            if (att.getNodeName().equals("name")) {
                name = new String(att.getNodeValue());
            } else if (att.getNodeName().equals("joinCondition")) {
                joinCondition = new String(att.getNodeValue());
            } else if (att.getNodeName().equals("suppressJoinFailure")) {
                suppressJoinFailure = new String(att.getNodeValue());
            }
        }

        //f.insertLine(" //org.objectweb.orchestra.tree.Empty s" + emptyNb + " = new
        // org.objectweb.orchestra.tree.Empty(" + emptyNb + ", \"" + name + "\", \"" +
        // joinCondition + "\", \"" + suppressJoinFailure + "\");");

        String[] tabParam = new String[4];
        tabParam[0] = "new java.lang.Integer(" + emptyNb + ")";
        tabParam[1] = "\"" + name + "\"";
        tabParam[2] = "\"" + joinCondition + "\"";
        tabParam[3] = "\"" + suppressJoinFailure + "\"";
        createInstance("org.objectweb.orchestra.tree.Empty", tabParam);

        analyseChilds(n, "Empty");
        //System.out.println("BPELCompilo : End AnalyseEmpty");
    }

    /**
     * analyseWhileOrRepeatUntil
     * analyse While or RepeatUntil node
     *
     * @param n The XML node representing this While or RepeatUntil
     * @throws BPELGeneratorException
     */
    private void analyseWhileOrRepeatUntil(Node n) throws BPELGeneratorException {

        int whileNb = ++this.statementNb;
        String whileCondition = null;
        String name = null;
        String joinCondition = null;
        String suppressJoinFailure = null;

        NamedNodeMap atts = n.getAttributes();
        for (int i = 0; i < atts.getLength(); i++) {
            Node att = atts.item(i);
            if (att.getNodeName().equals("name")) {
                name = new String(att.getNodeValue());
            } else if (att.getNodeName().equals("joinCondition")) {
                joinCondition = new String(att.getNodeValue());
            } else if (att.getNodeName().equals("suppressJoinFailure")) {
                suppressJoinFailure = new String(att.getNodeValue());
            } else if ( att.getNodeName().equals("condition")) { // BPEL 1.1 case
                if ( !this.BPELVersion.equals(BPELConstants.NS_BPEL_V1_1) )
                    throw new BPELGeneratorException("BPELCompilo.analyseWhileOrRepeatUntil : A BPEL 2.0 " + n.getNodeName() + " statement do not have a condition attribute", null) ;
                whileCondition = new String(att.getNodeValue());
            }
        }

        //f.insertLine(" //org.objectweb.orchestra.tree.While s" + whileNb + " = new
        // org.objectweb.orchestra.tree.While(" + whileNb + ", \"" + whileCondition + ", \"" + name
        // + "\", \"" + joinCondition + "\", \"" + suppressJoinFailure +
        // "\");");

        // BPEL 2.0
        if ( this.BPELVersion.equals(BPELConstants.NS_BPEL_V2_0) ) {
            // Retrieving condition expression
            Vector children = getElementChildren(n) ;
            Node child ;
            int i ;
            for ( i = 0; i < children.size(); i++ ) {
                child = (Node) children.get(i) ;
                if ( child.getNodeName().equals("condition") ) {
                    Node conditionchild = child.getFirstChild() ;
                    whileCondition = new String( replaceEndLine(conditionchild.getNodeValue()) ) ;
                    n.removeChild(child) ;
                    break ;
                }
            }
            if ( i == children.size() ) {
                throw new BPELGeneratorException("BPELCompilo.analyseWhileOrRepeatUntil : A BPEL 2.0 " + n.getNodeName() + " statement must have a condition child", null) ;
            }
        }

        String[] tabParam = new String[5];
        tabParam[0] = "new java.lang.Integer(" + whileNb + ")";
        tabParam[1] = "\"" + whileCondition + "\"";
        tabParam[2] = "\"" + name + "\"";
        tabParam[3] = "\"" + joinCondition + "\"";
        tabParam[4] = "\"" + suppressJoinFailure + "\"";

        String elementName = n.getLocalName();
        if (elementName == null) {
            elementName = n.getNodeName();
        }
        if ( elementName.equals("while") ) {
            createInstance("org.objectweb.orchestra.tree.While", tabParam);
            analyseChilds(n, "While");
        }
        else { // repeatUntil
            createInstance("org.objectweb.orchestra.tree.RepeatUntil", tabParam);
            analyseChilds(n, "RepeatUntil");
        }
    }

    /**
     * analyseForEach
     * analyse ForEach node
     * @author Alexandre Greco
     *
     * @param n The XML node representing this ForEach
     * @throws BPELGeneratorException
     */
    private void analyseForEach(Node n) throws BPELGeneratorException {
        int forEachNb = ++this.statementNb;

        String counterName = null;
        String name = null;
        String joinCondition = null;
        String suppressJoinFailure = null;
        String parallel = null ;
        String startCounterValue = null ;
        String finalCounterValue = null ;

        if ( !this.BPELVersion.equals(BPELConstants.NS_BPEL_V2_0) )
            throw new BPELGeneratorException ("BPELCompilo.analyseForEach : ForEach activity is a BPEL 2.0 activity only", null) ;

        NamedNodeMap atts = n.getAttributes();
        for (int i = 0; i < atts.getLength(); i++) {
            Node att = atts.item(i);
            if (att.getNodeName().equals("name")) {
                name = new String(att.getNodeValue());
            } else if (att.getNodeName().equals("joinCondition")) {
                joinCondition = new String(att.getNodeValue());
            } else if (att.getNodeName().equals("suppressJoinFailure")) {
                suppressJoinFailure = new String(att.getNodeValue());
            } else if ( att.getNodeName().equals("counterName")) {
                counterName = new String(att.getNodeValue());
            } else if ( att.getNodeName().equals("parallel")) {
                parallel = new String(att.getNodeValue()) ;
            }
        }

        // Retrieving ForEach iterator child
        Vector children = getElementChildren(n) ;
        Node child = null ;
        for ( int i = 0; i < children.size(); i++) {
            child = (Node) children.get(i) ;
            if ( child.getNodeName().equals("iterator")) break ;
        }
        if ( !child.getNodeName().equals("iterator") ) {
            throw new BPELGeneratorException ("BPELCompilo.analyseForEach : ForEach activity must have an iterator child", null) ;
        }

        // Retrieving start and final counter value
        Vector counterValue = getElementChildren(child) ;
        Node value = null ;
        for ( int i = 0; i < 2; i++) {
            value = (Node) counterValue.get(i) ;
            if ( value.getNodeName().equals("startCounterValue") ) {
                startCounterValue = new String(replaceEndLine(value.getFirstChild().getNodeValue())) ;
            }
            else if ( value.getNodeName().equals("finalCounterValue") ) {
                finalCounterValue = new String(replaceEndLine(value.getFirstChild().getNodeValue())) ;
            }
            else {
                throw new BPELGeneratorException ("BPELCompilo.analyseForEach : invalid iterator child", null) ;
            }
        }
        n.removeChild(child) ; // iterator child is now useless

        // Compilation check
        children = getElementChildren(n) ;
        child = null ;
        int j = 0 ;
        for ( int i = 0; i < children.size(); i++) {
            child = (Node) children.get(i) ;
            String childName = child.getNodeName() ;
            if ( childName.equals("scope")) {
                if ( j == 1 ) { // the foreach activity as more than one scope child
                    throw new BPELGeneratorException ("BPELCompilo.analyseForEach : ForEach activity must have only one scope child", null) ;
                }
                j++ ;
            }
            else if ( !(childName.equals("targets") || childName.equals("sources")) ){
                throw new BPELGeneratorException ("BPELCompilo.analyseForEach : " + childName + " child is not allowed for the ForEach activity", null) ;
            }
        }
        if ( j == 0 ) { // no scope child has been found
            throw new BPELGeneratorException ("BPELCompilo.analyseForEach : ForEach activity must have one scope child", null) ;
        }

        String[] tabParam = new String[8];
        tabParam[0] = "new java.lang.Integer(" + forEachNb + ")";
        tabParam[1] = "\"" + parallel + "\"";
        tabParam[2] = "\"" + counterName + "\"";
        tabParam[3] = "\"" + startCounterValue + "\"";
        tabParam[4] = "\"" + finalCounterValue + "\"";
        tabParam[5] = "\"" + name + "\"";
        tabParam[6] = "\"" + joinCondition + "\"";
        tabParam[7] = "\"" + suppressJoinFailure + "\"";

        createInstance("org.objectweb.orchestra.tree.ForEach", tabParam);
        analyseChilds(n, "ForEach");
    }

    /**
     * replaceEndLine
     * Replace newline characters by \n in the given String
     * @author Alexandre Greco
     *
     * @param s The String you want to process
     * @return String The String with newline characters replaced
     */
    private String replaceEndLine (String s) {
        return s.replaceAll("\n", "\\\\n") ;
    }

    /**
     * analyse node Source
     *
     * @param n
     *            the xml node representing this Source
     */
    private void analyseSource(Node n) throws BPELGeneratorException {
        //System.out.println("BPELCompilo : AnalyseSource");
        int sourceNb = ++this.statementNb;
        String linkName = null;
        String transitionCondition = null;
        NamedNodeMap atts = n.getAttributes();
        for (int i = 0; i < atts.getLength(); i++) {
            Node att = atts.item(i);
            if (att.getNodeName().equals("linkName")) {
                linkName = new String(att.getNodeValue());
            } else if (att.getNodeName().equals("transitionCondition")) {
                transitionCondition = getWellFormedXPathExpr(att.getNodeValue());
            }
        }
        //f.insertLine(" //org.objectweb.orchestra.tree.Source s" + sourceNb + " = new
        // org.objectweb.orchestra.tree.Source(" + sourceNb + ", \"" + linkName + "\", \"" +
        // transitionCondition + "\");");
        String[] tabParam = new String[3];
        tabParam[0] = "new java.lang.Integer(" + sourceNb + ")";
        tabParam[1] = "\"" + linkName + "\"";
        tabParam[2] = "\"" + transitionCondition + "\"";

        createInstance("org.objectweb.orchestra.tree.Source", tabParam);

        //System.out.println("BPELCompilo : End AnalyseSource");
    }

    /**
     * analyse node Target
     *
     * @param n
     *            the xml node representing this Target
     */
    private void analyseTarget(Node n) throws BPELGeneratorException {
        //System.out.println("BPELCompilo : AnalyseTarget");
        int targetNb = ++this.statementNb;
        String linkName = null;
        NamedNodeMap atts = n.getAttributes();
        for (int i = 0; i < atts.getLength(); i++) {
            Node att = atts.item(i);
            if (att.getNodeName().equals("linkName")) linkName = new String(att.getNodeValue());
        }
        //f.insertLine(" //org.objectweb.orchestra.tree.Target s" + targetNb + " = new
        // org.objectweb.orchestra.tree.Target(" + targetNb + ", \"" + linkName + "\");");
        String[] tabParam = new String[2];
        tabParam[0] = "new java.lang.Integer(" + targetNb + ")";
        tabParam[1] = "\"" + linkName + "\"";

        createInstance("org.objectweb.orchestra.tree.Target", tabParam);

        //System.out.println("BPELCompilo : End AnalyseTarget");
    }

    /**
     * analyse node Links
     *
     * @param n
     *            the xml node representing this Links
     */
    private void analyseLinks(Node n) throws BPELGeneratorException {
        //System.out.println("BPELCompilo : AnalyseLinks");
        int linksNb = ++this.statementNb;

        //f.insertLine(" //org.objectweb.orchestra.tree.Links s" + linksNb + " = new
        // org.objectweb.orchestra.tree.Links(" + linksNb + ");");
        String[] tabParam = new String[1];
        tabParam[0] = "new java.lang.Integer(" + linksNb + ")";

        createInstance("org.objectweb.orchestra.tree.Links", tabParam);

        analyseChilds(n, "Links");
        //System.out.println("BPELCompilo : End AnalyseLinks");
    }

    /**
     * analyse node Link
     *
     * @param n
     *            the xml node representing this Links
     */
    private void analyseLink(Node n) throws BPELGeneratorException {
        //System.out.println("BPELCompilo : AnalyseLink");
        int linkNb = ++this.statementNb;
        String name = null;
        NamedNodeMap atts = n.getAttributes();
        for (int i = 0; i < atts.getLength(); i++) {
            Node att = atts.item(i);
            if (att.getNodeName().equals("name")) name = new String(att.getNodeValue());
        }
        //f.insertLine(" //org.objectweb.orchestra.tree.Link s" + linkNb + " = new
        // org.objectweb.orchestra.tree.Link(" + linkNb + ", \"" + name + "\");");
        String[] tabParam = new String[2];
        tabParam[0] = "new java.lang.Integer(" + linkNb + ")";
        tabParam[1] = "\"" + name + "\"";

        createInstance("org.objectweb.orchestra.tree.Link", tabParam);
        //System.out.println("BPELCompilo : End AnalyseLink");
    }

    /**
     * analyse node PartnerLinks
     *
     * @param n
     *            the xml node representing this PartnerLinks
     */
    private void analysePartnerLinks(Node n) throws BPELGeneratorException {
        //System.out.println("BPELCompilo : AnalysePartnerLinks");
        int partnerLinksNb = ++this.statementNb;

        //f.insertLine(" //org.objectweb.orchestra.tree.PartnerLinks s" + partnerLinksNb + " = new
        // org.objectweb.orchestra.tree.PartnerLinks(" + partnerLinksNb + ");");
        String[] tabParam = new String[1];
        tabParam[0] = "new java.lang.Integer(" + partnerLinksNb + ")";

        createInstance("org.objectweb.orchestra.tree.PartnerLinks", tabParam);

        analyseChilds(n, "PartnerLinks");
        //System.out.println("BPELCompilo : End AnalysePartnerLinks");
    }

    /**
     * analyse node PartnerLink
     *
     * @param n
     *            the xml node representing this PartnerLink
     */
    private void analysePartnerLink(Node n) throws BPELGeneratorException {
        //System.out.println("BPELCompilo : AnalyseLink");
        int partnerLinkNb = ++this.statementNb;
        String name = null;
        String partnerLinkType = null;
        String myRole = null;
        String partnerRole = null;

        NamedNodeMap atts = n.getAttributes();
        for (int i = 0; i < atts.getLength(); i++) {
            Node att = atts.item(i);
            if (att.getNodeName().equals("name"))
                name = new String(att.getNodeValue());
            else if (att.getNodeName().equals("partnerLinkType"))
                partnerLinkType = new String(att.getNodeValue());
            else if (att.getNodeName().equals("myRole"))
                myRole = new String(att.getNodeValue());
            else if (att.getNodeName().equals("partnerRole")) partnerRole = new String(att.getNodeValue());
        }
        //namespaces management
        QName pltQN = Util.buildQname(partnerLinkType, this.proc);
        partnerLinkType = Util.buildStringFromQName(pltQN);
        if (this.plTypes.get(pltQN) == null) {
            throw new BPELGeneratorException("Undefined partnerLinkType found. Its name is '" + pltQN + "'");
        }

        //f.insertLine(" //org.objectweb.orchestra.tree.PartnerLink s" + partnerLinkNb + " = new
        // org.objectweb.orchestra.tree.PartnerLink(" + partnerLinkNb + ", \"" + name + "\", \"" +
        // partnerLinkType + "\", \"" + myRole + "\", \"" + partnerRole +
        // "\");");
        String[] tabParam = new String[5];
        tabParam[0] = "new java.lang.Integer(" + partnerLinkNb + ")";
        tabParam[1] = "\"" + name + "\"";
        tabParam[2] = "\"" + partnerLinkType + "\"";
        tabParam[3] = "\"" + myRole + "\"";
        tabParam[4] = "\"" + partnerRole + "\"";

        createInstance("org.objectweb.orchestra.tree.PartnerLink", tabParam);
        //System.out.println("BPELCompilo : End AnalyseLink");
    }

    /**
     * analyse node Partners
     *
     * @param n
     *            the xml node representing this Partners
     */
    private void analysePartners(Node n) throws BPELGeneratorException {
        //System.out.println("BPELCompilo : AnalysePartners");
        int partnersNb = ++this.statementNb;

        //f.insertLine(" //org.objectweb.orchestra.tree.Partners s" + partnersNb + " = new
        // org.objectweb.orchestra.tree.Partners(" + partnersNb + ");");
        String[] tabParam = new String[1];
        tabParam[0] = "new java.lang.Integer(" + partnersNb + ")";

        createInstance("org.objectweb.orchestra.tree.Partners", tabParam);

        analyseChilds(n, "Partners");
        //System.out.println("BPELCompilo : End AnalysePartners");
    }

    /**
     * analyse node Partner
     *
     * @param n
     *            the xml node representing this Partner
     */
    private void analysePartner(Node n) throws BPELGeneratorException {
        //System.out.println("BPELCompilo : AnalysePartner");
        int partnerNb = ++this.statementNb;
        String name = null;

        NamedNodeMap atts = n.getAttributes();
        for (int i = 0; i < atts.getLength(); i++) {
            Node att = atts.item(i);
            if (att.getNodeName().equals("name")) name = new String(att.getNodeValue());
        }
        //f.insertLine(" //org.objectweb.orchestra.tree.Partner s" + partnerNb + " = new
        // org.objectweb.orchestra.tree.Partner(" + partnerNb + ", \"" + name + "\");");
        String[] tabParam = new String[2];
        tabParam[0] = "new java.lang.Integer(" + partnerNb + ")";
        tabParam[1] = "\"" + name + "\"";

        createInstance("org.objectweb.orchestra.tree.Partner", tabParam);

        analyseChilds(n, "Partner");
        //System.out.println("BPELCompilo : End AnalysePartner");
    }

    /**
     * analyse node Copy
     *
     * @param n
     *            the xml node representing this Copy
     */
    private void analyseCopy(Node n) throws BPELGeneratorException {
        //System.out.println("BPELCompilo : AnalyseCopy");
        int copyNb = ++this.statementNb;

        //f.insertLine(" //org.objectweb.orchestra.tree.Copy s" + copyNb + " = new
        // org.objectweb.orchestra.tree.Copy(" + copyNb + ");");
        String[] tabParam = new String[1];
        tabParam[0] = "new java.lang.Integer(" + copyNb + ")";
        createInstance("org.objectweb.orchestra.tree.Copy", tabParam);

        analyseChilds(n, "Copy");

        //System.out.println("BPELCompilo : End AnalyseCopy");
    }

    /**
     * analyse node From
     *
     * @param n
     *            the xml node representing this From
     */
    private void analyseFrom(Node n) throws BPELGeneratorException {
        //System.out.println("BPELCompilo : AnalyseFrom");
        // type = 1 : variable, part
        // type = 2 : partnerLink, endpointReference
        // type = 3 : variable, property
        // type = 4 : expression
        int fromNb = ++this.statementNb;
        int type = -1;
        int newType = -1;

        String variable = null;
        String part = null;
        String partnerLink = null;
        String endpointReference = null;
        String property = null;
        String expression = null;
        String query = null;
        String opaque = null;

        NamedNodeMap atts = n.getAttributes();
        for (int i = 0; i < atts.getLength(); i++) {
            Node att = atts.item(i);
            if (att.getNodeName().equals("variable")) {
                newType = 0;
                variable = new String(att.getNodeValue());
            } else if (att.getNodeName().equals("part")) {
                newType = 1;
                part = new String(att.getNodeValue());
            } else if (att.getNodeName().equals("partnerLink")) {
                newType = 2;
                partnerLink = new String(att.getNodeValue());
            } else if (att.getNodeName().equals("endpointReference")) {
                endpointReference = new String(att.getNodeValue());
            } else if (att.getNodeName().equals("property")) {
                newType = 3;
                property = new String(att.getNodeValue());
                QName pQN = Util.buildQname(property, this.proc);
                property = Util.buildStringFromQName(pQN);
            } else if (att.getNodeName().equals("expression")) {
                newType = 4;
                expression = getWellFormedXPathExpr(att.getNodeValue());
            } else if (att.getNodeName().equals("query")) {
                newType = 5;
                query = new String(att.getNodeValue());
            } else if (att.getNodeName().equals("opaque")) {
                newType = 6;
                opaque = new String(att.getNodeValue());
            }
            if (type <= newType) {
                type = newType;
            }
        }

        //f.insertLine(" //org.objectweb.orchestra.tree.From s" + fromNb + " = new
        // org.objectweb.orchestra.tree.From(" + fromNb + ");");
        String[] tabParam = new String[1];
        tabParam[0] = "new java.lang.Integer(" + fromNb + ")";
        createInstance("org.objectweb.orchestra.tree.From", tabParam);

        switch (type) {
        case 0:
            //f.insertLine(" //s" + fromNb + ".setVariable(\"" + variable +
            // "\");");
            stringCall("org.objectweb.orchestra.tree.From", "setVariable", "s" + fromNb, variable);
            break;
        case 1:
            //f.insertLine(" //s" + fromNb + ".setVariable(\"" + variable +
            // "\");");
            stringCall("org.objectweb.orchestra.tree.From", "setVariable", "s" + fromNb, variable);
            //f.insertLine(" //s" + fromNb + ".setPart(\"" + part + "\");");
            stringCall("org.objectweb.orchestra.tree.From", "setPart", "s" + fromNb, part);
            break;
        case 2:
            //f.insertLine(" //s" + fromNb + ".setPartnerLink(\"" + partnerLink
            // + "\");");
            stringCall("org.objectweb.orchestra.tree.From", "setPartnerLink", "s" + fromNb, partnerLink);
            //f.insertLine(" //s" + fromNb + ".setEndpointReference(\"" +
            // endpointReference + "\");");
            stringCall("org.objectweb.orchestra.tree.From", "setEndpointReference", "s" + fromNb, endpointReference);

            break;
        case 3:
            //f.insertLine(" //s" + fromNb + ".setVariable(\"" + variable +
            // "\");");
            stringCall("org.objectweb.orchestra.tree.From", "setVariable", "s" + fromNb, variable);
            //f.insertLine(" //s" + fromNb + ".setProperty(\"" + property +
            // "\");");
            stringCall("org.objectweb.orchestra.tree.From", "setProperty", "s" + fromNb, property);
            break;
        case 4:
            //f.insertLine(" //s" + fromNb + ".setExpression(\"" + expression +
            // "\");");
            stringCall("org.objectweb.orchestra.tree.From", "setExpression", "s" + fromNb, expression);
            break;
        case 5:
            //f.insertLine(" //s" + fromNb + ".setVariable(\"" + variable +
            // "\");");
            stringCall("org.objectweb.orchestra.tree.From", "setVariable", "s" + fromNb, variable);
            //f.insertLine(" //s" + fromNb + ".setPart(\"" + part + "\");");
            stringCall("org.objectweb.orchestra.tree.From", "setPart", "s" + fromNb, part);
            //f.insertLine(" //s" + fromNb + ".setQuery(\"" + query + "\");");
            stringCall("org.objectweb.orchestra.tree.From", "setQuery", "s" + fromNb, query);
            break;
        case 6:
            //f.insertLine(" //s" + fromNb + ".setOpaque(\"" + opaque +
            // "\");");
            stringCall("org.objectweb.orchestra.tree.From", "setOpaque", "s" + fromNb, opaque);
            break;
        }
        //System.out.println("BPELCompilo : End AnalyseFrom");
    }

    /**
     * analyse node To
     *
     * @param n
     *            the xml node representing this To
     */
    private void analyseTo(Node n) throws BPELGeneratorException {
        //System.out.println("BPELCompilo : AnalyseTo");
        // type = 1 : variable, part
        // type = 2 : partnerLink
        // type = 3 : variable, property
        int toNb = ++this.statementNb;
        int type = -1;
        int newType = -1;
        String variable = null;
        String part = null;
        String partnerLink = null;
        String property = null;
        String query = null;

        NamedNodeMap atts = n.getAttributes();
        for (int i = 0; i < atts.getLength(); i++) {
            Node att = atts.item(i);
            if (att.getNodeName().equals("variable")) {
                newType = 0;
                variable = new String(att.getNodeValue());
            } else if (att.getNodeName().equals("part")) {
                newType = 1;
                part = new String(att.getNodeValue());
            } else if (att.getNodeName().equals("partnerLink")) {
                newType = 2;
                partnerLink = new String(att.getNodeValue());
            } else if (att.getNodeName().equals("property")) {
                newType = 3;
                property = new String(att.getNodeValue());
                QName pQN = Util.buildQname(property, this.proc);
                property = Util.buildStringFromQName(pQN);
            } else if (att.getNodeName().equals("query")) {
                newType = 4;
                query = new String(att.getNodeValue());
            }
            if (type <= newType) {
                type = newType;
            }
        }

        //f.insertLine(" //org.objectweb.orchestra.tree.To s" + toNb + " = new org.objectweb.orchestra.tree.To(" +
        // toNb + ");");
        String[] tabParam = new String[1];
        tabParam[0] = "new java.lang.Integer(" + toNb + ")";
        createInstance("org.objectweb.orchestra.tree.To", tabParam);

        switch (type) {
        case 0:
            //f.insertLine(" //s" + toNb + ".setVariable(\"" + variable +
            // "\");");
            stringCall("org.objectweb.orchestra.tree.To", "setVariable", "s" + toNb, variable);
            break;
        case 1:
            //f.insertLine(" //s" + toNb + ".setVariable(\"" + variable +
            // "\");");
            stringCall("org.objectweb.orchestra.tree.To", "setVariable", "s" + toNb, variable);
            //f.insertLine(" //s" + toNb + ".setPart(\"" + part + "\");");
            stringCall("org.objectweb.orchestra.tree.To", "setPart", "s" + toNb, part);
            break;
        case 2:
            //f.insertLine(" //s" + toNb + ".setPartnerLink(\"" + partnerLink +
            // "\");");
            stringCall("org.objectweb.orchestra.tree.To", "setPartnerLink", "s" + toNb, partnerLink);
            break;
        case 3:
            //f.insertLine(" //s" + toNb + ".setVariable(\"" + variable +
            // "\");");
            stringCall("org.objectweb.orchestra.tree.To", "setVariable", "s" + toNb, variable);
            //f.insertLine(" //s" + toNb + ".setProperty(\"" + property +
            // "\");");
            stringCall("org.objectweb.orchestra.tree.To", "setProperty", "s" + toNb, property);
            break;
        case 4:
            //f.insertLine(" //s" + toNb + ".setVariable(\"" + variable +
            // "\");");
            stringCall("org.objectweb.orchestra.tree.To", "setVariable", "s" + toNb, variable);
            //f.insertLine(" //s" + toNb + ".setPart(\"" + part + "\");");
            stringCall("org.objectweb.orchestra.tree.To", "setPart", "s" + toNb, part);
            //f.insertLine(" //s" + toNb + ".setQuery(\"" + query + "\");");
            stringCall("org.objectweb.orchestra.tree.To", "setQuery", "s" + toNb, query);
            break;
        }
        //System.out.println("BPELCompilo : End AnalyseTo");
    }

    /**
     * analyse node Assign
     *
     * @param n
     *            the xml node representing this Assign
     */
    private void analyseAssign(Node n) throws BPELGeneratorException {
        //System.out.println("BPELCompilo : AnalyseAssign");
        int assignNb = ++this.statementNb;
        String name = null;
        String joinCondition = null;
        String suppressJoinFailure = null;
        NamedNodeMap atts = n.getAttributes();
        for (int i = 0; i < atts.getLength(); i++) {
            Node att = atts.item(i);
            if (att.getNodeName().equals("name")) {
                name = new String(att.getNodeValue());
            } else if (att.getNodeName().equals("joinCondition")) {
                joinCondition = new String(att.getNodeValue());
            } else if (att.getNodeName().equals("suppressJoinFailure")) {
                suppressJoinFailure = new String(att.getNodeValue());
            }
        }

        //f.insertLine(" //org.objectweb.orchestra.tree.Assign s" + assignNb + " = new
        // org.objectweb.orchestra.tree.Assign(" + assignNb + ", \"" + name + "\", \"" +
        // joinCondition + "\", \"" + suppressJoinFailure + "\");");
        String[] tabParam = new String[4];
        tabParam[0] = "new java.lang.Integer(" + assignNb + ")";
        tabParam[1] = "\"" + name + "\"";
        tabParam[2] = "\"" + joinCondition + "\"";
        tabParam[3] = "\"" + suppressJoinFailure + "\"";
        createInstance("org.objectweb.orchestra.tree.Assign", tabParam);

        analyseChilds(n, "Assign");

        //System.out.println("BPELCompilo : End AnalyseAssign");
    }

    /**
     * analyse node Switch
     *
     * @param n
     *            the xml node representing this Switch
     */
    private void analyseSwitch(Node n) throws BPELGeneratorException {
        //System.out.println("BPELCompilo : AnalyseSwitch");
        int switchNb = ++this.statementNb;
        String name = null;
        String joinCondition = null;
        String suppressJoinFailure = null;

        NamedNodeMap atts = n.getAttributes();
        for (int i = 0; i < atts.getLength(); i++) {
            Node att = atts.item(i);
            if (att.getNodeName().equals("name")) {
                name = new String(att.getNodeValue());
            } else if (att.getNodeName().equals("joinCondition")) {
                joinCondition = new String(att.getNodeValue());
            } else if (att.getNodeName().equals("suppressJoinFailure")) {
                suppressJoinFailure = new String(att.getNodeValue());
            }
        }

        //f.insertLine(" //org.objectweb.orchestra.tree.Switch s" + switchNb + " = new
        // org.objectweb.orchestra.tree.Switch(" + switchNb + ", \"" + name + "\", \"" +
        // joinCondition + "\", \"" + suppressJoinFailure + "\");");
        String[] tabParam = new String[4];
        tabParam[0] = "new java.lang.Integer(" + switchNb + ")";
        tabParam[1] = "\"" + name + "\"";
        tabParam[2] = "\"" + joinCondition + "\"";
        tabParam[3] = "\"" + suppressJoinFailure + "\"";
        createInstance("org.objectweb.orchestra.tree.Switch", tabParam);

        analyseChilds(n, "Switch");

        //System.out.println("BPELCompilo : End AnalyseSwitch");
    }

    /**
     * analyseIf
     * Analyse node If : actually transforms the if statement
     * to a switch BPEL 1.1 statement by working on the XML tree.
     * Then it calls analyseSwitch.
     * @author Alexandre Greco
     *
     * @param n The XML node representing this If
     * @throws BPELGeneratorException
     * @throws ParserConfigurationException
     */
    private void analyseIf(Node n) throws BPELGeneratorException {

        if ( !this.BPELVersion.equals(BPELConstants.NS_BPEL_V2_0) )
            throw new BPELGeneratorException ("BPELCompilo.analyseIf : If activity is a BPEL 2.0 activity only", null) ;

        Node child, newChild ;
        Node firstCase = null ;
        Node activity = null ;
        String childName ;

        Node switchNode = this.document.createElement("switch") ;

        // Retrieving attributes
        NamedNodeMap atts = n.getAttributes();
        for (int i = 0; i < atts.getLength(); i++) {
            Node attribute = atts.item(i) ;
            ((Element) switchNode).setAttribute(attribute.getNodeName(), attribute.getNodeValue()) ;
        }

        // Switch node building
        Vector children = getElementChildren(n) ;
        for (int i = 0; i < children.size(); i++) {
            child = (Node) children.get(i) ;
            childName = child.getNodeName() ;
            if ( childName.equals("condition") ) {
                firstCase = this.document.createElement("case") ;
                ((Element) firstCase).setAttribute("condition", replaceEndLine(child.getFirstChild().getNodeValue())) ;
                switchNode.appendChild(firstCase) ;
            }
            else if ( childName.equals("elseif") ) {
                newChild = this.document.createElement("case") ;
                for ( Node node = child.getFirstChild(); node != null; node = node.getNextSibling() ) {
                    if ( node.getNodeName().equals("condition") ) {
                        ((Element) newChild).setAttribute("condition", replaceEndLine(node.getFirstChild().getNodeValue())) ;
                    }
                    else if ( node.getNodeType() == Node.ELEMENT_NODE ){
                        newChild.appendChild( node ) ;
                    }
                }
                switchNode.appendChild( newChild ) ;
            }
            else if ( childName.equals("else") ) {
                newChild = this.document.createElement("otherwise") ;
                newChild.appendChild( getFirstElementChild(child) ) ;
                switchNode.appendChild( newChild ) ;
            }
            else if ( childName.equals("targets") || childName.equals("sources") ) {
                switchNode.appendChild(child) ;
            }
            else {
                if ( activity != null )
                    throw new BPELGeneratorException ("BPELCompilo.analyseIf : A If statement must have only one activity to execute", null) ;
                activity = child ;
            }
        }

        if ( firstCase == null )
            throw new BPELGeneratorException ("BPELCompilo.analyseIf : An If statement must have a condition child", null) ;
        if ( activity == null )
            throw new BPELGeneratorException ("BPELCompilo.analyseIf : An If statement must have an activity to execute", null) ;

        firstCase.appendChild(activity) ;

        n.getParentNode().replaceChild(switchNode, n) ;
        analyseSwitch( switchNode ) ;
    }

    /**
     * getFirstElementChild
     * @author Alexandre Greco
     *
     * @param n An XML node
     * @return Node The first element node in the child list of n,
     * 			   and null if there is no such node
     */
    private Node getFirstElementChild (Node n) {
        Node result = null ;
        for (Node child = n.getFirstChild(); child != null; child = child.getNextSibling()) {
            if (child.getNodeType() == Node.ELEMENT_NODE) {
                result = child ;
                break ;
            }
        }
        return result;
    }

    /**
     * printIncrement
     * Print the string s nb times
     * @author Alexandre Greco
     *
     * @param nb Represent the indentation level
     * @param s String to print
     */
    public void printIncrement (int nb, String s) {
        for (int i = 0; i < nb; i++ ) {
            System.out.print(s) ;
        }
        System.out.print(" ") ;
    }

    /**
     * printAttributes
     * Print the attributes of the node n
     * @author Alexandre Greco
     *
     * @param n An XML node
     * @param nb Represent the indentation level
     */
    public void printAttributes (Node n, int nb) {
        if ( n.hasAttributes() ) {
            NamedNodeMap atts = n.getAttributes() ;
            printIncrement(nb, ">>>") ;
            for (int j = 0; j < atts.getLength(); j++) {
                Node attribute = atts.item(j) ;
                System.out.print(attribute.getNodeName() + "=" + attribute.getNodeValue() + " ");
            }
            System.out.println() ;
        }
    }

    /**
     * printXMLTree
     * Print the XML tree which root is n
     * @author Alexandre Greco
     *
     * @param n An XML node
     * @param nb Represent the indentation level
     */
    public void printXMLTree (Node n, int nb) {
        Node child ;
        Vector children = getElementChildren(n) ;
        for ( int i = 0; i < children.size(); i++ ) {
            child = (Node) children.get(i) ;
            printIncrement(nb, "###") ;
            System.out.println(child.getNodeName()) ;
            printAttributes(child, nb) ;
            printXMLTree(child, nb + 1) ;
        }
    }

    /**
     * printXMLTree
     * Print the XML tree which root is n
     * @author Alexandre Greco
     *
     * @param n An XML node
     */
    public void printXMLTree (Node n) {
        System.out.println("**************************************") ;
        System.out.println("printXMLTree : Node analyse") ;
        System.out.println(n.getNodeName()) ;
        printAttributes(n, 0) ;
        printXMLTree(n, 1) ;
        System.out.println("**************************************") ;
    }

    /**
     * analyse node Case
     *
     * @param n
     *            the xml node representing this Case
     */
    private void analyseCase(Node n) throws BPELGeneratorException {
        //System.out.println("BPELCompilo : AnalyseCase");
        int caseNb = ++this.statementNb;
        String caseCondition = null;

        NamedNodeMap atts = n.getAttributes();
        for (int i = 0; i < atts.getLength(); i++) {
            Node att = atts.item(i);
            if (att.getNodeName().equals("condition")) {
                caseCondition = getWellFormedXPathExpr(att.getNodeValue());
                //System.out.println("BPELCompilo : Case Condition : " +
                // caseCondition);
            }
        }

        //f.insertLine(" //org.objectweb.orchestra.tree.Case s" + caseNb + " = new
        // org.objectweb.orchestra.tree.Case(" + caseNb + ", \"" + caseCondition + "\");");
        String[] tabParam = new String[2];
        tabParam[0] = "new java.lang.Integer(" + caseNb + ")";
        tabParam[1] = "\"" + caseCondition + "\"";
        createInstance("org.objectweb.orchestra.tree.Case", tabParam);

        analyseChilds(n, "Case");

        //System.out.println("BPELCompilo : End AnalyseCase");
    }

    /**
     * analyse node Otherwise
     *
     * @param n
     *            the xml node representing this Otherwise
     */
    private void analyseOtherwise(Node n) throws BPELGeneratorException {
        //System.out.println("BPELCompilo : AnalyseOtherwise");
        int otherwiseNb = ++this.statementNb;

        //f.insertLine(" //org.objectweb.orchestra.tree.Otherwise s" + otherwiseNb + " = new
        // org.objectweb.orchestra.tree.Otherwise(" + otherwiseNb + ");");
        String[] tabParam = new String[1];
        tabParam[0] = "new java.lang.Integer(" + otherwiseNb + ")";
        createInstance("org.objectweb.orchestra.tree.Otherwise", tabParam);

        analyseChilds(n, "Otherwise");

        //System.out.println("BPELCompilo : End AnalyseOtherwise");
    }

    /**
     * analyse node Flow
     *
     * @param n
     *            the xml node representing this Flow
     */
    private void analyseFlow(Node n) throws BPELGeneratorException {
        //System.out.println("BPELCompilo : AnalyseFlow");
        int flowNb = ++this.statementNb;
        String name = null;
        String joinCondition = null;
        String suppressJoinFailure = null;

        NamedNodeMap atts = n.getAttributes();
        for (int i = 0; i < atts.getLength(); i++) {
            Node att = atts.item(i);
            if (att.getNodeName().equals("name")) {
                name = new String(att.getNodeValue());
            } else if (att.getNodeName().equals("joinCondition")) {
                joinCondition = new String(att.getNodeValue());
            } else if (att.getNodeName().equals("suppressJoinFailure")) {
                suppressJoinFailure = new String(att.getNodeValue());
            }
        }

        //f.insertLine(" //org.objectweb.orchestra.tree.Scope s" + flowNb + " = new
        // org.objectweb.orchestra.tree.Scope(" + flowNb + ", \"" + name+ "\", \"" + joinCondition
        // + "\", \"" + suppressJoinFailure + "\");");
        String[] tabParam = new String[4];
        tabParam[0] = "new java.lang.Integer(" + flowNb + ")";
        tabParam[1] = "\"" + name + "\"";
        tabParam[2] = "\"" + joinCondition + "\"";
        tabParam[3] = "\"" + suppressJoinFailure + "\"";
        createInstance("org.objectweb.orchestra.tree.Flow", tabParam);

        analyseChilds(n, "Flow");
        //System.out.println("BPELCompilo : End AnalyseFlow");
    }

    /**
     * analyse node Sequence
     *
     * @param n
     *            the xml node representing this Sequence
     */
    private void analyseSequence(Node n) throws BPELGeneratorException {
        //System.out.println("BPELCompilo : AnalyseSequence");
        int sequenceNb = ++this.statementNb;
        String name = null;
        String joinCondition = null;
        String suppressJoinFailure = null;

        NamedNodeMap atts = n.getAttributes();
        for (int i = 0; i < atts.getLength(); i++) {
            Node att = atts.item(i);
            if (att.getNodeName().equals("name")) {
                name = new String(att.getNodeValue());
            } else if (att.getNodeName().equals("joinCondition")) {
                joinCondition = new String(att.getNodeValue());
            } else if (att.getNodeName().equals("suppressJoinFailure")) {
                suppressJoinFailure = new String(att.getNodeValue());
            }
        }

        //f.insertLine(" //org.objectweb.orchestra.tree.Sequence s" + sequenceNb + " = new
        // org.objectweb.orchestra.tree.Sequence(" + sequenceNb + ", \"" + name + "\", \"" +
        // joinCondition + "\", \"" + suppressJoinFailure + "\");");
        String[] tabParam = new String[4];
        tabParam[0] = "new java.lang.Integer(" + sequenceNb + ")";
        tabParam[1] = "\"" + name + "\"";
        tabParam[2] = "\"" + joinCondition + "\"";
        tabParam[3] = "\"" + suppressJoinFailure + "\"";
        createInstance("org.objectweb.orchestra.tree.Sequence", tabParam);

        analyseChilds(n, "Sequence");

        //System.out.println("BPELCompilo : End AnalyseSequence");
    }

    /**
     * analyse node Receive
     *
     * @param n
     *            the xml node representing this Receive
     */
    private void analyseReceive(Node n) throws BPELGeneratorException {
        //System.out.println("BPELCompilo : AnalyseReceive");
        String inoutTypeSave = this.inoutType;
        this.inoutType = "Receive";
        int receiveNb = ++this.statementNb;
        String partnerLink = null;
        String portType = null;
        String operation = null;
        String variable = null;
        String createInstance = "no";
        String name = null;
        String joinCondition = null;
        String suppressJoinFailure = null;

        NamedNodeMap atts = n.getAttributes();
        for (int i = 0; i < atts.getLength(); i++) {
            Node att = atts.item(i);
            if (att.getNodeName().equals("partnerLink")) {
                partnerLink = new String(att.getNodeValue());
            } else if (att.getNodeName().equals("portType")) {
                portType = new String(att.getNodeValue());
                //namespaces management
                QName ptQN = Util.buildQname(portType, this.proc);
                portType = Util.buildStringFromQName(ptQN);
            } else if (att.getNodeName().equals("operation")) {
                operation = new String(att.getNodeValue());
            } else if (att.getNodeName().equals("variable")) {
                variable = new String(att.getNodeValue());
            } else if (att.getNodeName().equals("createInstance")) {
                createInstance = new String(att.getNodeValue());
            } else if (att.getNodeName().equals("name")) {
                name = new String(att.getNodeValue());
            } else if (att.getNodeName().equals("joinCondition")) {
                joinCondition = new String(att.getNodeValue());
            } else if (att.getNodeName().equals("suppressJoinFailure")) {
                suppressJoinFailure = new String(att.getNodeValue());
            }
        }
        boolean receivePrintedSave = this.receivePrinted;
        setReceiveInfo("receive", partnerLink, portType, operation, createInstance,name);
        this.receivePrinted = false;
        //f.insertLine(" //org.objectweb.orchestra.tree.Receive s" + receiveNb + " = new
        // org.objectweb.orchestra.tree.Receive(" + receiveNb + ", \"" + partnerLink + "\", \"" +
        // portType + "\", \"" + operation + "\", \"" + variable + "\", \"" +
        // createInstance + "\", \"" + name + "\", \"" + joinCondition + "\",
        // \"" + suppressJoinFailure + "\");");

        String[] tabParam = new String[9];
        tabParam[0] = "new java.lang.Integer(" + receiveNb + ")";
        tabParam[1] = "\"" + partnerLink + "\"";
        tabParam[2] = "\"" + portType + "\"";
        tabParam[3] = "\"" + operation + "\"";
        tabParam[4] = "\"" + variable + "\"";
        tabParam[5] = "\"" + createInstance + "\"";
        tabParam[6] = "\"" + name + "\"";
        tabParam[7] = "\"" + joinCondition + "\"";
        tabParam[8] = "\"" + suppressJoinFailure + "\"";

        createInstance("org.objectweb.orchestra.tree.Receive", tabParam);

        analyseChilds(n, "Receive");
        if (!this.receivePrinted) {
            setReceiveInfo("receive", partnerLink, portType, operation, createInstance,name);
            this.processMapFile.insertLine(this.receiveInfo);
        }
        this.receivePrinted = receivePrintedSave;
        this.inoutType = inoutTypeSave;
        //System.out.println("BPELCompilo : End AnalyseReceive");
    }

    /**
     * Method to set the Receive Info. Used for the ProcessMap
     *
     * @param type
     *            the type of the statement : receive, onMessage
     * @param partnerLink
     *            the partnerLink attribute of this statement
     * @param portType
     *            the portType attribute of this statement
     * @param operation
     *            the operation attribute of this statement
     * @param createInstance
     *            the createInstance attribute of this statement
     */
    private void setReceiveInfo(String type, String partnerLink, String portType, String operation, String createInstance,String name) {
        this.receiveInfo[0] = type;
        this.receiveInfo[1] = partnerLink;
        this.receiveInfo[2] = portType;
        this.receiveInfo[3] = operation;
        this.receiveInfo[4] = createInstance;
        this.receiveInfo[5] = name;
    }

    /**
     * analyse node Reply
     *
     * @param n
     *            the xml node representing this Reply
     */
    private void analyseReply(Node n) throws BPELGeneratorException {
        //System.out.println("BPELCompilo : AnalyseReply");
        String inoutTypeSave = this.inoutType;
        this.inoutType = "Reply";
        int replyNb = ++this.statementNb;
        String partnerLink = null;
        String portType = null;
        String operation = null;
        String variable = null;
        String name = null;
        String joinCondition = null;
        String suppressJoinFailure = null;

        NamedNodeMap atts = n.getAttributes();
        for (int i = 0; i < atts.getLength(); i++) {
            Node att = atts.item(i);
            if (att.getNodeName().equals("partnerLink")) {
                partnerLink = new String(att.getNodeValue());
            } else if (att.getNodeName().equals("portType")) {
                portType = new String(att.getNodeValue());
                //namespaces management
                QName ptQN = Util.buildQname(portType, this.proc);
                portType = Util.buildStringFromQName(ptQN);
            } else if (att.getNodeName().equals("operation")) {
                operation = new String(att.getNodeValue());
            } else if (att.getNodeName().equals("variable")) {
                variable = new String(att.getNodeValue());
            } else if (att.getNodeName().equals("name")) {
                name = new String(att.getNodeValue());
            } else if (att.getNodeName().equals("joinCondition")) {
                joinCondition = new String(att.getNodeValue());
            } else if (att.getNodeName().equals("suppressJoinFailure")) {
                suppressJoinFailure = new String(att.getNodeValue());
            }
        }
        //f.insertLine(" //org.objectweb.orchestra.tree.Reply s" + replyNb + " = new
        // org.objectweb.orchestra.tree.Reply(" + replyNb + ", \"" + partnerLink + "\", \"" +
        // portType + "\", \"" + operation + "\", \"" + variable + "\", \"" +
        // name + "\", \"" + joinCondition + "\", \"" + suppressJoinFailure +
        // "\");");

        String[] tabParam = new String[8];
        tabParam[0] = "new java.lang.Integer(" + replyNb + ")";
        tabParam[1] = "\"" + partnerLink + "\"";
        tabParam[2] = "\"" + portType + "\"";
        tabParam[3] = "\"" + operation + "\"";
        tabParam[4] = "\"" + variable + "\"";
        tabParam[5] = "\"" + name + "\"";
        tabParam[6] = "\"" + joinCondition + "\"";
        tabParam[7] = "\"" + suppressJoinFailure + "\"";
        createInstance("org.objectweb.orchestra.tree.Reply", tabParam);

        analyseChilds(n, "Reply");
        this.inoutType = inoutTypeSave;
        //System.out.println("BPELCompilo : End AnalyseReply");
    }

    /**
     * analyse node Invoke
     *
     * @param n
     *            the xml node representing this Invoke
     */
    private void analyseInvoke(Node n) throws BPELGeneratorException {
        //System.out.println("BPELCompilo : AnalyseInvoke");
        String inoutTypeSave = this.inoutType;
        this.inoutType = "Invoke";
        int invokeNb = ++this.statementNb;
        String partnerLink = null;
        String portType = null;
        String operation = null;
        String inputVariable = null;
        String outputVariable = null;
        String port = null;
        String name = null;
        String joinCondition = null;
        String suppressJoinFailure = null;
        NamedNodeMap atts = n.getAttributes();
        for (int i = 0; i < atts.getLength(); i++) {
            Node att = atts.item(i);
            if (att.getNodeName().equals("partnerLink")) {
                partnerLink = new String(att.getNodeValue());
            } else if (att.getNodeName().equals("portType")) {
                portType = new String(att.getNodeValue());
                //namespaces management
                QName ptQN = Util.buildQname(portType, this.proc);
                portType = Util.buildStringFromQName(ptQN);
            } else if (att.getNodeName().equals("operation")) {
                operation = new String(att.getNodeValue());
            } else if (att.getNodeName().equals("inputVariable")) {
                inputVariable = new String(att.getNodeValue());
            } else if (att.getNodeName().equals("outputVariable")) {
                outputVariable = new String(att.getNodeValue());
            } else if (att.getNodeName().equals("port")) {
                port = new String(att.getNodeValue());
            } else if (att.getNodeName().equals("name")) {
                name = new String(att.getNodeValue());
            } else if (att.getNodeName().equals("joinCondition")) {
                joinCondition = new String(att.getNodeValue());
            } else if (att.getNodeName().equals("suppressJoinFailure")) {
                suppressJoinFailure = new String(att.getNodeValue());
            }
        }
        //f.insertLine(" //org.objectweb.orchestra.tree.Invoke s" + invokeNb + " = new
        // org.objectweb.orchestra.tree.Invoke(" + invokeNb + ", \"" + partnerLink + "\", \"" +
        // portType + "\", \"" + operation + "\", \"" + inputVariable + "\", \""
        // + outputVariable + "\", \"" + port + "\", \"" + name + "\", \"" +
        // joinCondition + "\", \"" + suppressJoinFailure + "\");");

        String[] tabParam = new String[10];
        tabParam[0] = "new java.lang.Integer(" + invokeNb + ")";
        tabParam[1] = "\"" + partnerLink + "\"";
        tabParam[2] = "\"" + portType + "\"";
        tabParam[3] = "\"" + operation + "\"";
        tabParam[4] = "\"" + inputVariable + "\"";
        tabParam[5] = "\"" + outputVariable + "\"";
        tabParam[6] = "\"" + port + "\"";
        tabParam[7] = "\"" + name + "\"";
        tabParam[8] = "\"" + joinCondition + "\"";
        tabParam[9] = "\"" + suppressJoinFailure + "\"";
        createInstance("org.objectweb.orchestra.tree.Invoke", tabParam);

        analyseChilds(n, "Invoke");
        this.inoutType = inoutTypeSave;
        //System.out.println("BPELCompilo : End AnalyseInvoke");
    }

    /**
     * analyse node Correlations
     *
     * @param n
     *            the xml node representing this Correlations
     */
    private void analyseCorrelations(Node n) throws BPELGeneratorException {
        //System.out.println("BPELCompilo : AnalyseCorrelations");
        int correlationsNb = ++this.statementNb;

        //f.insertLine(" //org.objectweb.orchestra.tree.Correlations s" + correlationsNb + " = new
        // org.objectweb.orchestra.tree.Correlations(" + correlationsNb + ");");
        String[] tabParam = new String[1];
        tabParam[0] = "new java.lang.Integer(" + correlationsNb + ")";
        createInstance("org.objectweb.orchestra.tree.Correlations", tabParam);

        analyseChilds(n, "Correlations");

        //System.out.println("BPELCompilo : End AnalyseCorrelations");
    }

    /**
     * analyse node Correlation
     *
     * @param n
     *            the xml node representing this Correlation
     */
    private void analyseCorrelation(Node n) throws BPELGeneratorException {
        //System.out.println("BPELCompilo : AnalyseCorrelation");
        int correlationNb = ++this.statementNb;
        String correlationSet = null;
        String initiate = null;
        String pattern = null;

        NamedNodeMap atts = n.getAttributes();
        for (int i = 0; i < atts.getLength(); i++) {
            Node att = atts.item(i);
            if (att.getNodeName().equals("set")) {
                correlationSet = new String(att.getNodeValue());
            } else if (att.getNodeName().equals("initiate")) {
                initiate = new String(att.getNodeValue());
            } else if (att.getNodeName().equals("pattern")) {
                pattern = new String(att.getNodeValue());
            }
        }
        CorrelationSet cs = findCS(correlationSet);
        if ((this.inoutType.equals("Receive")) || (this.inoutType.equals("OnMessage"))) {
            this.processMapFile.insertLine(this.receiveInfo, cs);
            this.receivePrinted = true;
        }
        //f.insertLine(" //org.objectweb.orchestra.tree.Correlation s" + correlationNb + " = new
        // org.objectweb.orchestra.tree.Correlation(" + correlationNb + ", \"" + correlationSet +
        // "\", \"" + initiate + "\", \"" + pattern + "\");");
        String[] tabParam = new String[4];
        tabParam[0] = "new java.lang.Integer(" + correlationNb + ")";
        tabParam[1] = "\"" + correlationSet + "\"";
        tabParam[2] = "\"" + initiate + "\"";
        tabParam[3] = "\"" + pattern + "\"";

        createInstance("org.objectweb.orchestra.tree.Correlation", tabParam);
        //System.out.println("BPELCompilo : End AnalyseCorrelation");
    }

    /**
     * analyse node Scope
     *
     * @param n
     *            the xml node representing this Scope
     */
    private void analyseScope(Node n) throws BPELGeneratorException {
        //System.out.println("BPELCompilo : AnalyseScope");
        int scopeNb = ++this.statementNb;
        String name = null;
        String joinCondition = null;
        String suppressJoinFailure = null;
        int scopeIdSave = this.scopeId++;

        NamedNodeMap atts = n.getAttributes();
        for (int i = 0; i < atts.getLength(); i++) {
            Node att = atts.item(i);
            if (att.getNodeName().equals("name")) {
                name = new String(att.getNodeValue());
            } else if (att.getNodeName().equals("joinCondition")) {
                joinCondition = new String(att.getNodeValue());
            } else if (att.getNodeName().equals("suppressJoinFailure")) {
                suppressJoinFailure = new String(att.getNodeValue());
            }
        }

        //f.insertLine(" //org.objectweb.orchestra.tree.Scope s" + scopeNb + " = new
        // org.objectweb.orchestra.tree.Scope(" + scopeNb + ", \"" + name+ "\", \"" + joinCondition
        // + "\", \"" + suppressJoinFailure + "\");");
        String[] tabParam = new String[4];
        tabParam[0] = "new java.lang.Integer(" + scopeNb + ")";
        tabParam[1] = "\"" + name + "\"";
        tabParam[2] = "\"" + joinCondition + "\"";
        tabParam[3] = "\"" + suppressJoinFailure + "\"";

        createInstance("org.objectweb.orchestra.tree.Scope", tabParam);

        analyseChilds(n, "Scope");
        removeCS(this.scopeId);
        this.scopeId = scopeIdSave;
        //System.out.println("BPELCompilo : End AnalyseScope");
    }

    /**
     * analyse node Compensate
     *
     * @param n
     *            the xml node representing this Compensate
     */
    private void analyseCompensate(Node n) throws BPELGeneratorException {
        //System.out.println("BPELCompilo : AnalyseCompensate");
        int compensateNb = ++this.statementNb;
        String scopeName = null;
        String name = null;
        String joinCondition = null;
        String suppressJoinFailure = null;

        NamedNodeMap atts = n.getAttributes();
        for (int i = 0; i < atts.getLength(); i++) {
            Node att = atts.item(i);
            if (att.getNodeName().equals("scope")) {
                scopeName = new String(att.getNodeValue());
            } else if (att.getNodeName().equals("name")) {
                name = new String(att.getNodeValue());
            } else if (att.getNodeName().equals("joinCondition")) {
                joinCondition = new String(att.getNodeValue());
            } else if (att.getNodeName().equals("suppressJoinFailure")) {
                suppressJoinFailure = new String(att.getNodeValue());
            }
        }

        //f.insertLine(" //org.objectweb.orchestra.tree.Compensate s" + compensateNb + " = new
        // org.objectweb.orchestra.tree.Compensate(" + compensateNb + ", \"" + scopeName + ", \"" +
        // name+ "\", \"" + joinCondition + "\", \"" + suppressJoinFailure +
        // "\");");
        String[] tabParam = new String[5];
        tabParam[0] = "new java.lang.Integer(" + compensateNb + ")";
        tabParam[1] = "\"" + scopeName + "\"";
        tabParam[2] = "\"" + name + "\"";
        tabParam[3] = "\"" + joinCondition + "\"";
        tabParam[4] = "\"" + suppressJoinFailure + "\"";

        createInstance("org.objectweb.orchestra.tree.Compensate", tabParam);

        analyseChilds(n, "Compensate");

        //System.out.println("BPELCompilo : End AnalyseCompensate");
    }

    /**
     * analyse node Variables
     *
     * @param n
     *            the xml node representing this Variables
     */
    private void analyseVariables(Node n) throws BPELGeneratorException {
        //System.out.println("BPELCompilo : AnalyseVariables");
        int variablesNb = ++this.statementNb;

        //f.insertLine(" //org.objectweb.orchestra.tree.Variables s" + variablesNb + " = new
        // org.objectweb.orchestra.tree.Variables(" + variablesNb + ");");
        String[] tabParam = new String[1];
        tabParam[0] = "new java.lang.Integer(" + variablesNb + ")";

        createInstance("org.objectweb.orchestra.tree.Variables", tabParam);

        analyseChilds(n, "Variables");
        //System.out.println("BPELCompilo : End AnalyseVariables");
    }

    /**
     * analyse node Variable
     *
     * @param n
     *            the xml node representing this Variable
     */
    private void analyseVariable(Node n) throws BPELGeneratorException {
        //System.out.println("BPELCompilo : AnalyseVariable");
        int variableNb = ++this.statementNb;
        String variableName = null;
        String messageType = null;
        String type = null;
        String element = null;
        String BPELJType = null;

        NamedNodeMap atts = n.getAttributes();
        for (int i = 0; i < atts.getLength(); i++) {
            Node att = atts.item(i);
            String nodeName = att.getNodeName();
            if (nodeName.equals("name")) {
                variableName = new String(att.getNodeValue());
                //System.out.println("BPELCompilo : Variable Name : " +
                // variableName);
            } else if (nodeName.equals("messageType")) {
                messageType = new String(att.getNodeValue());
                QName mtQN = Util.buildQname(messageType, this.proc);
                messageType = Util.buildStringFromQName(mtQN);
                if (messageType == null) {
                    throw new BPELGeneratorException("Undefined WSDL messageType found into <variable> definition."
                            + " Its name is '" + mtQN + "'");
                }
            } else if (nodeName.equals("type")) {
                String nodeValue = new String(att.getNodeValue());
                if (hasPrefixOfBPELJNS(nodeValue)) {
                    BPELJType = BPELString.removePrefix(nodeValue);
                    if (WSDLJavaMatching.isJavaPrimitiveType(BPELJType)) {
                        throw new BPELGeneratorException("Java primitive types are not authorized as BPELJ variable types.\nUse XML Schema simple types instead");
                    }
                    Class variableType = null;
                    try {
                        variableType = Class.forName(BPELJType);
                    } catch (ClassNotFoundException e) {
                        throw new BPELGeneratorException("The class " + BPELJType + " cannot be found.", e);
                    }
                    if (!java.io.Serializable.class.isAssignableFrom(variableType)) {
                        throw new BPELGeneratorException("The class " + BPELJType + " must implement the java.io.Serializable interface.");
                    }
                } else {
                    QName tQN = Util.buildQname(nodeValue, this.proc);
                    //type = Util.buildStringFromQName(tQN);
                    type = Util.getJavaType(this.symbolTable, tQN, null);
                    if (type == null) {
                        throw new BPELGeneratorException("Undefined XML Schema type found into <variable> definition."
                                + " Its name is '" + tQN + "'");
                    }
                }
            } else if (nodeName.equals("element")) {
                element = new String(att.getNodeValue());
                QName elQN = Util.buildQname(element, this.proc);
                //element = Util.buildStringFromQName(elQN);
                element = Util.getJavaType(this.symbolTable, null, elQN);
                if (element == null) {
                    throw new BPELGeneratorException("Undefined XML Schema element found into <variable> definition."
                            + " Its name is '" + elQN + "'");
                }
            }
            // We don't need to look at the xmlBinding attribute
            // because getVariableType() already does it
        }

        //f.insertLine(" //org.objectweb.orchestra.tree.Variable s" + variableNb + " = new
        // org.objectweb.orchestra.tree.Variable(" + variableNb + ", \"" + variableName + "\", \""
        // + messageType + "\", \"" + type + "\", \"" + element + "\");");
        String[] tabParam = {
                "new java.lang.Integer(" + variableNb + ")",
                Util.toQuotedString(variableName),
                Util.toQuotedString(messageType),
                Util.toQuotedString(type),
                Util.toQuotedString(element),
                Util.toQuotedString(BPELJType)
        };
        createInstance("org.objectweb.orchestra.tree.Variable", tabParam);

        analyseChilds(n, "Variable");
        //System.out.println("BPELCompilo : End AnalyseVariable");
    }

    /**
     * analyse node CorrelationSets
     *
     * @param n
     *            the xml node representing this CorrelationSets
     */
    private void analyseCorrelationSets(Node n) throws BPELGeneratorException {
        //System.out.println("BPELCompilo : analyseCorrelationSets");
        int correlationSetsNb = ++this.statementNb;

        //f.insertLine(" //org.objectweb.orchestra.tree.CorrelationSets s" + correlationSetsNb + "
        // = new org.objectweb.orchestra.tree.CorrelationSets(" + correlationSetsNb + ");");
        String[] tabParam = new String[1];
        tabParam[0] = "new java.lang.Integer(" + correlationSetsNb + ")";

        createInstance("org.objectweb.orchestra.tree.CorrelationSets", tabParam);

        analyseChilds(n, "CorrelationSets");
        //System.out.println("BPELCompilo : End analyseCorrelationSets");
    }

    /**
     * analyse node CorrelationSet
     *
     * @param n
     *            the xml node representing this CorrelationSet
     */
    private void analyseCorrelationSet(Node n) throws BPELGeneratorException {
        //System.out.println("BPELCompilo : AnalyseCS");
        int correlationSetNb = ++this.statementNb;
        String CSName = null;
        String CSProperties = null;

        NamedNodeMap atts = n.getAttributes();
        for (int i = 0; i < atts.getLength(); i++) {
            Node att = atts.item(i);
            if (att.getNodeName().equals("name")) {
                CSName = new String(att.getNodeValue());
                //System.out.println("BPELCompilo : CS Name : " + CSName);
            } else if (att.getNodeName().equals("properties")) {
                CSProperties = new String(att.getNodeValue());
                //System.out.println("BPELCompilo : CS Properties : " +
                // CSProperties);
            }
        }
        String[] props = CSProperties.split(" ");
        CSProperties = "";
        //namespaces management
        for (int i = 0; i < props.length; i++) {
            QName propQN = Util.buildQname(props[i], this.proc);
            String prop = Util.buildStringFromQName(propQN);
            if (i != 0) {
                CSProperties += " ";
            }
            CSProperties += prop;
        }
        //f.insertLine(" //org.objectweb.orchestra.tree.CorrelationSet s" + correlationSetNb + " =
        // new org.objectweb.orchestra.tree.CorrelationSet(" + correlationSetNb + ", \"" + CSName +
        // "\", \"" + CSProperties + "\");");
        String[] tabParam = new String[3];
        tabParam[0] = "new java.lang.Integer(" + correlationSetNb + ")";
        tabParam[1] = "\"" + CSName + "\"";
        tabParam[2] = "\"" + CSProperties + "\"";
        createInstance("org.objectweb.orchestra.tree.CorrelationSet", tabParam);

        CorrelationSet cs = new CorrelationSet(this.scopeId, CSName, CSProperties);
        this.CS.add(cs);
        //System.out.println("BPELCompilo : End AnalyseCS");
    }

    /**
     * analyse node BPELJSnippet
     *
     * @param n the xml node representing this BPELJSnippet
     */
    private void analyseBPELJSnippet(Node n) throws BPELGeneratorException {
        int stmtNb = ++this.statementNb;
        String name = null;
        String joinCondition = null;
        String suppressJoinFailure = null;
        NamedNodeMap atts = n.getAttributes();
        for (int i = 0; i < atts.getLength(); i++) {
            Node att = atts.item(i);
            if (att.getNodeName().equals("name")) {
                name = new String(att.getNodeValue());
            } else if (att.getNodeName().equals("joinCondition")) {
                joinCondition = new String(att.getNodeValue());
            } else if (att.getNodeName().equals("suppressJoinFailure")) {
                suppressJoinFailure = new String(att.getNodeValue());
            }
        }

        String[] tabParam = {
                "new java.lang.Integer(" + stmtNb + ")",
                Util.toQuotedString(name),
                Util.toQuotedString(joinCondition),
                Util.toQuotedString(suppressJoinFailure)
        };
        createInstance("org.objectweb.orchestra.tree.BPELJSnippet", tabParam);

        analyseChilds(n, "BPELJSnippet");
    }

    /**
     * analyse node BPELJ code
     * @param n the xml node representing this BPELJCode
     * @throws BPELGeneratorException If the node does not contain exactly one child or
     * if this child is different from a text node
     */
    private void analyseBPELJCode(Node n) throws BPELGeneratorException {
        //System.out.println("BPELCompilo : AnalyseBPELJCode");
        int stmtNb = ++this.statementNb;
        NodeList nl = n.getChildNodes();
        Node javaCodeNode = null;
        if (nl.getLength() > 0) {
            javaCodeNode = nl.item(0);
        }
        if (javaCodeNode != null) {
            Map accessVars = new HashMap();
            Map accessPLs = new HashMap();
            getAccessibleVars(n, accessVars);
            getAccessiblePLs(n, accessPLs);

            String javaCode = javaCodeNode.getNodeValue();
            String methodName = this.BPELJFile.insertBPELJCode(stmtNb, javaCode, accessVars, accessPLs);

            String[] tabParam = {
                    // makes the javaCode to be a quoted string. To do this, it replaces the newline
                    // characters ('\n') with '\', and 'n' chars and the double quote characters
                    // with '\' and '"' chars
                    Util.toQuotedString(javaCode.replaceAll("\n", "\\\\n").replaceAll("\"", "\\\\\"")),
                    Util.toQuotedString(this.processName),
                    Util.toQuotedString(this.proc.getBPELJPackage() + ".BPELJ" + this.processName),
                    Util.toQuotedString(methodName)
            };
            createInstance("org.objectweb.orchestra.tree.JavaSnippet", tabParam);
            createStatementInstance("org.objectweb.orchestra.tree.BPELJCode", stmtNb, "org.objectweb.orchestra.tree.JavaSnippet");
        }
        //System.out.println("BPELCompilo : End AnalyseBPELJCode");
    }

    /**
     * analyse node BPELJ value
     * @param n the xml node representing this BPELJValue
     * @throws BPELGeneratorException
     */
    private void analyseBPELJValue(Node n) throws BPELGeneratorException {
        //System.out.println("BPELCompilo : AnalyseBPELJValue");
        int stmtNb = ++this.statementNb;
        NodeList nl = n.getChildNodes();
        Node javaCodeNode = null;
        if (nl.getLength() > 0) {
            javaCodeNode = nl.item(0);
        }
        if (javaCodeNode != null) {
            boolean isBlock = false;
            NamedNodeMap nnm = n.getAttributes();
            for (int i=0; i<nnm.getLength(); ++i) {
                if (nnm.item(i).getNodeName().equals("style") && nnm.item(i).getNodeValue().equals("block")) {
                    isBlock = true;
                }
            }

            String returnType = getVariableType(n);

            Node startNode;
            Map accessVars = new HashMap();
            Map accessPLs = new HashMap();
            if (n.getPreviousSibling() != null) {
                startNode = n.getPreviousSibling();
            } else {
                startNode = n.getParentNode();
            }
            getAccessibleVars(startNode, accessVars);
            getAccessiblePLs(startNode, accessPLs);

            String javaCode = javaCodeNode.getNodeValue();
            String methodName = this.BPELJFile.insertBPELJValue(stmtNb, javaCode, returnType, isBlock, accessVars, accessPLs);

            String[] tabParam = {
                    // makes the javaCode to be a quoted string. To do this, it replaces the newline
                    // characters ('\n') with '\', and 'n' chars and the double quote characters
                    // with '\' and '"' chars
                    Util.toQuotedString(javaCode.replaceAll("\n", "\\\\n").replaceAll("\"", "\\\\\"")),
                    Util.toQuotedString(this.processName),
                    Util.toQuotedString(this.proc.getBPELJPackage() + ".BPELJ" + this.processName),
                    Util.toQuotedString(methodName)
            };
            createInstance("org.objectweb.orchestra.tree.JavaSnippet", tabParam);
            createStatementInstance("org.objectweb.orchestra.tree.BPELJValue", stmtNb, "org.objectweb.orchestra.tree.JavaSnippet", isBlock);
        }
        //System.out.println("BPELCompilo : End AnalyseBPELJValue");
    }

    /**
     * Returns the Java type associated to a BPEL variable node
     * @param n the <code><bpelj:value></code> node from which to start the search
     * @return String the Java type associated to the BPEL variable
     * @throws BPELGeneratorException if no xmlBinding is specified or the v
     */
    private final String getVariableType(Node n) throws BPELGeneratorException {
        Node variableNode = n.getParentNode();
        if (variableNode.getNodeName().equals("variable")) {
            String variableName = null;
            String messageType = null;
            String type = null;
            String element = null;
            String BPELJType = null;
            String xmlBinding = null;

            NamedNodeMap atts = variableNode.getAttributes();
            for (int i = 0; i < atts.getLength(); i++) {
                Node att = atts.item(i);
                String nodeName = att.getNodeName();
                if (nodeName.equals("name")) {
                    variableName = new String(att.getNodeValue());
                    // We don't need to verify that the name is not a Java keyword,
                    // because getAccessibleVar() already does it
                } else if (nodeName.equals("messageType")) {
                    messageType = new String(att.getNodeValue());
                    QName mtQN = Util.buildQname(messageType, this.proc);
                    String packag = Util.getPackageFromNamespace(mtQN.getNamespaceURI());
                    String messageClass = BPELConstants.BM + this.processName + mtQN.getLocalPart();
                    messageType = packag + "." + messageClass;
                } else if (nodeName.equals("type")) {
                    String nodeValue = new String(att.getNodeValue());
                    if (hasPrefixOfBPELJNS(nodeValue)) {
                        BPELJType = BPELString.removePrefix(nodeValue);
                        // We don't need to verify that the type is java.io.Serializable
                        // because analyseVariable() already does it
                    } else {
                        QName tQN = Util.buildQname(nodeValue, this.proc);
                        //type = Util.buildStringFromQName(tQN);
                        type = Util.getJavaType(this.symbolTable, tQN, null);
                    }
                } else if (nodeName.equals("element")) {
                    element = new String(att.getNodeValue());
                    QName elQN = Util.buildQname(element, this.proc);
                    element = Util.getJavaType(this.symbolTable, null, elQN);
                } else if (hasPrefixOfBPELJNS(nodeName)
                        && (BPELString.removePrefix(nodeName).equals("xmlBinding")) ) {
                    xmlBinding = new String(att.getNodeValue());
                }
            }

            checkVariableXMLBinding(variableName, xmlBinding, BPELJType!=null);

            String returnType = null;
            if (messageType != null) {
                returnType = messageType;
            } else if (type != null) {
                returnType = type;
            } else if (element != null) {
                returnType = element;
            } else if (BPELJType != null) {
                returnType = BPELJType;
            }
            return returnType;
        }
        throw new BPELGeneratorException("BPELCompilo.getVariableType: the node "
                + n.getNodeName() + " is not the child of a <variable> node.");
    }

    /**
     * Gives the BPEL variables accessible from the specified node.
     * These variables are put into the Map parameter.
     * @param n the node from which to search accessible variables
     * @param foundVariables the accessible variables found under the form of tree.Variable objects
     * @throws BPELGeneratorException if any variable has a name which is a Java keyword,
     * or if no XML binding is specified for a variable
     * or if the XML binding specified for a variable is not supported by Orchestra
     */
    private final void getAccessibleVars(Node n, Map foundVariables) throws BPELGeneratorException {
        if (n == null) {
            return;
        }

        Node tmp = n;
        while (tmp.getPreviousSibling() != null) {
            tmp = tmp.getPreviousSibling();
            if (tmp.getNodeType() == Node.ELEMENT_NODE) {
                if (tmp.getNodeName().equals("variables")) {
                    tmp = tmp.getLastChild();
                } else if (tmp.getNodeName().equals("variable")) {
                    getAccessibleVar(tmp, foundVariables);
                }
            }
        }

        getAccessibleVars(n.getParentNode(), foundVariables);
    }

    /**
     * Analyses the BPEL variable node. If there is no variable with the same
     * name into the Map, then the variable is added to the Map under the form
     * of a tree.Variable object.
     * @param variableNode the variable node to analyse
     * @param foundVariables the accessible variables found
     * @throws BPELGeneratorException if any variable has a name which is a Java keyword,
     * or if no XML binding is specified for a variable
     * or if the XML binding specified for a variable is not supported by Orchestra
     */
    private final void getAccessibleVar(Node variableNode, Map foundVariables) throws BPELGeneratorException {
        String variableName = null;
        String messageType = null;
        String type = null;
        String element = null;
        String BPELJType = null;
        String xmlBinding = null;

        NamedNodeMap atts = variableNode.getAttributes();
        for (int i = 0; i < atts.getLength(); i++) {
            Node att = atts.item(i);
            String nodeName = att.getNodeName();
            if (nodeName.equals("name")) {
                variableName = new String(att.getNodeValue());
                // We only keep the variable of lesser scope
                if (foundVariables.get(variableName) != null) {
                    variableName = null;
                    break;
                }
                if (WSDLJavaMatching.isJavaKeyword(variableName)) {
                    throw new BPELGeneratorException("BPELCompilo.getAccessibleVar : the variable "
                            + variableName + " cannot be used inside a Java snippet: its name is a Java keyword.");
                }
            } else if (nodeName.equals("messageType")) {
                messageType = new String(att.getNodeValue());
                QName mtQN = Util.buildQname(messageType, this.proc);
                String packag = Util.getPackageFromNamespace(mtQN.getNamespaceURI());
                String messageClass = BPELConstants.BM + this.processName + mtQN.getLocalPart();
                messageType = packag + "." + messageClass;
            } else if (nodeName.equals("type")) {
                String nodeValue = new String(att.getNodeValue());
                if (hasPrefixOfBPELJNS(nodeValue)) {
                    BPELJType = BPELString.removePrefix(nodeValue);
                    // We don't need to verify that the type is java.io.Serializable
                    // because analyseVariable() has already done it
                } else {
                    QName tQN = Util.buildQname(nodeValue, this.proc);
                    type = Util.getJavaType(this.symbolTable, tQN, null);
                }
            } else if (nodeName.equals("element")) {
                element = new String(att.getNodeValue());
                QName elQN = Util.buildQname(element, this.proc);
                element = Util.getJavaType(this.symbolTable, null, elQN);
            } else if (hasPrefixOfBPELJNS(nodeName)
                    && (BPELString.removePrefix(nodeName).equals("xmlBinding")) ) {
                xmlBinding = new String(att.getNodeValue());
            }
        }

        checkVariableXMLBinding(variableName, xmlBinding, BPELJType!=null);

        if ( (variableName != null) && (foundVariables.get(variableName) == null) ) {
            // Here, we do not care about which XML to Java binding is used
            Variable v = new Variable(new Integer(0), variableName, messageType, type, element, BPELJType);
            foundVariables.put(variableName, v);
        }
    }

    /**
     * Gives the BPEL partner links accessible from the specified node.
     * These partner links are put into the Map parameter.
     * @param partnerLinkNode the partner link node to analyse
     * @param foundPLs the accessible partner links found
     * @throws BPELGeneratorException  if no XML binding is specified for the process
     * or if the XML binding specified is not supported by Orchestra
     */
    private final void getAccessiblePLs(Node n, Map foundPLs)  throws BPELGeneratorException {
        if (n == null) {
            return;
        }

        Node tmp = n;
        while (tmp.getPreviousSibling() != null) {
            tmp = tmp.getPreviousSibling();
            if (tmp.getNodeType() == Node.ELEMENT_NODE) {
                if (tmp.getNodeName().equals("partnerLinks")) {
                    tmp = tmp.getLastChild();
                } else if (tmp.getNodeName().equals("partnerLink")) {
                    getAccessiblePL(tmp, foundPLs);
                }
            }
        }

        getAccessiblePLs(n.getParentNode(), foundPLs);
    }

    /**
     * Analyses the BPEL partner link node. If there is no partner link with the
     * same name into the Map, then the variable is added to the Map under the
     * form a data.PartnerLinkData object.
     * @param partnerLinkNode the partner link node to analyse
     * @param foundPLs the accessible partner links found
     * @throws BPELGeneratorException  if no XML binding is specified for the process
     * or if the XML binding specified is not supported by Orchestra
     */
    private final void getAccessiblePL(Node partnerLinkNode, Map foundPLs) throws BPELGeneratorException {
        String plName = null;
        String plType = null;
        QName pltQN = null;
        String myRoleName = null;
        String partnerRoleName = null;

        NamedNodeMap atts = partnerLinkNode.getAttributes();
        for (int i = 0; i < atts.getLength(); i++) {
            Node att = atts.item(i);
            if (att.getNodeName().equals("name")) {
                plName = "p_" + att.getNodeValue();
                // We only keep the PL of lesser scope
                if (foundPLs.get(plName) != null) {
                    plName = null;
                    break;
                }
            } else if (att.getNodeName().equals("partnerLinkType")) {
                plType = att.getNodeValue();
                pltQN = Util.buildQname(plType, this.wsdl);
                plType = Util.buildStringFromQName(pltQN);
            } else if (att.getNodeName().equals("myRole")) {
                myRoleName = att.getNodeValue();
            } else if (att.getNodeName().equals("partnerRole")) {
                partnerRoleName = att.getNodeValue();
            }
        }

        if ( (pltQN != null) && (foundPLs.get(pltQN) == null) ) {
            // Test whether the XML binding used is good
            String xmlBinding = this.proc.getXMLBinding();
            if (xmlBinding == null) {
                throw new BPELGeneratorException("The XML to Java binding of this process is unspecified."
                        + " Please add the attribute 'bpelj:xmlBinding=\"JAXRPC11\"' to the process element.");
            } else if (!BPELConstants.BPELJ_XML_BINDING.equals(xmlBinding)) {
                throw new BPELGeneratorException("The XML to Java binding named "
                        + Util.toQuotedString(xmlBinding)
                        + " is not supported by Orchestra."
                        + " Please consider using the \"JAXRPC11\" XML to Java binding.");
            }

            PartnerLinkType plt = (PartnerLinkType) this.plTypes.get(pltQN);
            Role myRole = plt.getRole(myRoleName);
            Role partnerRole = plt.getRole(partnerRoleName);
            PartnerLinkData pld = new PartnerLinkData(0,
                    plName,
                    this.processName,
                    pltQN,
                    (myRole == null)? null : myRole.getName(),
                            (partnerRole == null)? null : partnerRole.getName(),
                                    (myRole == null)? null : myRole.getPortType(),
                                            (partnerRole == null)? null : partnerRole.getPortType(),
                                                    null,
                                                    null,
                                                    0
            );
            foundPLs.put(pltQN, pld);
        }
    }

    /**
     * Checks whether the given variable has its xmlBinding specified.
     * It only works if this is a BPELJ process and if the variable is not a Java variable.
     * @param variableName The name of this BPEL variable
     * @param xmlBindingAtt The xmlBinding attribute of this BPEL variable
     * @param isJavaVariable Whether this is a Java variable or not
     * @throws BPELGeneratorException
     */
    private final void checkVariableXMLBinding(String variableName, String xmlBindingAtt, boolean isJavaVariable) throws BPELGeneratorException {
        if (this.proc.isBPELJProcess() && !isJavaVariable) {
            // If no XML binding is specified, we retrieve the process XML binding
            if (xmlBindingAtt == null) {
                xmlBindingAtt = this.proc.getXMLBinding();
            }
            // There is not any XML binding specified
            if (xmlBindingAtt == null) {
                throw new BPELGeneratorException("BPELCompilo.analyseVariable : the variable "
                        + Util.toQuotedString(variableName)
                        + " must specify its XML to Java binding since it is used inside a BPELJ Snippet.");
            } else if (!BPELConstants.BPELJ_XML_BINDING.equals(xmlBindingAtt)) {
                throw new BPELGeneratorException("BPELCompilo.analyseVariable : the XML to Java binding named "
                        + Util.toQuotedString(xmlBindingAtt)
                        + " is not supported by Orchestra for the variable "
                        + Util.toQuotedString(variableName)
                        + ".\nPlease consider using the \"JAXRPC11\" XML to Java binding.");
            }
        }
    }

    /**
     * analyseChilds recursively calls analyseNode for each child of n
     *
     * @param n
     *            the node to analyse children
     * @param type
     *            the type of node n
     */
    private void analyseChilds(Node n, String type) throws BPELGeneratorException {
        int fatherNb = this.statementNb;

        Vector children = getElementChildren(n);

        int childNb = children.size();
        int[] childStatementNbs = new int[childNb];

        String[] temp = new String[childNb];
        String createInstanceSave = this.mCreateInstance;
        /*
        System.out.println("BPELCompilo, analyseChilds, n = "+n+", type = "+type);
        System.out.println("BPELCompilo, analyseChilds, n.getNodeName() = "+n.getNodeName());
        System.out.println("BPELCompilo, analyseChilds, n.getChildNodes().getLength() = "+n.getChildNodes().getLength());
        System.out.println("BPELCompilo, analyseChilds, childNb = " + childNb);
         */

        //f.insertLine(" //s" + fatherNb + ".setChildNb(" + childNb + ");");
        intCall("org.objectweb.orchestra.tree." + type, "setChildNb", "s" + fatherNb, childNb);

        //for (Node child = n.getFirstChild(); child != null; child = child.getNextSibling()) {
        for (int i=0;i<children.size();i++) {
            Node child = (Node) children.get(i);
            int nb = analyseNode(child);
            this.mCreateInstance = createInstanceSave;
            if (!this.text_node) {
                //System.out.println("BPELCompilo, analyseChilds, in if (!text_node), nb = " + nb);
                //System.out.println("BPELCompilo, analyseChilds, in if (!text_node), i = " + i);

                temp[i] = this.node_type;
                childStatementNbs[i] = nb;

                //f.insertLine(" //s" + fatherNb + ".addChild(s" + nb + ");");
                statementCall("addChild", "s" + fatherNb, "s" + nb);

                //f.insertLine(" //s" + nb + ".setFather(s" + fatherNb + ");");
                statementCall("setFather", "s" + nb, "s" + fatherNb);
            }
        }
        //childNb = i;

        for (int cmpt = childNb - 1; cmpt > 0; cmpt--) {
            //System.out.println("BPELCompilo : s" + childStatementNbs[cmpt-1]
            // + ".setNext(s" + childStatementNbs[cmpt] + ");");

            //f.insertLine(" //s" + childStatementNbs[cmpt-1] + ".setNext(s" +
            // childStatementNbs[cmpt] + ");");
            statementCall("setNext", "s" + childStatementNbs[cmpt - 1], "s" + childStatementNbs[cmpt]);

        }
    }

    /**
     * remove a correlationSet based on its scopeId
     *
     * @param sId
     *            the scopeId
     */
    private void removeCS(int sId) {
        for (int i = 0; i < this.CS.size(); i++) {
            CorrelationSet cs = (CorrelationSet) this.CS.get(i);
            if (cs.getScopeId() == sId) this.CS.remove(cs);
        }
    }

    /**
     * search a correlationSet based on its name
     *
     * @param csName
     *            the CorrelationSet name
     * @return CorrelationSet the CorrelationSet found or null if not found
     */
    private CorrelationSet findCS(String csName) {
        int i = 0;
        while (i < this.CS.size()) {
            CorrelationSet cs = (CorrelationSet) this.CS.get(i);
            if (csName.equals(cs.getName())) return cs;
            i++;
        }
        return null;
    }

    private Vector getElementChildren(Node n) {
        Vector result = new Vector();
        for (Node child = n.getFirstChild(); child != null; child = child.getNextSibling()) {
            if (child.getNodeType() == Node.ELEMENT_NODE) {
                result.addElement(child);
            }
        }
        return result;
    }

    /**
     * Returns true if the prefix of the qname parameter (given as a String)
     * is a BPELJ namespace prefix, false otherwise
     * @param qname the qname which prefix is tested
     * @param process the BPEL process in which the prefix of the qname is tested
     * @return boolean true if the prefix of the qname parameter (given as a String)
     * is a BPELJ namespace prefix, false otherwise
     */
    private boolean hasPrefixOfBPELJNS(String qname) {
        return this.proc.isBPELJProcess()
        && BPELConstants.NS_BPELJ.equals(this.proc.getNS(BPELString.getPrefix(qname)));
    }

    private String getWellFormedXPathExpr(String expr) {
        char[] tab = new char[]{'\\','\\','"'};
        String st = String.copyValueOf(tab);
        return new String(expr.replaceAll("\"", st));
    }
}