/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.objectweb.orchestra.generator;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.PrintWriter;
import java.util.Map;

import javax.xml.namespace.QName;

import org.objectweb.orchestra.data.PartnerLinkData;
import org.objectweb.orchestra.exception.BPELGeneratorException;
import org.objectweb.orchestra.tree.Variable;
import org.objectweb.orchestra.util.BPELConstants;
import org.objectweb.orchestra.util.BPELString;
import org.objectweb.orchestra.util.Util;
import org.objectweb.orchestra.util.WSDLJavaMatching;



/**
 * BPELJProcessFile.java, generate the java representation of the bpelj file
 *
 * @author Jean-Noel Rouvignac
 * @version $Id: BPELJProcessFile.java 94 2006-12-21 10:12:50Z souillac $
 */
public class BPELJProcessFile {

    private static String myPath = BPELConstants.GENERATOR_DIR;
    private PrintWriter fileWriter = null;
    private String fileName = null;
    private String directoryName = null;
    private String processName = null;
    private String filePath = null;
    private Process proc = null;
    private String javaPackage = null;

    /**
     * Constructor
     * @param processName the process name
     */
    public BPELJProcessFile(String processName, Process p) throws BPELGeneratorException {

        this.directoryName = processName;
        this.processName = processName;
        this.fileName = "BPELJ" + this.processName;
        this.proc = p;

        try {
            this.javaPackage = this.proc.getBPELJPackage();

            if (this.javaPackage != null) {
                //create the file
                String dir = Util.getDirectoryFromPackage(this.javaPackage);
                this.filePath = myPath + File.separator
                + this.directoryName + File.separator
                + dir + File.separator
                + this.fileName + ".java";

                new File(this.filePath).getAbsoluteFile().getParentFile().mkdirs();
                this.fileWriter = new PrintWriter(new BufferedWriter(new FileWriter(this.filePath)));

                this.fileWriter.println(Util.autograph(this.fileName + ".java"));
                if (!this.javaPackage.matches("\\s*")) {
                    this.fileWriter.println("package " + this.javaPackage + ";");
                    this.fileWriter.println();
                }
                this.fileWriter.println("import java.lang.*;");
                this.fileWriter.println("import java.util.*;");
                this.fileWriter.println("import javax.xml.namespace.QName;");
                this.fileWriter.println("import org.objectweb.orchestra.util.ClassLoaderMap;");
                this.fileWriter.println();
                this.fileWriter.println("public class " + this.fileName + " {");
                this.fileWriter.println();
            }

        } catch (Exception e) {
            throw new BPELGeneratorException("BPELJProcessFile.init : ",e);
        }

    }

    /**
     * creates a new static void method with the java code associated to a <code>bpelj:code</code>
     * in the BPELJ process file
     *
     * @param stmtNb the statement number of this BPELJCode
     * @param javaCode the java code to insert in the static void method (only statements, no return)
     * @param variables Information over the variables accessible from this snippet
     * @param partnerLinks Information over the partnerLinks accessible from this snippet
     * @return the method name corresponding to this snippet
     */
    public String insertBPELJCode(int stmtNb, String javaCode,
            Map variables, Map partnerLinks) {
        String methodName = insertJavaSnippet(stmtNb, javaCode, "void", true,
                variables, partnerLinks);
        return methodName;
    }

    /**
     * creates a new static method with the java code associated to a <code>bpelj:value</code>
     * in the BPELJ process file. The return type is specified.
     *
     * @param stmtNb the statement number of this BPELJCode
     * @param javaCode the java code to insert in the static method
     * @param returnType the return type of this snippet
     * @param isBlock Whether or not this statement is a "block" statement
     * @param variables Information over the variables accessible from this snippet
     * @param partnerLinks Information over the partnerLinks accessible from this snippet
     * @return the method name corresponding to this snippet
     */
    public String insertBPELJValue(int stmtNb, String javaCode, String returnType, boolean isBlock,
            Map variables, Map partnerLinks) {
        String methodName = insertJavaSnippet(stmtNb, javaCode, returnType, isBlock,
                variables, partnerLinks);
        return methodName;
    }

    /**
     * Inserts a Java snippet method into the BPELJ process file.
     * @param stmtNb the statement number of this BPELJCode
     * @param javaCode the Java code  of the Java snippet
     * @param returnType the return type of this snippet
     * @param isBlock Whether or not this statement is a "block" statement
     * @param variables Information over the variables accessible from this snippet
     * @param partnerLinks Information over the partnerLinks accessible from this snippet
     * @return the method name corresponding to this snippet
     * @see "Snippet section of BPELJ specification (page 10)"
     */
    private String insertJavaSnippet(int stmtNb, String javaCode, String returnType, boolean isBlock,
            Map variables, Map partnerLinks) {
        String methodName = "snippet" + stmtNb;
        String className = "Snippet" + stmtNb + "Class";

        // init variable names and variable types arrays
        Variable[] vars = (Variable[]) variables.values().toArray(new Variable[variables.values().size()]);
        PartnerLinkData[] plds = (PartnerLinkData[]) partnerLinks.values().toArray(new PartnerLinkData[partnerLinks.values().size()]);

        this.fileWriter.println("  public static " + returnType + " " + methodName + "(");
        this.fileWriter.println("           org.objectweb.orchestra.engine.EngineLocal engine,");
        this.fileWriter.println("           int bpiId, int scopeId,");
        this.fileWriter.println("           org.objectweb.orchestra.engine.PLType plType,");
        this.fileWriter.println("           org.objectweb.orchestra.engine.DbUpdater dbu) throws Exception {");

        // Creates a class for this snippet
        this.fileWriter.println("    class " + className + " {");
        insertClassDeclarations(vars, plds);
        insertClassConstructor(className, vars, plds);
        insertExecMethod(javaCode, returnType, isBlock);
        insertSaveMethod(vars);
        this.fileWriter.println("    }"); // end of class
        this.fileWriter.println();

        // Creates an object of this class and calls methods on it
        this.fileWriter.println("    " + className + " snippet = new " + className + "(engine, bpiId, scopeId, plType);");
        boolean hasResult = !("void".equals(returnType));
        if (hasResult) {
            this.fileWriter.print("    " + returnType + " result = ");
        }
        this.fileWriter.println("    snippet.__e_x_e_c__();");
        this.fileWriter.println("    snippet.__s_a_v_e__(dbu, bpiId);");
        if (hasResult) {
            this.fileWriter.println("    return result;");
        }
        this.fileWriter.println("  }"); // end of snippetX() static method
        this.fileWriter.println();
        this.fileWriter.println();
        this.fileWriter.println();

        return methodName;
    }

    /**
     * Inserts the attributes declarations in the inner class of a Java snippet method.
     * @param variables Information over the variables accessible from this snippet
     * @param partnerLinks Information over the partnerLinks accessible from this snippet
     */
    private void insertClassDeclarations(Variable[] variables, PartnerLinkData[] partnerLinks) {
        // Declares BPEL variables
        for (int i=0; i<variables.length; ++i) {
            String javaType = null;
            if (variables[i].getMessageType() != null) {
                javaType = variables[i].getMessageType();
            } else if (variables[i].getVarType() != null) {
                javaType = variables[i].getVarType();
            } else if (variables[i].getElement() != null) {
                javaType = variables[i].getElement();
            } else if (variables[i].getBPELJType() != null) {
                javaType = variables[i].getBPELJType();
            }
            if (WSDLJavaMatching.isJavaPrimitiveType(javaType)) {
                // the javaType is wrapped
                this.fileWriter.println("      private " + javaType + " " + variables[i].getVariableName() + ";");
            } else {
                this.fileWriter.println("      private " + javaType + " " + variables[i].getVariableName() + " = null;");
            }
        }
        // This is necessary to avoid searching the variableData and the BPELVariable twice:
        // when retrieving the variable values and when saving them
        this.fileWriter.println("      org.objectweb.orchestra.data.VariableData[] varsData = "
                + "new org.objectweb.orchestra.data.VariableData[" + variables.length + "];");
        this.fileWriter.println("      org.objectweb.orchestra.engine.BPELVariable[] bpelVars = "
                + "new org.objectweb.orchestra.engine.BPELVariable[" + variables.length + "];");
        // Declares partnerLinks
        for (int i=0; i<partnerLinks.length; ++i) {
            if (partnerLinks[i] != null) {
                this.fileWriter.println("      private javax.bpelj.PartnerLink " + partnerLinks[i].getPlName() + ";");
            }
        }
        this.fileWriter.println();
    }

    /**
     * Inserts the inner class constructor of a Java snippet method.
     * @param className the name of the inner class
     * @param variables Information over the variables accessible from this snippet
     * @param partnerLinks Information over the partnerLinks accessible from this snippet
     */
    private void insertClassConstructor(String className, Variable[] variables, PartnerLinkData[] partnerLinks) {
        this.fileWriter.println("      public " + className + "("
                + "org.objectweb.orchestra.engine.EngineLocal engine,"
                + "int bpiId, int scopeId,"
                + "org.objectweb.orchestra.engine.PLType plType"
                + ") throws Exception {");
        insertInitVariables(variables);
        insertInitPLs(partnerLinks);
        this.fileWriter.println("      }"); // end of class constructor
    }

    private void insertInitVariables(Variable[] variables) {
        // Initializes variables
        for (int i=0; i<variables.length; ++i) {
            this.fileWriter.println("        { // " + variables[i].getVariableName());
            this.fileWriter.println("          this.varsData[" + i + "] = engine.getVar(bpiId, " + Util.toQuotedString(variables[i].getVariableName()) + ", scopeId);");
            String javaType = null;
            if (variables[i].getMessageType() != null) {
                javaType = variables[i].getMessageType();
                this.fileWriter.println("          this." + variables[i].getVariableName() + " = (" + javaType + ") "
                        + "org.objectweb.orchestra.util.Util.getValueObject(");
                this.fileWriter.println("                (byte[]) this.varsData[" + i + "].getValue(), " + Util.toQuotedString(this.processName) + ");");
            } else {
                if (variables[i].getVarType() != null) {
                    javaType = variables[i].getVarType();
                } else if (variables[i].getElement() != null) {
                    javaType = variables[i].getElement();
                } else if (variables[i].getBPELJType() != null) {
                    javaType = variables[i].getBPELJType();
                }
                this.fileWriter.println("          this.bpelVars[" + i + "] = (org.objectweb.orchestra.engine.BPELVariable) "
                        + "org.objectweb.orchestra.util.Util.getValueObject(");
                this.fileWriter.println("                (byte[]) this.varsData[" + i + "].getValue(), " + Util.toQuotedString(this.processName) + ");");
                if (WSDLJavaMatching.isJavaPrimitiveType(javaType)) {
                    // the javaType is wrapped
                    this.fileWriter.println("          Object tmp = this.bpelVars[" + i + "].getValue();");
                    this.fileWriter.println("          if (tmp != null) {");
                    this.fileWriter.println("            " + variables[i].getVariableName() + " = " +
                            "((" + WSDLJavaMatching.getWrapper(javaType) + ") tmp)." + javaType + "Value();");
                    this.fileWriter.println("          }");
                } else {
                    this.fileWriter.println("          " + variables[i].getVariableName() + " = (" + javaType + ") " + "this.bpelVars[" + i + "].getValue();");
                }
            }
            this.fileWriter.println("        }");
        }
    }

    private void insertInitPLs(PartnerLinkData[] partnerLinks) {
        // Initializes partnerLinks
        for (int i=0; i<partnerLinks.length; ++i) {
            if (partnerLinks[i] != null) {
                QName plMyRolePTQN = partnerLinks[i].getPlMyRolePT();
                String plMyRoleJavaType = null;
                if (plMyRolePTQN != null) {
                    plMyRoleJavaType = Util.getPackageFromNamespace(plMyRolePTQN.getNamespaceURI())
                    + "." + BPELString.toUpper(plMyRolePTQN.getLocalPart());
                }
                QName plPartnerRolePTQN = partnerLinks[i].getPlPartnerRolePT();
                String plPartnerRoleJavaType = null;
                if (plPartnerRolePTQN != null) {
                    plPartnerRoleJavaType = Util.getPackageFromNamespace(plPartnerRolePTQN.getNamespaceURI())
                    + "." + BPELString.toUpper(plPartnerRolePTQN.getLocalPart());
                }

                String plClassName = "PlImpl" + partnerLinks[i].getPlName();
                this.fileWriter.println("        class " + plClassName + " implements javax.bpelj.PartnerLink {");
                if (plMyRoleJavaType != null) {
                    this.fileWriter.println("          " + plMyRoleJavaType + " myRole = null;");
                    this.fileWriter.println("          public Object getMyRole() { return this.myRole; }");
                } else {
                    this.fileWriter.println("          public Object getMyRole() { return null; }");
                }
                if (plPartnerRoleJavaType != null) {
                    this.fileWriter.println("          " + plPartnerRoleJavaType + " partnerRole = null;");
                    this.fileWriter.println("          public void setPartnerRole(Object partner) { this.partnerRole = (" + plPartnerRoleJavaType + ") partner; }");
                } else {
                    this.fileWriter.println("          public Object partnerRole = null;");
                    // TODO when setPartnerRole() is used, really change the partnerRole in the database
                    // To be able to do this, you must change bindingFramework.BCSoapWork.invoke()
                    this.fileWriter.println("          public void setPartnerRole(Object partner) { this.partnerRole = partner; }");
                }
                this.fileWriter.println("          public Object getPartnerRole() { return this.partnerRole; }");
                this.fileWriter.println();
                this.fileWriter.println("          " + plClassName + "(org.objectweb.orchestra.engine.PLType plType) throws Exception {");
                QName plTypeQN = partnerLinks[i].getPlType();
                insertGetPlRole("myRole", plMyRolePTQN, plMyRoleJavaType, partnerLinks[i].getPlMyRole(), plTypeQN);
                insertGetPlRole("partnerRole", plPartnerRolePTQN, plPartnerRoleJavaType, partnerLinks[i].getPlPartnerRole(), plTypeQN);
                this.fileWriter.println("          }");
                this.fileWriter.println("        }");
                this.fileWriter.println("        this." + partnerLinks[i].getPlName() + " = new " + plClassName + "(plType);");
            }
        }
    }

    /**
     * Inserts the code to get the partnerLink roles
     * @param roleType "myRole" or "partnerRole"
     * @param plRolePTQN the partnerLink role QName
     * @param plRoleJavaType the partnerLink role Java type in the partnerLink class
     * @param plRoleName the role name
     * @param plTypeQN the partnerLink type QName
     */
    private void insertGetPlRole(String roleType, QName plRolePTQN, String plRoleJavaType, String plRoleName, QName plTypeQN) {
        if (plRoleName != null && plRolePTQN != null) {
            // To understand what follows, see engine.EngineBean.invoke() and engine.EngineBean.executeBFInvoke()
            this.fileWriter.println("            { // " + roleType);
            this.fileWriter.println("              String[] uddiService = org.objectweb.orchestra.util.UDDIDirectory.getService("
                    + Util.toQuotedString(this.processName) + ", " + Util.toQuotedString(Util.buildStringFromQName(plRolePTQN)) + ");");
            this.fileWriter.println("              String serviceName = uddiService[0];");
            //this.fileWriter.println("              String endpoint = uddiService[1];");
            //this.fileWriter.println("              String port = uddiService[2];");

            // Get the business process locator class constructor
            String packag = Util.getPackageFromNamespace(plRolePTQN.getNamespaceURI());
            this.fileWriter.println("              String locatorClsName = \"" + packag + ".\" + "
                    + "serviceName.substring(0, 1).toUpperCase() + serviceName.substring(1)"
                    + " + \"Locator\";");
            this.fileWriter.println("              Class locatorCls = Class.forName(locatorClsName, true, org.objectweb.orchestra.util.ClassLoaderMap.getCL("
                    + Util.toQuotedString(this.processName) + "));");
            this.fileWriter.println("              Object locatorObj = locatorCls.getConstructor((Class [])null).newInstance((Object [])null);");
            // Get a portType object
            this.fileWriter.println("              QName plTypeQN = new QName(" + Util.toQuotedString(plTypeQN.getNamespaceURI()) + ", " + Util.toQuotedString(plTypeQN.getLocalPart()) + ");");
            this.fileWriter.println("              QName rolePT = plType.getPortType(plTypeQN, " + Util.toQuotedString(plRoleName) + ");");
            //this.fileWriter.println("              java.lang.reflect.Method met = locatorObj.getClass().getMethod(\"get\"+rolePT.getLocalPart(),(Class [])null);");
            this.fileWriter.println("              java.util.Iterator ports = (java.util.Iterator) locatorObj.getClass().getMethod(\"getPorts\",(Class [])null).invoke(locatorObj, (Object [])null);");
            this.fileWriter.println("              QName port = (QName) ports.next();");
            this.fileWriter.println("              String portLP = port.getLocalPart();");
            this.fileWriter.println("              java.lang.reflect.Method met = locatorObj.getClass().getMethod(\"get\"+portLP,(Class [])null);");

            // Assign the class attribute
            this.fileWriter.println("              this." + roleType + " = (" + plRoleJavaType + ") met.invoke(locatorObj, (Object [])null);");
            this.fileWriter.println("            }");
        }
    }

    /**
     * Inserts the exec method of a Java snippet inner class
     * @param javaCode the Java code  of the Java snippet
     * @param returnType the return type of this snippet
     * @param isBlock Whether or not this statement is a "block" statement
     */
    private void insertExecMethod(String javaCode, String returnType, boolean isBlock) {
        // Executes the snippet
        this.fileWriter.println("      public " + returnType + " __e_x_e_c__() throws Exception {");
        if (isBlock) {
            this.fileWriter.println(javaCode);
        } else {
            this.fileWriter.println("        return ( " + javaCode + ");");
        }
        this.fileWriter.println("      }");
    }

    /**
     * Inserts the save method of a Java snippet inner class
     * @param variables Information over the variables accessible from this snippet
     */
    private void insertSaveMethod(Variable[] variables) {
        // Saves BPEL variables in the vars array
        this.fileWriter.println("      public void __s_a_v_e__(org.objectweb.orchestra.engine.DbUpdater dbu, int bpiId) throws Exception {");
        this.fileWriter.println("        try {");
        this.fileWriter.println("          dbu.startTx();");
        insertSaveVariables(variables);
        insertSavePLs(null);
        this.fileWriter.println("          dbu.commitTx();");
        this.fileWriter.println("        } finally {");
        this.fileWriter.println("          dbu.roolbackTx();");
        this.fileWriter.println("        }");
        this.fileWriter.println("      }"); // end of __s_a_v_e__
    }

    private void insertSaveVariables(Variable[] variables) {
        for (int i=0; i<variables.length; ++i) {
            this.fileWriter.println("          { // " + variables[i].getVariableName());
            if (variables[i].getMessageType() != null) {
                this.fileWriter.println("            org.objectweb.orchestra.data.VariableCache.setVariableValue(" +
                        "this.varsData[" + i + "].getVariableId(), this." + variables[i].getVariableName() + ", bpiId);");
            } else {
                String javaType = null;
                if (variables[i].getVarType() != null) {
                    javaType = variables[i].getVarType();
                } else if (variables[i].getElement() != null) {
                    javaType = variables[i].getElement();
                } else if (variables[i].getBPELJType() != null) {
                    javaType = variables[i].getBPELJType();
                }
                if (WSDLJavaMatching.isJavaPrimitiveType(javaType)) {
                    // the javaType is wrapped
                    this.fileWriter.println("            this.bpelVars[" + i + "].setValue(new "
                            + WSDLJavaMatching.getWrapper(javaType) + "(this." + variables[i].getVariableName() + "));");
                } else {
                    this.fileWriter.println("            this.bpelVars[" + i + "].setValue(this." + variables[i].getVariableName() + ");");
                }
                this.fileWriter.println("            org.objectweb.orchestra.data.VariableCache.setVariableValue(" +
                        "this.varsData[" + i + "].getVariableId(), this.bpelVars[" + i + "], bpiId);");
            }
            this.fileWriter.println("          }");
        }
    }

    private void insertSavePLs(PartnerLinkData[] partnerLinks) {
        // TODO
    }

    /**
     * insert the last instructions of the file
     */
    public void finish() throws BPELGeneratorException {
        String BPELJPackage = this.proc.getBPELJPackage();
        if (BPELJPackage != null) {
            BPELJPackage = "\"" + BPELJPackage + "\"";
        }
        try {
            this.fileWriter.println("}");
            this.fileWriter.close();
        } catch (Exception e) {
            throw new BPELGeneratorException("BPELJProcessFile.finish : ",e);
        }

        System.out.println("File " + this.filePath + " created");
    }
}