/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.objectweb.orchestra.generator;

import java.io.File;
import java.io.FileWriter;
import java.util.Vector;

import org.objectweb.orchestra.exception.BPELGeneratorException;
import org.objectweb.orchestra.util.BPELConstants;
import org.objectweb.orchestra.util.BPELString;
import org.objectweb.orchestra.util.Util;
import org.objectweb.orchestra.util.WSDLJavaMatching;



/**
 * BPELMessageFile.java, build a file for each message declared in wsdl file,
 * extends org.objectweb.orchestra.engine.BPELMessage
 *
 * @author Goulven Le Jeune & Charles Souillard
 * @version $Id: BPELMessageFile.java 5 2006-10-20 14:37:44Z lejeuneg $
 *
 */
public class BPELMessageFile {

    private static String myPath = BPELConstants.GENERATOR_DIR;

    // -------------------------------------------------------------------------
    // Members
    // -------------------------------------------------------------------------
    private FileWriter fileWriter = null;
    private String fileName = null;
    private String directoryName = null;
    private String processName = null;

    // Message properties
    private String messageName = null;
    private String messageNS = null;
    private Vector messageParts = null;
    private Vector messageCS = null;
    private String f = null;
    private String packag = null;

    /**
     * Constructor
     *
     * @param process
     *            the process name
     * @param name
     *            the message name
     * @param parts
     *            the vector representing all parts of this message
     * @param cs
     *            the vector representing correlationsSets associated to this
     *            message
     * @param w
     *            a wsdl object giving namespaces mapping
     */
    public BPELMessageFile(String process, String name, Vector parts, Vector cs) {
        this.processName = process;
        this.messageName = BPELString.removeNS(name);
        this.messageNS = BPELString.getNS(name);
        this.messageParts = parts;
        this.messageCS = cs;
    }

    /**
     * print the BPELMessage file
     *
     */
    public void print() throws BPELGeneratorException {
        createFile();
        insertDeclaration();
        insertConstructor();
        insertGetter();
        insertSetPart();
        insertGetPropertyValue();
        insertSetPropertyValue();
        insertCompare();
        insertGetStructure();
        finish();
    }

    /**
     * creates the BPELMessage file
     *
     */
    private void createFile() throws BPELGeneratorException {
        this.directoryName = this.processName;
        this.fileName = BPELConstants.BM + this.processName + this.messageName;

        //make the directory based on the namespace of the message
        this.packag = Util.getPackageFromNamespace(this.messageNS);
        String dir = Util.getDirectoryFromPackage(this.packag);
        try {
            this.f = myPath + File.separator + this.directoryName + File.separator + dir + File.separator + this.fileName + ".java";
            new File(this.f).getAbsoluteFile().getParentFile().mkdirs();
            this.fileWriter = new FileWriter(this.f);
        } catch (Exception e) {
            throw new BPELGeneratorException("BPELMessageFile.createFile : ",e);
        }
        insertLine(Util.autograph(this.fileName + ".java"));
    }

    /**
     * insert the beginning of the file : package, import, attributes
     *
     */
    private void insertDeclaration() throws BPELGeneratorException {
        insertLine("package " + this.packag + ";");
        insertLine("");
        insertLine("import org.objectweb.orchestra.engine.*;");
        insertLine("import org.objectweb.orchestra.util.*;");
        insertLine("");
        insertLine("");
        insertLine("public class " + this.fileName + " extends " + BPELConstants.BPELMESSAGE + " {");
        insertLine("");
        //for each part of the message, add an attribute
        for (int i = 0; i < this.messageParts.size(); i++) {
            MessagePart mp = (MessagePart) this.messageParts.get(i);
            //String type =
            // WSDLJavaMatching.getJavaType(mp.getPartType(),processName);
            String type = WSDLJavaMatching.getWrapperType(mp.getPartType());

            insertLine("  private " + type + " " + mp.getPartName() + BPELConstants.PART_SUFFIX + " = null;");
        }
        insertLine("  private org.objectweb.orchestra.engine.MessageStructure ms;");
        insertLine("");
    }

    /**
     * insert the constructor
     *
     */
    private void insertConstructor() throws BPELGeneratorException {
        insertLine("public " + BPELConstants.BM + this.processName + this.messageName + "() {");
        insertLine("  setMessageName(\"" +this.messageName+ "\");");
        insertLine("  ms = new org.objectweb.orchestra.engine.MessageStructure(\"" + this.messageName + "\");");
        insertLine("  ms.setPartNb(" + this.messageParts.size() + ");");
        //for each part of the message, create a partStructure and add it to
        // the messageStructure
        for (int i = 0; i < this.messageParts.size(); i++) {
            MessagePart mp = (MessagePart) this.messageParts.get(i);
            insertLine("  org.objectweb.orchestra.engine.PartStructure ps" + i + " = new org.objectweb.orchestra.engine.PartStructure(\"" + mp.getPartName() + "\", \"" + mp.getPartType() + "\", \"" + mp.getPartElement() + "\");");
            insertLine("  ms.addPart(ps" + i + ");");
        }
        insertLine("}");
        insertLine("");
    }

    /**
     * insert a getter method to retreive the messageStructure
     *
     */
    private void insertGetStructure() throws BPELGeneratorException {
        insertLine("public org.objectweb.orchestra.engine.MessageStructure getStructure() {");
        insertLine("  return ms;");
        insertLine("}");
        insertLine("");
    }

    /**
     * insert a getter method to retreive a part value based on the part name
     *
     */
    private void insertGetter() throws BPELGeneratorException {
        insertLine("public java.lang.Object getPart(java.lang.String partName) {");
        //for each part of the message, return the right value
        for (int i = 0; i < this.messageParts.size(); i++) {
            MessagePart mp = (MessagePart) this.messageParts.get(i);
            insertLine("  if (partName.equals(\"" + mp.getPartName() + "\")) {");
            insertLine("    return " + mp.getPartName() + BPELConstants.PART_SUFFIX + ";");
            insert("  } else ");
        }
        insertLine("{");
        insertLine("    System.out.println(\"Error in BPELMessage (getPart) : no part with name : \"+partName+\"  in BPELMessage Class : \"+this.getClass());");
        insertLine("    return null;");
        insertLine("  }");
        insertLine("}");
        insertLine("");
    }

    /**
     * insert a setter method to set a part value based on the part name
     *
     */
    private void insertSetPart() throws BPELGeneratorException {
        insertLine("public void setPart(java.lang.String partName, java.lang.Object partValue) {");
        //for each part of the message, set the right value
        for (int i = 0; i < this.messageParts.size(); i++) {
            MessagePart mp = (MessagePart) this.messageParts.get(i);
            insertLine("  if (partName.equals(\"" + mp.getPartName() + "\")) {");
            //String type =
            // WSDLJavaMatching.getJavaType(mp.getPartType(),processName);
            String type = WSDLJavaMatching.getWrapperType(mp.getPartType());
            insertLine("    " + mp.getPartName() + BPELConstants.PART_SUFFIX + " = (" + type + ")partValue;");
            insert("  } else ");
        }
        insertLine("{");
        insertLine("    System.out.println(\"Error in BPELMessage (setPart) : no part with name : \"+partName+\"  in BPELMessage Class : \"+this.getClass());");
        insertLine("  }");
        insertLine("}");
        insertLine("");
    }

    /**
     * insert a getter method to get the value of the CorrelationSet of this
     * message
     *
     */
    private void insertGetPropertyValue() throws BPELGeneratorException {
        insertLine("public java.lang.Object getPropertyValue(java.lang.String prop) {");
        if (this.messageCS.size() > 0) {
            //for each property of the message, return the right value (part,
            // query)
            for (int i = 0; i < this.messageCS.size(); i++) {
                MessageCS mcs = (MessageCS) this.messageCS.get(i);
                insertLine("  if (prop.equals(\"" + mcs.getCSName() + "\")) {");
                String csPart = mcs.getCSPart() + BPELConstants.PART_SUFFIX;
                if (mcs.getCSQuery() != null) {
                    insertLine("    Object o = null;");
                    insertLine("    try {");
                    insertLine("        o = Expression.executeGetQuery("+csPart+",\""+mcs.getCSQuery()+"\",-1,-1,\""+this.processName+"\");");
                    insertLine("    } catch (Exception theEx) {");
                    insertLine("        theEx.printStackTrace();");
                    insertLine("    }");
                    insertLine("    return o;");
                } else {
                    insertLine("    return "+csPart+";");
                }
                insert("  } else ");
            }
            insertLine("{");
            insertLine("    System.out.println(\"Error in BPELMessage (getPropertyValue) : no property with name : \"+prop+\"  in BPELMessage Class : \"+this.getClass());");
            insertLine("    return null;");
            insertLine("  }");
        } else {
            insertLine("    System.out.println(\"Error in BPELMessage (getPropertyValue) : no property with name : \"+prop+\"  in BPELMessage Class : \"+this.getClass());");
            insertLine("    return null;");
        }
        insertLine("}");
        insertLine("");
    }

    /**
     * insert a getter method to get the value of the CorrelationSet of this
     * message
     *
     */
    private void insertSetPropertyValue() throws BPELGeneratorException {
        insertLine("public void setPropertyValue(java.lang.String prop, Object value) {");
        if (this.messageCS.size() > 0) {
            //for each property of the message, set the value (part, query)
            for (int i = 0; i < this.messageCS.size(); i++) {
                MessageCS mcs = (MessageCS) this.messageCS.get(i);
                insertLine("  if (prop.equals(\"" + mcs.getCSName() + "\")) {");
                String csPart = mcs.getCSPart() + BPELConstants.PART_SUFFIX;
                if (mcs.getCSQuery() != null) {
                    insertLine("    Object o = null;");
                    insertLine("    try {");
                    insertLine("        if (" + csPart + " == null)");
                    insertLine("            initializePart(\"" + csPart.substring(0,csPart.lastIndexOf("_")) + "\", \"" + getPartType(csPart.substring(0,csPart.lastIndexOf("_"))) + "\", \"" + this.processName + "\");");
                    insertLine("        Expression.executeSetQuery(-1,-1,"+csPart+",\""+mcs.getCSQuery()+"\",value,\""+this.processName+"\",\""+this.messageName+"\",\""+this.packag+"\");");
                    insertLine("    } catch (Exception theEx) {");
                    insertLine("        theEx.printStackTrace();");
                    insertLine("    }");
                } else {
                    insertLine("    this.setPart(\""+csPart.substring(0,csPart.lastIndexOf("_"))+"\", value);");
                }
                insert("  }");
            }
        }
        insertLine("}");
        insertLine("");
    }

    /**
     * insert a comparaison method to know if the correlation value of cs (name)
     * is equals or not to a given correlation value
     *
     */
    private void insertCompare() throws BPELGeneratorException {
        insertLine("public boolean compare(java.lang.String cs, java.lang.Object corrValue) {");
        insertLine("  return ((corrValue).equals(getPropertyValue(cs)));");
        insertLine("}");
        insertLine("");
    }

    /**
     * insert the end of the file
     *
     */
    private void finish() throws BPELGeneratorException {
        insertLine("}");
        try {
            this.fileWriter.close();
        } catch (Exception e) {
            throw new BPELGeneratorException("BPELMessageFile.finish : ",e);
        }
        System.out.println("File " + this.f + " created");
    }

    /**
     * insert a string in the file
     *
     * @param commandLine
     *            the string to insert
     *
     */
    private void insert(String commandLine) throws BPELGeneratorException {
        try {
            this.fileWriter.write(commandLine);
        } catch (Exception e) {
            throw new BPELGeneratorException("BPELMessageFile.insert : ",e);
        }
    }

    /**
     * insert a line in the file
     *
     * @param commandLine
     *            the line to insert
     *
     */
    private void insertLine(String commandLine) throws BPELGeneratorException {
        try {
            this.fileWriter.write(commandLine + "\n");
        } catch (Exception e) {
            throw new BPELGeneratorException("BPELMessageFile.insertLine : ",e);
        }
    }

    /**
     * Type of a Part
     *
     */
    private String getPartType(String partName) {
        String partType = null;
        //System.out.println("getPartType partName : " + partName);
        //for each part of the message, add an attribute
        for (int i = 0; i < this.messageParts.size() && partType == null; i++) {
            MessagePart mp = (MessagePart) this.messageParts.get(i);
            //System.out.println("getPartType mp.getPartName() : " + mp.getPartName());
            if (mp.getPartName().equals(partName))
                partType = WSDLJavaMatching.getWrapperType(mp.getPartType());
        }
        return partType;
    }

}