/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.objectweb.orchestra.generator;

import java.io.File;
import java.io.FileWriter;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.wsdl.Definition;
import javax.wsdl.Fault;
import javax.wsdl.Input;
import javax.wsdl.Message;
import javax.wsdl.Output;
import javax.wsdl.xml.WSDLReader;
import javax.xml.namespace.QName;

import org.objectweb.orchestra.exception.BPELGeneratorException;
import org.objectweb.orchestra.util.BPELConstants;
import org.objectweb.orchestra.util.Util;

import com.ibm.wsdl.OperationImpl;
import com.ibm.wsdl.PortTypeImpl;
import com.ibm.wsdl.factory.WSDLFactoryImpl;

/**
 * WSDLBindingFile.java, writes a file giving the association between a wsdl
 * service and its binding and their information
 *
 * @author Goulven Le Jeune & Charles Souillard
 * @version $Id: WSDLBindingFile.java 5 2006-10-20 14:37:44Z lejeuneg $
 *
 */
public class WSDLBindingFile {

    private static String myPath = BPELConstants.WSDLBINDING_DIR;
    private FileWriter fileWriter = null;
    private String fileName = null;
    private String processName = null;
    private String f = null;

    /**
     * constructor
     *
     * @param process
     *            the name of the process
     */
    public WSDLBindingFile(String process) throws BPELGeneratorException {

        this.processName = process;
        this.fileName = this.processName;

        try {
            this.f = myPath + File.separator + this.fileName + ".txt";
            //System.out.println("File : " + f);
            new File(this.f).getAbsoluteFile().getParentFile().mkdirs();
            this.fileWriter = new FileWriter(this.f);
        } catch (Exception e) {
            throw new BPELGeneratorException("WSDLBindingFile.init : ",e);
        }
    }

    public void analyseWsdl(String wsdlFile) throws BPELGeneratorException {
        Definition def = null;
        WSDLFactoryImpl wsdlFactory = new WSDLFactoryImpl();
        WSDLReader wsdlReader = wsdlFactory.newWSDLReader();
        try {
            def = wsdlReader.readWSDL(wsdlFile);
        } catch (javax.wsdl.WSDLException e) {
            throw new BPELGeneratorException("WSDLException WSDLCompilo analyseFile :\n" , e);
        }
        Map portTypes = def.getPortTypes();
        Set ptSet = portTypes.entrySet();
        for (Iterator i = ptSet.iterator(); i.hasNext();) {
            Map.Entry entry0 = (Map.Entry) i.next();
            PortTypeImpl pt = (PortTypeImpl) entry0.getValue();
            String ptName = Util.buildStringFromQName(pt.getQName());
            List operationsL = pt.getOperations();
            for (Iterator j = operationsL.iterator(); j.hasNext();) {
                OperationImpl op = (OperationImpl) j.next();
                String operationName = op.getName();

                Input in = op.getInput();
                String input = null;
                if (in != null) {
                    Message mIn = in.getMessage();
                    input = Util.buildStringFromQName(mIn.getQName());
                }
                Output out = op.getOutput();
                String output = null;
                if (out != null) {
                    Message mOut = out.getMessage();
                    output = Util.buildStringFromQName(mOut.getQName());
                }
                Map faultsMap = op.getFaults();
                insertOperationLine(ptName, operationName, input, output, faultsMap);
            }
        }
    }


    /**
     * constructor
     *
     * @param process
     *            the name of the process
     */
    public void insertLine(String property, String value) throws BPELGeneratorException {
        try {
            this.fileWriter.write(property + "=" + value + "\n");
        } catch (Exception e) {
            throw new BPELGeneratorException("WSDLBindingFile.insertLine : ",e);
        }
    }

    /**
     * insert a line in the file
     *
     * @param serviceName
     *            the name of the service
     * @param portName
     *            the port concerned by this line for this service
     * @param binding
     *            the binding associated to this service
     * @param addressLocation
     *            the location uri where the wsdl file can be found
     * @param bindingType
     *            the portType associated to this binding
     */
    public void insertServiceLine(QName serviceName, String portName, QName binding, String addressLocation, QName bindingType) throws BPELGeneratorException {
        try {
            String st = "";
            st += "serviceName=" + Util.buildStringFromQName(serviceName);
            st += " ";
            st += "portName=" + portName;
            st += " ";
            st += "binding=" + Util.buildStringFromQName(binding);
            st += " ";
            st += "addressLocation=" + addressLocation;
            st += " ";
            st += "bindingType=" + Util.buildStringFromQName(bindingType);
            st += "\n";

            this.fileWriter.write(st);
        } catch (Exception e) {
            throw new BPELGeneratorException("WSDLBindingFile.insertServiceLine : ",e);
        }
    }

    /**
     * insert a line in the file
     *
     * @param portTypeName
     *            the portType concerned
     * @param operationName
     *            the operation concerned
     * @param inputMessageName
     *            the input message name of the operation
     * @param outputMessageName
     *            the output message name of the operation
     */
    public void insertOperationLine(String portTypeName, String operationName, String inputMessageName, String outputMessageName, Map faults) throws BPELGeneratorException {
        try {
            String st = "";
            st += "portTypeName=" + portTypeName;
            st += " ";
            st += "operationName=" + operationName;
            st += " ";
            st += "inputMessageName=" + inputMessageName;
            st += " ";
            st += "outputMessageName=" + outputMessageName;
            st += " ";
            st += "Faults=";
            for (Iterator it = faults.values().iterator(); it.hasNext();) {
                Fault f = (Fault) it.next();
                Message faultMessage = f.getMessage();
                String fm = Util.buildStringFromQName(faultMessage.getQName());
                st += fm.substring(0,fm.lastIndexOf(",")+1) + f.getName() + BPELConstants.FAULTINFO_SEPARATOR + fm + ";";
            }
            st += "\n";

            this.fileWriter.write(st);
        } catch (Exception e) {
            throw new BPELGeneratorException("WSDLBindingFile.insertOperationline : ",e);
        }
    }

    /**
     * write the last instructions of the file
     *
     */
    public void finish() throws BPELGeneratorException {
        try {
            this.fileWriter.close();
        } catch (Exception e) {
            throw new BPELGeneratorException("WSDLBindingFile.finish : ",e);
        }
        System.out.println("File " + this.f + " created");
    }

}