/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.objectweb.orchestra.monitoring;

import java.io.File;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.Calendar;
import java.util.Collection;
import java.util.GregorianCalendar;
import java.util.Iterator;
import java.util.Vector;

import javax.ejb.CreateException;
import javax.ejb.RemoveException;
import javax.ejb.SessionBean;
import javax.ejb.SessionContext;
import javax.xml.namespace.QName;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.objectweb.orchestra.data.BPICache;
import org.objectweb.orchestra.data.BPIData;
import org.objectweb.orchestra.data.BPUCache;
import org.objectweb.orchestra.data.BPUData;
import org.objectweb.orchestra.data.ScopeCache;
import org.objectweb.orchestra.data.ScopeData;
import org.objectweb.orchestra.data.VariableData;
import org.objectweb.orchestra.engine.BPELEngine;
import org.objectweb.orchestra.engine.BPELMessage;
import org.objectweb.orchestra.engine.BPELVariable;
import org.objectweb.orchestra.engine.EngineHomeLocal;
import org.objectweb.orchestra.engine.EngineLocal;
import org.objectweb.orchestra.engine.EngineUtil;
import org.objectweb.orchestra.exception.BPELEngineException;
import org.objectweb.orchestra.exception.BPELMonitoringException;
import org.objectweb.orchestra.exception.BPELMonitoringObjectNotFoundException;
import org.objectweb.orchestra.exception.BPELUtilException;
import org.objectweb.orchestra.tree.Activity;
import org.objectweb.orchestra.tree.BPELProcess;
import org.objectweb.orchestra.tree.Case;
import org.objectweb.orchestra.tree.Catch;
import org.objectweb.orchestra.tree.Compensate;
import org.objectweb.orchestra.tree.Correlation;
import org.objectweb.orchestra.tree.CorrelationSet;
import org.objectweb.orchestra.tree.From;
import org.objectweb.orchestra.tree.Invoke;
import org.objectweb.orchestra.tree.OnAlarm;
import org.objectweb.orchestra.tree.OnMessage;
import org.objectweb.orchestra.tree.PartnerLink;
import org.objectweb.orchestra.tree.Pick;
import org.objectweb.orchestra.tree.Receive;
import org.objectweb.orchestra.tree.Reply;
import org.objectweb.orchestra.tree.Source;
import org.objectweb.orchestra.tree.Statement;
import org.objectweb.orchestra.tree.Target;
import org.objectweb.orchestra.tree.Throw;
import org.objectweb.orchestra.tree.To;
import org.objectweb.orchestra.tree.Variable;
import org.objectweb.orchestra.tree.Wait;
import org.objectweb.orchestra.tree.While;
import org.objectweb.orchestra.util.BPELConstants;
import org.objectweb.orchestra.util.BPELTrace;
import org.objectweb.orchestra.util.ClassLoaderMap;
import org.objectweb.orchestra.util.Util;
import org.objectweb.orchestra.util.WSDLJavaMatching;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 * BpelMonitoringSessionBean.java.
 *
 * @author Goulven Le Jeune & Charles Souillard
 * @version $Id: BpelMonitoringSessionBean.java,v 1.5 2004/10/20 13:35:52
 *          souillac Exp $
 *
 * @ejb:bean name="BpelMonitoringSession" display-name="Session
 *           BpelMonitoringSession Bean" type="Stateless"
 *           transaction-type="Container"
 *           jndi-name="bpel/monitoring/BpelMonitoringSessionHome"
 *           local-jndi-name="bpel/monitoring/BpelMonitoringSessionHome_L"
 *           view-type="both"
 * @ejb.home local-class="org.objectweb.orchestra.monitoring.BpelMonitoringSessionHomeLocal"
 *           remote-class="org.objectweb.orchestra.monitoring.BpelMonitoringSessionHome"
 * @ejb.interface local-class="org.objectweb.orchestra.monitoring.BpelMonitoringSessionLocal"
 *                remote-class="org.objectweb.orchestra.monitoring.BpelMonitoringSession"
 * @ejb:transaction type="Supports"
 *
 * @jonas.max-cache-size @BpelMonitoringSessionBean.max-cache-size@
 */
public class BpelMonitoringSessionBean implements SessionBean {

    private SessionContext mContext;
    private Vector[] msTab = null;
    private BPELProcess bpelProcess = null;
    private Document doc = null;
    private boolean running = true;

    public SessionContext getContext() {
        return this.mContext;
    }

    /**
     * Create the All Projects Session Bean
     *
     * @throws CreateException
     * @ejb:create-method
     */
    public void ejbCreate() throws CreateException {
        // BEAN METHOD
    }

    public void setSessionContext(final SessionContext context) {
        this.mContext = context;
    }

    public void ejbRemove() {
        //BEAN METHOD
    }

    public void ejbActivate() {
        //BEAN METHOD
    }

    public void ejbPassivate() {
        //BEAN METHOD
    }

    // -------------------------------------------------------------------------
    // Methods - public
    // -------------------------------------------------------------------------

    /**
     * buildStatementsTree
     *
     * @param bpiId :
     *            Id of the instance to analyse
     * @ejb:interface-method
     */
    public Document getStatementsTree(int bpiId) throws BPELMonitoringException {
        return getStatementsTree(bpiId, -1);
    }

    /**
     * buildStatementsTree
     *
     * @param bpiId :
     *            Id of the instance to analyse
     * @param lastMaxTimeStamp :
     *            The last maximum timestamp
     * @ejb:interface-method
     */
    public Document getStatementsTree(int bpiId, long lastMaxTimeStamp) throws BPELMonitoringException {
        if (BPELTrace.isMonitoringDebugEnabled()) BPELTrace.traceMonitoring("BpelMonitoringSession buildStatementsTree, bpiId = " + bpiId);
        getBpelProcess(bpiId);
        this.msTab = new Vector[this.bpelProcess.getTotalStatementNb() + 1];
        //if the bpi is running

        if (!this.running) {
            //else, read xml file representing this instance
            readXmlFile(bpiId);
        } else {
            long maxTimeStamp = -1;
            maxTimeStamp = collapseBPUsInfos(bpiId);
            if (BPELTrace.isMonitoringDebugEnabled()) BPELTrace.traceMonitoring("BpelMonitoringSession buildStatementsTree, bpiId = " + bpiId + ", maxTimeStamp = " + maxTimeStamp);
            // sort ms by timestamp
            for (int i = 0; i < this.msTab.length; i++) {
                Vector v = this.msTab[i];
                Vector sorted = sortByTimeStamp(v);
                this.msTab[i] = sorted;
            }
            if (BPELTrace.isMonitoringDebugEnabled()) BPELTrace.traceMonitoring("BpelMonitoringSession buildStatementsTree, bpiId = " + bpiId + ", after sort, this.msTab = " + this.msTab + "\n, this.msTab.size = " + this.msTab.length);
            if (maxTimeStamp > lastMaxTimeStamp) {
                initDocument();
                buildDomDocument(maxTimeStamp, bpiId);
            }
        }
        if (BPELTrace.isMonitoringDebugEnabled()) BPELTrace.traceMonitoring("BpelMonitoringSession end buildStatementsTree");
        return this.doc;
    }

    private void readXmlFile(int bpiId) throws BPELMonitoringException {
        String dirName = BPELConstants.BUILD_XML_DIR + File.separator + this.bpelProcess.getName();
        File dir = new File(dirName);
        if (!dir.exists()) { throw new BPELMonitoringException("No Instance stored on hard disk with processName : " + this.bpelProcess.getName() + ". Perhaps the first instance of this process is currently being flushed on hard disk, try again in a few seconds.", null); }
        File theFile = null;
        File[] list = dir.listFiles();
        for (int i = 0; i < list.length; i++) {
            File f = list[i];
            if (f.getName().startsWith(Integer.toString(bpiId) + "_")) {
                theFile = f;
                break;
            }
        }
        if (theFile == null) { throw new BPELMonitoringException("No Instance stored on hard disk with id : " + bpiId + ". Perhaps this instance is currently being flushed on hard disk, try again in a few seconds.", null); }
        this.doc = getDoc(theFile);
    }

    private Document getDoc(File f) throws BPELMonitoringException {
        Document document = null;
        try {
            DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
            DocumentBuilder docBuilder = dbf.newDocumentBuilder();
            document = docBuilder.parse(f);
        } catch (Exception e) {
            throw new BPELMonitoringException("BPELMonitoring readXmlFile : ", e);
        }
        return document;
    }

    private Vector sortByTimeStamp(Vector v) {
        if (v != null) {
            for (int i = 0; i < v.size() - 1; i++) {
                int im = i;
                for (int j = i + 1; j < v.size(); j++) {
                    MonitoringStatement msj = (MonitoringStatement) v.get(j);
                    MonitoringStatement msim = (MonitoringStatement) v.get(im);
                    if (msj.getTimeStamp() < msim.getTimeStamp()) {
                        im = j;
                    }
                }
                Object tmp = v.get(i);
                Object tmp2 = v.get(im);
                v.remove(i);
                v.add(i, tmp2);
                v.remove(im);
                v.add(im, tmp);
            }
        }
        return v;
    }

    /**
     * getEndFlows
     *
     * @param bpiId : Id of the instance to analyse
     * @ejb:interface-method
     */
    public Vector getEndFlows(int bpiId) throws BPELMonitoringException {
        if (BPELTrace.isMonitoringDebugEnabled()) BPELTrace.traceMonitoring("BpelMonitoringSession getEndFlows, bpiId = " + bpiId);
        Vector endFlows = null;
        getBpelProcess(bpiId);
        endFlows = getEndFlowsFromBPI(bpiId);
        if (BPELTrace.isMonitoringDebugEnabled()) BPELTrace.traceMonitoring("BpelMonitoringSession end getEndFlows");
        return endFlows;
    }

    /**
     * getTargets
     *
     * @param bpiId :
     *            Id of the instance to analyse
     * @ejb:interface-method
     */
    public Vector getTargets(int bpiId) throws BPELMonitoringException {
        if (BPELTrace.isMonitoringDebugEnabled()) BPELTrace.traceMonitoring("BpelMonitoringSession getTargets, bpiId = " + bpiId);
        Vector targets = null;
        getBpelProcess(bpiId);
        targets = getTargetsFromBPI(bpiId);
        if (BPELTrace.isMonitoringDebugEnabled()) BPELTrace.traceMonitoring("BpelMonitoringSession end getTargets");
        return targets;
    }

    private long collapseBPUsInfos(int bpiId) throws BPELMonitoringException {
        if (BPELTrace.isMonitoringDebugEnabled()) BPELTrace.traceMonitoring("BpelMonitoringSession collapseBPUsInfos, bpiId = " + bpiId);
        Collection bpus;
        try {
            bpus = BPICache.getBPUs(bpiId);
        } catch (BPELEngineException e) {
            throw new BPELMonitoringException("BPELEngineException BpelMonitoringSession collapseBPUsInfos :\n", e);
        }
        long maxTimeStamp = -1;
        for (Iterator it = bpus.iterator(); it.hasNext();) {
            BPUData bpu = (BPUData) it.next();
            Vector monitor = bpu.getMonitor();
            String st = null;
            for (int i = 0; i < monitor.size(); i++) {
                st = (String) monitor.get(i);
                maxTimeStamp = analyseLine(st, bpu.getBPUId(), maxTimeStamp);
            }
        }
        if (BPELTrace.isMonitoringDebugEnabled()) BPELTrace.traceMonitoring("BpelMonitoringSession end collapseBPUsInfos, bpus = " + bpus + ", maxTimeStamp = " + maxTimeStamp);
        return maxTimeStamp;
    }

    private long analyseLine(String st, int bpuId, long maxTimeStamp) {
        if (BPELTrace.isMonitoringDebugEnabled()) BPELTrace.traceMonitoring("BpelMonitoringSession analyseLine, st = " + st + ", bpuId = " + bpuId + ", maxTimeStamp = " + maxTimeStamp);
        String[] split = st.split(BPELConstants.MONITORING_INFO_SEP);
        if (split.length >= 3) {
            MonitoringStatement statement = new MonitoringStatement(bpuId, split);
            int statementNb = statement.getStatementNb();
            Vector v = initVector(this.msTab[statementNb]);
            v.addElement(statement);
            this.msTab[statementNb] = v;
            if (statement.getTimeStamp() > maxTimeStamp) {
                maxTimeStamp = statement.getTimeStamp();
            }
        }
        if (BPELTrace.isMonitoringDebugEnabled()) BPELTrace.traceMonitoring("BpelMonitoringSession end analyseLine, maxTimeStamp = " + maxTimeStamp);
        // else it is an endFlow
        return maxTimeStamp;
    }

    private Vector initVector(Vector v) {
        if (v == null) {
            v = new Vector();
        }
        return v;
    }

    private void getBpelProcess(int bpiId) throws BPELMonitoringException {
        if (BPELTrace.isMonitoringDebugEnabled()) BPELTrace.traceMonitoring("BpelMonitoringSession getBpelProcess, bpiId = " + bpiId);
        this.running = true;
        String process = null;
        String targetNamespace = null;
        try {
            BPIData bd = BPICache.findBPI(bpiId);
            if (bd != null) {
                process = bd.getProcessName();
                targetNamespace = bd.getTargetNamespace();
            } else {
                //DO NOTHING, it means this instance is not running
                this.running = false;
            }
        } catch (BPELEngineException e) {
            //DO NOTHING, it means this instance is not running
            this.running = false;
        }
        if (process == null && !BPELEngine.flushInstance()) {
            throw new BPELMonitoringObjectNotFoundException("Instance with id "+bpiId+" is not running and the engine is not flushing finished instances.", null);
        } else if (process == null ){
            Document doc = getDocumentOfBpi(bpiId);
            Node first = doc.getFirstChild();
            NamedNodeMap attrs = first.getAttributes();
            process = attrs.getNamedItem(BPELConstants.MONITORING_PROCESSNAME).getNodeValue();
            targetNamespace = attrs.getNamedItem(BPELConstants.MONITORING_TARGETNAMESPACE).getNodeValue();
        }

        if (process == null) { throw new BPELMonitoringException("BPELMonitoring : no instance found with the id : " + bpiId, null); }
        try {
            String packag = Util.getPackageFromNamespace(targetNamespace);
            ClassLoader classL = ClassLoaderMap.getCL(process);
            Class cl = Class.forName((packag + ".BPEL" + process), true, classL);
            Method met = cl.getMethod("getProcess", (Class[]) null);
            this.bpelProcess = (BPELProcess) met.invoke(null, (Object[]) null);
        } catch (NoSuchMethodException e) {
            throw new BPELMonitoringException("NoSuchMethodException BpelMonitoringSession getBpelProcess :\n", e);
        } catch (ClassNotFoundException e) {
            throw new BPELMonitoringException("ClassNotFoundException BpelMonitoringSession getBpelProcess :\n", e);
        } catch (IllegalAccessException e) {
            throw new BPELMonitoringException("IllegalAccessException BpelMonitoringSession getBpelProcess :\n", e);
        } catch (InvocationTargetException e) {
            throw new BPELMonitoringException("InvocationTargetException BpelMonitoringSession getBpelProcess :\n", e);
        } catch (BPELUtilException e) {
            throw new BPELMonitoringException("BPELUtilException BpelMonitoringSession getBpelProcess :\n", e);
        }
        if (BPELTrace.isMonitoringDebugEnabled()) BPELTrace.traceMonitoring("BpelMonitoringSession getBpelProcess end, this.bpelProcess = " + this.bpelProcess);
    }

    private Document getDocumentOfBpi(int bpiId, int mode) throws BPELMonitoringException {
        //mode = 1 => Inst document
        //mode = 2 => Vars document
        //mode = 3 => Mess document
        Document document = null;
        if (mode == 2) {
            document = MonitoringMap.getVariable(bpiId);
        } else if (mode == 1) {
            document = MonitoringMap.getInstance(bpiId);
        }
        if (document == null) {
            String dirName = BPELConstants.BUILD_XML_DIR;
            File xml = new File(dirName);
            File[] files = xml.listFiles();
            all: for (int i = 0; i < files.length; i++) {
                File processDir = files[i];
                File[] instances = processDir.listFiles();
                for (int j = 0; j < instances.length; j++) {
                    File instance = instances[j];
                    String name = instance.getName();
                    boolean ok = name.startsWith(Integer.toString(bpiId) + "_");
                    if (mode == 2) {
                        ok = ok && name.endsWith("_Vars.xml");
                    } else if (mode == 1) {
                        ok = ok && name.endsWith("_Inst.xml");
                    }
                    if (ok) {
                        document = getDoc(instance);
                        break all;
                    }
                }
            }
            if (mode == 2) {
                MonitoringMap.addVariable(bpiId, document);
            } else if (mode == 1) {
                MonitoringMap.addInstance(bpiId, document);
            }
        }
        return document;
    }

    private Document getDocumentOfBpi(int bpiId) throws BPELMonitoringException {
        return getDocumentOfBpi(bpiId, 1);
    }

    private Document getDocumentVariableOfBpi(int bpiId) throws BPELMonitoringException {
        return getDocumentOfBpi(bpiId, 2);
    }

    private void initDocument() throws BPELMonitoringException {
        DocumentBuilder docBuilder = null;
        DocumentBuilderFactory docFactory = DocumentBuilderFactory.newInstance();
        try {
            docBuilder = docFactory.newDocumentBuilder();
        } catch (ParserConfigurationException e) {
            throw new BPELMonitoringException("ParserConfigurationException caught in BpelMonitoringSessionBean.buildDOMDocument : \n", e);
        }
        this.doc = docBuilder.newDocument();
    }

    private void buildDomDocument(long maxTimeStamp, int bpiId) throws BPELMonitoringException {
        if (BPELTrace.isMonitoringDebugEnabled()) BPELTrace.traceMonitoring("BpelMonitoringSession buildDomDocument, maxTimeStamp = " + maxTimeStamp);
        BPIData bpid = null;
        try {
            bpid = BPICache.findBPI(bpiId);
        } catch (BPELEngineException e) {
            throw new BPELMonitoringException("BPELEngineException caught in BPELMonitoring.buildDomDocument : ", e);
        }

        Element elem = this.doc.createElement(this.bpelProcess.getName());
        addAttribute(elem, BPELConstants.MONITORING_MAXTIMESTAMP, Long.toString(maxTimeStamp));
        addAttribute(elem, BPELConstants.MONITORING_CREATIONDATE, Long.toString(bpid.getCreationDate().getTime()));
        addAttribute(elem, BPELConstants.MONITORING_BPIID, Integer.toString(bpid.getBPIId()));
        addAttribute(elem, BPELConstants.MONITORING_PROCESSID, Integer.toString(bpid.getProcessId()));
        addAttribute(elem, BPELConstants.MONITORING_PROCESSNAME, bpid.getProcessName());
        addAttribute(elem, BPELConstants.MONITORING_TARGETNAMESPACE, bpid.getTargetNamespace());
        addAttribute(elem, BPELConstants.MONITORING_STATE, bpid.getState());
        this.doc.appendChild(elem);
        Statement s = this.bpelProcess.findStatement(1);
        fillDocument(elem, s);
        if (BPELTrace.isMonitoringDebugEnabled()) BPELTrace.traceMonitoring("BpelMonitoringSession end buildDomDocument");
    }

    private void fillDocument(Node father, Statement statementToAdd) {
        if (BPELTrace.isMonitoringDebugEnabled()) BPELTrace.traceMonitoring("BpelMonitoringSession fillDocument, father = " + father + ", statementToAdd = " + statementToAdd);
        Element child = this.doc.createElement("_" + statementToAdd.getNb().toString());
        fillChild(child, statementToAdd);
        father.appendChild(child);
        Statement[] children = statementToAdd.getChilds();
        for (int i = 0; i < statementToAdd.getChildNb(); i++) {
            Statement next = children[i];
            fillDocument(child, next);
        }
        if (BPELTrace.isMonitoringDebugEnabled()) BPELTrace.traceMonitoring("BpelMonitoringSession end fillDocument");
    }

    private void fillChild(Element el, Statement s) {
        addAttribute(el, "type", s.getType());
        addAttribute(el, "nb", s.getNb().toString());
        Vector v = this.msTab[s.getNb().intValue()];
        MonitoringStatement ms = getMonitoringStatement(v);
        if (ms != null) {
            addAttribute(el, BPELConstants.MONITORING_EXECUTED_NODE_NAME, "true");
            addAttribute(el, BPELConstants.MONITORING_TIMESTAMP_NODE_NAME, Long.toString(ms.getTimeStamp()));
            addAttribute(el, BPELConstants.MONITORING_LOOPS_NODE_NAME, Integer.toString(ms.getExecutionNb()));
            addAttribute(el, BPELConstants.MONITORING_BPUID_NODE_NAME, Integer.toString(ms.getBpuId()));
            Vector args = ms.getArgs();
            for (int i = 0; i < args.size(); i++) {
                MonitoringArg arg = (MonitoringArg) args.get(i);
                addAttribute(el, arg.getName(), arg.getValue());
            }
        } else {
            addAttribute(el, "executed", "false");
        }

        if (Activity.class.isAssignableFrom(s.getClass())) {
            fillNodeWithActivity(el, (Activity) s);
        }
        /* List of statements having nothing more to get */
        /*
         * Assign CatchAll CompensationHandler Copy Correlations CorrelationSets
         * Empty EventHandler FaultHandler Flow Link Links Otherwise Partner
         * PartnerLinks Partners Scope Sequence Switch Terminate Variables
         */

        if (s.getType().equals("Case")) {
            Case tmp = (Case) s;
            addAttribute(el, "condition", tmp.getCondition());
        } else if (s.getType().equals("Catch")) {
            Catch tmp = (Catch) s;
            addAttribute(el, "faultName", tmp.getFaultName());
            addAttribute(el, "faultVariable", tmp.getFaultVariable());
        } else if (s.getType().equals("Compensate")) {
            Compensate tmp = (Compensate) s;
            addAttribute(el, "scopeName", tmp.getScopeName());
        } else if (s.getType().equals("Correlation")) {
            Correlation tmp = (Correlation) s;
            addAttribute(el, "correlationSet", tmp.getCorrelationSet());
            addAttribute(el, "initiate", tmp.getInitiate());
            addAttribute(el, "pattern", tmp.getPattern());
        } else if (s.getType().equals("CorrelationSet")) {
            CorrelationSet tmp = (CorrelationSet) s;
            addAttribute(el, "properties", tmp.getProperties());
        } else if (s.getType().equals("From")) {
            From tmp = (From) s;
            addAttribute(el, "part", tmp.getPart());
            addAttribute(el, "partnerLink", tmp.getPartnerLink());
            addAttribute(el, "property", tmp.getProperty());
            addAttribute(el, "query", tmp.getQuery());
            addAttribute(el, "variable", tmp.getVariable());
            addAttribute(el, "endpointReference", tmp.getEndpointReference());
            addAttribute(el, "expression", tmp.getExpression());
            addAttribute(el, "literal", tmp.getLiteral());
            addAttribute(el, "opaque", tmp.getOpaque());
        } else if (s.getType().equals("Invoke")) {
            Invoke tmp = (Invoke) s;
            addAttribute(el, "inputVariable", tmp.getInputVariable());
            addAttribute(el, "operation", tmp.getInvokeOperation());
            addAttribute(el, "partnerLink", tmp.getInvokePartnerLink());
            addAttribute(el, "port", tmp.getInvokePort());
            addAttribute(el, "portType", tmp.getInvokePortType());
            addAttribute(el, "outputVariable", tmp.getOutputVariable());
        } else if (s.getType().equals("OnAlarm")) {
            OnAlarm tmp = (OnAlarm) s;
            addAttribute(el, "for", tmp.getFor());
            addAttribute(el, "until", tmp.getUntil());
        } else if (s.getType().equals("OnMessage")) {
            OnMessage tmp = (OnMessage) s;
            addAttribute(el, "operation", tmp.getOnMessageOperation());
            addAttribute(el, "partnerLink", tmp.getOnMessagePartnerLink());
            addAttribute(el, "portType", tmp.getOnMessagePortType());
            addAttribute(el, "variable", tmp.getOnMessageVariable());
        } else if (s.getType().equals("PartnerLink")) {
            PartnerLink tmp = (PartnerLink) s;
            addAttribute(el, "partnerLinkType", tmp.getLinkType());
            addAttribute(el, "myRole", tmp.getMyRole());
            addAttribute(el, "partnerRole", tmp.getPartnerRole());
        } else if (s.getType().equals("Pick")) {
            Pick tmp = (Pick) s;
            addAttribute(el, "createInstance", tmp.getCreateInstance());
        } else if (s.getType().equals("Receive")) {
            Receive tmp = (Receive) s;
            addAttribute(el, "createInstance", tmp.getReceiveCreateInstance());
            addAttribute(el, "operation", tmp.getReceiveOperation());
            addAttribute(el, "partnerLink", tmp.getReceivePartnerLink());
            addAttribute(el, "portType", tmp.getReceivePortType());
            addAttribute(el, "variable", tmp.getReceiveVariable());
        } else if (s.getType().equals("Reply")) {
            Reply tmp = (Reply) s;
            addAttribute(el, "operation", tmp.getReplyOperation());
            addAttribute(el, "partnerLink", tmp.getReplyPartnerLink());
            addAttribute(el, "portType", tmp.getReplyPortType());
            addAttribute(el, "variable", tmp.getReplyVariable());
        } else if (s.getType().equals("Source")) {
            Source tmp = (Source) s;
            addAttribute(el, "linkName", tmp.getLinkName());
            addAttribute(el, "transitionCondition", tmp.getTransitionCondition());
        } else if (s.getType().equals("Target")) {
            Target tmp = (Target) s;
            addAttribute(el, "linkName", tmp.getLinkName());
        } else if (s.getType().equals("Throw")) {
            Throw tmp = (Throw) s;
            addAttribute(el, "faultName", tmp.getFaultName());
            addAttribute(el, "faultVariable", tmp.getFaultVariable());
        } else if (s.getType().equals("To")) {
            To tmp = (To) s;
            addAttribute(el, "part", tmp.getPart());
            addAttribute(el, "partnerLink", tmp.getPartnerLink());
            addAttribute(el, "property", tmp.getProperty());
            addAttribute(el, "query", tmp.getQuery());
            addAttribute(el, "variable", tmp.getVariable());
        } else if (s.getType().equals("Variable")) {
            Variable tmp = (Variable) s;
            addAttribute(el, "element", tmp.getElement());
            addAttribute(el, "messageType", tmp.getMessageType());
            addAttribute(el, "name", tmp.getVariableName());
        } else if (s.getType().equals("Wait")) {
            Wait tmp = (Wait) s;
            addAttribute(el, "for", tmp.getFor());
            addAttribute(el, "until", tmp.getUntil());
        } else if (s.getType().equals("While")) {
            While tmp = (While) s;
            addAttribute(el, "condition", tmp.getCondition());
        }
        if (BPELTrace.isMonitoringDebugEnabled()) BPELTrace.traceMonitoring("BpelMonitoringSession end fillChild");
    }

    private void fillNodeWithActivity(Element el, Activity a) {
        addAttribute(el, "name", a.getName());
        addAttribute(el, "joinCondition", a.getJoinCondition());
        addAttribute(el, "suppressJoinFailure", a.getSuppressJoinFailure());
    }

    private MonitoringStatement getMonitoringStatement(Vector v) {
        MonitoringStatement ms = null;
        if (v != null) {
            if (v.size() > 0) {
                ms = new MonitoringStatement();
                MonitoringStatement last = (MonitoringStatement) v.get(v.size() - 1);
                ms.setBpuId(last.getBpuId());
                ms.setExecutionNb(v.size());
                ms.setTimeStamp(last.getTimeStamp());
                ms.setStatementNb(last.getStatementNb());
                ms.setType(last.getType());

                Vector allArgs = new Vector();
                for (int i = 0; i < v.size(); i++) {
                    MonitoringStatement tmp = (MonitoringStatement) v.get(i);
                    Vector args = tmp.getArgs();
                    for (int j = 0; j < args.size(); j++) {
                        MonitoringArg ma = (MonitoringArg) args.get(j);
                        allArgs.addElement(new MonitoringArg(ma.getName() + "_loop" + (i + 1), ma.getValue()));
                    }
                }
                ms.setArgs(allArgs);
            }
        }
        return ms;
    }

    private void addAttribute(Element el, String attrName, String attrValue) {
        el.setAttribute(attrName, attrValue);
    }

    /**
     * get a var thanks to its bpi, name and statementNb using this variable
     *
     * @ejb:interface-method
     */
    public org.objectweb.orchestra.monitoring.Variable getVar(int bpiId, String name, int statementNb) throws BPELMonitoringException {
        if (BPELTrace.isMonitoringDebugEnabled()) BPELTrace.traceMonitoring("BpelMonitoringSession getVar, bpiId = " + bpiId + ", name = " + name + ", statementNb = " + statementNb);
        org.objectweb.orchestra.monitoring.Variable var = null;
        int scopeId = -1;
        getBpelProcess(bpiId);
        Statement s = this.bpelProcess.findStatement(statementNb);
        int scopeNb = findStatementOfType("Scope", s);
        String scopeNbName = Integer.toString(bpiId) + scopeNb;
        if (this.running) {
            try {
                Collection scopes = ScopeCache.findByNb(scopeNbName);
                int max = -1;
                for (Iterator it = scopes.iterator(); it.hasNext();) {
                    ScopeData sd = (ScopeData) it.next();
                    int iter = sd.getIter();
                    if (iter > max) {
                        max = iter;
                        scopeId = sd.getScopeId();
                    }
                }
                var = getVarFromScope(bpiId, name, scopeId);
            } catch (BPELEngineException e) {
                //DO NOTHING, it means this instance is not running
                this.running = false;
            }
        }
        if (var == null && !BPELEngine.flushInstance()) {
            throw new BPELMonitoringObjectNotFoundException("Instance with id "+bpiId+" is not running and the engine is not flushing finished instances.", null);
        } else if (var == null) {
            Document varDoc = getDocumentVariableOfBpi(bpiId);
            var = getVarFromDocument(varDoc, name, scopeNbName);
        }
        if (BPELTrace.isMonitoringDebugEnabled()) BPELTrace.traceMonitoring("BpelMonitoringSession getVar end, var = " + var);
        return var;
    }

    private org.objectweb.orchestra.monitoring.Variable getVarFromDocument(Document doc, String varName, String scopeNb) throws BPELMonitoringException {
        org.objectweb.orchestra.monitoring.Variable var = null;
        if (BPELTrace.isMonitoringDebugEnabled()) BPELTrace.traceMonitoring("BpelMonitoringSession getVarFromDocument, doc = " + doc + ", varName = " + varName + ", scopeNb = " + scopeNb);
        Node varNode = null;
        int iter = -1;
        NodeList variables = doc.getDocumentElement().getChildNodes();
        for (int i = 0; i < variables.getLength(); i++) {
            Node n = variables.item(i);
            if (n.getNodeName().equals(varName)) {
                NamedNodeMap attrs = n.getAttributes();
                Node scopeNbNode = attrs.getNamedItem("scopeNb");
                if (scopeNbNode.getNodeValue().equals(scopeNb)) {
                    Node iterNode = attrs.getNamedItem("scopeIter");
                    int iterTmp = Integer.parseInt(iterNode.getNodeValue());
                    if (iterTmp > iter) {
                        iter = iterTmp;
                        varNode = n;
                    }
                }
            }
        }
        var = buildVariableFromNode(varNode);
        if (BPELTrace.isMonitoringDebugEnabled()) BPELTrace.traceMonitoring("BpelMonitoringSession getVarFromDocument end, var = " + var);
        return var;
    }

    private org.objectweb.orchestra.monitoring.Variable buildVariableFromNode(Node n) throws BPELMonitoringException {
        org.objectweb.orchestra.monitoring.Variable var = null;
        if (BPELTrace.isMonitoringDebugEnabled()) BPELTrace.traceMonitoring("BpelMonitoringSession buildVariableFromNode, n = " + n + ", n.localName = " + n.getLocalName());
        Object value = null;
        String type = null;
        String varName = n.getNodeName();
        NamedNodeMap attrs = n.getAttributes();
        String varMessageType = attrs.getNamedItem("messageType").getNodeValue();
        if (!varMessageType.equals("")) {
            type = BPELConstants.BPELMESSAGE;
            try {
                BPELMessage bm = Util.buildJM(this.bpelProcess.getName(), Util.buildQname(varMessageType));
                NodeList parts = n.getChildNodes();
                for (int i = 0; i < parts.getLength(); i++) {
                    Node part = parts.item(i);
                    NamedNodeMap partAttrs = part.getAttributes();
                    String partName = partAttrs.getNamedItem("partName").getNodeValue();
                    Node partValue = partAttrs.getNamedItem("value");
                    if (partValue != null) {
                        //this part is null or its type is a final wsdl type
                        if (!partValue.getNodeValue().equals("")) {
                            String partValueClass = partAttrs.getNamedItem("class").getNodeValue();
                            bm.setPart(partName, getValueFromString(partValue.getNodeValue(), getClass(partValueClass)));
                        }
                    } else {
                        Object o = fillObject(part);
                        bm.setPart(partName, o);
                    }
                }
                value = bm;
            } catch (BPELUtilException e) {
                throw new BPELMonitoringException("BPELUtilException caught in BPELMonitoring.buildVariableFromNode : ", e);
            }
        } else {
            type = BPELConstants.BPELVARIABLE;
            String varType = attrs.getNamedItem("type").getNodeValue();
            String varElement = attrs.getNamedItem("element").getNodeValue();
            String varValue = attrs.getNamedItem("value").getNodeValue();
            String varValueClass = attrs.getNamedItem("class").getNodeValue();
            String t = null;
            if (varType != null) {
                t = varType;
            } else {
                t = varElement;
            }

            Class cl = null;
            try {
                ClassLoader classL = ClassLoaderMap.getCL(this.bpelProcess.getName());
                //cl = Class.forName(WSDLJavaMatching.getWrapper(t), true, classL);
                cl = Class.forName(varValueClass, true, classL);
            } catch (BPELUtilException e) {
                throw new BPELMonitoringException("BPELUtilException caught in BPELMonitoring.buildVariableFromNode : ", e);
            } catch (ClassNotFoundException e) {
                throw new BPELMonitoringException("ClassNotFoundException caught in BPELMonitoring.buildVariableFromNode : ", e);
            }

            // TODO There is duplicate code with EngineBean.createVariable()
            BPELVariable v = new BPELVariable(varName, t, cl);
            if (!varValue.equals("")) {
                v.setValue(getValueFromString(varValue, cl));
            }
            value = v;
        }
        var = new org.objectweb.orchestra.monitoring.Variable(value, type);
        if (BPELTrace.isMonitoringDebugEnabled()) BPELTrace.traceMonitoring("BpelMonitoringSession buildVariableFromNode end, var = " + var);
        return var;
    }

    private Constructor getCompleteConstructor(Class cl) {
        Constructor result = null;
        int max = -1;
        Constructor[] tab = cl.getDeclaredConstructors();
        for (int i = 0; i < tab.length; i++) {
            Constructor tmp = tab[i];
            if (tmp.getParameterTypes().length > max) {
                max = tmp.getParameterTypes().length;
                result = tmp;
            }
        }
        return result;
    }

    private Object fillObject(Node n) throws BPELMonitoringException {
        Object result = null;
        NamedNodeMap attrs = n.getAttributes();
        Node value = attrs.getNamedItem("value");
        Node valueClass = attrs.getNamedItem("class");
        Class valueCl = getClass(valueClass.getNodeValue());
        if (value != null) {
            if (value.getNodeValue() != null) {
                result = getValueFromString(value.getNodeValue(), valueCl);
            }
        } else {
            //Node valueClass = attrs.getNamedItem("class");
            //Class valueCl = getClass(valueClass.getNodeValue());
            Constructor c = getCompleteConstructor(valueCl);
            Object[] params = new Object[c.getParameterTypes().length];
            NodeList children = n.getChildNodes();
            for (int i = 0; i < children.getLength(); i++) {
                Node child = children.item(i);
                params[i] = fillObject(child);
            }
            try {
                result = c.newInstance(params);
            } catch (Exception e) {
                throw new BPELMonitoringException("Exception caught in BPELMonitoring.fillObject : ", e);
            }
        }

        return result;
    }

    private Class getClass(String className) throws BPELMonitoringException {
        Class cl = null;
        try {
            ClassLoader classL = ClassLoaderMap.getCL(this.bpelProcess.getName());
            cl = Class.forName(className, true, classL);
        } catch (ClassNotFoundException e) {
            throw new BPELMonitoringException("ClassNotFoundException caught in BPELMonitoring.buildVariableFromNode : ", e);
        } catch (BPELUtilException e) {
            throw new BPELMonitoringException("BPELUtilException caught in BPELMonitoring.buildVariableFromNode : ", e);
        }
        return cl;
    }

    private Object getValueFromString(String stValue, Class cl) throws BPELMonitoringException {
        Object result = null;
        try {
            if (WSDLJavaMatching.isFinalWsdlType(cl)) {
                if (cl.equals(QName.class)) {
                    String ns = null;
                    String localPart = null;
                    int index = stValue.indexOf("}");
                    if (index != -1) {
                        ns = stValue.substring(1, index);
                        localPart = stValue.substring(index + 1);
                        result = new QName(ns, localPart);
                    }
                } else if (Calendar.class.isAssignableFrom(cl)) {
                    //int year, int month, int date, int hour, int minute, int second
                    String[] param = stValue.split(",");
                    if (param.length == 6) {
                        result = new GregorianCalendar(Integer.parseInt(param[0]), Integer.parseInt(param[1]), Integer.parseInt(param[2]), Integer.parseInt(param[3]), Integer.parseInt(param[4]), Integer.parseInt(param[5]));
                    }
                } else {
                    //there is a constructor with a string
                    Constructor constr = cl.getConstructor(new Class[] {String.class});
                    String param = stValue;
                    if (param.equals("")) {
                        param = "0";
                    }
                    result = constr.newInstance(new Object[] {param});
                }
            } else {
                //cl represents a complex type with a null value
                //result = cl.newInstance();
                result = null;
            }
        } catch (Exception e) {
            throw new BPELMonitoringException("Exception caught in BPELMonitoring.getValueFromString : ", e);
        }
        return result;
    }

    private org.objectweb.orchestra.monitoring.Variable getVarFromScope(int bpiId, String name, int scopeId) throws BPELMonitoringException {
        org.objectweb.orchestra.monitoring.Variable monitoringVar = null;
        if (BPELTrace.isMonitoringDebugEnabled()) BPELTrace.traceMonitoring("BpelMonitoringSession getVarFromScope, bpiId = " + bpiId + ", name = " + name + ", scopeId = " + scopeId);
        try {
            EngineHomeLocal eh = EngineUtil.getLocalHome();
            EngineLocal engine = eh.create();
            String processName = getProcessName(bpiId);
            ClassLoader classL = ClassLoaderMap.getCL(processName);
            Thread.currentThread().setContextClassLoader(classL);
            VariableData v = engine.getVar(bpiId, name, scopeId);
            engine.remove();
            Object value = Util.getValueObject(v.getValue(), processName);
            String type = null;
            if (BPELMessage.class.isAssignableFrom(value.getClass())) {
                type = BPELConstants.BPELMESSAGE;
            } else if (BPELVariable.class.isAssignableFrom(value.getClass())) {
                type = BPELConstants.BPELVARIABLE;
            } else {
                throw new BPELMonitoringException("BpelMonitoringSessionBean Wrong value type in variable : " + name + " : " + value.getClass().getName(), null);
            }
            monitoringVar = new org.objectweb.orchestra.monitoring.Variable(value, type);
        } catch (javax.naming.NamingException e) {
            throw new BPELMonitoringException("NamingException BpelMonitoringSessionBean getVar :\n", e);
        } catch (javax.ejb.CreateException e) {
            throw new BPELMonitoringException("CreateException BpelMonitoringSessionBean getVar :\n", e);
        } catch (BPELEngineException e) {
            throw new BPELMonitoringException("BPELEngineException BpelMonitoringSessionBean getVar :\n", e);
        } catch (BPELUtilException e) {
            throw new BPELMonitoringException("BPELUtilException BpelMonitoringSessionBean getVar :\n", e);
        } catch (RemoveException e) {
            throw new BPELMonitoringException("RemoveException BpelMonitoringSessionBean getVar :\n", e);
        }
        if (BPELTrace.isMonitoringDebugEnabled()) BPELTrace.traceMonitoring("BpelMonitoringSession getVarFromScope end, monitoringVar = " + monitoringVar);
        return monitoringVar;
    }

    private int findStatementOfType(String type, Statement s) {
        if (BPELTrace.isMonitoringDebugEnabled()) BPELTrace.traceMonitoring("BpelMonitoringSession findStatementOfType, statement s = " + s + ", type = " + type);
        int result = -1;
        Statement father = s.getFather();
        while (result == -1 && father != null) {
            if (father.getType().equals(type)) {
                result = father.getNb().intValue();
            } else {
                father = father.getFather();
            }
        }
        if (BPELTrace.isMonitoringDebugEnabled()) BPELTrace.traceMonitoring("BpelMonitoringSession findStatementOfType end, result = " + result);
        return result;
    }

    /**
     * @ejb:interface-method
     */
    public String getProcessName(int bpiId) throws BPELMonitoringException {
        getBpelProcess(bpiId);
        return this.bpelProcess.getName();
    }

    /**
     * @ejb:interface-method
     */
    public int getLastFromBpu(int bpiId, int bpuId) throws BPELMonitoringException {
        if (BPELTrace.isMonitoringDebugEnabled()) BPELTrace.traceMonitoring("BpelMonitoringSession getLastFromBpu, bpiId = " + bpiId + ", bpuId = " + bpuId);
        int statementNb = -1;
        try {
            BPUData bpu = BPUCache.findBPU(bpuId);
            Vector monitor = bpu.getMonitor();
            int size = monitor.size();
            if (size > 0) {
                String st = (String) monitor.get(size - 1);
                int index = st.indexOf(BPELConstants.MONITORING_INFO_SEP);
                if (index != -1) {
                    statementNb = Integer.parseInt(st.substring(0, index));
                } else if (st.startsWith(BPELConstants.MONITORING_ENDFLOW)) {
                    // the last of this bpu is an endFlow
                    statementNb = -2;
                }
            }
        } catch (BPELEngineException e) {
            throw new BPELMonitoringException("BPELEngineException BpelMonitoringSession getLastFromBpu :\n", e);
        }
        if (BPELTrace.isMonitoringDebugEnabled()) BPELTrace.traceMonitoring("BpelMonitoringSession end getLastFromBpu, statementNb = " + statementNb);
        return statementNb;
    }

    private Vector getEndFlowsFromBPI(int bpiId) throws BPELMonitoringException {
        Collection bpus;
        try {
            bpus = BPICache.getBPUs(bpiId);
        } catch (BPELEngineException e) {
            throw new BPELMonitoringException("BPELEngineException BpelMonitoringSession getEndFlowsFromBPI :\n", e);
        }
        Vector endFlows = new Vector();
        this.msTab = new Vector[this.bpelProcess.getTotalStatementNb() + 1];
        for (Iterator it = bpus.iterator(); it.hasNext();) {
            BPUData bpu = (BPUData) it.next();
            Vector monitor = bpu.getMonitor();
            String st = null;
            for (int i = 0; i < monitor.size(); i++) {
                st = (String) monitor.get(i);
                String[] split = st.split(BPELConstants.MONITORING_INFO_SEP);
                if (split.length == 1 && split[0].startsWith(BPELConstants.MONITORING_ENDFLOW)) {
                    int size = BPELConstants.MONITORING_ENDFLOW.length();
                    Integer flowNb = new Integer(split[0].substring(size));
                    endFlows.addElement(flowNb);
                }
                // else it is a monitoring statement
            }
        }
        return endFlows;
    }

    private Vector getTargetsFromBPI(int bpiId) throws BPELMonitoringException {
        Collection bpus;
        try {
            bpus = BPICache.getBPUs(bpiId);
        } catch (BPELEngineException e) {
            throw new BPELMonitoringException("BPELEngineException BpelMonitoringSession getEndFlowsFromBPI :\n", e);
        }
        Vector targets = new Vector();
        this.msTab = new Vector[this.bpelProcess.getTotalStatementNb() + 1];
        for (Iterator it = bpus.iterator(); it.hasNext();) {
            BPUData bpu = (BPUData) it.next();
            Vector monitor = bpu.getMonitor();
            String st = null;
            for (int i = 0; i < monitor.size(); i++) {
                st = (String) monitor.get(i);
                String[] split = st.split(BPELConstants.MONITORING_INFO_SEP);
                if (split.length == 1 && split[0].startsWith(BPELConstants.MONITORING_TARGETS)) {
                    int size = BPELConstants.MONITORING_TARGETS.length();
                    Integer activityNb = new Integer(split[0].substring(size));
                    targets.addElement(activityNb);
                }
                // else it is a monitoring statement
            }
        }
        return targets;
    }
}
