/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.objectweb.orchestra.tasks;

import java.io.File;
import java.util.Vector;

import org.apache.tools.ant.DefaultLogger;
import org.apache.tools.ant.Project;
import org.objectweb.orchestra.util.BPELConstants;



/**
 * ProcessProject.java
 *
 * @author Goulven Le Jeune & Charles Souillard
 * @version $Id: ProcessProject.java 5 2006-10-20 14:37:44Z lejeuneg $
 */
public class ProcessProject {

    private String sep = File.separator;

    private Project p = null;

    private InitPropertiesTask initPropertiesTask = null;
    private CleanProcessTask cleanProcessTask = null;
    private DeployProcessTask deployProcessTask = null;
    private UndeployProcessTask undeployProcessTask = null;
    private LaunchProcessTask launchProcessTask = null;
    private JunitTestsTask junitTestsTask = null;
    private LoadProcessWSTask loadProcessWSTask = null;
    private CompileClientTask compileClientTask = null;

    public ProcessProject() {
        initProject();
    }

    public void initProject() {
        this.p = new Project();

        DefaultLogger dl = new DefaultLogger();
        dl.setMessageOutputLevel(Project.MSG_INFO);
        dl.setErrorPrintStream(System.out);
        dl.setOutputPrintStream(System.out);
        dl.setEmacsMode(false);
        this.p.addBuildListener(dl);
        this.p.init();
    }

    public void initInitProperties(String processName) {
        initInitProperties();
        this.initPropertiesTask.setProcessName(processName);
    }

    public void initInitProperties() {
        this.initPropertiesTask = new InitPropertiesTask();
        this.initPropertiesTask.setProject(this.p);
    }

    public void initClean(String processName) {
        this.cleanProcessTask = new CleanProcessTask();
        this.cleanProcessTask.setProject(this.p);
        this.cleanProcessTask.setProcessName(processName);
    }

    public void initDeploy(String processName) {
        String processPath = BPELConstants.BUILD_PROCESSFILES_DIR + this.sep + processName;
        String bpelFile = processPath + this.sep + processName + ".bpel";
        String wsdlFile = processPath + this.sep + processName + ".wsdl";

        File f = new File(processPath);
        String[] processFiles = f.list();
        Vector tmp = new Vector();
        for (int i=0;i<processFiles.length;i++) {
            if (processFiles[i].startsWith(BPELConstants.ADMIN_EXTWSDLFILE) && processFiles[i].endsWith(".wsdl")) {
                tmp.add(processPath + this.sep + processFiles[i]);
            }
        }
        String[] extwsdlFiles = new String[tmp.size()];
        for (int i=0;i<tmp.size();i++) {
            extwsdlFiles[i] = (String) tmp.get(i);
        }
        if (extwsdlFiles.length != 0)
            this.deployProcessTask = new DeployProcessTask(extwsdlFiles, false);
        else
            this.deployProcessTask = new DeployProcessTask(null, false);
        this.deployProcessTask.setProject(this.p);
        this.deployProcessTask.setProcessName(processName);
        this.deployProcessTask.setBpelFile(bpelFile);
        this.deployProcessTask.setWsdlFile(wsdlFile);
        this.deployProcessTask.setSrcDir(null);
    }

    public void initDeploy(String processName, String bpelFile, String wsdlFile, String[] outServices, String srcDir, boolean skipwsdl) {
        this.deployProcessTask = new DeployProcessTask(outServices, skipwsdl);
        this.deployProcessTask.setProject(this.p);
        this.deployProcessTask.setProcessName(processName);
        this.deployProcessTask.setBpelFile(bpelFile);
        this.deployProcessTask.setWsdlFile(wsdlFile);
        this.deployProcessTask.setSrcDir(srcDir);
    }

    public void initUndeploy(String processName) {
        this.undeployProcessTask = new UndeployProcessTask();
        this.undeployProcessTask.setProject(this.p);
        this.undeployProcessTask.setProcessName(processName);
    }

    public void initLaunch(String processName, String clientName, String args) {
        this.launchProcessTask = new LaunchProcessTask();
        this.launchProcessTask.setProject(this.p);
        this.launchProcessTask.setProcessName(processName);
        this.launchProcessTask.setClientname(clientName);
        this.launchProcessTask.setArgs(args);
    }

    public void initJunit() {
        this.junitTestsTask = new JunitTestsTask();
        this.junitTestsTask.setProject(this.p);
    }

    public void initLoadProcessWS(String processName) {
        this.loadProcessWSTask = new LoadProcessWSTask();
        this.loadProcessWSTask.setProject(this.p);
        this.loadProcessWSTask.setProcessName(processName);
    }

    public void initCompileClient(String processName, String dir, String includes, String excludes, String classpath) {
        this.compileClientTask = new CompileClientTask();
        this.compileClientTask.setProject(this.p);
        this.compileClientTask.setProcessName(processName);
        this.compileClientTask.setDir(dir);
        this.compileClientTask.setIncludes(includes);
        this.compileClientTask.setExcludes(excludes);
        this.compileClientTask.setClasspath(classpath);
    }

    public void cleanProcess(String processName) {
        initInitProperties(processName);
        initClean(processName);
        this.initPropertiesTask.execute();
        this.cleanProcessTask.execute();
    }

    public void deployProcess(String processName) {
        initInitProperties(processName);
        initDeploy(processName);
        this.initPropertiesTask.execute();
        this.deployProcessTask.execute();
    }

    public void deployProcess(String processName, String bpelFile, String wsdlFile, String[] outServices, String srcDir, boolean skipwsdl) {
        initInitProperties(processName);
        initDeploy(processName, bpelFile, wsdlFile, outServices, srcDir, skipwsdl);
        this.initPropertiesTask.execute();
        this.deployProcessTask.execute();
    }

    public void undeployProcess(String processName) {
        initInitProperties(processName);
        initUndeploy(processName);
        this.initPropertiesTask.execute();
        this.undeployProcessTask.execute();
    }

    public void launchProcess(String processName, String clientName, String args) {
        initInitProperties(processName);
        initLaunch(processName, clientName, args);
        this.initPropertiesTask.execute();
        this.launchProcessTask.execute();
    }

    public void launchJunit() {
        initInitProperties();
        initJunit();
        this.initPropertiesTask.execute();
        this.junitTestsTask.execute();
    }


    public void deployProcessWS(String processName) {
        initInitProperties(processName);
        initLoadProcessWS(processName);
        this.initPropertiesTask.execute();
        this.loadProcessWSTask.execute();
    }

    public void compileClient(String processName, String dir, String includes, String excludes, String classpath) {
        initInitProperties(processName);
        initCompileClient(processName,dir, includes, excludes, classpath);
        this.initPropertiesTask.execute();
        this.compileClientTask.execute();
    }
}