/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.objectweb.orchestra.util;

import javax.xml.namespace.QName;

import org.apache.axis.utils.JavaUtils;
import org.apache.axis.wsdl.toJava.Utils;

/**
 * WSDLJavaMatching.java, give utility methods to represents the mapping between
 * wsdl types and java types
 *
 * @author Goulven Le Jeune & Charles Souillard
 * @version $Id: WSDLJavaMatching.java 5 2006-10-20 14:37:44Z lejeuneg $
 *
 */
public class WSDLJavaMatching {

    public static boolean printDebug = false;

    /**
     * print a message if printDebug variable is set to true
     *
     * @param mess
     *            the message to print
     */
    private static void printMess(String mess) {
        if (printDebug) {
            System.out.println(mess);
        }
    }


    public static String getWrapperConstructor(String t) {
        printMess("WSDLJavaMatching, getWrapperConstructor, t=" + t);
        String result = t;
        if (t.equals("java.lang.Integer")) {
            result = "java.lang.Integer(0)";
        } else if (t.equals("java.lang.Long")) {
            result = "java.lang.Long(0)";
        } else if (t.equals("java.lang.Short")) {
            result = "java.lang.Short((short)0)";
        } else if (t.equals("java.lang.Float")) {
            result = "java.lang.Float(0.0)";
        } else if (t.equals("java.lang.Double")) {
            result = "java.lang.Double(0.0)";
        } else if (t.equals("java.lang.Boolean")) {
            result = "java.lang.Boolean(false)";
        } else if (t.equals("java.lang.Byte")) {
            result = "java.lang.Byte((byte)0)";
        } else if (t.equals("java.lang.Character")) {
            result = "java.lang.Character('0')";
        } else if (t.equals("java.math.BigDecimal")) {
            result = "java.math.BigDecimal(0.0)";
        } else if (t.equals("java.math.BigInteger")) {
            result = "java.math.BigInteger(\"0\")";
        } else if (t.equals("java.net.URI")) {
            result = "java.net.URI(\"http://localhost:9000\")";
        } else if (t.equals("javax.xml.namespace.QName")) {
            result = "javax.xml.namespace.QName(\"\",\"\")";
        }
        printMess("WSDLJavaMatching, end getWrapperConstructor, result=" + result);
        return result;
    }

    /**
     * get the primitive type of a java wrapper type : for example for
     * java.lang.Integer, return int
     *
     * @param t
     *            the name of a java class
     * @return String the primitive type associated to t
     */
    public static String getPrimitiveType(String t) {
        printMess("WSDLJavaMatching, getPrimitiveType, t=" + t);
        String result = t;
        if (t.equals("java.lang.Integer")) {
            result = "int";
        } else if (t.equals("java.lang.Long")) {
            result = "long";
        } else if (t.equals("java.lang.Short")) {
            result = "short";
        } else if (t.equals("java.lang.Float")) {
            result = "float";
        } else if (t.equals("java.lang.Double")) {
            result = "double";
        } else if (t.equals("java.lang.Boolean")) {
            result = "boolean";
        } else if (t.equals("java.lang.Byte")) {
            result = "byte";
        } else if (t.equals("java.lang.Character")) {
            result = "char";
        }
        printMess("WSDLJavaMatching, end getPrimitiveType, result=" + result);
        return result;
    }

    public static String getWrapper(String type) {
        String result = type;
        if (JavaUtils.getWrapper(type) != null) {
            result = "java.lang." + JavaUtils.getWrapper(type);
        }
        return result;
    }

    public static String getWrapperType(String type) {
        return Utils.getWrapperType(type);
    }

    /**
     * get the wsdl type associated to class cl
     *
     * @param cl the java class we want to find the wsdl type
     * @return Class the wsdl java class corresponding to cl, cl if there is no
     *         association between wsdl types and cl
     */
    public static Class getPrimitiveClass(Class cl) {
        printMess("WSDLJavaMatching, getWsdlType, cl=" + cl);
        Class result = cl;
        Class cl2 = JavaUtils.getPrimitiveClass(cl);
        if (cl2 != null) {
            result = cl2;
        }
        printMess("WSDLJavaMatching, end getWsdlType, result=" + result);
        return result;
    }

    /**
     * Returns the Java type associated to a XML Schema built-in base type 
     * (aka XML Schema built-in datatype) following the JAX RPC 1.1 specification 
     * @param qname a XML Schema base type
     * @return the Java type associated to the XML Schema base type or null 
     * if the QName does not represent a XML Schema base type
     * @see chapter 4 "WSDL/XML to Java Mapping" of JAX-RPC 1.1 specification (page 31) 
     */
    public static String getJavaType(QName qname) {
        printMess("WSDLJavaMatching, getJavaType, qname=" + qname);
        String result = null;
        if (BPELConstants.NS_XSD.equals(qname.getNamespaceURI())) {
            String t = qname.getLocalPart();
            if (t.equals("integer")) {
                result = "java.math.BigInteger";
            } else if (t.equals("int")) {
                result = "int";
            } else if (t.equals("long")) {
                result = "long";
            } else if (t.equals("short")) {
                result = "short";
            } else if (t.equals("decimal")) {
                result = "java.math.BigDecimal";
            } else if (t.equals("float")) {
                result = "float";
            } else if (t.equals("double")) {
                result = "double";
            } else if (t.equals("boolean")) {
                result = "boolean";
            } else if (t.equals("byte")) {
                result = "byte";
            } else if (t.equals("unsignedInt")) {
                result = "long";
            } else if (t.equals("unsignedShort")) {
                result = "int";
            } else if (t.equals("unsignedByte")) {
                result = "short";
            } else if (t.equals("QName")) {
                result = "javax.xml.namespace.QName";
            } else if (t.equals("anyURI")) {
                result = "java.net.URI";
            } else if (t.equals("base64Binary")) {
                result = "byte[]";
            } else if (t.equals("hexBinary")) {
                result = "byte[]";
            } else if ((t.equals("dateTime")) ||
                    (t.equals("date")) ||
                    (t.equals("time"))) {
                result = "java.util.Calendar";
            } else if ((t.equals("nonPositiveInteger")) ||
                    (t.equals("negativeInteger")) ||
                    (t.equals("nonNegativeInteger")) ||
                    (t.equals("unsignedLong")) ||
                    (t.equals("positiveInteger"))) {
                result = "java.math.BigInteger";
            } else if ((t.equals("string")) ||
                    (t.equals("anySimpleType")) ||
                    (t.equals("duration")) ||
                    (t.equals("gYearMonth")) ||
                    (t.equals("gYear")) ||
                    (t.equals("gMonthDay")) ||
                    (t.equals("gDay")) ||
                    (t.equals("gMonth")) ||
                    (t.equals("normalizedString")) ||
                    (t.equals("token")) ||
                    (t.equals("language")) ||
                    (t.equals("Name")) ||
                    (t.equals("NCName")) ||
                    (t.equals("ID")) ||
                    (t.equals("NMTOKEN")) ||
                    (t.equals("NMTOKENS"))) {
                result = "java.lang.String";
            }
        }
        printMess("WSDLJavaMatching, end getJavaType, result=" + result);
        return result;
    }

    public static boolean isFinalWsdlType(Class cl) {
        printMess("WSDLJavaMatching, isFinalWsdlType, cl=" + cl);
        //wrapper, tableaux

        if (cl.isArray()) {
            cl = cl.getComponentType();
        }
        boolean result = false;

        if (java.math.BigInteger.class.isAssignableFrom(cl)) {
            result = true;
        } else if (java.lang.Integer.class.isAssignableFrom(cl)) {
            result = true;
        } else if (java.lang.Long.class.isAssignableFrom(cl)) {
            result = true;
        } else if (java.lang.Short.class.isAssignableFrom(cl)) {
            result = true;
        } else if (java.math.BigDecimal.class.isAssignableFrom(cl)) {
            result = true;
        } else if (java.lang.Float.class.isAssignableFrom(cl)) {
            result = true;
        } else if (java.lang.Double.class.isAssignableFrom(cl)) {
            result = true;
        } else if (java.lang.Boolean.class.isAssignableFrom(cl)) {
            result = true;
        } else if (java.lang.Byte.class.isAssignableFrom(cl)) {
            result = true;
        } else if (javax.xml.namespace.QName.class.isAssignableFrom(cl)) {
            result = true;
        } else if (java.net.URI.class.isAssignableFrom(cl)) {
            result = true;
        } else if (java.util.Calendar.class.isAssignableFrom(cl)) {
            result = true;
        } else if (java.lang.String.class.isAssignableFrom(cl)) {
            result = true;
        }

        printMess("WSDLJavaMatching, end isFinalWsdlType, result=" + result);
        return result;
    }

    public static String getHolderType(String type) {
        printMess("WSDLJavaMatching, getHolderType, type=" + type);
        String t = type;
        String result = null;

        if (t.equals("java.awt.Image")) {
            result = "org.apache.axis.holders.ImageHolder";
        } else if (t.equals("javax.mail.internet.MimeMultipart")) {
            result = "org.apache.axis.holders.MimeMultipartHolder";
        } else if (t.equals("javax.xml.transform.Source")) {
            result = "org.apache.axis.holders.SourceHolder";
        } else if (t.equals("byte[]")) {
            result = "javax.xml.rpc.holders.ByteArrayHolder";
        } else if (t.equals("java.lang.Object")) {
            result = "javax.xml.rpc.holders.ObjectHolder";
        } else if (t.equals("java.lang.String")) {
            result = "javax.xml.rpc.holders.StringHolder";
        } else if (t.equals("int")) {
            result = "javax.xml.rpc.holders.IntHolder";
        } else if (t.equals("long")) {
            result = "javax.xml.rpc.holders.LongHolder";
        } else if (t.equals("short")) {
            result = "javax.xml.rpc.holders.ShortHolder";
        } else if (t.equals("float")) {
            result = "javax.xml.rpc.holders.FloatHolder";
        } else if (t.equals("double")) {
            result = "javax.xml.rpc.holders.DoubleHolder";
        } else if (t.equals("boolean")) {
            result = "javax.xml.rpc.holders.BooleanHolder";
        } else if (t.equals("byte")) {
            result = "javax.xml.rpc.holders.ByteHolder";
        } else if (t.startsWith("java.lang.")) {
            result = "javax.xml.rpc.holders" + t.substring(t.lastIndexOf(".")) + "WrapperHolder";
        } else if (t.indexOf(".") < 0) {
            result = "javax.xml.rpc.holders" + t + "WrapperHolder";
        } else if (t.equals("java.math.BigDecimal")) {
            result = "javax.xml.rpc.holders.BigDecimalHolder";
        } else if (t.equals("java.math.BigInteger")) {
            result = "javax.xml.rpc.holders.BigIntegerHolder";
            //} else if (t.equals("java.util.Date")) {
            //    result = "org.apache.axis.holders.DateHolder";
        } else if (t.equals("java.net.URI")) {
            result = "org.apache.axis.holders.jaxrpc.URIHolder";
        } else if (t.equals("java.util.Calendar")) {
            result = "javax.xml.rpc.holders.CalendarHolder";
        } else if (t.equals("javax.xml.namespace.QName")) {
            result = "javax.xml.rpc.holders.QNameHolder";
        } else if (t.equals("javax.activation.DataHandler")) {
            result = "org.apache.axis.holders.DataHandlerHolder";
            //} else if (t.startsWith("org.apache.axis.types.")) {
            //    int i = t.lastIndexOf('.');
            //    String tmp = t.substring(i + 1);
            //    result = "org.apache.axis.holders." + tmp + "Holder";
        } else {
            result = Utils.addPackageName(t, "holders") + "Holder";
        }
        printMess("WSDLJavaMatching, end getHolderType, result=" + result);
        return result;
    }

    public static boolean isWrapper(Class cl) {
        boolean result = false;
        if (cl.equals(Integer.class)) {
            result = true;
        } else if (cl.equals(Boolean.class)) {
            result = true;
        } else if (cl.equals(Double.class)) {
            result = true;
        } else if (cl.equals(Float.class)) {
            result = true;
        } else if (cl.equals(Long.class)) {
            result = true;
        } else if (cl.equals(Byte.class)) {
            result = true;
        } else if (cl.equals(Character.class)) {
            result = true;
        } else if (cl.equals(Short.class)) {
            result = true;
        }
        return result;
    }

    /**
     * Returns true if the specified string is the name of a Java primitive type, false otherwise
     * @see http://java.sun.com/docs/books/tutorial/java/nutsandbolts/datatypes.html @ Primitive Data Types
     * @return boolean true if the specified string is the name of a Java primitive type, false otherwise
     */
    public static final boolean isJavaPrimitiveType(String s) {
        if (s == null) {
            return false;
        }
        return s.equals("int")
        || s.equals("char")
        || s.equals("byte")
        || s.equals("boolean")
        || s.equals("double")
        || s.equals("float")
        || s.equals("long")
        || s.equals("short");
    }

    /**
     * Returns true if the specified string is a Java keyword, false otherwise
     * @see http://java.sun.com/docs/books/tutorial/java/nutsandbolts/_keywords.html
     * @return boolean true if the String is a Java keyword, false otherwise
     */
    public static final boolean isJavaKeyword(String s) {
        if (s == null) {
            return false;
        }
        return s.equals("abstract")
        || s.equals("assert")
        || s.equals("boolean")
        || s.equals("break")
        || s.equals("byte")
        || s.equals("case")
        || s.equals("catch")
        || s.equals("char")
        || s.equals("class")
        || s.equals("const")
        || s.equals("continue")
        || s.equals("default")
        || s.equals("do")
        || s.equals("double")
        || s.equals("else")
        || s.equals("enum")
        || s.equals("extends")
        || s.equals("final")
        || s.equals("finally")
        || s.equals("float")
        || s.equals("for")
        || s.equals("goto")
        || s.equals("if")
        || s.equals("implements")
        || s.equals("import")
        || s.equals("instanceof")
        || s.equals("int")
        || s.equals("interface")
        || s.equals("long")
        || s.equals("native")
        || s.equals("new")
        || s.equals("package")
        || s.equals("private")
        || s.equals("protected")
        || s.equals("public")
        || s.equals("return")
        || s.equals("short")
        || s.equals("static")
        || s.equals("strictfp")
        || s.equals("super")
        || s.equals("switch")
        || s.equals("synchronized")
        || s.equals("this")
        || s.equals("throw")
        || s.equals("throws")
        || s.equals("transient")
        || s.equals("try")
        || s.equals("void")
        || s.equals("volatile")
        || s.equals("while");
    }
}